# ifndef _GroupStructure_h_
# define _GroupStructure_h_



typedef VECTOR<int> Vint;
typedef LIST<VECTOR<int> >  LVint;
typedef LIST< LIST<VECTOR<int> >* >  LLVint;


class ListWithCommonNode{
public:
	LVint *List;
	NODE *commonNode;
};


class OED_common{
public:
	VECTOR<NODE*> p;
	VECTOR<NODE*> A; // one A variable for each group; A(i) is the A variable for group i 
	LIST<ListWithCommonNode*> *LLN;
	void make_p(VECTOR<int> &isCommon, DAG &G);
	void make_A(int nGroups, DAG &G); 
	void make_interGroup_t(LIST<fourIntegers> &L, DAG &G);
	void make_intraGroup_t(VECTOR<int> &nInd, DAG &G);
	bool findNode(NODE *&n, int group, int ind); 
};

void OED_common::make_p(VECTOR<int> &isCommon, DAG &G){

	// A common variable p_i is shared by all individuals.
	p = VECTOR<NODE*>(isCommon.Size());
	for(int i=1; i<=p.Size(); i++){
		
		if( isCommon(i) )
			p(i)=MakeVar(&G, "p"+ToString(i)+"_CV");
		else
 			p(i)=0;
	} 
}
	

void OED_common::make_A(int nGroups, DAG &G){
	
	// A common variable A_i is shared by all individuals in group i.
	A = VECTOR<NODE*>( nGroups );
	for(int i=1; i<=A.Size(); i++)
	{ 
		A(i) =MakeVar(&G, "A_g"+ToString(i)+"CV");
	}
}




////////////////////////////////////////////////////////////////////////
// Parse Common Times
////////////////////////////////////////////////////////////////////////

bool PCT_areEqual(Vint &x, Vint &y){	


// return 1 if x and y are equal, return 0 otherwise 
	
	for(int i=1; i<=x.Size(); i++)
	{
		if( x(i) != y(i) ) return 0;
	}
	return 1;
}

bool PCT_isIn(LVint *A, Vint &a){

// return 1 if a in A, return 0, otherwise. 

	A->MoveTop();
	while( !A->eol())
	{
		if( PCT_areEqual( A->Data(), a ) ) return 1;
		A->MoveDown();
	}
	return 0;
}

void PCT_find(LVint *&A, Vint &a, LLVint *ListCollection)
{

//	if we find a list which contains a, then set A pointing to it.
//	else set A to zero.


	A=0; //if ListCollection is empty;
	
	ListCollection->MoveTop();
	while(!ListCollection->eol())
	{
		A=ListCollection->Data();

		if ( PCT_isIn(A, a) )	break; 
		else 				A=0;
		ListCollection->MoveDown();
	}
	
}	


template<typename T>
bool PCT_RemoveFirst(T &element, LIST<T> &L){

L.MoveTop();
while( !L.eol() ){
	if(element == L.Data()) { L.Remove(); return 1;}
	else L.MoveDown();
}
return 0;
}


void OED_common::make_interGroup_t(LIST<fourIntegers> &F, DAG &G){

LLVint *L = new LLVint;
fourIntegers f;
LVint *A = new LVint;
LVint *B = new LVint;
Vint a(2), b(2);


while( !F.empty() )
{
	
	//remove a pair 
	f=F.pop();
	a(1)=f.group_a;
	a(2)=f.variable_a;
	
	b(1)=f.group_b;
	b(2)=f.variable_b;
	
	PCT_find(A, a, L);
	PCT_find(B, b, L);

	if(A==0 && B==0)
	//We did not find a or b in L,
	//there fore a and b are stored in a new list 
	//which is pushed on L
	{
		
		LVint *p = new LVint;
		p->push(a);
		p->push(b);
		L->push(p);
	}
	
	else if(A!=0 && B!=0)
	{
		if (A != B){ Push(*A, *B); }
		//delete B and remove it from L
		PCT_RemoveFirst<LVint*>(B,*L);
		B->Clean();
		//if(A==B) do nothing
	}
	
	else if(A!=0  && B==0) A->push(b);
	else if(A==0 && B!=0)  B->push(a);

}


//1. Associate a unique node to each sublist of L.

LLN = new LIST<ListWithCommonNode*>;
L->MoveTop();
int count(0);
while(!L->eol())
{
	
	NODE *n = MakeVar(&G, "t_CV"+ToString(++count));
	ListWithCommonNode *LN = new ListWithCommonNode;
	LN -> List = L->Data();
	LN->commonNode = n;
	LLN->push(LN);
	L->MoveDown();
	
}

 
}

void OED_common::make_intraGroup_t(VECTOR<int> &nTime, DAG &G)
{

NODE *n;
for(int group_nr =1; group_nr<=nTime.Size(); group_nr++)
{
	for(int time_nr=1; time_nr<=nTime(group_nr); time_nr++)
	{
		if( !findNode(n,group_nr, time_nr) )
		{
			ListWithCommonNode *LN = new ListWithCommonNode;
			LN->List = new LVint;
			Vint v(2); 
			v(1)=group_nr; 
			v(2)=time_nr;
			LN->List->push(v);
			LN->commonNode = MakeVar(&G, "t"+ToString(time_nr)+"_g"+ToString(group_nr)+"CV");
			LLN->push(LN);
		}	
	}
}

}


bool OED_common::findNode(NODE *&n, int group_nr, int time_nr){

// Search through LLN for the (group,time) element.
// If found --- the associated node is stored in n and 1 is returned.
// If not found --- 0 is returned.

ListWithCommonNode *p;
LLN->MoveTop();
while( !LLN->eol() )
{
	p=LLN->Data();
	p->List->MoveTop();
	while(!p->List->eol())
	{
		
		if(p->List->Data()(1)==group_nr && p->List->Data()(2)==time_nr)
		{
			n=p->commonNode;
			return 1;
		}
		p->List->MoveDown();
	}
	LLN->MoveDown();
}

return 0;
}





class OED_individual{

public:
	VECTOR<NODE*> p;
	VECTOR<NODE*> t;
	NODE* A;
	VECTOR<NODE*> y; 

	typedef NODE*(*modelFunction)(NODE *t, VECTOR<NODE*> &p, NODE *A);	 	
	modelFunction model;
	
	OED_individual(int p_size, int t_size): p(p_size), t(t_size), y(t_size){}; 
	OED_individual(int p_size, int t_size, modelFunction f): p(p_size), t(t_size), y(t_size),model(f){}; 
	
	void makeSourceNodes(DAG &G, OED_common &common, int group, int ind);
	void makeDAG(DAG &G);
	
	void initialize(Filter &F,
			const VECTOR<interval> &r_t,
			const VECTOR<interval> &r_p,
			const interval &r_A); 

};   


void OED_individual::makeSourceNodes(DAG &G, OED_common &common, int group, int ind){

	string name = "_g"+ToString(group)+"i"+ToString(ind);

//p	

	for(int i=1; i<=p.Size();i++)
	{
		if( common.p(i) ) 
			p(i)=common.p(i);
		else 	
		
			p(i)=MakeVar(&G, "p"+ToString(i) +name);
	}

//A

	A = common.A(group); 

 
//t	
	
	NODE *node;
	for(int i=1; i<=t.Size(); i++)
	{
		 
		if( common.findNode(node, group, i) ) t(i)=node;  
		//Search for (group time ) pair in common and if found take its common node
		//else make a new time variable 
	
		else
		t(i)=MakeVar(&G, "t"+ToString(i) + name);
	}

}



void OED_individual::makeDAG(DAG &G){

	for(int i=1; i<=t.Size(); i++){
		
		y(i) = model( t(i), p, A); 
	
	}

}


void OED_individual::initialize(Filter &F,
		const VECTOR<interval> &r_t,
		const VECTOR<interval> &r_p,
		const interval &r_A){
		
		SetConstants(*F.G, F.V);
		F.Write(t,  r_t);
		F.Write(p, r_p);
		F.Write(A, r_A);
		//F.SweepForwardIO();
}   






class OED_group{
public:
VECTOR<OED_individual*> individuals;
};


//ostream vector
template<class T>
std::ostream & operator<<(std::ostream & stream, std::vector<T> &v){

typename std::vector<T>::iterator it;

for(it=v.begin(); it!=v.end(); it++)
      	stream << "\n"<<*it;
return stream;
}

  

//Print an individual
std::ostream & operator<<(std::ostream & stream, OED_individual &ind){

stream<<"\np:";
for(int i=1; i<=ind.p.Size(); i++)
       	stream << "\n"<<ind.p(i);

stream<<"\nt:";       	
for(int i=1; i<=ind.t.Size(); i++)
       	stream << "\n"<<ind.t(i);

stream<<"\nA:\n"<<ind.A;

stream<<"\ny:";
for(int i=1; i<=ind.y.Size(); i++)
       	stream << "\n"<<ind.y(i);

return stream;
}



#endif
