////////////////////////////////////////////////////////////////////////
// 1.	Reads parameters from an input file.
//		Detects absence or incorrect input.
//      Sets default values for absent parameters.
//		OED parameters are not checked for 
//		correctness.
// 2.	Stores results that are input to
//		the main algorithm.
////////////////////////////////////////////////////////////////////////

class INPUT{
 public:

  //OED
  int 							nModel; // which model to consider
  int 							OED_nParameter;
  VECTOR<int> 			OED_nTime;
  int 							OED_nGroups;
  VECTOR<int> 			OED_nInd;
  VECTOR<interval> 	OED_A;
  VECTOR<VECTOR<interval> > OED_t; //(group)(time)
  VECTOR<int>			OED_p_isCommon;
  VECTOR<interval>	OED_psharp;
  VECTOR<interval>	OED_psearch;
  VECTOR<VECTOR<VECTOR<interval> > >	OED_pstar; //(group)(individual)(pstar)
  void OED_make_pstar();
  
  int OED_numberOfCommonTimes;
  LIST<fourIntegers> OED_common_time; 
  
  
  //RNG
  string RNGmode;
  unsigned RNGseed; 


  //DataPerturbation
  bool DataPerturbation_Enable;

  string DataPerturbation_Distribution; 
  string DataPerturbation_Mode;
  double DataPerturbation_epsilon;
  double DataPerturbation_mean;
  double	DataPerturbation_std;

  //ParameterPerturbation
  bool ParameterPerturbation_Enable;

  VECTOR<string> ParameterPerturbation_Distribution; 
  VECTOR<string> ParameterPerturbation_Mode;
  VECTOR<double> ParameterPerturbation_epsilon;
  VECTOR<double> ParameterPerturbation_mean;
  VECTOR<double> ParameterPerturbation_std;

  //Filter
  double Filter_ContractionRatio;
  double Filter_MinWidth;
  double Filter_MaxIter;
  double Filter_MinIter;

  //Split
  string	Split_CUT_cutFunction;
  double	Split_CUT_safeGM_r;
  string	Split_COMP_componentSelector;

  //Expander
  string 	Expander_Mode;
  double Expander_Rate;
  double Expander_absTerm;
  double	Expander_RateAdjuster;
  double	Expander_RateMin;
  double	Expander_MinCount;
  double Expander_MaxWidth;

  //BAP

  int 		BAP_CC_enable;
  int 		BAP_IOVR_enable;
  int 		BAP_CON_enable;
  int 		BAP_SMALL_enable;
  int		BAP_PRUNE_enable;

  int 			BAP_RangeListMaxSize;
  int 			BAP_MaxIter;
  double 	BAP_SMALL_width;
  double	BAP_SMALL_abswidth;
  string		BAP_SMALL_width_measure;

  int			BAP_scale_update_period;

  int			BAP_CC_minCoverSize;
  double 	BAP_CC_contractionRatio;
  double 	BAP_CC_sizeRatio;
  double		BAP_IOVR;
  int			BAP_IOVR_iter_check_period;
  double		BAP_CON_softDegree;
  bool			BAP_CON_inconsistencyTest;
  double		BAP_reExpansionRate;
  string 		BAP_pruner;


  bool get(const string &filename);
  void ParseAndCorrect(ParameterReader &pr, const string &p);
  void Incorrect(const string &p, const string &q);

  void DataPerturbation_Make(VECTOR<interval> &y);
  void DataPerturbation_Make(interval &y);
  void ParameterPerturbation_Make(VECTOR<interval> &p);
  void ParameterPerturbation_Make(VECTOR<double> &p);
  void setRNG();
  void setPruner(BranchAndPrune &BAP);

  void set(BranchAndPrune &BAP);
  void info(ostream &o = cout);
};






void INPUT::ParseAndCorrect(ParameterReader &pr, const string &p){

  string q;
  q = pr.find(p);




  //ParameterPertubation




  if(p=="ParameterPerturbation_Enable")
    {

      if(q.empty()) ParameterPerturbation_Enable=0;
      else ParameterPerturbation_Enable=parseString<bool>(q);
      return;

    }		 



  for(int i=1; i<=OED_nParameter; i++){


    if(p=="ParameterPerturbation_Distribution"+ToString(i))
      {

	if(q.empty()) ParameterPerturbation_Distribution(i)="normal";
	else ParameterPerturbation_Distribution(i)=q;
	if( !( ParameterPerturbation_Distribution(i)=="normal" || 
	       ParameterPerturbation_Distribution(i)=="lognormal" ||
	       ParameterPerturbation_Distribution(i)=="uniform" 
	       ) ) Incorrect(p,q);
	return;
      }		 

    if(p=="ParameterPerturbation_Mode"+ToString(i))
      {
	if(q.empty()) ParameterPerturbation_Mode(i)="relative";
	else ParameterPerturbation_Mode(i)=q;
	if( !( ParameterPerturbation_Mode(i)=="absolute" || 
	       ParameterPerturbation_Mode(i)=="relative") ) Incorrect(p,q);
	return;
      }

    if(p=="ParameterPerturbation_epsilon"+ToString(i))
      {
	if(q.empty()) ParameterPerturbation_epsilon(i)= 0.1;
	else ParameterPerturbation_epsilon(i)=parseString<double>(q);
	if( !( ParameterPerturbation_epsilon(i)>=0.0) ) Incorrect(p,q);
	return;
      }

    if(p=="ParameterPerturbation_mean"+ToString(i))
      {
	if(q.empty()) ParameterPerturbation_mean(i)= 0.0;
	else ParameterPerturbation_mean(i)=parseString<double>(q);
	return;
      }

    if(p=="ParameterPerturbation_std"+ToString(i))
      {
	if(q.empty()) ParameterPerturbation_std(i)= 0.0;
	else ParameterPerturbation_std(i)=parseString<double>(q);

	if(!(ParameterPerturbation_std(i)>=0.0)) Incorrect(p,q);	
	return;
      }
	
  }	




  //RNG




  if(p=="RNGmode")
    {

      if (q.empty()){
	RNGmode ="variable";
      }
      else RNGmode = q; 
	
      if( !(q=="variable" || q=="fixed") )  Incorrect(p,q);

    }


  else if(p=="RNGseed")
    {
      if(RNGmode=="fixed"){
	
	if (q.empty()){
	  RNGseed =1111111;
	}
	else RNGseed = parseString<int>(q); 
      }
    }




  //DataPerturbation




  else if(p=="DataPerturbation_Enable")
    {

      if(q.empty()) DataPerturbation_Enable=0;
      else DataPerturbation_Enable=parseString<bool>(q);
      return;

    }		 




  else if(p=="DataPerturbation_Distribution")
    {
      if(q.empty()) DataPerturbation_Distribution="normal";
      else DataPerturbation_Distribution=q;
	
      if( !(DataPerturbation_Distribution=="uniform" || DataPerturbation_Distribution=="normal") ) Incorrect(p,q);

    } 
  else if(p=="DataPerturbation_Mode")
    {
      if(q.empty()) DataPerturbation_Mode="relative";
      else DataPerturbation_Mode=q;
	
      if ( !(DataPerturbation_Mode=="absolute" || DataPerturbation_Mode=="relative") ) Incorrect(p,q);
    } 
  else if(p=="DataPerturbation_epsilon")
    {
	
      if(q.empty() ) DataPerturbation_epsilon=0.1;
      else DataPerturbation_epsilon	=	parseString<double>(q);
	
      if( !(DataPerturbation_epsilon>=0.0) ) Incorrect(p,q);
    } 

  else if(p=="DataPerturbation_mean")
    {
	
      if(q.empty() ) DataPerturbation_mean=0.0;
      else DataPerturbation_mean =	parseString<double>(q);
		
    } 
  else if(p=="DataPerturbation_std")
    {
	
      if(q.empty() ) DataPerturbation_std=0.1;
      else DataPerturbation_std	=	parseString<double>(q);
	
      if( !(DataPerturbation_std>=0.0) ) Incorrect(p,q);
	
    } 




  //Filter




  else if(p=="Filter_ContractionRatio")
    {
	
      if(q.empty() ) Filter_ContractionRatio=0.9999;
      else Filter_ContractionRatio	=	parseString<double>(q);
	
      if( !( Filter_ContractionRatio>=0.0  && Filter_ContractionRatio<=1.0) ) Incorrect(p,q);
	
	
    }
  else if(p=="Filter_MinWidth")
    {

      if(q.empty() ) Filter_MinWidth=0.0;
      else Filter_MinWidth	=	parseString<double>(q);
	
      if( !( Filter_MinWidth>=0.0 ) )Incorrect(p,q);
	
	
	
    }
  else if(p=="Filter_MaxIter")
    {
	
	
      if(q.empty() ) Filter_MaxIter=9999;
      else Filter_MaxIter	=	parseString<int>(q);
	
      if( !( Filter_MaxIter >=0 ) ) Incorrect(p,q);
	
    }

  else if(p=="Filter_MinIter")
    {
	
	
      if(q.empty() ) Filter_MinIter=1;
      else Filter_MinIter	=	parseString<int>(q);
	
      if( !( Filter_MinIter>=0 ) ) Incorrect(p,q);
	
    }



  //Split



  else if(p=="Split_CUT_cutFunction")
    {
      if(q.empty() ) Split_CUT_cutFunction="safeGM";
      else Split_CUT_cutFunction	=	q;
	
      if( !( Split_CUT_cutFunction=="safeAM" || Split_CUT_cutFunction=="safeGM" ) ) Incorrect(p,q);


    }
  else if(p=="Split_CUT_safeGM_r")
    {
	
      if(q.empty() ) Split_CUT_safeGM_r=0.3;
      else Split_CUT_safeGM_r=parseString<double>(q);
	
      if( !( Split_CUT_safeGM_r>=0.0 &&  Split_CUT_safeGM_r<=1.0) ) Incorrect(p,q);

    }
  else if(p=="Split_COMP_componentSelector")
    {
	
      if(q.empty() ) Split_COMP_componentSelector="max_scaled_width";
      else Split_COMP_componentSelector = q;
	
      if( !( Split_COMP_componentSelector=="max_scaled_width" ||
	     Split_COMP_componentSelector=="max_width"||
	     Split_COMP_componentSelector=="max_rel_width"||
	     Split_COMP_componentSelector=="bestBranch") ) Incorrect(p,q);
    }



  //Expander



  else if(p=="Expander_Mode")
    {
      if(q.empty() ) Expander_Mode	="relative";
      else Expander_Mode	=	q;
	
      if( !( Expander_Mode=="absolute" || Expander_Mode=="relative" ) ) Incorrect(p,q);


    }
  else if(p=="Expander_Rate")
    {
	
      if(q.empty() ) Expander_Rate	= 0.1;
      else Expander_Rate	=	parseString<double>(q);
	
      if( !( Expander_Rate>=0 ) ) Incorrect(p,q);
    }

  else if(p=="Expander_absTerm")
    {
      if(q.empty() ) Expander_absTerm	=1e-10;
      else Expander_absTerm	=	parseString<double>(q);
	
      if( !( Expander_absTerm>=0 ) ) Incorrect(p,q);
    }
  else if(p=="Expander_RateAdjuster")
    {

      if(q.empty() ) Expander_RateAdjuster	=0.1;
      else Expander_RateAdjuster	=	parseString<double>(q);
	
      if( !( Expander_RateAdjuster>=0.0 ) ) Incorrect(p,q);
    }
  else if(p=="Expander_RateMin")
    {

      if(q.empty() ) Expander_RateMin	=1e-5;
      else Expander_RateMin	=	parseString<double>(q);
	
      if( !( Expander_RateMin>=0.0 ) ) Incorrect(p,q);
	
    }	

  else if(p=="Expander_MinCount")
    {

      if(q.empty() ) Expander_MinCount	=10;
      else Expander_MinCount	=	parseString<int>(q);
	
      if( !( Expander_MinCount>=0 ) ) Incorrect(p,q);
	
    }	
  else if(p=="Expander_MaxWidth")
    {

      if(q.empty() ) Expander_MaxWidth	= FP::PosInf;
      else Expander_MaxWidth	=	parseString<double>(q);
	
      if( !( Expander_MaxWidth>=0 ) ) Incorrect(p,q);
	
    }



  //BAP


	
  else if(p=="BAP_CC_enable")
    {
	
      if(q.empty() ) BAP_CC_enable		= 1;
      else BAP_CC_enable	=	parseString<bool>(q);
	
      if( !( BAP_CC_enable==0 || BAP_CC_enable==1) ) Incorrect(p,q);
	
    }	
  else if(p=="BAP_IOVR_enable")
    {
      if(q.empty() ) BAP_IOVR_enable		= 1;
      else BAP_IOVR_enable	=	parseString<bool>(q);
	
      if( !( BAP_IOVR_enable==0 || BAP_IOVR_enable==1) ) Incorrect(p,q);
	
    }	
  else if(p=="BAP_CON_enable")
    {

      if(q.empty() ) BAP_CON_enable		= 1;
      else BAP_CON_enable	=	parseString<bool>(q);
	
      if( !( BAP_CON_enable==0 || BAP_CON_enable==1) ) Incorrect(p,q);
	
    }	
  else if(p=="BAP_SMALL_enable")
    {

      if(q.empty() ) BAP_SMALL_enable		= 1;
      else BAP_SMALL_enable	=	parseString<bool>(q);
	
      if( !( BAP_SMALL_enable==0 || BAP_SMALL_enable==1) ) Incorrect(p,q);
	
    }	
  else if(p=="BAP_PRUNE_enable")
    {
	
      if(q.empty() ) BAP_PRUNE_enable		= 1;
      else BAP_PRUNE_enable	=	parseString<bool>(q);
	
      if( !( BAP_PRUNE_enable==0 || BAP_PRUNE_enable==1) ) Incorrect(p,q);
	
    }	

  else if(p=="BAP_RangeListMaxSize")
    {

      if(q.empty() ) BAP_RangeListMaxSize		= 1000;
      else BAP_RangeListMaxSize	=	parseString<int>(q);
	
      if( !( BAP_RangeListMaxSize>=0 ) ) Incorrect(p,q);
	
    }	
  else if(p=="BAP_MaxIter")
    {

      if(q.empty() ) BAP_MaxIter		= 1000;
      else BAP_MaxIter	=	parseString<int>(q);
	
      if( !( BAP_MaxIter>=0 ) ) Incorrect(p,q);
	
    }	
  else if(p=="BAP_SMALL_width")
    {
      if(q.empty() ) BAP_SMALL_width		= 0.0;
      else BAP_SMALL_width	=	parseString<double>(q);
	
      if( !( BAP_SMALL_width>=0 ) ) Incorrect(p,q);
	
    }	
  else if(p=="BAP_SMALL_abswidth")
    {
      if(q.empty() ) BAP_SMALL_abswidth		= 0.0;
      else BAP_SMALL_abswidth	=	parseString<double>(q);
	
      if( !( BAP_SMALL_abswidth>=0 ) ) Incorrect(p,q);
	
    }	
  else if(p=="BAP_SMALL_width_measure")
    {

      if(q.empty() ) BAP_SMALL_width_measure		= "scaled";
      else BAP_SMALL_width_measure	=	q;
	
      if( !( BAP_SMALL_width_measure == "scaled" ||
	     BAP_SMALL_width_measure == "absolute"||
	     BAP_SMALL_width_measure == "relative") ) Incorrect(p,q);
	
    }	
    
  else if(p=="BAP_scale_update_period")
    {
      if(q.empty() ) BAP_scale_update_period		= 10;
      else 	BAP_scale_update_period =	parseString<double>(q);
	
      if( !( BAP_scale_update_period>=1 ) ) Incorrect(p,q);
	
    }	
  else if(p=="BAP_CC_minCoverSize")
    {

      if(q.empty() ) BAP_CC_minCoverSize		= 100;
      else BAP_CC_minCoverSize	=	parseString<int>(q);
	
      if( !( BAP_CC_minCoverSize>=0 ) ) Incorrect(p,q);
	
    }	
  else if(p=="BAP_CC_contractionRatio")
    {

      if(q.empty() ) BAP_CC_contractionRatio		= 0.7;
      else BAP_CC_contractionRatio	=	parseString<double>(q);
	
      if( !( BAP_CC_contractionRatio>=0.0 && BAP_CC_contractionRatio<=1.0) ) Incorrect(p,q);
	
    }	
  else if(p=="BAP_CC_sizeRatio")
    {

      if(q.empty() ) BAP_CC_sizeRatio		= 2;
      else BAP_CC_sizeRatio	=	parseString<int>(q);
	
      if( !( BAP_CC_sizeRatio>=0 ) ) Incorrect(p,q);
	
    }	
  else if(p=="BAP_IOVR")
    {

      if(q.empty() ) BAP_IOVR	= 0.7;
      else BAP_IOVR	=	parseString<double>(q);
	
      if( !( BAP_IOVR>=0.0 && BAP_IOVR<=1.0) ) Incorrect(p,q);
	
    }	
  else if(p=="BAP_IOVR_iter_check_period")
    {

      if(q.empty() ) BAP_IOVR_iter_check_period		= 100;
      else BAP_IOVR_iter_check_period	=	parseString<int>(q);
	
      if( !( BAP_IOVR_iter_check_period>=0 ) ) Incorrect(p,q);
	
    }	
  else if(p=="BAP_CON_softDegree")
    {
	
      if(q.empty() ) BAP_CON_softDegree	= 0.1;
      else BAP_CON_softDegree	=	parseString<double>(q);
	
      if( !( BAP_CON_softDegree>=0.0) ) Incorrect(p,q);
	
    }	
  else if(p=="BAP_CON_inconsistencyTest")
    {

      if(q.empty() ) BAP_CON_inconsistencyTest	= 0;
      else BAP_CON_inconsistencyTest	=	parseString<bool>(q);
	
      if( !( BAP_CON_inconsistencyTest==0 || BAP_CON_inconsistencyTest==1) ) Incorrect(p,q);
	
    }	
  else if(p=="BAP_reExpansionRate")
    {

      if(q.empty() ) BAP_reExpansionRate	= 0.1;
      else BAP_reExpansionRate	=	parseString<double>(q);
	
      if( !( BAP_reExpansionRate>=0.0) ) Incorrect(p,q);
	
    }	

  else if(p=="BAP_pruner")
    {
      if(q.empty() ) BAP_pruner	= "LoopSweepForwardBackward";
      else BAP_pruner	=	q;
      if( !( BAP_pruner =="LoopSweepForwardBackward") ) Incorrect(p,q);
	
    }	

  else {cout<<"\nCannot parse"<<p; exit(1);}


};

void INPUT::Incorrect(const string &p, const string &q){
  cout<<"\n INCORRECT INPUT  "<<p<<"  "<<q;
  exit(1);
} 




//get

	   
bool INPUT::get(const string &filename){
	
  ////////////////////////////////////////////////////////////////////////  
  // 1. Read Input
  // 2. Set default values if input unavailable.
  // 3. Signal error if input is out of bound.
  // 4. Compute perturbations.  
  ////////////////////////////////////////////////////////////////////////
		
  ParameterReader pr;
  pr.readFile(filename);


  // OED 
  // No Default, No Correction

  nModel				= parseString<int>(pr.find("nModel"));
  OED_nParameter 	= parseString<int>(pr.find("OED_nParameter"));
  OED_nGroups		= parseString<int>(pr.find("OED_nGroups"));
  
 
  OED_nTime = VECTOR<int>(OED_nGroups);
  for(int i=1; i<=OED_nGroups; i++){
    OED_nTime(i) = parseString<int>(pr.find("OED_group"+ToString(i)+"_nTime"));
  }

  
  OED_nInd = VECTOR<int>(OED_nGroups);
  for(int i=1; i<=OED_nGroups; i++){
    OED_nInd(i)	= parseString<int>(pr.find("OED_group"+ToString(i)+"_nInd"));
  }

  OED_A=VECTOR<interval>(OED_nGroups);
  for(int i=1; i<=OED_nGroups; i++){
    OED_A(i)	= parseString<interval>(pr.find("OED_group"+ToString(i)+"_A"));
  }
 
 
 
 
 OED_t = VECTOR<VECTOR<interval> >(OED_nGroups);
 
for(int g=1; g<=OED_nGroups; g++)
{
	OED_t(g) = VECTOR<interval>(OED_nTime(g));
	for(int c=1; c<=OED_t(g).Size(); c++)
	{
	OED_t(g)(c)=parseString<interval>(pr.find("OED_group"+ToString(g)+"_time"+ToString(c)));
	}
}
  
 
  OED_p_isCommon = VECTOR<int>(OED_nParameter);		
  for(int i=1;i<=OED_nParameter;i++){
    OED_p_isCommon(i)= parseString<int>( pr.find("OED_p"+ToString(i)+"_isCommon") );
  }

  OED_psharp = VECTOR<interval>(OED_nParameter);		
  for(int i=1;i<=OED_nParameter;i++){
    OED_psharp(i)= parseString<interval>( pr.find("OED_p"+ToString(i)+"sharp") );
  }
	
  OED_psearch = VECTOR<interval>(OED_nParameter);
  for(int i=1;i<=OED_nParameter;i++){
    OED_psearch(i)= parseString<interval>(pr.find("OED_p"+ToString(i)+"search") ); 
  }


OED_numberOfCommonTimes = parseString<int>(pr.find("OED_numberOfCommonTimes"));
for(int i=1; i<=OED_numberOfCommonTimes; i++){
OED_common_time.push(parseString<fourIntegers>(pr.find("OED_common_time"+ToString(i))));
}
	  
 
  //	RNG

  ParseAndCorrect(pr, "RNGmode");
  ParseAndCorrect(pr, "RNGseed");
  setRNG();

  // DataPertubation

  ParseAndCorrect(pr,	"DataPerturbation_Enable");

  ParseAndCorrect(pr,	"DataPerturbation_Distribution");
  ParseAndCorrect(pr,	"DataPerturbation_Mode");
  ParseAndCorrect(pr,	"DataPerturbation_epsilon");
  ParseAndCorrect(pr,	"DataPerturbation_mean");
  ParseAndCorrect(pr,	"DataPerturbation_std");


  //	ParameterPerturbation


  ParameterPerturbation_Distribution 		= VECTOR<string>(OED_nParameter); 
  ParameterPerturbation_Mode				= VECTOR<string>(OED_nParameter);
  ParameterPerturbation_epsilon			= VECTOR<double>(OED_nParameter);
  ParameterPerturbation_mean				= VECTOR<double>(OED_nParameter);
  ParameterPerturbation_std					= VECTOR<double>(OED_nParameter);

  ParseAndCorrect(pr,	"ParameterPerturbation_Enable");

  for(int i=1; i<=OED_nParameter; i++){
    ParseAndCorrect(pr,	"ParameterPerturbation_Distribution"+ToString(i));
    ParseAndCorrect(pr,	"ParameterPerturbation_Mode"+ToString(i));
    ParseAndCorrect(pr,	"ParameterPerturbation_epsilon"+ToString(i));
    ParseAndCorrect(pr,	"ParameterPerturbation_mean"+ToString(i));
    ParseAndCorrect(pr,	"ParameterPerturbation_std"+ToString(i));
  }


  ///////////////////////////////
  OED_make_pstar();
  ///////////////////////////////


  //Filter

  ParseAndCorrect(pr,	"Filter_ContractionRatio");
  ParseAndCorrect(pr,	"Filter_MinWidth");
  ParseAndCorrect(pr,	"Filter_MaxIter");
  ParseAndCorrect(pr,	"Filter_MinIter");

  //Split

  ParseAndCorrect(pr,	"Split_CUT_cutFunction");
  ParseAndCorrect(pr,	"Split_CUT_safeGM_r");
  ParseAndCorrect(pr,	"Split_COMP_componentSelector");


  //Expander


  ParseAndCorrect(pr,	"Expander_Mode");
  ParseAndCorrect(pr,	"Expander_Rate");
  ParseAndCorrect(pr,	"Expander_absTerm");
  ParseAndCorrect(pr,	"Expander_RateAdjuster");
  ParseAndCorrect(pr,	"Expander_MaxWidth");
  ParseAndCorrect(pr,	"Expander_MinCount");
  ParseAndCorrect(pr,	"Expander_RateMin");


  //BAP

  ParseAndCorrect(pr,	"BAP_CC_enable");
  ParseAndCorrect(pr,	"BAP_IOVR_enable");
  ParseAndCorrect(pr,	"BAP_CON_enable");
  ParseAndCorrect(pr,	"BAP_SMALL_enable");
  ParseAndCorrect(pr,	"BAP_PRUNE_enable");
  ParseAndCorrect(pr,	"BAP_RangeListMaxSize");
  ParseAndCorrect(pr,	"BAP_MaxIter");
  ParseAndCorrect(pr,	"BAP_SMALL_width");
  ParseAndCorrect(pr,	"BAP_SMALL_abswidth");
  ParseAndCorrect(pr,	"BAP_SMALL_width_measure");
  ParseAndCorrect(pr,	"BAP_scale_update_period");
  ParseAndCorrect(pr,	"BAP_CC_minCoverSize");
  ParseAndCorrect(pr,	"BAP_CC_contractionRatio");
  ParseAndCorrect(pr,	"BAP_CC_sizeRatio");
  ParseAndCorrect(pr,	"BAP_IOVR");
  ParseAndCorrect(pr,	"BAP_IOVR_iter_check_period");
  ParseAndCorrect(pr,	"BAP_CON_softDegree");
  ParseAndCorrect(pr,	"BAP_CON_inconsistencyTest");
  ParseAndCorrect(pr,	"BAP_reExpansionRate");
  ParseAndCorrect(pr,	"BAP_pruner");

  return 1;
}



void INPUT::OED_make_pstar(){
	
  // p_star = perturbed p_sharp 


//new
	if( !ParameterPerturbation_Enable ) return;


	OED_pstar = VECTOR<VECTOR<VECTOR<interval> > >(OED_nGroups);   
	for(int g=1; g<=OED_nGroups; g++)
	{
		OED_pstar(g)=VECTOR<VECTOR<interval> >(OED_nInd(g)); 
		for(int i=1; i<=OED_nInd(g); i++)
		{
			OED_pstar(g)(i) =OED_psharp;
			ParameterPerturbation_Make(OED_pstar(g)(i));
			
		}
	}
	
}



bool pruner_SweepForwardIO(Filter &F){
  ////////////////////////////////////////////////////////////////////////
  // Does not test for INCONSISTENCY of the image 
  // with respect to the data, it always returns 0;
  ////////////////////////////////////////////////////////////////////////

  F.SweepForwardIO();
  return 0;
}

bool pruner_default(Filter &F){

  return F.LoopSweepForwardBackward();
}

void INPUT::setPruner(BranchAndPrune &BAP){

  if( BAP_pruner.empty() ||
      BAP_pruner == "Default" || 
      BAP_pruner == "LoopSweepForwardBackward")
    {
		 
      BAP.pruner = g_pruner_default;
      //BAP.pruner = pruner_default
    }

  else if(BAP_pruner == "SweepForwardIO"){
    BAP.pruner = pruner_SweepForwardIO;
  }
	
  else
    {
	
      cout<<"\n\nBAP_pruner not defined";
      cout<<"\nIs there a space?\n\n"; 
      exit(1);
    }
	

}




void INPUT::set(BranchAndPrune &BAP){
  ////////////////////////////////////////////////////////////////////////
  // set BAP input parameter values.
  ////////////////////////////////////////////////////////////////////////


  BAP.CC_enable 			= BAP_CC_enable;
  BAP.IOVR_enable 		= BAP_IOVR_enable;
  BAP.CON_enable 		= BAP_CON_enable;
  BAP.SMALL_enable 	= BAP_SMALL_enable;
  BAP.PRUNE_enable 	= BAP_PRUNE_enable;

  BAP.RangeListMaxSize 		= BAP_RangeListMaxSize;
  BAP.MaxIter						= BAP_MaxIter; 

  BAP.F.ContractionRatio		= Filter_ContractionRatio;
  BAP.F.MinWidth					= Filter_MinWidth;
  BAP.F.MaxIter						= Filter_MaxIter;
  BAP.F.MinIter						= Filter_MinIter;


  BAP.SMALL_width				= BAP_SMALL_width;
  BAP.SMALL_abswidth		= BAP_SMALL_abswidth;
  BAP.SMALL_width_measure	= BAP_SMALL_width_measure; 
  
  BAP.scale_update_period	=BAP_scale_update_period;
  
  BAP.CC_contractionRatio		=	BAP_CC_contractionRatio;
  BAP.CC_sizeRatio				=	BAP_CC_sizeRatio;
  BAP.CC_minCoverSize			=	BAP_CC_minCoverSize;	

  BAP.IOVR							=	BAP_IOVR;
  BAP.IOVR_iter_check_period	=	BAP_IOVR_iter_check_period;

  BAP.CON_softDegree					=	BAP_CON_softDegree;

  BAP.CON_inconsistencyTest = BAP_CON_inconsistencyTest;

  BAP.S.CUT_cutFunction				=	Split_CUT_cutFunction;
  BAP.S.CUT_safeGM_r					=	Split_CUT_safeGM_r;
  BAP.S.COMP_componentSelector = Split_COMP_componentSelector;

  BAP.E.Mode							= Expander_Mode;
  BAP.E.rate								= Expander_Rate;
  BAP.E.absTerm						= Expander_absTerm;
  BAP.E.rateAdjuster					= Expander_RateAdjuster; 	
  BAP.E.MaxWidth						= Expander_MaxWidth; 
  BAP.E.MinCount						= Expander_MinCount; 

  BAP.reExpansionRate = BAP_reExpansionRate;


  setPruner(BAP);

}



//RNG


void INPUT::setRNG(){
	
  if(RNGmode=="fixed"){
    srand(RNGseed);
  }
  else if(RNGmode=="variable"){
    RNGseed = time(0);
    srand(RNGseed); 
  }
}




//perturb

void INPUT::DataPerturbation_Make(interval &y){
  ////////////////////////////////////////////////////////////////////////
  //	Perturb y, using the
  //	the DataPerturbation information specified in input.
  ////////////////////////////////////////////////////////////////////////


  if(!DataPerturbation_Enable) return;


  //
  // uniform
  //

if(DataPerturbation_Distribution == "uniform")
{
    if(DataPerturbation_Mode=="absolute")
    { 
	y += DataPerturbation_epsilon*uniformRnd(-1.0,1.0);
	}
	else if(DataPerturbation_Mode=="relative")
	{
	y *=(1.0 + DataPerturbation_epsilon*uniformRnd(-1.0,1.0) );
	}
    else
    {
    cout<<"\nUnknown DataPerturbationMode\n"; 
    exit(1);
    }
}

  //
  // normal
  //
	
else if(DataPerturbation_Distribution == "normal")
{
	if(DataPerturbation_Mode=="absolute")
	{
	  //cout<<endl<<"yold="<<y;
	y += DataPerturbation_mean + gaussianRnd(DataPerturbation_std);
	//cout<<endl<<"y="<<y;
	}
	else if(DataPerturbation_Mode=="relative")
	{
	  //cout<<endl<<"yold="<<y;
	y *=(1.0 + gaussianRnd(DataPerturbation_std));
	//cout<<endl<<"y="<<y;
	}
    else
    {
	cout<<"\nUnknown DataPerturbationMode\n"; 
    exit(1);
    }				
}		

}


void INPUT::DataPerturbation_Make(VECTOR<interval> &y){
  ////////////////////////////////////////////////////////////////////////
  //	Perturb each component of y, using the
  //	the DataPerturbation information specified in input.
  ////////////////////////////////////////////////////////////////////////


  if(!DataPerturbation_Enable) return;


  //
  // uniform
  //

  if(DataPerturbation_Distribution == "uniform"){
	
    if(DataPerturbation_Mode=="absolute"){
      for(int i=1; i<=y.Size(); i++){ 
	y(i) += DataPerturbation_epsilon*uniformRnd(-1.0,1.0);
      }
    }
    else if(DataPerturbation_Mode=="relative"){
      for(int i=1; i<=y.Size(); i++){ 
	y(i) *=(1.0 + DataPerturbation_epsilon*uniformRnd(-1.0,1.0) );
      }
    }
    else{
      cout<<"\nUnknown DataPerturbationMode\n"; 
      exit(1);
    }
  }

  //
  // normal
  //
	
  else if(DataPerturbation_Distribution == "normal"){
	
    if(DataPerturbation_Mode=="absolute"){
      for(int i=1; i<=y.Size(); i++){ 
	//cout<<endl<<"yold="<<y(i);
	y(i) += DataPerturbation_mean + gaussianRnd(DataPerturbation_std);
	//cout<<endl<<"ynew="<<y(i);
      }
    }
    else if(DataPerturbation_Mode=="relative"){
      for(int i=1; i<=y.Size(); i++){ 
	y(i) *=(1.0 + gaussianRnd(DataPerturbation_std));
      }
    }
    else{
      cout<<"\nUnknown DataPerturbationMode\n"; 
      exit(1);
    }
					
  }		

}



void INPUT::ParameterPerturbation_Make(VECTOR<interval> &p){
  ////////////////////////////////////////////////////////////////////////
  //	Perturb each component of p, using the
  //	the ParameterPerturbation information specified in input.
  ////////////////////////////////////////////////////////////////////////

  //
  // uniform
  //
  
  
  for(int k=1; k<=OED_nParameter; k++){
	
    // Only perturb individual parameters
    if (OED_p_isCommon(k)) continue;

    if(ParameterPerturbation_Distribution(k) == "uniform"){
	
      if(ParameterPerturbation_Mode(k)=="absolute"){
	p(k) += ParameterPerturbation_epsilon(k)*uniformRnd(-1.0,1.0);
      }
      else if(ParameterPerturbation_Mode(k)=="relative"){ 
	p(k) *=(1.0 + ParameterPerturbation_epsilon(k)*uniformRnd(-1.0,1.0) );
      }
      else{
	cout<<"\nUnknown ParameterPerturbationMode\n"; 
	exit(1);
      }
    }

    // normal
    // is assumed ADDITIVE
    // elseif-branch, editerad av Peter 090515
    else if(ParameterPerturbation_Distribution(k) == "normal"){
	  
      if(ParameterPerturbation_Mode(k)=="absolute"){
	p(k) += ParameterPerturbation_mean(k) + 
	  gaussianRnd(ParameterPerturbation_std(k));
      }
      else if(ParameterPerturbation_Mode(k)=="relative"){
 	//p(k) *=(1.0 + gaussianRnd(ParameterPerturbation_std(k)));
	p(k) += ParameterPerturbation_mean(k) + 
	  gaussianRnd(mid(p(k))*ParameterPerturbation_std(k));
      }
      else{
	cout<<"\nUnknown ParameterPerturbationMode\n";
 	exit(1);
      }
	  
    }
	
    //
    // lognormal
    // is assumed MULTIPLICATIVE
    // elseif-branch, editerad av Peter 090515
    else if(ParameterPerturbation_Distribution(k) == "lognormal"){
	  
      if(ParameterPerturbation_Mode(k)=="absolute"){
	//p(k) += LogNormalSample(ParameterPerturbation_mean(k), 
	//ParameterPerturbation_std(k));
	p(k) *=LogNormalSample(ParameterPerturbation_mean(k), 
			       ParameterPerturbation_std(k));
      }
      else if(ParameterPerturbation_Mode(k)=="relative"){
	// p(k) *=(1.0 +   LogNormalSample(ParameterPerturbation_mean(k), 
	//ParameterPerturbation_std(k)) );
	p(k) *=LogNormalSample(ParameterPerturbation_mean(k), 
			       mid(p(k))*ParameterPerturbation_std(k));
      }
      else{
	cout<<"\nUnknown ParameterPerturbationMode\n";
	exit(1);
      }
	  
    }
	
	
	
	
  }
}


void INPUT::info(ostream &o){

  o<<"\nOED parameters:\n";

  o<<"\nOED_nParameter "	<< OED_nParameter;
  o<<"\nOED_nGroups "			<< OED_nGroups;
  
  for(int i=1;i<=OED_nGroups; i++){
	o<<"\nOED_group"+ToString(i)+"_nTime "<<OED_nTime(i);
  }

  for(int i=1;i<=OED_nGroups; i++){
	o<<"\nOED_group"+ToString(i)+"_nInd "<<OED_nInd(i);
  }
  for(int i=1;i<=OED_nGroups; i++){
	o<<"\nOED_group"+ToString(i)+"_A "<<OED_A(i);
  }


for(int g=1; g<=OED_t.Size(); g++)
{
	for(int i=1; i<=OED_t(g).Size(); i++)
	{
		
		o<<"\nOED_t_g"+ToString(g)+"t"+ToString(i)+" "<<OED_t(g)(i);
	
	} 
}

  for(int i=1;i<=OED_psearch.Size(); i++){
    o<<"\nOED_psearch"+ToString(i)+" "<<OED_psearch(i);
  }	

  for(int i=1;i<=OED_p_isCommon.Size(); i++){
	o<<"\nOED_p_isCommon"+ToString(i)+" "<<OED_p_isCommon(i);
  }

  for(int i=1;i<=OED_psharp.Size(); i++){
    o<<"\nOED_psharp"+ToString(i)+" "<<OED_psharp(i);
  }	

   
for(int g=1; g<=OED_pstar.Size(); g++)
{
	for(int i=1; i<=OED_pstar(g).Size(); i++)
	{
		o<<"\npstar_g"+ToString(g)+"i"+ToString(i)+"\n"<<OED_pstar(g)(i);	
		//cout<<"\npstar_g"+ToString(g)+"i"+ToString(i)+"\n"<<OED_pstar(g)(i);	
		}
}
	
	
 o<<"\nOED_numberOfCommonTimes "<<OED_numberOfCommonTimes;
 for(int i=1; i<=OED_common_time.Size(); i++){
 	//o<<"\nOED_common_time"+ToString(i)+" "<<OED_common_time(i);
}

OED_common_time.MoveTop();
while(!OED_common_time.eol()){ 
o<<"\nOED_common_time "<<OED_common_time.Data();
OED_common_time.MoveDown();
}



  o<<"\n\nStochastic parameters:\n";




  //RNG

  o<<"\nRNGmode "	<< RNGmode;
  o<<"\nRNGseed	 "	<< RNGseed; 	

  o<<"\n";
  //DataPerturbation
  o<<"\nDataPerturbation_Enable " << DataPerturbation_Enable;

  o<<"\nDataPerturbation_Distribution "		<< DataPerturbation_Distribution;
  o<<"\nDataPerturbation_Mode "				<< DataPerturbation_Mode;
  o<<"\nDataPerturbation_epsilon "			<< DataPerturbation_epsilon;
  o<<"\nDataPerturbation_mean "				<< DataPerturbation_mean;
  o<<"\nDataPerturbation_std "					<< DataPerturbation_std;

  o<<"\n";


  //ParameterPerturbation

  o<<"\nParameterPerturbation_Enable " << ParameterPerturbation_Enable;

  for(int i=1; i<=OED_nParameter; i++){

    // Only perturb individual parameters
    if (OED_p_isCommon(i)==1) continue;

    o<<"\n\nPARAMETER"+ToString(i)+"\n";
    o<<"\nParameterPerturbation_Distribution"+ToString(i)+" "	<< ParameterPerturbation_Distribution(i);
    o<<"\nParameterPerturbation_Mode"+ToString(i)+" "				<< ParameterPerturbation_Mode(i);
    o<<"\nParameterPerturbation_epsilon"+ToString(i)+" "			<< ParameterPerturbation_epsilon(i);
    o<<"\nParameterPerturbation_mean"+ToString(i)+" "			<< ParameterPerturbation_mean(i);
    o<<"\nParameterPerturbation_std"+ToString(i)+" "				<< ParameterPerturbation_std(i);

  }	

  o<<"\n\nBranch And Prune parameters:\n\n";


  //Filter
  o<<"\nFilter_ContractionRatio "		<< Filter_ContractionRatio;
  o<<"\nFilter_MinWidth "					<< Filter_MinWidth;
  o<<"\nFilter_MaxIter "						<< Filter_MaxIter;
  o<<"\nFilter_MinIter "						<< Filter_MinIter;

  //Split
  o<<"\nSplit_CUT_cutFunction "					<< Split_CUT_cutFunction;
  o<<"\nSplit_CUT_safeGM_r "					<< Split_CUT_safeGM_r;
  o<<"\nSplit_COMP_componentSelector "	<< Split_COMP_componentSelector;

  //Expansion 
  o<<"\nExpander_Mode "					<< Expander_Mode;
  o<<"\nExpander_Rate "					<< Expander_Rate;
  o<<"\nExpander_absTerm "			<< Expander_absTerm;
  o<<"\nExpander_RateAdjuster "		<< Expander_RateAdjuster;
  o<<"\nExpander_MaxWidth "			<< Expander_MaxWidth;
  o<<"\nExpander_MinCount "			<< Expander_MinCount;
  o<<"\nExpander_RateMin "				<< Expander_RateMin;

  //BAP
  o<<"\nBAP_CC_enable "						<<BAP_CC_enable;
  o<<"\nBAP_IOVR_enable "					<<BAP_IOVR_enable;
  o<<"\nBAP_CON_enable "					<<BAP_CON_enable;
  o<<"\nBAP_SMALL_enable "				<<BAP_SMALL_enable;
  o<<"\nBAP_PRUNE_enable "				<<BAP_PRUNE_enable;
  o<<"\nBAP_RangeListMaxSize "			<< BAP_RangeListMaxSize;
  o<<"\nBAP_MaxIter "							<< BAP_MaxIter;
  o<<"\nBAP_SMALL_width "					<< BAP_SMALL_width;
 o<<"\nBAP_SMALL_abswidth "					<< BAP_SMALL_abswidth;
 o<<"\nBAP_scale_update_period "				<< BAP_scale_update_period;

  o<<"\nBAP_SMALL_width_measure "	<< BAP_SMALL_width_measure;
  o<<"\nBAP_CC_minCoverSize "			<< BAP_CC_minCoverSize;
  o<<"\nBAP_CC_contractionRatio "		<< BAP_CC_contractionRatio;
  o<<"\nBAP_CC_sizeRatio "					<< BAP_CC_sizeRatio;
  o<<"\nBAP_IOVR "								<< BAP_IOVR;
  o<<"\nBAP_IOVR_iter_check_period "	<< BAP_IOVR_iter_check_period;
  o<<"\nBAP_CON_softDegree "				<< BAP_CON_softDegree;
  o<<"\nBAP_CON_inconsistencyTest "	<< BAP_CON_inconsistencyTest;
  o<<"\nBAP_reExpansionRate "				<< BAP_reExpansionRate;
  o<<"\nBAP_pruner "							<< BAP_pruner;


}		







