#include <stdio.h>
#include <iostream>
#include <fstream>  
#include <string>
#include "Structs/symUtilities.h"
#include "DAG/DAGUtilities.h"
#include "DAG/DAGFilters.h"
#include "Structs/Tools.h"
#include "../../cpp/Stochastic.cpp"
#include "../../cpp/time.h"
#include "Structs/ParameterReader.h"
#include "Structs/symUtilities.h"
#include "Structs/intervalUtilities.h"
#include "Structs/point.h"
#include "DAG/GraphViz.h"


#include "INPUT.h"
#include "OUTPUT.h"    
#include "GroupStructure.h" 
       	
       
using namespace std;  
                   	  
// Here we choose the model //////////////////////////////
  
#include "Input/Model.h"
string InputDIR  	= "Input/";
string OutputDIR	= "Output/";    
  
////////////////////////////////////////////////////////////////////////

   
typedef LIST<VECTOR<interval> > 	LVI;	// List of Vector of Interval
typedef VECTOR<NODE*> 				VNP;	// Vector of Node Pointer
typedef VECTOR<interval> 				VI;	// Vector of Interval
             
             
             
 void p_boxes(const string &filename, LIST<VECTOR<interval> > &L, VECTOR<OED_group*> &groups)
 {
	const string dir = "./Output/";
	
	LIST<VECTOR<interval> > OUTList;
		
	for(int g=1; g<=groups.Size(); g++)
	{
		for(int i=1; i<=groups(g)->individuals.Size(); i++)
		{
		
			OED_individual *ind = groups(g)->individuals(i);
			VECTOR<interval> w( ind->p.Size() );
			
			OUTList.Clean();
			L.MoveTop();
			while(!L.eol())
			{	
				for(int k=1; k<=w.Size(); k++)
				{
					w(k) = L.Data()( ind->p(k)->num );
				}
				
				OUTList.push(w);
				L.MoveDown();
			}
		
		WriteBoxes( OUTList, dir + filename + "_g" + ToString(g) + "i"+ToString(i)+".txt");
		}
	}
}




void GenerateDataAndInitialize(Filter &F, VECTOR<OED_group*> &groups, 
INPUT &input){
    
// DG, data generation
// mid points for t-variables
// mid points for A-variables
// input.pstar for p-variables

for(int g=1; g<=groups.Size(); g++)
{  
    //t
    VECTOR<interval> DG_t( input.OED_t(g) );
    for(int k=1; k<=DG_t.Size(); k++) DG_t(k)=imid(DG_t(k));
    //A
    interval DG_A(input.OED_A(g));
    DG_A=imid(DG_A);
    //p
    for(int i=1; i<=groups(g)->individuals.Size(); i++)
    {   
        if(input.ParameterPerturbation_Enable){
        groups(g)->individuals(i)->initialize( F, DG_t, 
input.OED_pstar(g)(i), DG_A );
        }
        else    groups(g)->individuals(i)->initialize( F, DG_t, 
input.OED_psharp, DG_A );
    }  
}


F.SweepForwardIO();
VECTOR<interval> Fcopy = F.V;

writedot(F,"filter.dot");

//initialize F with search domains (and data from Fcopy)

for(int k=1; k<=F.V.Size(); k++) F.V(k)=interval::Entire;

for(int g=1; g<=groups.Size(); g++)
{
    for(int i=1; i<=groups(g)->individuals.Size(); i++)
    {
        OED_individual *ind = groups(g)->individuals(i);
        ind->initialize( F, input.OED_t(g), input.OED_psearch, 
input.OED_A(g) );
                
                
                if(input.DataPerturbation_Enable)
                {
                    for(int k=1; k<=ind->y.Size(); k++)
                    {
                    input.DataPerturbation_Make(Fcopy( ind->y(k)->num ));
                    F.Write(ind->y(k), Fcopy( ind->y(k)->num ) );
                    }
                }
                else   
                {
                for(int k=1; k<=ind->y.Size(); k++)  F.Write(ind->y(k), 
Fcopy( ind->y(k)->num ) );
                }
    }
}


}





void makeDAG(DAG &G, VECTOR<OED_group*> &groups, INPUT &input){

OED_common common; 
common.make_p( input.OED_p_isCommon, G );
common.make_A( input.OED_nGroups, G);
common.make_interGroup_t(input.OED_common_time,G);
common.make_intraGroup_t(input.OED_nTime,G); //does it depend on order? inter before intra?? 
   
       
common.LLN->MoveTop();
 while(!common.LLN->eol())
   { 
     common.LLN->Data()->List->MoveTop();
     //cout<<"\nList and associated common node:\n";
     //cout<<"\nnode = "<<common.LLN->Data()->commonNode;
     //cout<<"\n";
     while(!common.LLN->Data()->List->eol()){
       //cout<<common.LLN->Data()->List->Data()<<"\n\n";
       common.LLN->Data()->List->MoveDown();
     }
     common.LLN->MoveDown();
   }
 


       
for(int i=1; i<=groups.Size(); i++)
{
	groups(i)=new OED_group;
	groups(i)->individuals = VECTOR<OED_individual*>(input.OED_nInd(i));
	for(int j=1; j<=groups(i)->individuals.Size(); j++)
	{
	  if (input.nModel==1) groups(i)->individuals(j) = new OED_individual(input.OED_nParameter, input.OED_nTime(i), Model1);
	  if (input.nModel==2) groups(i)->individuals(j) = new OED_individual(input.OED_nParameter, input.OED_nTime(i), Model2_Complemented);
	  if (input.nModel==22) groups(i)->individuals(j) = new OED_individual(input.OED_nParameter, input.OED_nTime(i), Model2);
		groups(i)->individuals(j)->makeSourceNodes(G, common,i,j);
		groups(i)->individuals(j)->makeDAG(G);
	}	  
}  

}






 
int main(){ 

string OutputName = OutputDIR + "Output.txt";
ofstream Output( OutputName.c_str() );
Output.precision(17);  
if (!Output){ 
	cout<<"\nERROR, Cannot access\n";
	cout<<OutputName<<"\n";
}  
else{   
	Output<<"#Date_Time: "<<Date_Time()<<"\n\n\n";
} 
     
////////////////////////////////////////////////////////////////////////  
// get Input
////////////////////////////////////////////////////////////////////////
  
INPUT input;	
string theInput = InputDIR; 
theInput +="Input.txt";
input.get(theInput);
//input.info();

dotInput dotParameters;
       
DAG G;  
DAG::sEnableRewrite=0;

VECTOR<OED_group*> groups(input.OED_nGroups);
makeDAG(G, groups, input);
//writedot(G,"DAG.dot");

Filter F(G);
GenerateDataAndInitialize(F,groups, input);
//writedot(F,"filter0.dot");
    
   
// Save the initial box.
	  
VI s_BOX_init(F.V);
  
  

	
input.info(Output);   

		
////////////////////////////////////////////////////////////////////////	
// Splitter: common p nodes taken from g1i1
////////////////////////////////////////////////////////////////////////
  
LIST<NODE*> SplitList;
for(int i=1;i<=input.OED_p_isCommon.Size();i++)
{
	
	if(input.OED_p_isCommon(i)) SplitList.push( groups(1)->individuals(1)->p(i) );  

}

 if (input.nModel==2 || input.nModel==22)
for(int g=1; g<=groups.Size(); g++)
{
    for(int i=1; i<=groups(g)->individuals.Size(); i++)
    {
        SplitList.push( groups(g)->individuals(i)->p(4) );
        SplitList.push( groups(g)->individuals(i)->p(5) );
        SplitList.push( groups(g)->individuals(i)->p(6) );
        
    }
}

Split S(G,SplitList);
 
	
////////////////////////////////////////////////////////////////////////	
// Expander: data nodes
////////////////////////////////////////////////////////////////////////
 
Expander E;

LIST<NODE*> expansionNodes;
for(int g=1; g<=groups.Size(); g++){
	for(int i=1; i<=groups(g)->individuals.Size(); i++){
		for(int j=1; j<=groups(g)->individuals(i)->y.Size(); j++ ){
		
			expansionNodes.push(  groups(g)->individuals(i)->y(j) );
		
		}
	}
}


E.setNodeNumber(expansionNodes);


////////////////////////////////////////////////////////////////////////	
// Branch And Prune
////////////////////////////////////////////////////////////////////////
   
BranchAndPrune BAP(F,S,E);
input.set(BAP);


////////////////////////////////////////////////////////////////////////
// SMALL	
////////////////////////////////////////////////////////////////////////

LIST<NODE*> SmallNodes;

for(int g=1; g<=groups.Size(); g++){
	for(int i=1; i<=groups(g)->individuals.Size(); i++){
		for(int j=1; j<=groups(g)->individuals(i)->p.Size(); j++ ){
		
			if( !input.OED_p_isCommon(j) ) SmallNodes.push(  groups(g)->individuals(i)->p(j) );
		
		}
	}
}

//common nodes taken from g1i1
for(int i=1;i<=input.OED_p_isCommon.Size();i++)
{
	
	if(input.OED_p_isCommon(i)) SmallNodes.push( groups(1)->individuals(1)->p(i) );  

}

BAP.SMALL_set_nodenumber(SmallNodes);
	  




BAP.RangeList.push(F.V);  
 
////////////////////////////////////////////////////////////////////////
// BAP1_Expand
////////////////////////////////////////////////////////////////////////    
cout.precision(20);  


if (BAP.BAP1_Expand()) {
	cout<<"\nINCONCISTENCY by BAP.BAP1_Expand()"; 
	return 0;  
}  


//OUTPUT	
	

LIST<VECTOR<interval> > boxList;
BAP.PushAllLists(boxList);
p_boxes( "p_boxes",boxList,groups); 

   


return 0;
}
   
 
