#ifndef __DAGFILTERS_H__
#define __DAGFILTERS_H__

#include "Structs/VECTOR.h"
#include "Structs/interval.h"
#include "Structs/symUtilities.h"
#include "DAG/NODE.h"
#include "DAG/DAG.h"

#include <stdlib.h>
#include <list>

////////////////////////////////////////////////////////////////////////
// Measure
////////////////////////////////////////////////////////////////////////

class Measure{
public:

static double Volume(LIST<VECTOR<interval> > &L, 
										VECTOR<double> &scale, 
										VECTOR<int> &select);

static double AbsRelWidth(double threshold, const interval &x);
static double NewWidth(const interval &x, double p);
	
static double RelWidth(const interval &x);
static double MaxWidth(const VECTOR<interval> &x, 
									const VECTOR<int> &select);
static double MaxRelWidth(const VECTOR<interval> &x, 
											const VECTOR<int> &select);
static double SumOfWidths(const VECTOR<interval> &x);
static double SumOfWidths(const VECTOR<interval> &x, 
											const VECTOR<int> &index);
static double Volume(const VECTOR<interval> &x);
static double Volume(const VECTOR<interval> &x, 
									const VECTOR<int> &index);
static double Volume(LIST<VECTOR<interval> > &L);
};



////////////////////////////////////////////////////////////////////////
// Split
////////////////////////////////////////////////////////////////////////

class Split{
public:

Split();
Split(DAG &G, VECTOR<NODE> &v);
Split(DAG &G, vNODE &v);
Split(DAG &G, LIST<NODE*> &L);

VECTOR<int> Count;
VECTOR<int> Select;

VECTOR<double> scale;

void SetSelect(VECTOR<NODE> &v);
void SetSelect(vNODE &v);
void SetSelect(LIST<NODE*> &L);
void Info(ostream &out = cout);

bool safeAM(double &cut, const interval &X);
bool safeGM(double &cut, 
							const double M, 
							const double r, 
							const interval &X);	

//CUT
string	CUT_cutFunction; //safeAM,safeGM
double	CUT_safeGM_M;
double CUT_safeGM_r;
bool 	CUT_getCut(double &cut, const interval &X);

//COMP
string 	COMP_componentSelector;
int 		COMP_getComponent(const VECTOR<interval> &X);

int COMP_max_width				(const VECTOR<interval> &X);
int COMP_max_rel_width			(const VECTOR<interval> &X);
int COMP_max_scaled_width	(const VECTOR<interval> &X);
int COMP_bestBranch				(const VECTOR<interval> &X);

void MultiSplit(list<VECTOR<interval> > &L, const VECTOR<int> &comp);

bool SplitAndPush(VECTOR<interval> &x, LIST< VECTOR<interval> > &L);	
};	

////////////////////////////////////////////////////////////////////////
// Expander
////////////////////////////////////////////////////////////////////////

class Expander{
public:
	
Expander();
std::string Mode; //"absolute","relative"
double rate;
double absTerm;
double rateAdjuster;


VECTOR<int> NodeNumber;

void setNodeNumber(LIST<NODE*> &L); // 24June2009

int Count;//Count counts the number of expansions(=number of rejections)	
int MinCount;//Minimum number of rejections before acceptance.
int MaxCount;// to get absolute stability, is currently not used!
double rateMin;
	
VECTOR<interval> X0;
VECTOR<interval> X1;

bool isPossibleX1();
double MaxWidth;
VECTOR<interval> Xinit;

void Expand(VECTOR<interval> &X);
void relMidExpand(VECTOR<interval> &X);
void absExpand(VECTOR<interval> &X);
void rateAdjustment(const string &direction);

void SymmetricExpand(VECTOR<interval> &V, const VECTOR<double> &data);
void SymmetricExpand(VECTOR<interval> &C, const VECTOR<interval> &D);

void Info(ostream &out = cout);
};


class Filter;
typedef bool (*Contractor)(Filter &F);

////////////////////////////////////////////////////////////////////////
// Filter
////////////////////////////////////////////////////////////////////////

class Filter{
public:


DAG *G;
VECTOR<interval> V;
VECTOR<NODE*> FORWARD;
VECTOR<NODE*> BACKWARD;
int MinIter;
int MaxIter;
int Iter;

double ContractionRatio;
double MinWidth;

//shave
double shave_ratio;
VECTOR<int> shave_components;
LIST<int> *shaveList;
Filter();
Filter(DAG &aG); 

void setParameters();

//Read
void	Read(NODE &n, interval &x);
interval Read(NODE &n);
void Read(VECTOR<NODE> &n, VECTOR<interval> &x);
VECTOR<interval> Read(VECTOR<NODE> &n);

//Write
void Write(NODE &n, const interval &x);
void Write(VECTOR<NODE> &n, const VECTOR<interval> &x);
void Write(vNODE &n, const VECTOR<interval> &x);
void Write(NODE* n, const interval &x); //20090610
void Write(VECTOR<NODE*> &n, const VECTOR<interval> &x);//20090610



bool SweepForward();
void SweepForwardIO();
void SweepForwardIO(VECTOR<interval> &X);
bool SweepBackward();
bool SweepForwardBackward(int nCycles=1);
bool SweepForwardBackward(VECTOR<interval> &X);
bool LoopSweepForwardBackward();
bool LoopSweepForwardBackward(VECTOR<interval> &X);

bool Shave();
bool ShaveLower_Unbounded(int i, double ratio);
bool ShaveUpper_Unbounded(int i, double ratio);
bool ShaveLower(int i, double ratio);
bool ShaveUpper(int i, double ratio);
bool ImprovedShaver(VECTOR<interval> &X, int i, double ratio);

bool isContraction(const interval &OLD, const interval &NEW);
bool isContraction(VECTOR<interval> &OLD);
bool isContraction(VECTOR<interval> &NEW, VECTOR<interval> &OLD);
bool isRelContraction(VECTOR<interval> &OLD);

bool CPH2(int i, double cut=FP::NegInf);

void Info(ostream &out = cout);	
};


////////////////////////////////////////////////////////////////////////
// BranchAndPrune
////////////////////////////////////////////////////////////////////////

typedef bool (*Pruner)(Filter &F);

class BranchAndPrune{
public:
	int MinIter;
	int MaxIter;
	int Iter;
	int RangeListMaxSize;
	int RangeListMinSize;
	
	
	//enable/disable stopping and box-removal criteria
	
	bool CC_enable;
	bool IOVR_enable;
	bool SMALL_enable;
	bool CON_enable;
	bool PRUNE_enable;
	
	//CC - Cover Contraction
	int 									CC_minCoverSize;
	double 							CC_sizeRatio;
	double								CC_contractionRatio;
	LIST<VECTOR<interval> > 	CC_prevCover;
	bool									CC_firstTime;
	bool CC_isLarge();
	bool CC_isLoosingCoverVolume();
	
	//IOVR - Inner Outer Volume Ratio
	double IOVR;
	bool IOVR_isLarge();
		//
		//IOVR checking
		//
	double IOVR_prev;
	int		IOVR_iter_check;
	int 		IOVR_iter_check_period;
	void 	IOVR_checking(double IOVR_prev);

	//CON - parameter-data consistency
	
	VECTOR<int> 		CON_components;		
	VECTOR<interval>	CON_ranges;				
	void CON_set_ranges(const VECTOR<interval> &X);
	bool CON_consistent();
	bool CON_inconsistencyTest;
	double CON_softExclusionDegree(interval &X, interval &Y);
	double CON_softDegree;
	
	BranchAndPrune();
	BranchAndPrune(Filter &aF, Split &aS);
	BranchAndPrune(Filter &aF, Split &aS, Expander &aE);
	Filter F;
	Split S;
	Expander E;

	
	LIST<VECTOR<interval> > RangeList;  
	LIST<VECTOR<interval> > ConsistencyList; 
   	LIST<VECTOR<interval> > IndefiniteList;
   	int CoverSize(){return RangeList.Size() + ConsistencyList.Size() +IndefiniteList.Size();} 
	void CleanAllLists();
  	void PushAllLists(LIST<VECTOR<interval> > &L);
	void HullAllLists(VECTOR<interval> &V);  	
	
	
	bool Contract(LIST<VECTOR<interval> > &L);
	bool Contract(list<VECTOR<interval> > &L);

	bool CPH(VECTOR<interval> &X, 
			const VECTOR<int> &components);
	bool CPH();

	//BAP1
	void BAP1_printInfo();
	void BAP1_initialize();
	bool BAP1();
	
	//BAP1_Expand
	int BAP1_Expand_nRejections;
	void BAP1_Expand_initialize();
	bool BAP1_Expand();
	
	//BAP1_Expand_Symmetric
	bool BAP1_Expand_Symmetric();	
	
	void get_point_image(VECTOR<interval> &point,
										VECTOR<interval> &image,
										vNODE &p1, NODE &p2, NODE &p3);

	bool BAP1_Expand_Symmetric(vNODE &p1, NODE &p2, NODE &p3);
	
	//for BAP1_Expand_Symmetric
	double reExpansionRate; 
	
	//SMALL
	bool SMALL();
	double SMALL_width;
	string SMALL_width_measure;
	VECTOR<int> SMALL_nodenumber; //24June2009
	void SMALL_set_nodenumber(LIST<NODE*> &L);//24June2009
	double SMALL_abswidth;//25June2009
	int SMALL_abswidth_count;//25June2009
	
		
	bool isInconsistent();
	
	Pruner pruner;
	bool Prune();
	void getNextBox();
	void splitBox();
	void setScale();
	
	VECTOR<double> scale; //24June2009
	int scale_update_period;

	bool Expand();
	bool ExpansionRate_downAdjustment();
	bool ExpansionRate_upAdjustment();
	
	void Info(ostream &out = cout);

};


////////////////////////////////////////////////////////////////////////
//default pruner
////////////////////////////////////////////////////////////////////////
extern bool g_pruner_default(Filter &F);
		
	
	
////////////////////////////////////////////////////////////////////////
// Globals:
// Are only used to visualize the 
// intermediate computations in 
//	BAP1_Expand_Symmetric.
////////////////////////////////////////////////////////////////////////

extern VECTOR<interval> 				gFirstHit;
extern VECTOR<interval>				gCFirstHit;
extern LIST<VECTOR<interval> > 	glist_CFirstHit;
extern VECTOR<interval>				gSymExp;
extern VECTOR<interval>				gpoint;
extern VECTOR<interval>				gimage;

		   
#endif


