
#ifndef __DAGUtilities_H__
#define __DAGUtilities_H__

#include "DAG/NODE.h"
#include "DAG/DAG.h"

#include <typeinfo> 
#include "Structs/interval.h"

#include "DAG/DAGFilters.h"

void Initialize(Filter &F,
		vNODE &t,		const VECTOR<interval> &r_t,
		vNODE &p1,	const VECTOR<interval> &r_p1,
		NODE &p2,	const interval& r_p2,
		NODE &p3,	const interval& r_p3);


class Derivative{
public:
	DAG *G;
	NODE *zero;
	NODE *one;
	VECTOR<NODE*> PATH;
	VECTOR<NODE> cD;
	VECTOR<interval> nV;
	VECTOR<interval> nD;
	

	LIST<NODE*> Used;
	void SetUsed(NODE *v);
	void SetUsed(VECTOR<NODE*> &N);
	void RemoveUnused();
	
	Derivative();
	Derivative(DAG &aDAG);

	void cInit(NODE* n);
	void nInit(NODE* n);
	void cBackward(NODE *y);
	void cBackward(NODE *y, NODE *x);
	void cForward(NODE *x);
	void cForward(NODE *y, NODE *x);
	
	void nBackward(NODE *y);
	void nBackward(NODE *y, NODE *x);
	void nForward(NODE *x);
	void nForward(NODE *y, NODE *x);
	
	
	void cBackwardMixed(VECTOR<NODE*> &df,  NODE *f, const VECTOR<NODE*> &x);
	void cForwardMixed(VECTOR<NODE*> &df,  NODE *f, const VECTOR<NODE*> &x);
	void cForwardDiff(VECTOR<NODE*> &z, NODE *y, NODE *x);
	void cBackwardDiff(VECTOR<NODE*> &z, NODE *y, NODE *x);
	
	void cProgressiveExpansion(	NODE &y,
					VECTOR<NODE> &t,
					VECTOR<NODE> &c,
					VECTOR<NODE> &x,
					NODE&(*f)(VECTOR<NODE>&)     ); 
	
	
};


















//NEW 
//   - Changes in DAG.cpp



//-------
bool isIn(NODE *n, VECTOR<NODE*> X); //Change to the template version
//void setZero(VECTOR<NODE*> &X);      //Change to the template version

//bool isThin(const interval &x);
//Moved To interval.cpp




bool   isSource      (NODE *n);
bool   isSink        (NODE *n);
bool   isIntermediate(NODE *n);

bool   isConstant    (NODE *n);
bool   isVariable    (NODE *n);  


bool   isArithmetic  (NODE *n);
bool   isCommutative (NODE *n);

bool isConstantAndZero(NODE *n);
bool isConstantAndOne(NODE *n); 
bool isConstantAndMinusOne(NODE *n);
bool containsZero(NODE *n);
bool isThinConstant(NODE *n);

bool isUSub(NODE *n);
bool isAdd(NODE *n);
bool isSub(NODE *n);
bool isMul(NODE *n);
bool isDiv(NODE *n);
bool isSquare(NODE *n);
bool isExp(NODE *n);
bool isLog(NODE *n);
bool isPow(NODE *n);
bool isSum(NODE *n);
bool isProd(NODE *n);
bool isEquality(NODE *n);
bool isInclusion(NODE *n);
bool isIntersection(NODE *n);
bool isLessOrEqual(NODE *n);



bool isAddition(NODE *n);
bool isMultiplication(NODE *n);

bool isUnary(NODE *n);
bool isBinary(NODE *n);
bool isk_ary(NODE *n);

bool isUnaryOp(NODE *n);
bool isBinaryOp(NODE *n);
bool isk_aryOp(NODE *n);

bool isArithmetic (NODE *n);
bool isCommutative(NODE *n);

bool areSymbolicallyEquivalent(NODE *n1, NODE* n2); 

bool isMulAddition(NODE *n);
bool isMulSub(NODE *n);

bool isInverted(NODE*n, NODE* &x);
bool isNegated(NODE* n, NODE* &x);
bool isNegatedAddition(NODE *n, NODE* &x);
bool isNegatedSubtraction(NODE *n, NODE* &x);

NODE* Negation(NODE *n);

void NegateAndPush(LIST<NODE*> *A,  LIST<NODE*> *B);
void DeFactorization(LIST<NODE*> *S, NODE *v);

void GetSISNodes(DAG &G,
		LIST<NODE*> &SOURCE,
		LIST<NODE*> &INTERMEDIATE,
		LIST<NODE*> &SINK);
void GetSourceNodes(LIST<NODE*> &SOURCE, DAG &G);
void GetSourceNodes(LIST<NODE*> &SOURCE, NODE *x); 
void GetSinkNodes(LIST<NODE*> &SINK, DAG &G);
void GetSymbolNodes(LIST<NODE*> &SYM, const string &NodeSymbol, DAG &G);
void NumberToPointer(VECTOR<NODE*> &N, DAG &G);
void GetNodeNumbers(VECTOR<int> &ind, VECTOR<NODE> &n);
void GetNodeNumbers(VECTOR<int> &ind, VECTOR<NODE*> &n);
void GetNodeNumbers(VECTOR<int> &ind, vNODE &n);

bool isSubGraphMatch( DAG &G, NODE *n, VECTOR<NODE*> &X, NODE *e);
bool CopyDAG(VECTOR<NODE*> &X, DAG &H);
bool FindExpression(NODE *e, DAG &G, LIST<NODE*> &EXPR);
bool FindExpression(NODE *e, DAG &G, LIST< VECTOR<NODE*> > &EXPR); 

int N_GPred(DAG &G, NODE *n); 	
void GPred(VECTOR<NODE*> &PRED, NODE *n);
void GPred(VECTOR<NODE*> &PRED, LIST<NODE*> &L);


void MakePredecessor(NODE*n, NODE*m);			//Not In Use
void MakePredecessors(NODE* n, LIST<NODE*> *L);		//Not In Use

bool RemoveEdge(NODE *m, NODE *n);
void InsertPredEdges(NODE* o);
void InsertSuccEdges(NODE* o);
void RemovePredEdges(NODE* o);
void RemoveSuccEdges(NODE* o);
void ReplaceSuccEdges(NODE *o, NODE *r);
void ReplacePredEdges(NODE *o, NODE *r);

void DAGNodeInsert(DAG *G, NODE *o);
void DAGNodeRemove(DAG *G, NODE *o);
void DAGNodeReplace(DAG *G, NODE *o, NODE *r);

void RemoveUnusedNodes(DAG &G, LIST<NODE*> &Used);

//FriMay22
NODE* SpecialFunction_OEDmodel2s(NODE* t, NODE *p1, NODE *p2);

void leq(NODE *a, NODE *b);
void geq(NODE *a, NODE *b);
bool hasLEQSucc(NODE* &leq, NODE *a, NODE *b);
void ConstructLessOrEqualNode(DAG *G, NODE *a, NODE *b);


void in(NODE*a, NODE *b);
void in(NODE *a, const interval &b);
bool hasInclusionSucc(NODE*& in, NODE *n);
void ConstructInclusionNode(DAG *G, NODE *x, NODE *y);
void InsertInclusionNode(DAG *G, NODE *x, NODE *y); //7FEB

void equal(NODE *m, NODE *n);
void equal(NODE &m, NODE &n);

void ConstructEqualityNode(DAG *G, NODE *m, NODE *n);
NODE_equal* InsertEqualityNode(DAG &G);  //Not In Use
bool hasEqSucc(NODE*& eq, NODE *n);
//bool hasEqSucc(NODE*& eq, NODE* n, NODE* m);
void MakeEqual(NODE *x, NODE *eq);
void MakeEqual(LIST<NODE*> &X, NODE *eq);
void SearchAndEquate(DAG &G, NODE *s, VECTOR<NODE> &arg, NODE& (*F)(VECTOR<NODE> &arg) );


void SetConstants(DAG &G, VECTOR<interval> &V);

void SuccLevel(NODE* n, VECTOR<int> &LEVEL);
void PredLevel(NODE *n, VECTOR<int> &LEVEL);
void ComputeSuccLevels(DAG &G, VECTOR<int> &LEVEL);

VECTOR<NODE*> Path(DAG &G, NODE *target );

VECTOR<NODE*> PathF(DAG &G, NODE *start); //ToDo

VECTOR<NODE*> Path(DAG &G, NODE *n1, NODE *n2 );
void ForwardByLevel(DAG &G, VECTOR<NODE*> &FORWARD, VECTOR<int> &LEVEL);
VECTOR<NODE*> ForwardOrderingByLevel(DAG &G); //NEW
VECTOR<NODE*> BackwardByReverseForward(const VECTOR<NODE*> &FORWARD);//NEW


void ForwardOrdering(DAG &G, VECTOR<NODE*> &FORWARD);
void BackwardOrdering(DAG &G, VECTOR<NODE*> &BACKWARD);

bool FP(const VECTOR<NODE*> &FORWARD, VECTOR<interval> &V);
void FPIO(const VECTOR<NODE*> &FORWARD, VECTOR<interval> &V);
bool BP(const VECTOR<NODE*> &BACKWARD, VECTOR<interval> &V);
void BPD(const VECTOR<NODE*> &BACKWARD, VECTOR<interval> &V, VECTOR<interval> &D);
void cBPD(const VECTOR<NODE*> &BACKWARD, VECTOR<NODE> &D);

VECTOR<int> nofSuccessors(DAG &G);
void NumberOfSuccessors(DAG &G, VECTOR<NODE*> &SORTED_NODEVECTOR, VECTOR<int> &NUM_OF_SUCC);
VECTOR<int> MaxSuccessor(DAG &G);
void MaxSuccessor(LIST<NODE*> &L, DAG &G);

void Print(VECTOR<NODE> &n, const VECTOR<interval> &V, ostream& out=cout);
void Print(vNODE &n, const VECTOR<interval> &V, ostream& out=cout);
void Print(const VECTOR<int> &n, const VECTOR<interval> &V, ostream& out=cout);


void PE(NODE *n);
void PE1(NODE *n, NODE *s, ostream& out=cout);
void PrintExpression(NODE *n, ostream& out=cout);
void PrintExpression(NODE &n, ostream& out=cout);
//void PrintExpression(vNODE &n);
//void PrintExpression(VECTOR<NODE> &n);
void PrintExpression(VECTOR<NODE> &n, ostream& out=cout);
void PrintExpression(const VECTOR<NODE*> &n, ostream& out=cout);
void PrintExpression(LIST<NODE*> &L, ostream& out=cout);
void PrintExpression(VECTOR<int> &NodeNumbers, DAG &G, ostream& out=cout);
void PrintSinks(DAG &G, ostream& out=cout);
void PrintCodeList(DAG &G, ostream& out=cout); 

void Operation(NODE *n, ostream& out=cout);
void DAGInformation(DAG &G, ostream& out=cout);
void DAGInformation(DAG &G, const VECTOR<interval> &RANGE, ostream& out=cout);


NODE* MakeVar(DAG *G, const string &name);
NODE& MakeSource(DAG &G, const string &name);
NODE& MakeConstant(DAG &G, interval c, const string &name);
void MakeConstant(NODE* &n, DAG &G, const interval &c, const string &name);
NODE& MakeConstant(DAG &G, interval c);

NODE& MakeNODE(DAG &G, const string &name);

void MakeSource(NODE &v, const string &name, DAG &G);
void MakeSource(VECTOR<NODE> &V, const string &name, DAG &G);
vNODE& MakeSource(DAG &G, const string &name, int dim);
 
    
#endif
