#ifndef __LIST_H__
#define __LIST_H__


#include<stdio.h>
using namespace std;
#include <stdlib.h>



template<class T>
class LIST{

private:
struct element{
       T data;
       element * prev;
};

 
element * top;
element * bot;
element * first_in;
element * curr;

int size;
 
 

public:
LIST();
LIST(const LIST & L);
LIST & operator=(const LIST & x);
~LIST();

int Size()const; 
void push(const T & data);
void PushTop(const T & data);      //071007
void PushBottom(const T & data);   //071007

T pop();
bool empty()const;
 
void init();         
bool FINT()const;    // Not Empty and Not on Top
bool eol();
bool IsLast(); // 110807

T & operator[](int i);
T & Data();
T & NextData();  // 110807
 
void MoveTop   ();
void MoveBottom(); //071007
void MoveDown  ();
void MoveUp    ();
void Insert    (const T & data, int i);
void Insert    (const T & d);
void Remove    (T & data, int i);
void Remove    ();

bool operator==(const LIST &x)const;   // 101007

void Clean();
void Sprint(); // Special print for debugging 
void Partition(LIST & L1, LIST & L2, int i); //130807
 
 

 
};


 
//-----------------------------------------------------------------------------




// CONSTRUCTORS----------------------------------------------------------------


template<class T>
LIST<T>::LIST(){
  top = 0;
  bot = 0;
  first_in = 0;
  curr = 0;
  size = 0;
}


template<class T>    
LIST<T>::LIST(const LIST<T> & L){

  /*
    COPY constructor  
  */


  if (L.top == 0){ top=0; bot=0; first_in=0; curr=0;size=0; } 
  else{
    
     element * ne = new element;                        // ne = new element; neprev = ne prev 
     top = ne;
     top->data = L.top->data;

     if (L.first_in == L.top){ first_in = top; }
     if (L.curr == L.top){ curr = top; }
     
     element * neprev;
     element * Lprev = L.top->prev;
     
     while ( Lprev != 0 ){
            neprev = new element;
	    ne->prev = neprev;
	    neprev->data = Lprev->data;
	    ne = neprev;
	    
            if (L.first_in == Lprev){ first_in = ne; }
            if (L.curr == Lprev){ curr = ne; }

	    Lprev = Lprev->prev;
     }
     bot = ne;
     ne->prev = 0;
     size = L.size;
  }
}



  
template<class T>
LIST<T>& LIST<T>::operator=(const LIST<T> & L){
  
  if (this == &L) return (*this);
  
  
  while( size>0 ){ pop(); }
  
  //Exact copy of the COPY constructor
  if (L.top == 0){ top=0; bot=0; first_in=0; curr=0; size=0; }
  else{
    
     element * ne = new element;                        // ne = new element; neprev = ne prev 
     top = ne;
     top->data = L.top->data;

     if (L.first_in == L.top){ first_in = top; }
     if (L.curr == L.top){ curr = top; }
     
     element * neprev;
     element * Lprev = L.top->prev;
     
     while ( Lprev != 0 ){
            neprev = new element;
	    ne->prev = neprev;
	    neprev->data = Lprev->data;
	    ne = neprev;
	    
            if (L.first_in == Lprev){ first_in = ne; }
            if (L.curr == Lprev){ curr = ne; }

	    Lprev = Lprev->prev;
     }
     bot = ne;
     ne->prev = 0;
     size = L.size;
  }

  
  return (*this);
}






template<class T>
LIST<T>::~LIST(){

 while (top!=0){
        element * tmp = top;
        top = top->prev;      
        delete tmp;
 }
}

//ACCESS ----------------------------------------------------------------------



template<class T>
T& LIST<T>::operator[](int i){               

  // Index Range  = 0,..,size-1
  // Costly. Not to be used when going down the list.

  if (size == 0){ cout<<"\n\n\nLIST IS EMPTY; NO ELEMENT IS ACCESSIBLE,  T& LIST<T>::operator[](int i)"; exit(1); }
  else if (i<0 || i>size-1){ cout<<"\n\n\nINDEX OUTSIDE LIST INDEX RANGE,  T& LIST<T>::operator[](int i)"; exit(1); }

  element * p = top;
  for(int j=1;j<=i;j++){ p = p->prev; }
  return p->data;
}

template<class T>
T& LIST<T>::Data(){return curr->data;}   // Returns a reference, so it can be used to read and write

template<class T>
T& LIST<T>::NextData(){return curr->prev->data;} 
 
template<class T>
void LIST<T>::MoveTop(){ curr = top; }                                            

template<class T>
void LIST<T>::MoveBottom(){ curr = bot; }  //071007


template<class T>
void LIST<T>::MoveDown(){

  // Moves curr down one step.
  // If curr is bottom, then it remains at bottom.
  
  if (curr != 0){ curr = curr->prev; }           
}
template<class T>
void LIST<T>::MoveUp(){

  // Moves curr up one step.
  // If curr is top, then it remains at top

  element * p = top;
  element * s = top;
  
  while( p != curr ){s = p; p = p->prev; }
  curr = s;
}
template<class T>
bool LIST<T>::eol(){ return (curr==0); }

template<class T>
bool LIST<T>::IsLast(){ return (curr->prev == 0); }


// STORING and REMOVING ----------------------------------------------------------

template<class T>
void LIST<T>::push(const T & d){    

  // push on bottom/top
  
element * p = new element; p->data = d;

// push on bottom 
if  (top==0){top=p; bot=p; p->prev = 0;}
else        {bot->prev=p; bot=p; p->prev=0; }


// push on top
//if   (top==0) {top=p; bot=p;p->prev=0;}
//else          {p->prev=top; top=p; }
 
 
if (first_in == 0){first_in = p;}
 
size++;
}

template<class T>
void LIST<T>::PushTop(const T & d){    

  // push on top
  
element * p = new element; p->data = d;



if   (top==0) {top=p; bot=p;p->prev=0;}
else          {p->prev=top; top=p; }
 
 
if (first_in == 0){first_in = p;}
 
size++;
}


template<class T>
void LIST<T>::PushBottom(const T & d){    

  // push on bottom
  
element * p = new element; p->data = d;


if  (top==0){top=p; bot=p; p->prev = 0;}
else        {bot->prev=p; bot=p; p->prev=0; }

if (first_in == 0){first_in = p;}
 
size++;
}




template<class T>
T LIST<T>::pop(){                        

// pop on top

if(top==0){ cout<<"\nError: Pop from empty list, T LIST<T>::pop()"; exit(1);}
  
T d = top->data;
element * t = top;     
top = top->prev;      
if(top==0){ bot=0; }
delete t;
 
size--;

return d; 
}

template<class T>
void LIST<T>::Insert(const T & d, int i){

  // Insertion at arbitrary position:
  // Insert at positions i=0,...,size
  // If insert at i=0, then the inserted element is the new top element.
  // If insert at i=size, then the inserted element is the new bottom element.
  // If insert at 0<i<size, the inserted element is squeezed in between
  // element i-1 and i, so the inserted element becomes element i in the new list.

  

  element * p = new element;
  p->data = d;
  
  if (i<0 || i>size){ cout << "\n\n\nATTEMPT TO INSERT OUTSIDE OF INDEX RANGE,   void LIST<T>::Insert(const T & d, int i)"; exit(1);}
  else if (size == 0){ top = p; bot = p; p->prev = 0; }
  else if ( i == 0 ){ p->prev = top; top = p;}
  else if ( i == size ){ bot->prev = p; bot = p; p->prev = 0;  }
  else { // size!=0 and 0<i<size
         element * r = top;
         for(int j=1;j<=i-1;j++){ r = r->prev; }
         p->prev = r->prev;
         r->prev = p;  
  }
  size++;
}

template<class T>
void LIST<T>::Insert(const T & d){

  // 8 MAJ
  // 11 Maj hade glmt argumentet d !
  
// Insert d at current position (actually, on top of current element). The inserted element becomes the new current.
// If curr == top, then the inserted element is the new top element.
// If curr == 0, then the inserted element is the new bottom element.
// If current element is not top or 0, then the element is squeezed between the current element and its above element
// current element.

  element * p = new element;
  p->data = d;

  if      (size == 0)   { top = p; bot = p; p->prev = 0; curr=top;}
  else if (curr == top) { p->prev = top; top = p; curr = p;}
  else if (curr == 0)   { bot->prev = p; bot = p; p->prev = 0; curr = p; }
  else { // size!=0 and 0<curr<top
         element * r = top;
         while( r->prev != curr ){ r = r->prev; }
         p->prev = r->prev;//=curr
         r->prev = p;
	 curr = p;
  }
  size++;
}


template<class T>
void LIST<T>::Remove(T & d, int i){

  // Removal at position i:
  // Remove element at position i=0,...,size-1 and return its data in d.
  // If list is empty, return ERROR message, we cannot remove anything.
  // If i = 0, the top element is removed.
  // If i = size-1, the bottom element is removed.


  
   
  if (size==0){ cout << "\n\n\nATTEMPT TO REMOVE FROM EMPTY LIST,     void LIST<T>::Remove(T & d, int i) "; exit(1); }
  else if (i<0 || i>size-1){ cout << "\n\n\nATTEMPT TO REMOVE OUTSIDE OF INDEX RANGE,   void LIST<T>::Remove(T & d, int i)"; exit(1);}
  else if (i == 0){ d = top->data; element * t = top; top = top->prev; delete t; if(top==0){bot=0;} }
  else { // 1<=i<=size-1
         element * r = top;
         for(int j=1;j<=i-1;j++){ r = r->prev; }
         element * s = r;
	 r = r->prev;
	 s->prev = r->prev;
	 d=r->data;                          
	 delete r;
	 if ( i==(size-1) ){ bot = s; }   
	 
  }
  size--;  
}

template<class T>
void LIST<T>::Remove(){

  // Remove current element.
  // New curr points the element after the removed element
  // Remove and Insert are inverses of eachother.
  
  
  if(curr==0){cout<<"\n\n\nATTEMPT TO REMOVE FROM EMPTY LIST,void LIST<T>::Remove()";exit(1); }
  if(curr==top){ element * t = top; top=top->prev; curr=top; if(top==0){bot=0;}; delete t;}
  else        {
               element * r = top;
               while((r->prev) != curr){r = r->prev;}
               //here we have r->prev =curr
	       if(curr==bot){ bot = r; }
               r->prev=(curr->prev);
               r=curr;
               curr=(curr->prev);
               delete r;
  }
  size--; 
  
}



// -------------------------------------------------------------------------------

template<class T>
//bool LIST<T>::empty()const{ return(top==0); }
bool LIST<T>::empty()const{ return(size==0); }


template<class T>
void LIST<T>::init(){ first_in = 0; }

template<class T>
bool LIST<T>::FINT()const{
  
  // First Is Not on Top :
  // returns 1 iff list is empty and first_in element is not on top
  
  return ( (first_in != top) && (top != 0) ) ;  
}
template<class T>
int LIST<T>::Size()const{ return size; }


template<class T>
bool LIST<T>::operator==(const LIST &x)const{
  if ( this == &x ){ return 1; }
  if (size != x.size){ return 0; }
  
  element * p = top;
  element *xp = x.top;
  while(p!=0){
    if ( !( p->data == xp->data ) ){ return 0;}
    p=p->prev;
    xp=xp->prev;
    
  }
  return 1;
  
}



template<class T>
void LIST<T>::Clean(){
              while( top!=0 ){ pop(); };
}



//-------------------------------------------------------------------------------

// Om jag har ostream utanfr klassen s fungerar det. Men jag
// fr unresolved external error om jag har den innanfr klassen. 
// Vet inte varfr. Har att gra med typparametriseringen, men vet inte vad.
// Fungerar inte med friend heller.


template<class T>
ostream & operator<<(ostream & stream, LIST<T> & L){

  if ( L.empty() ){ stream <<"EMPTY"; }
  else{
  	    
  	    
  	    L.MoveTop();
  		while( !L.eol() ){
  			stream<<"\n"<<L.Data();
			L.MoveDown();	 
		}
		
       //stream << L[0];
       //for(int i=1; i<L.Size(); i++){ stream << endl << L[i]; }
  }
  return stream;
}

//-------------------------------------------------------------------------------


template<class T>
void LIST<T>::Sprint(){

  // Special print for debugging.
  
  if (size == 0){ cout << "EMPTY"; }
  else {
        element * p = top;
	for(int i=0; i<(size); i++){
	    cout<< endl << "element " << i << ",  pointer "<< p;
	    p = p->prev;
	}
  }
}
						  




template<class T>
void LIST<T>::Partition(LIST & L1, LIST & L2, int i){


  
  //***********************************************************
  // Partition the list L into two sublists L1, L2, where the first i elements of L
  // is placed in L1 and the last n-i elements is placed in L2.
  //
  // L = e1-e2-...-en ->  L1 = e1-..-ei ,  L2 = e(i+1)-...-en
  //
  // Check that 1<=i<=n-1.
  // Pointers to the correct locations of the current list L are copied and 
  // then the pointers of L is set to NULL. So the current list L
  // is effectively deleted.
  //***********************************************************


  /*
  cout<<endl<<"L=  "<<endl<<*this;
  cout<<endl<<"the adresses =";
  curr = top;
  while (!eol()){
    cout << endl << curr;
    curr = curr->prev;
  }
  */ 
  
  
  if (size==0) return;
  if (size==1){ L1.push( pop() ); return; }
  
  if (i<1 ||i>=size){
    cout<<endl<<"index does not split the list: void LIST<T>::Partition(LIST & L1, LIST & L2, int i)";
    exit(1);
  }
  if (!L1.empty() || !L2.empty() ){
   cout<<endl<<"Sublists L1 and L2 must be empty when calling: void LIST<T>::Partition(LIST & L1, LIST & L2, int i)";
    exit(1);
}

    
  
  int count = 0;
  curr = top;
  while( ++count < i ){  curr = curr->prev; }
  
  
  
  L1.top = top;
  L1.bot = curr;
  L2.top = curr->prev;
  curr->prev = 0; // (L1.bot)->prev=0
  L1.curr = top; //tex
  L1.size = i;

  L2.bot = bot;
  L2.curr = L2.top; //tex
  L2.size = size-i;
  
  top  = 0;
  bot  = 0;
  first_in = 0;
  curr = 0;
  size = 0;

  
  
  
}



#endif

