#include "Structs/LIST.h"
#include "Structs/VECTOR.h"
#include <iomanip> //for MatrixForm
#ifndef __TOOLS_CPP__
#define __TOOLS_CPP__







template<typename T>
void MatrixForm(VECTOR<T> &X, const int nRow, const int nCol, ostream &o=cout);

template<typename T>
void setZero(VECTOR<T> &X);

template<typename T>
void set(VECTOR<T> &X, const T &x);

template<typename T>
void Push(LIST<T> *L1, LIST<T> *L0 );

template<typename T>
void Push(LIST<T> &L, const VECTOR<T> &V );

template<typename T>
void Push(LIST<T> &L, LIST<T> &M );

//Push If Absent
template<typename T>
void PushIA(LIST<T> &L, LIST<T> &M );

template <typename T>	
void Common(LIST<T> *I, LIST<T> *L1, LIST<T> *L2);

template<typename T>
bool areEqual(LIST<T> *L1, LIST<T> *L2);

template<typename T>
bool In(T &x, LIST<T> &L);

template<typename T1, typename T2>
bool In(VECTOR<T1> &v1, VECTOR<T2> &v2);

template<typename T>
bool Replace(T o, T r, LIST<T> &L);

template<typename T>
bool Remove(T e, LIST<T> &L);

template<typename T>
void RemoveMultiples(LIST<T> &U, LIST<T> &L);

template<typename T>
void Reverse(VECTOR<T> &V);
  
template<typename T>
void Copy_LIST_To_VECTOR(LIST<T> *L, VECTOR<T> &V);






template<typename T>
void MatrixForm(VECTOR<T> &X, const int nRow, const int nCol, ostream &o){

	if(nRow*nCol != X.Size()){ cout<<"Error in MatrixForm"; exit(1); }
	
	const int width = 10;
	for(int row = 1;row <= nRow; row++) {
		
		for(int col = 1;col<=nCol;col++) {
			o << setw(width) << X((row-1)*nCol + col );
		}
		o << endl;
	}

}










template<typename T>
void setZero(VECTOR<T> &X){

	for(int i=1; i<=X.Size(); i++){ X(i) = 0; } 
}

template<typename T>
void set(VECTOR<T> &X, const T &x){

	for(int i=1; i<=X.Size(); i++){ X(i) = x; } 
}

template<typename T>
void Push(LIST<T> *L1, LIST<T> *L0 ){

	// Push on L1 all elements of L0.

	L0 -> MoveTop();
	while( !L0->eol() )
	{ 
		L1 -> push( L0->Data() ); 
		L0 -> MoveDown();
	}
}

template<typename T>
void Push(LIST<T> &L, const VECTOR<T> &V ){

	//Push on L all elements of V
	for(int i=1; i<=V.Size(); i++){ L.push(V(i)); }
}

template<typename T>
void Push(LIST<T> &L, LIST<T> &M ){

	//Push on L all elements of M
	
	M.MoveTop();
	while(!M.eol()){
		L.push(M.Data());
		M.MoveDown();
	}
}





template<typename T>
void PushIA(LIST<T> &L, LIST<T> &M){

	//Push on L all elements of M
	
	M.MoveTop();
	while(!M.eol()){
		if(!In(M.Data(),L)){
			L.push(M.Data());
		}
		M.MoveDown();
	}
}












template <typename T>
void Common(LIST<T> *I, LIST<T> *L1, LIST<T> *L2){
     
     // Common items in L1 and L2.
     // I <- Common(L1,L2)
     // Copies distinguished
     // Ex: Common ( {a,a,a,b,c}, {b,a,e,a} ) = {a,a,b}
     
     // Algorithm: 
     // 1) while both L1 and L2 nonempty
     // 2) Remove an element from L1
     // 3) Search for this element in L2, if found remove it 
     //    from L2 and place it in I.
     
     

	LIST<T> C1 = *L1;
	LIST<T> C2 = *L2;
	T item;
	 
	bool continue_search;
	C1.MoveTop();
	while ( !C1.empty() && !C2.empty() ){
	        item = C1.pop();
	        C2.MoveTop();
	        continue_search = 1;
	        while ( !C2.eol() && continue_search ){ 
	        
	        	if ( item == C2.Data() ){
	        		 C2.Remove(); 
	        		 I->push(item); 
	        		 continue_search = 0;
	        	}
	        	C2.MoveDown();
		}

	}
}


 	
template<typename T>
bool areEqual(LIST<T> *L1, LIST<T> *L2){

	// L1 and L2 are equal iff they contain the same elements, 
	// copies included.	

	if (L1->Size() == L2->Size()){
		LIST<T> I;
		Common(&I,L1,L2);
		if (I.Size() == L1->Size()){ return 1; }
	}
return 0;
}


template<typename T>
bool In(T &x, LIST<T> &L){
	L.MoveTop();
	while( !L.eol() ){ 
		if ( L.Data() == x) return 1;
		L.MoveDown();
	}

return 0;
}

template<typename T1, typename T2>
bool In(VECTOR<T1> &v1, VECTOR<T2> &v2){
	
	for(int i=1; i<=v1.Size(); i++){
		if(!(v1(i)<<=v2(i)) ){ return 0;}
	}
	return 1;
}


template<typename T>
bool Replace(T o, T r, LIST<T> &L){
	
	/*
	
	Search for o in L and, if found, replace it
	with r.
	
	Output: 1 iff o exists in L, 0 otherwise
	
	Replaces only \emph{one} o.
	
	*/
	
	L.MoveTop();
	while( !L.eol() ){

 		if (L.Data() == o){ 
 			L.Data() = r; 
 			return 1; 
 		}
 		else L.MoveDown();
 	}
 	return 0; 

}

template<typename T>
bool Remove(T e, LIST<T> &L){
	
	/*
	
	Remove \emph{one} element e from L
	Output: 1 iff e is removed, 0 otherwise.
	
	*/
	
	L.MoveTop();
	while( !L.eol() ){

 		if (L.Data() == e){ 
 			L.Remove(); 
 			return 1; 
 		}
 		else L.MoveDown();
 	}
 	
 return 0; 
}

template<typename T>
void RemoveMultiples(LIST<T> &U, LIST<T> &L){

	L.MoveTop();
	while ( !L.eol() ){ 
		if ( !In(L.Data(), U) ){ U.push(L.Data()); }
		L.MoveDown();
	}
}


/*
template<typename T>
int Remove( LIST<VECTOR<T> > &K, bool(*Condition)(const VECTOR<T>& ) ){

		int count(0);
		K.MoveTop();
		while(!K.eol()){
		
			if ( Condition(K.Data()) ){ 
				K.Remove(); 
				count++;
			}
			else K.MoveDown();
		}

return count;
}

*/


template<typename T>
void Reverse(VECTOR<T> &V){
  
  // Reverses the elements of V:
  // Before: [V(1),V(2),...,V(s)]
  // After:  [V(s),...,V(2),V(1)]

  // not computationally efficient!

  int s = V.Size();  
  VECTOR<T> TEMP( s );
  for(int i=1; i<=s; i++){  TEMP(i) = V(s-i+1); }
  for(int i=1; i<=s; i++){  V(i) = TEMP(i); }


}


template<typename T>
void Copy_LIST_To_VECTOR(LIST<T> *L, VECTOR<T> &V){
	
	/*
	
	k'th element of V is copy of the k'th element of L
	( counting in L starts at Top ).  
	
	*/
	
	if(L->Size() != V.Size()){cout<<endl<<"Copy_LIST_To_VECTOR error"; exit(1);}
	
	int i = 1;
	L->MoveTop();
	while(!L->eol()){
		
		V(i) = L->Data();
		i++;
		L->MoveDown();
	}
}



#endif




