#ifndef __INTERVAL_H__
#define __INTERVAL_H__


#include <stdio.h>
#include <string>
#include <iostream>
#include <float.h>
#include <math.h>  
#include "Structs/round.h"


using namespace std;

#include <string>
#include <stdlib.h>

class FP{
	public:
	static const int _Max []; 
	static const double Max;	

	static const int _Min [];
	static const double Min;
	
	static const int _PosInf []; 
	static const double PosInf; 
	static const double& PosInfinity();
	
	static const int _NegInf [];
	static const double NegInf; 

	// a NaN
	static const int _aNaN []; 
	static const double aNaN; 

	static const int _Ita []; 
	static const double Ita; 

	static const int _Epsilon []; 	
	static const double Epsilon; 
	
	static double Shift(double d, int n);

	static void PrintHex(double w);
	
	static bool isInf(double x);   
	static bool isNaN(double x);

	static double RoundMode();

	static void RoundDown();
	static void RoundUp(); 
	static void RoundNear(); 
	static void RoundToZero(); 	
	
	static double minimum(double x, double y);
	static double maximum(double x, double y);

	
};






class interval{
  
  
private:
double lo;
double hi;
bool   e ;
 
public:   
interval():lo(FP::NegInf), hi(FP::PosInf), e(0) {}   
interval(double a, double b):lo(a),hi(b),e(b<a) {};
interval(double a):lo(a),hi(a),e(0) {};                
interval(const interval &a):lo(a.lo),hi(a.hi),e(a.e) {};      

static const interval Empty;
static const interval Entire;
static const interval Finite;
static const interval NonNegative;
static const interval NonPositive;
static const interval PositiveTail;
static const interval NegativeTail;

//Ingen destruktor
//~interval(){};
 


friend bool	empty (const interval &a);
friend bool	isEmpty(const interval &a); // 11mars  ser bættre ut æn empty
friend bool	isThin(const interval &a);
friend bool	isAtom(const interval &a);
friend bool	isFinite(const interval &a);
//friend bool	isNaI(const interval &a); 
friend double	inf(const interval & a);
friend double	sup(const interval &a);
friend interval iinf(const interval &a);      
friend interval isup(const interval &a);
friend double	width(const interval & a);
friend double	mid(const interval & a);
friend double	mig(const interval & a);
friend double	mag(const interval & a); 
friend double	HausdorffDistance(const interval & a, const interval & b);
friend double	SeparationDistance(const interval &a, const interval &b);

friend interval	iwidth(const interval & x);  
friend interval	imid(const interval & a);
friend interval	abs(const interval & a);
friend interval	gap(const interval &a, const interval &b);
//friend interval min(const interval & a, const interval & b);
//friend interval max(const interval & a, const interval & b); 
friend interval max(const interval &X, const interval &Y);
friend interval maxInv(const interval &Z, const interval &X);
friend interval min(const interval &X, const interval &Y);
friend interval minInv(const interval &Z, const interval &X); 

friend interval hull(const interval &X, const interval &Y);
friend interval hullInv(const interval &Z, const interval &X);

void set(double a, double b);
void setlo(double a);
void sethi(double a);
  
friend interval integer(const interval &X);  
friend double sign2(const interval &X);

interval & operator = (const interval & a);
 
interval operator        +  (const interval & a)const;
interval operator        +  (const double s)const;                               
friend interval operator +  (const double s, const interval & a);                   
interval operator        +  ()const; 
 
interval operator        - (const interval & a)const;
interval operator        - (const double s)const;     
friend interval operator - (const double s, const interval & a);   
interval operator        - ()const; 

friend interval         DSUB (const interval &a, const interval &b);  
friend interval         DADD (const interval &a, const interval &b);
 

void operator += (const interval & a);
void operator -= (const interval & a); 

 
interval operator         *     (const interval & a)const;
interval operator         *     (const double s)const;
friend interval operator  *     (const double s, const interval & a);
interval                  DMUL  (const interval & y);
 

interval operator         /     (const interval & a)const;
interval operator         /     (const double s)const;  
friend interval operator  /     (const double s, const interval & a);  
friend void               xdiv  (const interval & x, const interval & y, interval & D, interval & D1);
 
 
void operator *=(const interval &a);  
void operator *=(const double &a);  
void operator /=(const interval &a); 
void operator /=(const double &a); 
 

friend interval     sqrt    (const interval & a);
interval operator   ^       (int n)const; 
friend bool isEven(const interval &x);
friend bool isOdd(const interval &x);

interval operator   ^       (const interval & n)const;    //NONRIGOROUS 
friend interval     exp     (const interval & x);  //NONRIGOROUS
friend interval     log     (const interval & x);  //NONRIGOROUS 
 

    
bool operator         <= (const interval & b)const; 
bool operator         <= (const double b)const;
friend bool operator  <= (const double a, const interval & b);
bool operator         <  (const interval & b)const;
bool operator         <  (const double b)const;
friend bool operator  <  (const double a, const interval & b);
bool operator         >= (const interval & b)const;
bool operator         >= (const double b)const;
friend bool operator  >= (const double a, const interval & b);
bool operator         >  (const interval & b)const;
bool operator         >  (const double b)const;
friend bool operator  >  (const double a, const interval & b);
 

 
bool operator  != (const interval & b)const;
bool operator  == (const interval & b)const;
bool operator  == (const double b)const;
  


interval operator  &     (const interval & a)const;
interval operator  |     (const interval & b)const; 
void operator      &=    (const interval & a);
void operator      |=    (const interval & a);
 

 
bool operator        <<= (const interval & b)const;
bool operator        <<= (double s)const;
friend bool operator <<= (double s, const interval & b);
bool operator        >>= (const interval & b)const;
bool operator        >>= (double s)const;
friend bool operator >>= (double s, const interval & b);
 
bool operator        << (const interval & b)const;
bool operator        << (double s)const;
friend bool operator << (double s, const interval & b);
bool operator        >> (const interval & b)const;
bool operator        >> (double s)const;
friend bool operator >> (double s, const interval & b);
  

friend void PrintHex(const interval & x);
friend ostream & operator<<(ostream & stream, const interval & a);
friend istream & operator>>(istream & stream, interval & a);



//Propagation
friend bool Intersect(interval &h, const interval &u);
friend bool FPaddition(interval &h, const interval &u, const interval &v);
friend bool BPaddition(const interval &h, interval &u, interval &v);
friend bool FPsubtraction(interval &h, const interval &u, const interval &v);
friend bool BPsubtraction(const interval &h, interval &u, interval &v);
friend bool FPusubtraction(interval &h, const interval &u);
friend bool BPusubtraction(const interval &h, interval &u);
friend bool FPmultiplication(interval &h, const interval &u, const interval &v);
friend bool BPmultiplication(const interval &h, interval &u, interval &v);
friend bool FPdivision(interval &h, const interval &u, const interval &v);
friend bool BPdivision(const interval &h, interval &u, interval &v);
friend bool FPexp(interval &h, const interval &u);
friend bool BPexp(const interval &h, interval &u);
friend bool FPlog(interval &h, const interval &u);
friend bool BPlog(const interval &h, interval &u);
friend bool FPpow(interval &h, const interval &u, const interval &v);
friend bool BPpow(const interval &h, interval &u, interval &v);

friend bool FPabs(interval &h, const interval &u);
friend bool BPabs(const interval &h, interval &u);

}; 


void DivideAndHull(interval &result, const interval &num, const interval &denum);
void DivideAndHullIO(interval &result, const interval &num, const interval &denum);
 





/*

Neumaier: realHull

The following addition to my proposal:

There is an operation realHull(l,u) that returns
for two floating-point numbers l and u the interval [l',u']
with
    l' = NaN, u' = NaN  if l or u is NaN,
    l' = min(l,u,HUGEPOS), u' = max(l,u,HUGENEG)  otherwise
where HUGEPOS and HUGENEG are the finite floating-point numbers
closest to +inf and -inf.



interval realHull(double r, double s){

	double l,u;
	if( isNaN(l) || isNaN(u) ){return interval(aNaN,aNaN);}
	else{
		l=FP::minimum( FP::minimum(r.s), FP::MaxReal); 
		u=FP::minimum( FP::minimum(r.s), FP::MaxReal);  
	}
	return interval(l,u);
}


*/









#endif








 
