#ifndef _intervalUtilities_h_
#define _intervalUtilities_h_

#include <string>
#include <sstream>
#include "Structs/interval.h"
#include "Structs/VECTOR.h"
#include "Structs/LIST.h"

#include <stdlib.h>
#include <list>

VECTOR<interval> Hull(list<VECTOR<interval> > &L);



interval hull(double a, double b);
double relativeDistance(interval &X, interval &Y);

////////////////////////////////////////////////////////////////////////
// Help functions
////////////////////////////////////////////////////////////////////////

void DoubleToInterval(VECTOR<interval> &Ix, 
							const VECTOR<double> &x);

void setEmpty(interval &x);
void setEmpty(VECTOR<interval> &x);
void RelWidthInflate(interval &x, double eps);
void RelMidInflate(interval &x, double eps);

bool 	isEmpty(const VECTOR<interval> &H);
bool 	isFinite(const VECTOR<interval> &X);

double SumOfWidths(const VECTOR<interval> & X);
double Volume(const VECTOR<interval> &X);

////////////////////////////////////////////////////////////////////////
//	Intersect
////////////////////////////////////////////////////////////////////////


void Intersection(VECTOR<interval> &I,
	  	  const VECTOR<interval> &X);
void Intersection(LIST<VECTOR<interval> > &L, 
		const VECTOR<interval> &x);
void Intersection(VECTOR< VECTOR<interval> > &U,
	  	  const VECTOR<interval> &X);
bool IntersectBoth(VECTOR< VECTOR<interval> > &U,
	  	   VECTOR<interval> &X);
	  	   
////////////////////////////////////////////////////////////////////////
// Hull
////////////////////////////////////////////////////////////////////////

void Hull(interval &x, LIST<interval> &L);
void Hull(VECTOR<interval> &x, LIST<VECTOR<interval> > &L);
void Hull(VECTOR<interval> &H,
	  const VECTOR<interval> &X1,
	  const VECTOR<interval> &X2);
void Hull(VECTOR<interval> &H,
	  const VECTOR<interval> &X);
void Hull(VECTOR<interval> &H, const VECTOR< VECTOR<interval> > &U);

////////////////////////////////////////////////////////////////////////
// Mid
////////////////////////////////////////////////////////////////////////

void mid(VECTOR<double> &m, VECTOR<interval> &V);

////////////////////////////////////////////////////////////////////////
//leq,geq
////////////////////////////////////////////////////////////////////////

interval leq(double x);
interval geq(double x);

////////////////////////////////////////////////////////////////////////
//	Vertex
////////////////////////////////////////////////////////////////////////

void Vertex(int i, 
				const VECTOR<interval> &x, 
				VECTOR<double> &z, 
				LIST<VECTOR<double> > &L);

////////////////////////////////////////////////////////////////////////
//	Set Difference
////////////////////////////////////////////////////////////////////////

interval LComplement(const interval &x);
interval UComplement(const interval &x);
void SetDifference(LIST<VECTOR<interval> > &L,
							 VECTOR<interval> y, 
							 VECTOR<interval> x);
void SetDifference(LIST<VECTOR<interval> > &R,
							 const VECTOR<interval> &x, 
							 LIST<VECTOR<interval> > &L);
void SetDifference(LIST<interval> &L, 
							const interval &y, 
							const interval &x);

bool isIncluded(const VECTOR<interval> &x, 
						LIST<VECTOR<interval> > &L);

void DisjointUnion(LIST<VECTOR<interval> > &L, 
							VECTOR<interval> x, 
							VECTOR<interval> y);
void DisjointUnion(LIST<VECTOR<interval> > &L);
void DisjointUnion(LIST<VECTOR<interval> > & L1, 
							LIST<VECTOR<interval> > & L2);


////////////////////////////////////////////////////////////////////////
// old
////////////////////////////////////////////////////////////////////////


void old_ConnectedComponents(LIST<VECTOR<interval> >  L, 
											LIST< LIST< VECTOR<interval > > > &COMP);
VECTOR<interval> old_Hull(LIST<VECTOR<interval> > &L);
bool old_Intersection(const VECTOR<interval> &a, 
								const VECTOR<interval> &b);
bool old_Intersection (const VECTOR<interval> &x, 
								LIST<VECTOR<interval> > &L);
bool old_Intersection (LIST<VECTOR<interval> > &A,
								 LIST<VECTOR<interval> > &B);





////////////////////////////////////////////////////////////////////////
// Save to file
////////////////////////////////////////////////////////////////////////

bool WriteBoxesVertexFormat		(LIST<VECTOR<interval> > L, const std::string &filename);
bool WriteBoxesInlineFormat		(LIST<VECTOR<interval> > &L, const std::string &filename);
bool ReadBoxesInlineFormat		(LIST<VECTOR<interval> > &L, const std::string filename);
bool WriteRealVectorsInlineFormat	(LIST<VECTOR<double> > &L, const std::string filename);

//27Apr
bool WriteBoxes(LIST<VECTOR<interval> > &L, const std::string &filename);
bool ReadBoxes(LIST<VECTOR<interval> > &L, const std::string filename, int dim);

////////////////////////////////////////////////////////////////////////
// Merge
////////////////////////////////////////////////////////////////////////

void MergeAdjacent(LIST<interval> &M);
bool Merge(VECTOR<interval> &a, VECTOR<interval> &b);
void MergeAdjacent(LIST<VECTOR<interval> > &L);
	


#include <fstream>
////////////////////////////////////////////////////////////////////////
// StoreVECTOR
////////////////////////////////////////////////////////////////////////

template <typename T>
bool StoreVECTOR(const VECTOR<T> &v, const string &filename){
 
	ofstream o(filename.c_str());
	if(!o){cout<<endl<<"Cannot write to file " + filename; return 0;}
		
	for(int i=1; i<=v.Size(); i++){ 
		o<<v(i)<<endl; 
	}
	o.close();
	
  return 1;
}

bool StoreIntervalVECTOR(const VECTOR<interval> &v, const string &filename);

#endif

