#ifndef __multiindex_h__
#define __multiindex_h__

#include <iostream>
#include <stdlib.h>
#include "Structs/symUtilities.h"
#include "Structs/VECTOR.h"

//to implement ++ for lexicographic ordering, we must store the maximum order as well (maxOrder).

class multiindex{

// Array of int

private:
int	*vec;  
int	size;

public:
multiindex();       
multiindex(int dimension);
multiindex(const multiindex& x);                 
~multiindex();

int		Size()const{return size;};
int		Order()const;
int 		max()const;
int 		min()const;

int&		operator()(int component);
const int&	operator()(int component)const;

multiindex&	operator=(const multiindex& x);
multiindex&	operator=(const int i);
multiindex	operator+(const multiindex &alpha)const;                                        
multiindex	operator-(const multiindex &alpha)const;
multiindex	operator*(const multiindex &alpha)const;

multiindex	operator+(const int i)const;                                        
multiindex	operator-(const int i)const;
multiindex	operator*(const int i)const;                                       

void		operator+=(const multiindex &a);
void		operator-=(const multiindex &a);
void		operator*=(const multiindex &a);

void		operator+=(const int i);
void		operator-=(const int i);
void		operator*=(const int i);

bool 		operator<=(const multiindex &b)const; 
bool 		operator<=(const int b)const;
friend bool 	operator<=(const int a, const multiindex &b);
bool 		operator<(const multiindex &b)const;
bool 		operator<(const int b)const;
friend bool 	operator<(const int a, const multiindex &b);
bool 		operator>=(const multiindex &b)const;
bool 		operator>=(const int b)const;
friend bool 	operator>=(const int a, const multiindex &b);
bool 		operator>(const multiindex &b)const;
bool 		operator>(const int b)const;
friend bool 	operator>(const int a, const multiindex &b);
 

bool 		operator!=(const multiindex &b)const;
bool 		operator==(const multiindex &b)const;
bool 		operator==(const int b)const;
friend bool 	operator==(const int a, const multiindex &b);
 
int ArgMinPos()const;
int Arg(const int a);

void MakeCanonical(const int i);

/*
static VECTOR<interval> FactorialTable;
static void MakeFactorialTable(const int k);
interval Factorial()const;
static interval c(const multiindex &alpha, const multiindex &beta); 
static interval b(const multiindex &alpha, const multiindex &beta); 
static interval Binomial(const multiindex &alpha, const multiindex &beta);
static interval Binomial(const int n, const int k);
static void Tables(const int n, const int d, ostream &o=cout);
*/

static VECTOR<double> FactorialTable;
static void MakeFactorialTable(const int k);
double Factorial()const;
static double c(const multiindex &alpha, const multiindex &beta); 
static double b(const multiindex &alpha, const multiindex &beta); 
static double Binomial(const multiindex &alpha, const multiindex &beta);
static double Binomial(const int n, const int k);
static void Tables(const int n, const int d, ostream &o=cout);

friend ostream& operator<<(ostream &o, const multiindex &V);
friend istream& operator>>(istream &i, multiindex &a);
};


#endif
