#ifndef _point_h_
#define _point_h_

#include <math.h>
#include <float.h>	
#include "Structs/interval.h"
using namespace std;

////////////////////////////////////////////////////////////////////////

double MaxDistance(const VECTOR<double> &p, 
								const VECTOR<interval> &box, 
								const VECTOR<double> &scale);

void KeepClosePoints(LIST<VECTOR<double> > &ClosePointList,
 								LIST<VECTOR<double> > &PointList, 
 								LIST <VECTOR<interval> > &BoxList, 
 								const VECTOR<double> &scale, 
 								double small);
 								
void Vertex(int i, const VECTOR< VECTOR <double> > &x, 
				VECTOR<double> &z, 
				LIST<VECTOR<double> > &L);
				
void MakeGrid(LIST<VECTOR<double> > &Grid, 
								const VECTOR<interval> &x, 
								const VECTOR<double> &scale, 
								double epsilon);


//--Main Function
void PointRepresentation(LIST<VECTOR<double> > &ClosePointList, 
										LIST<VECTOR<interval> > &BoxList,  
										double epsilon);								
								

////////////////////////////////////////////////////////////////////////
// Mean and Std
////////////////////////////////////////////////////////////////////////

void		Mean(VECTOR<double> &m, LIST<VECTOR<double> > &L);
double Mean(LIST<double> &L);
double Mean(VECTOR<double> &X);
void		MeanAndStd(double &m, double &s, 
						LIST<VECTOR<double> > &L, int index);
void		MeanAndStd(double &m, double &s, 
						const VECTOR<double> &X);

void		EvenlyDistributed(VECTOR<double> &y, const interval &x);
					




////////////////////////////////////////////////////////////////////////			
// bias_rmse			
////////////////////////////////////////////////////////////////////////
void bias_rmse(double &bias, double &rmse, 
							const double trueParameterValue, 
							const VECTOR<double> &estimatedParameterValues);

			
			
								
								

double MaxDistance(const VECTOR<double> &p, 
								const VECTOR<interval> &box, 
								const VECTOR<double> &scale){

////////////////////////////////////////////////////////////////////////
// Scaled Max Distance between a point and a box.
////////////////////////////////////////////////////////////////////////

double d(0.0);
for(int i=1; i<=p.Size(); i++){  
		
	d = FP::maximum(   d,   mig(p(i) - box(i))/scale(i)    );   
} 

return d;
}



void KeepClosePoints(LIST<VECTOR<double> > &ClosePointList,
 								LIST<VECTOR<double> > &PointList, 
 								LIST <VECTOR<interval> > &BoxList, 
 								const VECTOR<double> &scale, 
 								double small){

////////////////////////////////////////////////////////////////////////
// Keep points in PointList that are close to BoxList:  
//
//	Input:		PointList, BoxList, scale,small
//	Output: 	ClosePointList.
//
//	For each point p in PointList,
//	1.	Compute the (scaled-) distance d 
//		from p to BoxList.  
//	2.	If d is small, push p on ClosePointList. 
//
//
//	This is a bottle neck.   
////////////////////////////////////////////////////////////////////////


if(PointList.empty() ){
cout<<"\nKeepClosePoints: PointList is empty.";
exit(1);
}
if(BoxList.empty() ){
cout<<"\nKeepClosePoints: BoxList is empty.";
exit(1);
}
	

bool NEAR(0);
PointList.MoveTop();
VECTOR<double> p( PointList.Data().Size() );

while( !PointList.empty() ){
    
	p = PointList.pop();
	BoxList.MoveTop();
	NEAR = 0;
	while( !BoxList.eol() && !NEAR ){
      
		if (MaxDistance(p ,BoxList.Data(), scale) <= small ) { 
			ClosePointList.push(p); 
			NEAR = 1; 
		}
      	BoxList.MoveDown();    
    }
}


}	


void Vertex(int i, const VECTOR< VECTOR <double> > &x, 
				VECTOR<double> &z, 
				LIST<VECTOR<double> > &L){

////////////////////////////////////////////////////////////////////////
// Store vertices of x in the list L.
// Call: Vertex(0,x,z,L)
////////////////////////////////////////////////////////////////////////

  if (i == z.Size()){ L.push(z); }
  else {
    for(int j=1; j<=x(i+1).Size(); j++){
        z(i+1) = x(i+1)(j); 
        Vertex(i+1,x,z,L);
    }
  }
}



void MakeGrid(LIST<VECTOR<double> > &Grid, 
								const VECTOR<interval> &x, 
								const VECTOR<double> &scale, 
								double epsilon){

////////////////////////////////////////////////////////////////////////
// Make a scaled epsilon rectangular point grid  
// within the bounds of the box x. 
//
// 1.	Computing (width-)scaled coordinate values 
//		for each component x(i).
// 2.	Compute coordinates: Start at vertex inf(x). 
//		For each side i, take (positive) scaled epsilon 
//		steps along that side until the right bound 
//		sup(x(i)) is crossed. 
// 3.	When all coordinates have been determined, 
//		call Vertex, which outputs all possible points.
////////////////////////////////////////////////////////////////////////

double s;  
int d(x.Size());
VECTOR<int> n_steps(d); 
VECTOR<double> step(d); 
	
for(int i=1; i<=d; i++){ 
	
	n_steps(i) 	= ceil( width(x(i))/( epsilon*scale(i) )  );  
	step(i) 		= width(x(i)) / n_steps(i); 
	n_steps(i)++;				
}

VECTOR<VECTOR<double>  > coord(d);
for(int i=1; i<=d ; i++){  
	
		coord(i) = VECTOR<double>(n_steps(i));  
}

for(int i = 1; i<=d; i++){

	s = inf(x(i));
	for(int j=1; j<=n_steps(i);j++){
     			
     			coord(i)(j)= s;
     			s+=step(i); 	
   		}
 }
  
VECTOR<double> z(d);	
Vertex(0, coord, z, Grid);
}


void PointRepresentation(LIST<VECTOR<double> > &ClosePointList, 
										LIST<VECTOR<interval> > &BoxList,  
										double epsilon){



if (BoxList.empty()) {
	cout<<"\nPointRepresentation: The box list is empty.\n";
	exit(1);
}

////////////////////////////////////////////////////////////////////////	
//	*	Input: 		BoxList = Box List, epsilon
//	*	Output: 	ClosePointList
//
// 1.	H=Hull(BoxList)
// 2.	Expand H to a proper enclosure of BoxList.
// 3.	Define scale. 
// 4.	Make a grid in H.
//		(MakeGrid)
// 5.	Keep gridpoints that are close to BoxList.
//		(KeepClosePoints)
////////////////////////////////////////////////////////////////////////


// H = Hull(BoxList) 
BoxList.MoveTop(); 
VECTOR<interval> H(BoxList.Data().Size());
setEmpty(H);
Hull(H,BoxList);
//cout.precision(20);cout<<endl<<H;
//cin.get();
VECTOR<double> scale(H.Size());

//----------scale is defined on the hull H:
//for(int i=1; i<=H.Size(); i++){ 
//	scale(i) = width(H(i)); 
//	H(i) += scale(i)*interval(-epsilon, epsilon);
//}

//-----------scale is defined on the expansion of H:
for(int i=1; i<=H.Size(); i++){ 
	H(i) += width(H(i)) * interval(-epsilon, epsilon);
	scale(i) = width(H(i)); 
}


//Make grid points
LIST<VECTOR<double> > Grid;  
MakeGrid(Grid, H, scale, epsilon);


//Keep grid points that are close to Boxlist
double small(epsilon/2.0); 
KeepClosePoints(ClosePointList, Grid, BoxList, scale, small); 


}






////////////////////////////////////////////////////////////////////////
// Mean and Std
////////////////////////////////////////////////////////////////////////

void Mean(VECTOR<double> &m, LIST<VECTOR<double> > &L){
	
	//m = Sum_{i=1..n} L(i)/n

	int n = L.Size();
	L.MoveTop();
	while(!L.eol()){
		
		for(int j=1; j<=m.Size(); j++){
			m(j) += L.Data()(j) / (double)n; 
		}
		L.MoveDown();
	}    
}


double Mean(LIST<double> &L){

	//return Sum_{i=1..n} L(i)/n

	double m(0.0);
	int n = L.Size();
	L.MoveTop();
	while(!L.eol()){
		m += L.Data() / (double)n; 
		L.MoveDown();    
	}
	return m;
}

double Mean(VECTOR<double> &X){

	// returns Sum_{i=1..n} X(i)/n

	double m(0.0);
	int n = X.Size();
	
	for(int i=1; i<=n; i++){
		m += X(i) / (double)n;     
	}
	return m;
}

void MeanAndStd(double &m, double &s, LIST<VECTOR<double> > &L, int index){


	/*
	
	Mean and Std of the set { X(index) : X in L }
		
	Unbiased std estimator:
	s = (1/(n-1)) sqrt(Sum_{i=1..n} (x_i-Mean(L))^2)
	
	*/
	
	m=0.0;
	int n= L.Size();
	L.MoveTop();
	while(!L.eol()){
		m += L.Data()(index) / (double)n; 
		L.MoveDown();
	}
	
	s=0.0;
	L.MoveTop();
	while(!L.eol()){
		s += pow(L.Data()(index)-m, 2) ;  
		L.MoveDown();
	}
	if (n>1) s=sqrt(s)/(double(n-1));
	else if (n==0){ cout<<endl<<"In MeanAndStd: The list of points L is empty."; exit(1); }  
}


void MeanAndStd(double &m, double &s, const VECTOR<double> &X){


	/*
	
	Mean and Std of the set { X(i) : i=1,...,X.Size() }
		
	std estimator:
	s = sqrt ( (1/(n-1)) (Sum_{i=1..n} (X(i)-Mean(X))^2) )
	
	*/
	
	
	int n=X.Size();

	if (n==0){ cout<<endl<<"In MeanAndStd: The vector X is empty"; exit(1); } 
	if (n==1){ m=X(1); s=0.0; return; } 
	
	
	//MEAN

	m=0.0;
	for(int i=1; i<=n; i++){
		m += X(i) / (double)n; 
	}
	
	
	//STANDARD DEVIATION	
	s=0.0;
	for(int i=1; i<=n; i++){
		s += pow(X(i)-m, 2);  
	}
	
	s=sqrt( s/(double(n-1)) );
	
	/*
	
	Bias Correction for Normal distributed samples 
	http://en.wikipedia.org/wiki/Unbiased_estimation_of_standard_deviation
	
	*/
	
	double BiasCorrection = 1.0 - ( 1.0/(4.0*n) ) - ( 7.0/( 32.0*pow(n,2) ) );
	s=s/BiasCorrection;
	
}

void bias_rmse(double &bias, double &rmse, 
							const double trueParameterValue, 
							const VECTOR<double> &estimatedParameterValues)

//bias and root mean square error
//Definitions from: 
//A.HOOKER et.al 2003, 
//An Evaluation of Population D-Optimal Designs via 
//Pharmacokinetic Simulations 

{
	double s(0.0);
	for(int i=1; i<=estimatedParameterValues.Size();i++)
	{
		s	+=	(estimatedParameterValues(i)-trueParameterValue);
	}
	bias	=	(1.0/trueParameterValue) * ( s/(double)estimatedParameterValues.Size() );
	
	s=0.0;
	for(int i=1; i<=estimatedParameterValues.Size();i++)
	{
		s	+=	pow(estimatedParameterValues(i)-trueParameterValue,2);
	}
	rmse	= (1.0/trueParameterValue)*sqrt( s/(double)estimatedParameterValues.Size() );
}



void EvenlyDistributed(VECTOR<double> &y, const interval &x){
	
	// Generates y.Size() "evenly" distrubuted points y(1),..,y(y.Size()) on interval x
	// NONRIGOROUS
		
	if(y.Size() == 1){ y(1) = mid(x); }   
	else{ 
		double step = width(x)/(y.Size()-1); 	
		for(int i=1; i<=y.Size(); i++){ 
			y(i) = inf(x) + (i-1)*step;    
		}
	}	
}


#endif



