#include "DAG/DAG.h"
#include "DAG/NODE.h"
#include "Structs/Tools.h"
#include "DAG/DAGUtilities.h"
#include "Structs/symUtilities.h"
#include "DAG/DAGRewrite.h"




bool DAG::sEnableRewrite = 1;
bool DAG::sEnableReuse = 1;
DAG* DAG::sCurrDAG;

   
DAG::DAG(){
	nodelist = new LIST<NODE*>;
 	sCurrDAG = this;  
}

DAG::~DAG(){
//18-03
	if(this == sCurrDAG){sCurrDAG =0;}
	delete nodelist;
}



int DAG::Size()const{ 
	return nodelist->Size(); 
}
void DAG::NodeEnumeration(){

	// Enumerate the nodes from 1 to nodelist.Size()
	
	nodelist->MoveTop();
	unsigned int i = 0;
	while ( !nodelist->eol() ){
		i++;
		nodelist->Data()->num = i; 
		nodelist->MoveDown();
	} 
}
void DAG::NumberToPointer(VECTOR<NODE*> &V ){

  	/*
	
	Pass vector V with the size of the DAG ( the size of the nodelist )
	and obtain a vector of NODE pointers.
 
	V(i) = pointer to node number i
	
	*/ 
  
	nodelist->MoveTop();
	while (!nodelist->eol() ){
		V( nodelist->Data()->num ) = nodelist->Data();  
		nodelist->MoveDown();
	}
}

           
	/*
	All insertions go via InsertNode(NODE*) 
	The inserted node's number becomes the number of the new DAG size. 
	*/
void  DAG::InsertNode(NODE *n){ 

	//Unconditionally
	
	nodelist->push(n); 
	n->num = nodelist->Size(); 
}     
void  DAG::RemoveNode(NODE *n){ 
	
	//Unconditionally
	
	nodelist->MoveTop();
	while (!nodelist->eol()){
	
		//NEW
		if(nodelist->Data() == n) { nodelist->Remove(); return; }
		else nodelist->MoveDown();
		 
	}     
}

	/*
	
	ConstructNODE
	Search for a constant node with range r.
	If such a node is found, then it is returned, 
	else a new constant node with range r is created and returned.  
	
	*/  

NODE* DAG::ConstructNODE( const interval &r ){   
//27Jan
//15Feb

  if(width(r)==0.0){
 	map<interval,NODE*>::iterator iter;
	iter = constants.find(r);
	if(iter != constants.end()){ return iter->second; }
  }
	// Not found, construct a new constant node with range r and return it.
 
	NODE *c = new NODE_con();
	InsertNode(c);        
	constants.insert(pair<interval,NODE*>(r,c));
	if(inf(r)==sup(r)) { Symbol(c) = ToString<double>(inf(r)) ;}
	else Symbol(c) = ToString<interval>(r);
	c->range = r;
return c;	
}



NODE* DAG::ConstructNODE(const string &nodesym, LIST<NODE*> *predList){

predList->MoveTop();

if(sEnableReuse){

	if (    nodesym == "+" || 
		nodesym == "*" || 
		nodesym == "-" || 
		nodesym =="/"  || 
		nodesym =="^"){

		//p1 op p2
 
		NODE *p1 = predList->Data(); 
		predList->MoveDown(); 
		NODE *p2 = predList->Data();

		LIST<NODE*> *p1succ = p1->Succ(); 
		p1succ->MoveTop(); 
		NODE *candidate = 0;
		while(!p1succ->eol()){
		
			candidate = p1succ->Data();
			if ( isEquivalent(candidate,nodesym,p1,p2) ){ return candidate; }
			p1succ->MoveDown();
			
		}	
	}
	
	else if(nodesym=="u-" || nodesym=="exp" || nodesym=="log"){
	
		//op p
 
		NODE *p = predList->Data();

		LIST<NODE*> *psucc = p->Succ(); 
		psucc->MoveTop(); 
		while(!psucc->eol()){
			
			if(Symbol(psucc->Data()) == nodesym){ return psucc->Data(); }
			psucc->MoveDown();
		}
	}
	
	
	else if(nodesym=="="||nodesym=="sum"||nodesym=="&"){
	
		NODE *p = predList->Data(); 

		LIST<NODE*> *psucc = p->Succ(); 
		psucc->MoveTop(); 
		NODE *e = 0;
		while(!psucc->eol()){
	
			e = psucc->Data();
			
			if( (Symbol(e) == nodesym) && areEqual(predList,e->Pred()) ){
			
					return e;
			}
			
			psucc->MoveDown();
		}
	}	
		
}





	//----Simplify----
	if(sEnableRewrite){     
        
  		//NODE *s = Simplify(*sCurrDAG, nodesym, predList); 
  		//if (s){ return s; }
  	
  		NODE *s;
  		for(int i=1; i<=RewriteSystem.Size(); i++){
  			s = RewriteSystem(i)(*sCurrDAG, nodesym, predList);
  			if(s){ return s;}
  		}
  	}
	
	
	
	
  
//	No equivalent node was found, so we construct a new node.
  
	

 NODE *c;
 if         (nodesym == "+")        c = new NODE_add    ("+");
 else if    (nodesym == "-")        c = new NODE_sub    ("-");
 else if    (nodesym == "u-")       c = new NODE_usub   ("u-");
 else if    (nodesym == "*")        c = new NODE_mul    ("*");
 else if    (nodesym == "/")        c = new NODE_div    ("/");
 else if    (nodesym == "^")        c = new NODE_power  ("^");
 //else if    (nodesym == "<=")       c = new NODE_leq    ("<=");
 //else if    (nodesym == ">=")       c = new NODE_geq    (">=");
 else if    (nodesym == "exp")      c = new NODE_exp    ("exp");
 else if    (nodesym == "log")      c = new NODE_log    ("log");
 else if    (nodesym == "sum")      c = new NODE_sum    ("sum");
 else if    (nodesym == "=")        c = new NODE_equal    ("=");
 else if    (nodesym == "&")        c = new NODE_intersect("&");
 else{}
 

 InsertNode(c);  

 	
 	
// 	MakeEdges:
// 	1) Each node in predList becomes a predecessor of c.
// 	2) c becomes a successor of each node in predList 
// 	   (c is inserted into the successorlist of each node in L).
  	
  	

 
	//1)
	c->setPred(predList); 

	//2)
	LIST<NODE*> U;
	RemoveMultiples(U, *predList);     
	U.MoveTop();
	while( !U.eol() ){
		U.Data()->Succ()->push(c); 
		U.MoveDown();
	}
 

return c;	
}


bool DAG::isEquivalent(NODE* &candidate, const string &nodesym, NODE *p1, NODE *p2){


	if ( !( Symbol(candidate) == nodesym ) ) return 0;
	
	
	
		NODE *c1,*c2;
		candidate->Pred()->MoveTop();
		c1 = candidate->Pred()->Data();
		candidate->Pred()->MoveDown();
		c2 = candidate->Pred()->Data();
		
		if(nodesym=="+" || nodesym=="*"){
			
			if(p1==c1 && p2==c2 || p2==c1 && p1==c2){return 1;}
		
		}
		else if(nodesym=="-"){


			if( p1==c1 && p2==c2){ return 1; }
			if( p2==c1 && p1==c2){ candidate = &(- *candidate); return 1;}
		}
		else if(nodesym =="/"){
		
			if( p1==c1 && p2==c2){ return 1; }
			if( p2==c1 && p1==c2){ candidate = &(1.0/ *candidate); return 1;}
		
		}
		else if(nodesym =="^"){
		
			if( p1==c1 && p2==c2){ return 1; }
		}
		
	
return 0;	
	

}

bool DAG::isEquivalentTo(const string &nodesym, LIST<NODE*> *predList){

  
	/*
  
	Test if current node (in nodelist) has:
	1) identical nodesymbol to nodesym
	2) equivalent predecessor list to predList
	*/
  

	NODE *n = nodelist->Data();
  	if ( !( Symbol(n) == nodesym ) ) return 0;
	if ( !isCommutative(n) ){
		return ( *n->Pred() == *predList   );
	}
  	else{// commutative
    
		return areEqual(n->Pred(), predList); 
 
	}
}


void DAG::GetNodeList(LIST<NODE*>* &L){
	  L = nodelist;
}






ostream & operator<<(ostream& stream, const DAG & G){

  
	G.nodelist->MoveTop();
	while ( !G.nodelist->eol() ){
      
	stream << endl <<  (G.nodelist)->Data();
	stream << endl <<  (G.nodelist)->Data()->num;
	stream << endl <<  Symbol(  (G.nodelist)->Data()  );

	stream << endl  << "Pred:" << endl << *(    (G.nodelist)->Data()->Pred()     ); 	
	stream << endl  << "Succ:" << endl << *(    (G.nodelist)->Data()->Succ()     );
    	stream << endl;
    	
    	G.nodelist->MoveDown();
  	
  	}
  
  
return stream; 
}












