using namespace std;
#include "DAG/DAGFilters.h"
#include "DAG/DAGUtilities.h"
#include "Structs/VECTOR.h"
#include "Structs/Tools.h"
#include "Structs/intervalUtilities.h"




////////////////////////////////////////////////////////////////////////
// Measure
////////////////////////////////////////////////////////////////////////




double Measure::Volume(LIST<VECTOR<interval> > &L, 
										VECTOR<double> &scale, 
										VECTOR<int> &select){

////////////////////////////////////////////////////////////////////////
//
// scaled volume: PROD{ width(x_i)/s_i } 
//
////////////////////////////////////////////////////////////////////////



double vol(0.0),v;

L.MoveTop();
while(!L.eol()){
		
	v=1.0;				  
	for(int i=1; i<=scale.Size(); i++){
	
		v *= width( L.Data()(select(i)) )  /  scale(i) ;
	
	}
	
	vol += v;
	L.MoveDown();	
}

return vol;
}



double Measure::AbsRelWidth(double threshold, const interval &x){
	
	
	double w = width(x);
	if(w==FP::PosInf) return w;
	double m(mid(abs(x)));
	
	if(m<threshold)	return w;
	else return w/m;
}


double Measure::RelWidth(const interval &x){
	
	double w = width(x);
	if (w<FP::PosInf){
	  //cout<<"\n RelWidth = "<<w/ (mid(abs(x)) + FP::Ita);
		return w/ (mid(abs(x)) + FP::Ita);
	}
	else return FP::PosInf;
}

double Measure::NewWidth(const interval &x, double p){
	
	double w = width(x);
	if (w<FP::PosInf){
		if(inf(x)>0.0 || sup(x)<0.0) p=1.0;
		//cout<<"\n RelWidth = "<<pow(w,p)/ (mid(abs(x)) + FP::Ita);
			return pow(w,p)/ (mid(abs(x)) + FP::Ita);
		} 
	else return FP::PosInf;
}


double Measure::MaxWidth(const VECTOR<interval> &x, 
										const VECTOR<int> &select){

	double w(0.0);
	for(int i=1;i<=select.Size();i++){ 
			if(width(x(select(i))) > w) w=width(x(select(i))) ; 
		}
	return w;
}

double Measure::MaxRelWidth(const VECTOR<interval> &x, 
											const VECTOR<int> &select){

	double m(0.0), rwi;
	for(int i=1;i<=select.Size();i++){ 
			rwi=RelWidth(x(select(i)));
			if(rwi > m) m=rwi; 
	}
	return m;
}

double Measure::SumOfWidths(const VECTOR<interval> &x){

	double s(0.0);
	for(int i=1; i<=x.Size(); i++){ s += width(x(i)); }
	return s;
}

double Measure::SumOfWidths(const VECTOR<interval> &x, const VECTOR<int> &index){

	double s(0.0);
	for(int i=1; i<=index.Size(); i++){ s += width(x(index(i))); }
	return s;
}

double Measure::Volume(const VECTOR<interval> &x){
	
	double v(1.0);
	for(int i=1; i<=x.Size(); i++){ v *= (width(x(i))); }
	return v;
} 
double Measure::Volume(const VECTOR<interval> &x, 
									const VECTOR<int> &index){
	
	double v(1.0);
	for(int i=1; i<=index.Size(); i++){ v *= (width(x(index(i)))); }
	return v;
} 

double Measure::Volume(LIST<VECTOR<interval> > &L){
	
	double v(0.0);
	L.MoveTop();
	while(!L.eol()){
		v += Volume(L.Data());
		L.MoveDown();
	}
	return v;
}


////////////////////////////////////////////////////////////////////////
//	Expander
////////////////////////////////////////////////////////////////////////

Expander::Expander():
		Mode("relative"),
		MaxWidth(FP::PosInf),
		rate(0.01),
		absTerm(FP::Ita),
		rateAdjuster(0.01), 
		Count(0),
		MinCount(10),
		MaxCount(999999),
		rateMin(1e-10){} 




void Expander::setNodeNumber(LIST<NODE*> &L){

// set the NodeNumber vector from the list L of nodes

NodeNumber = VECTOR<int>(L.Size());
L.MoveTop();
int index(0);
while( !L.eol() )
{
	index++;	
	NodeNumber(index)=L.Data()->num;
	L.MoveDown();
}

}

bool Expander::isPossibleX1(){
	
	// is it possible to expand X1 more ?
	
	for(int i=1; i<=NodeNumber.Size(); i++){
		
		if( width(X1(NodeNumber(i)))< MaxWidth ) return 1;
	}
	return 0;
}


void Expander::Expand(VECTOR<interval> &X){
	
if			(Mode =="absolute")	{ absExpand(X); }
else if	(Mode =="relative")	{ relMidExpand(X); }

}

void Expander::rateAdjustment(const string &direction){

	if(direction=="down") {
		
		rate *= rateAdjuster;
	}
	else if (direction=="up") rate /= rateAdjuster;
	
}

void Expander::relMidExpand(VECTOR<interval> &X){
			
	for(int i=1; i<=NodeNumber.Size();i++){
		
	    X(NodeNumber(i)) += mid(abs(X(NodeNumber(i))))*interval(-rate,rate) + interval(-absTerm, absTerm);	
	
	}
}

void Expander::absExpand(VECTOR<interval> &X){
			
	for(int i=1; i<=NodeNumber.Size();i++){
	
		X(NodeNumber(i)) += rate*interval(-1.0,1.0);	
	}
}

void Expander::SymmetricExpand(VECTOR<interval> &C, 
													const VECTOR<interval> &D){

/////////////////////////////////////////////////////////////////////////
// Expand C(NodeNumber(i)) symmetrically 
// about its mid point to an interval that encloses
// D(NodeNumber(i)).
////////////////////////////////////////////////////////////////////////

double m(0.0), r(0.0); 
for(int k=1; k<=NodeNumber.Size(); k++){    
		
	if( !( D(NodeNumber(k)) <<= C(NodeNumber(k)) ) ){
			
		m 	=	mid ( C(NodeNumber(k)) );
		r		= 	mag( D(NodeNumber(k)) - m );
		C(NodeNumber(k))	=	interval(m-r, m+r);		
	}	
 }

}


void Expander::SymmetricExpand(VECTOR<interval> &C, 
						const VECTOR<double> &d){

/////////////////////////////////////////////////////////////////////////
// Expand C(NodeNumber(i)) symmetrically 
// about its mid point to an interval that encloses
// d(NodeNumber(i)).
////////////////////////////////////////////////////////////////////////
	  
double m(0.0), r(0.0); 
for(int k=1; k<=NodeNumber.Size(); k++){    
		
if ( !( d(k) <<= C( NodeNumber(k) )) ) {
			
			m =    mid ( C(NodeNumber(k)) );
			r	=    fabs( d(k)-m );//nonrigorous
			C(NodeNumber(k))	=  interval(m-r,m+r);			
		}	
	}
}


	
void Expander::Info(ostream &o){

o<<endl<<"#Expander::Info()";
o<<endl<<"Expander::Mode "				<<Mode;
o<<endl<<"Expander::rate"					<<rate;
o<<endl<<"Expander::rateAdjuster "	<<rateAdjuster;
o<<endl<<"Expander::MinCount "		<<MinCount;
o<<endl<<"Expander::MaxCount "		<<MaxCount;
}

////////////////////////////////////////////////////////////////////////
// Filter
////////////////////////////////////////////////////////////////////////

Filter::Filter():G(DAG::sCurrDAG), 
		Iter(0),
		MinIter(1), 
		MaxIter(999),
		ContractionRatio(0.999),
		MinWidth(0.001),	
		shave_ratio(0.05){} //FP::Ita
		
Filter::Filter(DAG &aG){
	G = &aG;
	V = VECTOR<interval>(G->Size());
	G->NodeEnumeration(); 
	for(int i=1; i<=V.Size(); i++){V(i) = interval::Entire;}
	SetConstants(*G,V);
	
	FORWARD 	= ForwardOrderingByLevel(*G);
	BACKWARD = BackwardByReverseForward(FORWARD);
	
	setParameters();

}

void Filter::setParameters(){
	Iter					=	0;
	MinIter				=	1;
	MaxIter				=	999;
	ContractionRatio	=	0.999;
	MinWidth			=	0.000000001;
	shave_ratio		= 0.05;
}

////////////////////////////////////////////////////////////////////////
// Read, Write
////////////////////////////////////////////////////////////////////////

void Filter::Read(NODE &n, interval &x){
	x = V(n.NodePtr()->num);
}

interval Filter::Read(NODE &n){
	return V(n.NodePtr()->num);
}

void Filter::Read(VECTOR<NODE> &n, VECTOR<interval> &x){
	
	for(int i=1; i<=n.Size(); i++){
		x(i) = V(n(i).NodePtr()->num);
	}
}       			

VECTOR<interval> Filter::Read(VECTOR<NODE> &n){
	VECTOR<interval> x(n.Size());
	for(int i=1; i<=n.Size(); i++){
		x(i) = V(n(i).NodePtr()->num);
	}
	return x;
}       			

void Filter::Write(NODE &n, const interval &x){
	V(n.NodePtr()->num) = x;
}
void Filter::Write(VECTOR<NODE> &n, const VECTOR<interval> &x){
	
	for(int i=1; i<=n.Size(); i++){
		V(n(i).NodePtr()->num) = x(i);
	}
}
void Filter::Write(vNODE &n, const VECTOR<interval> &x){
	
	
	for(int i=1; i<=n.Size(); i++){
		V(n(i).NodePtr()->num) = x(i);
	}
}


void Filter::Write(NODE *n, const interval &x){
	V(n->num) = x;
}
void Filter::Write(VECTOR<NODE*> &n, const VECTOR<interval> &x){
	
	for(int i=1; i<=n.Size(); i++){
		V(n(i)->num) = x(i);
	}
}





////////////////////////////////////////////////////////////////////////
// Sweeps
////////////////////////////////////////////////////////////////////////

bool Filter::SweepForward(){

	int s = FORWARD.Size();
    for(int i=1; i<=s; i++){ if( FORWARD(i)  -> FP(V) ){ return 1; }  }
	return 0;
}


void Filter::SweepForwardIO(){

  int s = FORWARD.Size();
  for(int i = 1; i<=s; i++){ FORWARD(i) -> FPIO(V);}
}

void Filter::SweepForwardIO(VECTOR<interval> &X){

  int s = FORWARD.Size();
  for(int i = 1; i<=s; i++){ FORWARD(i) -> FPIO(X);}
}

bool Filter::SweepBackward(){

  int s = BACKWARD.Size();  
  for(int i=1; i<=s; i++){ if( BACKWARD(i)  -> BP(V) ){return 1;}  }
  return 0;
}

bool Filter::SweepForwardBackward(int nCycles){

	int f  = FORWARD.Size();
	int b  = BACKWARD.Size();

	for(int cycle = 1; cycle<=nCycles; cycle++){
		for(int i=1; i<=f; i++){ if( FORWARD(i)  -> FP(V) ){return 1;}  }
		for(int i=1; i<=b;i++){ if( BACKWARD(i) -> BP(V) ){return 1;} }
	}
  	return 0;
}

bool Filter::SweepForwardBackward(VECTOR<interval> &X){

  int f  = FORWARD.Size();
  int b  = BACKWARD.Size();

  for(int i = 1; i<=f; i++){ if( FORWARD(i)  -> FP(X) ){ return 1; }  }
  for(int i = 1; i<=b; i++){ if( BACKWARD(i) -> BP(X) ){ return 1; }  }
   
  return 0;

}

bool Filter::LoopSweepForwardBackward(){

	VECTOR<interval> V_OLD( V.Size() );

	Iter = 0;
	do{
		V_OLD = V;     
		Iter++; 
		if(SweepForwardBackward()){return 1;}
	}while ( Iter<MinIter || (Iter < MaxIter  &&  isContraction(V_OLD)) );
  
return 0;  
}

bool Filter::LoopSweepForwardBackward(VECTOR<interval> &X){

VECTOR<interval> V_OLD( X.Size() );

	Iter = 0;
	do{
		V_OLD = X;     
		Iter++; 
		if(SweepForwardBackward(X)){return 1;} 
	 }while ( Iter<MinIter || Iter<MaxIter  &&  isContraction(X, V_OLD) );
  
return 0;  
}


////////////////////////////////////////////////////////////////////////
// isContraction
//
// 
// 1. Does not depend on component number. 
//		(MinWidth and ContractionRatio should depend on i)
// 2. The infinite width case needs to be fixed.
////////////////////////////////////////////////////////////////////////



bool Filter::isContraction(const interval &NEW, const interval &OLD){
////////////////////////////////////////////////////////////////////////
// Ensure that NEW and OLD are non-empty.
////////////////////////////////////////////////////////////////////////

//	Infinite width

double r(0.0);
if( width(NEW) == FP::PosInf){

	if( inf(NEW)>inf(OLD) + r || sup(NEW)<sup(OLD) - r ) return 1;
	return 0;
}	

//	Finite width

if (width(NEW) < ContractionRatio * width(OLD) && 
	width(OLD) >= MinWidth) return 1;  
return 0;

}

bool Filter::isContraction(VECTOR<interval> &NEW, VECTOR<interval> &OLD){

	for(int i = 1; i<=NEW.Size(); i++){
		
		if( isContraction(NEW(i), OLD(i)) ) return 1;
  	}
	return 0;
}

bool Filter::isContraction(VECTOR<interval> &OLD){
// V variant 

	for(int i = 1; i<=V.Size(); i++){
		
		if( isContraction(V(i), OLD(i)) ) return 1;
 
  	}
  	return 0;
}


bool Filter::isRelContraction(VECTOR<interval> &OLD){
////////////////////////////////////////////////////////////////////////
// NOT USED
//
// Returns 1 iff some variable has relative 
// contraction and its initial width was above Low_Width.
////////////////////////////////////////////////////////////////////////

for(int i=1;i<=V.Size();i++){
	if(Measure::RelWidth(V(i))<ContractionRatio * Measure::RelWidth(OLD(i)) && 
		Measure::RelWidth(OLD(i)) >= MinWidth){
		return 1;
	} 
}
return 0;
}


bool Filter::CPH2(int i, double cut){
////////////////////////////////////////////////////////////////////////	
//	Contract Parts and Hull on 2 subboxes.
//	Split V(i) at cut.
//	SubContractor : LoopSweepForwardBackward
//	
//	Make sure that cut is in V(i) , otherwise there
//	will be expansion.
////////////////////////////////////////////////////////////////////////	   

	
	VECTOR<interval> R(V);
	R(i)	= interval(cut, sup(V(i)));	
	V(i)	= interval(inf(V(i)), cut); 
		
	VECTOR<bool> e(2);
	
	e(1) = LoopSweepForwardBackward();
	e(2) = LoopSweepForwardBackward(R);
	
	if(e(1) && e(2))		{return 1;}
	else if	(e(1))		{V=R;}	//return CPH2(i);
	else if	(e(2))		{}		//return CPH2(i); 
	else						{Hull(V, R);}
	
return 0;
}

////////////////////////////////////////////////////////////////////////
// Shaving
//
//	Gives disproportionate intervals. I should have used 
//	geometric series instead. DO THIS. Check out
//  Neumaiers safeguarded GM. 
////////////////////////////////////////////////////////////////////////

bool Filter::Shave(){
		
	shaveList->MoveTop();
	while(!shaveList->eol()){
		
		
		if(ShaveUpper(shaveList->Data(), shave_ratio)) return 1;
		if(ShaveLower(shaveList->Data(), shave_ratio)) return 1;
		shaveList->MoveDown();
	}
	return 0;
}



double cut_lower(const interval &x, double ratio){

/////////////////////////////////////////////
double step = ratio*FP::Max;
/////////////////////////////////////////////

double cut;

if ( inf(x) == FP::NegInf ){
	
	cut	= FP::Min;
}
else if( sup(x) == FP::PosInf ){ // here inf(x)>NegInf	

	if(inf(x)<FP::Max){
		cut	=	min(FP::Max, inf(x) +step); 
 	}
 	else{	//inf(x)=FP::Max
 		cut	= FP::Max;	}
}

else{
	cut	= min(sup(x), inf(x)+step);	
}

return cut;

}

double cut_upper(const interval &x, double ratio){

return -cut_lower(-x, ratio);
}

bool Filter::ShaveLower_Unbounded(int i, double ratio){
////////////////////////////////////////////////////////////////////////
//	*	The width of V(i)=[l,u]  is evaluated to infinity 
//		( u-l > FP::Max ). 
//	*	We can not use the ratio*width to define a proper 
//		shave off (SO) interval. 
//	*	Three basic cases:
// 
//  step = ratio*FP::Max;
//
//	The step is default  value in the infinite 
//	dimensional case, but I will define it as a filter 
//	parameter so that it can be set to 
//	any other smaller value as well.
//
//	1) l=NegInf  
//		-->	SO=NegativeTail
//				exit
//
//  2) l>NegInf && u=PosInf
//
//		2.1) l<FP::Max
//				-->	SO=[l, min{FP::Max, l+step}]
//						exit
//		2.2) inf(V(i))== FP::Max
//				-->	SO=PostiveTail
//						exit
//
//	3l>NegInf && u<PosInf
//  			-->	SO=[l, min{u, l+step} ]
//
////////////////////////////////////////////////////////////////////////

bool success;
VECTOR<interval> R(V.Size());  
double cut;

do{
	
	R		=	V; 
	cut 	= cut_lower(V(i), ratio);
	R(i)	=	interval( inf(V(i)), cut ); 
	V(i)	=	interval( cut, sup(V(i)) );

	success = LoopSweepForwardBackward(R);  
	//cout<<"\nShaveLower_Unbounded "<< success <<"     " <<V(i); 
}
while( success   &&   width(V(i))==FP::PosInf);
   	
if(success){ return LoopSweepForwardBackward();}
else			{ V(i) |= R(i); return 0;}

}

bool Filter::ShaveUpper_Unbounded(int i, double ratio){

bool success;
VECTOR<interval> R(V.Size());  
double cut;

do{
	
	R		=	V; 
	cut 	= cut_upper(V(i), ratio);
	R(i)	=	interval( cut, sup(V(i)) ); 
	V(i)	=	interval( inf(V(i)), cut );

	success = LoopSweepForwardBackward(R);  

}
while( success   &&   width(V(i))==FP::PosInf);
   	
if(success){ return LoopSweepForwardBackward();}
else			{ V(i) |= R(i); return 0;}
}




bool Filter::ShaveLower(int i, double ratio){
////////////////////////////////////////////////////////////////////////
// Greedy Shaving off lower part of V(i)
////////////////////////////////////////////////////////////////////////

//
//	Special Cases: 
//	1) thin, 2) unbounded.
//
if( width(V(i)) == 0.0){return LoopSweepForwardBackward();}

if( width(V(i)) == FP::PosInf ){ 
	if( ShaveLower_Unbounded(i,ratio) )		{ return 1;} 
	if( width(V(i)) == FP::PosInf)				{ return 0;}
	if( width(V(i)) == 0.0) 							{return LoopSweepForwardBackward();}
} 

//
// Positive and Finite Width	
//
bool success;
VECTOR<interval> R(V.Size());  
double u		= sup(V(i));
double cut	= inf(V(i));
double step	= ratio*width(V(i));
   	
do{ 
	
	R		=	V;
    cut	=	min(cut + step, u);       
    R(i)	=	interval(inf(V(i)), cut); 
   	V(i)	=	interval(cut, u);

   	success = LoopSweepForwardBackward(R);  
   	//cout<<"\nShave Lower "<< success <<"     " <<V(i); 		
}
while( success && cut != u);

 if(success){
   //cout<<"\nInterval rejected by ShaveLower\n";
   return 1;
 }   
else{ 
	V(i) |= R(i); 
	return 0;
}
}


bool Filter::ShaveUpper(int i, double ratio){
////////////////////////////////////////////////////////////////////////
// Greedy Shaving of ratio off upper 
// part of V(i)
////////////////////////////////////////////////////////////////////////

//
//	Special Cases: 
//	1) thin, 2) unbounded.
//

if( width(V(i)) == 0.0){return LoopSweepForwardBackward();}

if( width(V(i)) == FP::PosInf ){ 
	if( ShaveUpper_Unbounded(i,ratio) )		{ return 1;} 
	if( width(V(i)) == FP::PosInf)				{ return 0;}
	if( width(V(i)) == 0.0) 							{return LoopSweepForwardBackward();}
} 

//
// Positive 	
//

bool success;
VECTOR<interval> R(V.Size());
double l		= inf(V(i));
double cut	= sup(V(i)); 
double step = ratio*width(V(i));   	

 do{ 
	R		= V;
	cut	= max(cut - step, l);        
	R(i)	= interval(cut, sup(V(i)) ); 
	V(i)	= interval(l, cut);
   		
	success = LoopSweepForwardBackward(R);
	//cout<<"\nShave Upper "<< success <<"     " <<V(i);

}
while( success && cut != l);

 if(success) {//cout<<"\n Interval rejected by ShaveUpper\n";
return 1;}
else{ V(i)|=R(i); return 0;}
}



bool Filter::ImprovedShaver(VECTOR<interval> &X, int i, double ratio){

//
// Positive and Finite Width
//

bool success;
double bound	= sup(X(i));
double cut		= inf(X(i));
double step		= ratio*width(X(i));
VECTOR<interval> L(X.Size()), R(X.Size());  

// SHAVE LOWER SIDE OF X(i) 
   	
do
{ 
	
	// SHAVE OFF [inf(X(i)), cut]
	L		=	X;
    cut	=	min(cut + step, bound);       
    L(i)	=	interval(inf(X(i)), cut); 
   	X(i)	=	interval(cut, bound);

   	success = LoopSweepForwardBackward(L);  
   	//cout<<"\nShave_Lower sucess =  "<< success <<",   new interval " <<X(i); 		
}
while( success && cut  != bound);


if(success)
{
  //cout<<"\nX rejected by Shaver_Lower\n";
	return 1;
}   
else if(cut==bound)
{
	X = L; 
	return 0;
}
		
// HERE: L-nonempty X-nonempty

////////////////////////////////////////////////////////////////////////
// gap:
//
// If a gap ( sup(L(i)) < inf(X(i))  ) was formed then 
// we can try to expand it by shaving the left side 
// of X(i). 
// This should be done before we start shaving the
// right side of X(i). 

// ModeAgressive:
// Expand every detected gap, immediately. 

// bool gap = ( sup(L(i)) < inf(X(i)) );
// if( ShaveLower(X,i,ratio) ){X=L; return 0;}
////////////////////////////////////////////////////////////////////////

// SHAVE UPPER SIDE OF X(i)

bound	= inf(X(i));
cut 		= sup(X(i));

do
{ 
	// SHAVE OFF [cut, sup(X(i))]
	R		=	X;
    cut	=	max(bound, cut - step);       
    R(i)	=	interval(cut, sup(X(i))); 
   	X(i)	=	interval(bound, cut);

   	success = LoopSweepForwardBackward(R);  
   	//cout<<"\nShave_Upper "<< success <<"     " <<X(i); 		
}
while( success && cut != bound);


if(success)
{
  //cout<<"\nX rejected by Shaver_Upper\n";
	X = L;  
	return 0;
}   
else if(cut==bound)
{
	X = R; 
	//possible gap ( sup(L(i)) <inf(R(i)) )
	Hull(X, L); 
	return 0;
}

//HERE: L-nonempty, X-nonempty, R-nonempty
		
//CONTRACT X
success = LoopSweepForwardBackward(X); 
if (success) 
	{
		X=L; 
		Hull(X,R);
	}
else
	{
	Hull(X,L);
	Hull(X,R); 
}
return 0;

}



////////////////////////////////////////////////////////////////////////
// Info
////////////////////////////////////////////////////////////////////////

void Filter::Info(ostream &o){

o<<endl<<"#Filter::Info()";
o<<endl<<"Filter::ContractionRatio "<<ContractionRatio;
o<<endl<<"Filter::MinWidth "			<<MinWidth;
o<<endl<<"Filter::MinIter "				<<MinIter;
o<<endl<<"Filter::MaxIter "				<<MaxIter;
}

////////////////////////////////////////////////////////////////////////
// Splitter
////////////////////////////////////////////////////////////////////////

Split::Split(){

CUT_cutFunction = "safeGM";
CUT_safeGM_M 	= FP::Max;
CUT_safeGM_r	= 0.3;
}

Split::Split(DAG &G, VECTOR<NODE> &v){

SetSelect(v);
Count = VECTOR<int> (G.Size());
for(int i=1; i<=Count.Size(); i++) Count(i)=0;

CUT_cutFunction = "safeGM";
CUT_safeGM_M 	= FP::Max;
CUT_safeGM_r	= 0.3;

}

Split::Split(DAG &G, vNODE &v){

SetSelect(v);
Count = VECTOR<int> (G.Size());
for(int i=1; i<=Count.Size(); i++) Count(i)=0;

CUT_cutFunction = "safeGM";
CUT_safeGM_M 	= FP::Max;
CUT_safeGM_r	= 0.3;

}

Split::Split(DAG &G, LIST<NODE*> &L){

SetSelect(L);
Count = VECTOR<int> (G.Size());
for(int i=1; i<=Count.Size(); i++) Count(i)=0;

CUT_cutFunction = "safeGM";
CUT_safeGM_M 	= FP::Max;
CUT_safeGM_r	= 0.3;

}

////////////////////////////////////////////////////////////////////////
// SetSelect : Define the candidate set of nodes 
// among whom we decide which to split. 
////////////////////////////////////////////////////////////////////////

void Split::SetSelect(VECTOR<NODE> &v){

Select = VECTOR<int>(v.Size());
for(int i=1; i<=v.Size(); i++){ Select(i)=v(i).NodePtr()->num; }
}

void Split::SetSelect(vNODE &v){
	Select = VECTOR<int>(v.Size());
	for(int i=1; i<=v.Size(); i++){ Select(i)=v(i).NodePtr()->num; }
}
void Split::SetSelect(LIST<NODE*> &L){

	Select = VECTOR<int>(L.Size());
	L.MoveTop();
	int i(0);
	while(!L.eol()){
		i++;
		Select(i)=L.Data()->num;
		L.MoveDown();
	}
}



bool BranchAndPrune::CPH(VECTOR<interval> &X, 
									const VECTOR<int> &components)
{

	list<VECTOR<interval> > L;
	L.push_front(X);
	S.MultiSplit(L, components);

	if( Contract(L) ) return 1;
 	X=Hull(L);
 	return 0;
 	
}

bool BranchAndPrune::CPH(){
return CPH(F.V, S.Select);
}


bool Split::CUT_getCut(double &cut, const interval &X){

if(CUT_cutFunction == "safeAM")
{
	cut = mid(X);  //Need Fix  ( safeAM )
	return 1;
}

else if(CUT_cutFunction == "safeGM")
{
	
  if( safeGM(cut, CUT_safeGM_M, CUT_safeGM_r, X) ){ //cout<<"safeGM"; 
return 1;}
  else{//cout<<"---------------------------------safeGM FAILURE"; 
return 0;}
	

}
else
{
  if( safeGM(cut, CUT_safeGM_M, CUT_safeGM_r, X) ){ //cout<<"safeGM"; 
	  return 1;}
	else{//cout<<"---------------------------------safeGM FAILURE"; 
	  return 0;}
	

}

}

//
//	COMP_getComponent
//
int Split::COMP_getComponent(const VECTOR<interval> &X){

if(COMP_componentSelector == "max_width")
{
	return COMP_max_width(X);
}	
if(COMP_componentSelector == "max_rel_width")
{
	return COMP_max_rel_width(X);
}	
if(COMP_componentSelector == "max_scaled_width")
{
	
	return COMP_max_scaled_width(X);
}
if(COMP_componentSelector == "bestBranch")
{
	return COMP_bestBranch(X);
}



}



int Split::COMP_max_width(const VECTOR<interval> &X){

int index_max; double width_max(-1.0), w; 

for(int i=1; i<=Select.Size(); i++){

	w= width(X(Select(i)));
	if (width_max < w){width_max = w; index_max = Select(i);}      
}

return index_max;
}

int Split::COMP_max_rel_width(const VECTOR<interval> &X){

int index_max;    double width_max(-1.0), w;

for(int i=1; i<=Select.Size(); i++){

	w = Measure::RelWidth(X(Select(i)));
	if (width_max < w){width_max = w; index_max = Select(i);}      
}

return index_max;
}

int Split::COMP_max_scaled_width(const VECTOR<interval> &X){

int index_max; double width_max(-1.0), w; 

for(int i=1; i<=Select.Size(); i++){

	if(scale(i) !=0.0) w=width(X(Select(i)))/ scale(i);
	else w=0.0;
	if (width_max < w){width_max = w; index_max = Select(i);}      
}

return index_max;
}

int Split::COMP_bestBranch(const VECTOR<interval> &X){

return 1;
}



//
//	safeAM
//

bool  Split::safeAM(double &cut, const interval &X){
////////////////////////////////////////////////////////////////////////
//	Arithmetic Mean
//
// 	X is must be non-atomic.
////////////////////////////////////////////////////////////////////////	

	cut = max(-FP::Max, inf(X))*0.5 + min(FP::Max, sup(X))*0.5;

	if(cut==inf(X) || cut==sup(X)) return 0;
	return 1;
 	

}



double sgn(double x){
	if(x<0) return -1.0;
	return 1.0;
}


bool Split::safeGM(double &cut, 
							const double M, 
							const double r, 
							const interval &X){
////////////////////////////////////////////////////////////////////////

double lo( inf(X) ), hi(sup(X)), p(lo*hi);


if( 0.0 < p && p<FP::PosInf) cut=sgn(lo)*sqrt(p);  

//0 inside

//0 in interior
else if(lo <0.0 && 0.0<hi) cut=0.0;
//0 on boundary
else if(lo==0.0) cut = min(M, r*hi);
else if(hi==0.0) cut = max(-M, r*lo);


//0 not inside 
//p=FP::PosInf  
else if(lo>0.0) 	cut = min(M, lo/r);
else 				cut = max(-M, hi/r); //hi<0 

if(cut==lo || cut==hi) return 0;
return 1;
}




bool Split::SplitAndPush(VECTOR<interval> &X, 
									LIST<VECTOR<interval> > &L){
////////////////////////////////////////////////////////////////////////
//	Split And Push:
// 	1.	X(c) is atomic 		-->	
//				X is pushed on L, 
//				0 is returned
//  2.	X(ic) is nonatomic	--> 
//				X is split and the parts are 
//				pushed on L, 
//  			Count is updated, 
//				1 is returned.
////////////////////////////////////////////////////////////////////////
	
double cut;

//get component
int c = COMP_getComponent(X); 	
interval X_c = X(c);


if( isAtom(X_c) )
{ 
	L.push(X); 
	return 0; 
}

//get cut

else
{
	
	if( !CUT_getCut(cut, X_c) ) return 0;

	X(c)	=   interval( inf(X_c), cut );
	L.push(X);
	X(c)	=   interval( cut, sup(X_c) );
	L.push(X);
}
	
Count(c)++;  
return 1;
}

  
void Split::MultiSplit(list<VECTOR<interval> > &L, 
							const VECTOR<int> &comp){

////////////////////////////////////////////////////////////////////////
//	Every box in L is split in every component specified
//	by comp, 
////////////////////////////////////////////////////////////////////////

	
	double cut;
	interval X_ci;
	
	VECTOR<interval> X( L.front().Size() );
	
	list<VECTOR<interval> > *pL =&L;
	list<VECTOR<interval> > K;
	list<VECTOR<interval> > *pK = &K;
	list<VECTOR<interval> > *tempL;

	
	for(int i=1; i<=comp.Size(); i++)
		{
	
			while(!pL->empty())
				{
			
					X		=	pL->front(); 
					pL->pop_front();
					
					X_ci	=	X(comp(i));
				
				    CUT_getCut(cut, X_ci);
					
					X(comp(i)) = interval( inf(X_ci), cut );
					pK->push_back(X);
					X(comp(i)) = interval( cut, sup(X_ci) );
					pK->push_back(X);
				}
	
			tempL	=	pL;
			pL			=	pK;
			pK		=	tempL;
				
		}


L.assign(pL->begin(), pL->end()); 
}


void Split::Info(ostream &o){

	o<<"\n#Split::Info()";
	o<<"\nSplit::CUT_cutFunction "				<< CUT_cutFunction;
	o<<"\nSplit::CUT_safeGM_M "					<< CUT_safeGM_M;
	o<<"\nSplit::CUT_safeGM_r "					<< CUT_safeGM_r;
	o<<"\nSplit::COMP_componentSelector "	<< COMP_componentSelector;
	
	for(int i=1; i<=Select.Size(); i++){
	o<<"\nSplit::Select(" + ToString(i) + ") "<<Select(i);
	}
	
	for (int i=1; i<=Count.Size(); i++){
   		if( Count(i)!=0 ){
     			o<<"\nSplit::Count(" + ToString(i) + ") " <<Count(i);
   		}
 	}
 	
}



////////////////////////////////////////////////////////////////////////
// BAP
////////////////////////////////////////////////////////////////////////


////////////////////////////
//	Default Pruner
////////////////////////////
bool g_pruner_default(Filter &F){
return F.LoopSweepForwardBackward();
}



BranchAndPrune::BranchAndPrune(){}

BranchAndPrune::BranchAndPrune(Filter &aF, Split &aS):
	F(aF),
	S(aS),
	MinIter(1),
	MaxIter(999),
	Iter(0),
	RangeListMaxSize(500),
	RangeListMinSize(0),
	SMALL_width(0.1),
	SMALL_abswidth(0.0),
	SMALL_abswidth_count(0),
	SMALL_width_measure("scaled"),
	scale_update_period(1),
	pruner(&g_pruner_default),
	CC_enable(1),
	IOVR_enable(1),
	CON_enable(1),
	SMALL_enable(1),
	PRUNE_enable(1),
	CON_inconsistencyTest(0),
		CON_softDegree(0.0),
		BAP1_Expand_nRejections(10){ 
			
			scale = VECTOR<double>(aF.V.Size());
}

BranchAndPrune::BranchAndPrune(Filter &aF, Split &aS, Expander &aE):
	F(aF),
	S(aS),
	E(aE),
	MinIter(1),
	MaxIter(999),
	Iter(0),
	RangeListMaxSize(500),
	RangeListMinSize(0),
	SMALL_width(0.1),
	SMALL_abswidth(0.0),
	SMALL_abswidth_count(0),
	SMALL_width_measure("scaled"),
	scale_update_period(1),
	pruner(&g_pruner_default),
	CC_enable(1),
	IOVR_enable(1),
	CON_enable(1),
	SMALL_enable(1),
	PRUNE_enable(1),
	CON_inconsistencyTest(0),
		CON_softDegree(0.0),
		BAP1_Expand_nRejections(10){
			
			scale = VECTOR<double>(aF.V.Size());
}
	

bool BranchAndPrune::SMALL(){
////////////////////////////////////////////////////////////////////////
//	If current box F.V is small then push it on
//	IndefiniteList and return 1.
//
//	-- smallness is determined in current scale 
//	-- only split nodes are considered before, now
//	its the general case.
////////////////////////////////////////////////////////////////////////
	
if (!SMALL_enable) return 0;	

int n;
double w;


bool smallabswidth(1);
for( int i=1; i<=SMALL_nodenumber.Size(); i++)
{
	n = SMALL_nodenumber(i);
	if(  width( F.V(n) )   > SMALL_abswidth   ) {smallabswidth = 0; break;}
}

if( smallabswidth ){
  //cout<<"\n\n SMALL_abswidth \n\n";
	SMALL_abswidth_count++;
	IndefiniteList.push(F.V);
	return 1;
}



if(SMALL_width_measure=="scaled")
{

	for(int i=1; i<=SMALL_nodenumber.Size(); i++)
	{ 
		n = SMALL_nodenumber(i);
		if (scale(n)>0.0)
		{
					
		  //cout<<"width( F.V(n) ) / scale(n)"<<width( F.V(n) ) / scale(n);
			if(  width( F.V(n) ) / scale(n)  > SMALL_width   ) return 0;
			
		}
	}
}



else if (SMALL_width_measure=="relative")
{
	
	double d;
	for(int i=1; i<=SMALL_nodenumber.Size(); i++)
	{ 	
		n=SMALL_nodenumber(i);
		d = mid( abs(F.V( n )) );
		
		if(d<SMALL_width )
		{	
			if(  width(F.V( n ))   > SMALL_width   ) return 0;
		}
		else
		{
			if(  width(F.V( n )) / d  > SMALL_width   ) return 0;
		}
	}
}


else if (SMALL_width_measure=="absolute")
{	
	for(int i=1; i<=SMALL_nodenumber.Size(); i++)
	{ 
		n=SMALL_nodenumber(i);
		if(  width(F.V( n ))  > SMALL_width   ) return 0;
	}
}





//cout<<"\n\n SMALL \n\n";
IndefiniteList.push(F.V);
return 1;


}

void BranchAndPrune::SMALL_set_nodenumber(LIST<NODE*> &L)
{
	
SMALL_nodenumber = VECTOR<int>(L.Size());
L.MoveTop();
int i=0;
while(!L.eol())
{
	i++;
	SMALL_nodenumber(i)=L.Data()->num;
	L.MoveDown();
}	

}


bool BranchAndPrune::isInconsistent(){ 

return(		RangeList.Size()==0 &&
				IndefiniteList.Size()==0 && 
				ConsistencyList.Size()==0 );
}

void BranchAndPrune::CleanAllLists(){

	RangeList.Clean(); 
	IndefiniteList.Clean(); 
	ConsistencyList.Clean(); 
}

void BranchAndPrune::PushAllLists(LIST<VECTOR<interval> > &L){
	
	Push(L, RangeList); 
	Push(L, IndefiniteList); 
	Push(L, ConsistencyList);
}
      	
void BranchAndPrune::HullAllLists(VECTOR<interval> &V){
	
	Hull(V, RangeList); 
	Hull(V, IndefiniteList); 
	Hull(V, ConsistencyList);
}


bool BranchAndPrune::Contract(LIST<VECTOR<interval> > &L){
////////////////////////////////////////////////////////////////////////
//	Contract each box in the list L 
//	and if a box becomes empty remove it from L.
//	
//	Output:	1 if all boxes contract to empty.
//				0 otherwise.
////////////////////////////////////////////////////////////////////////

L.MoveTop();
while(!L.eol()){
		
		if(F.LoopSweepForwardBackward(L.Data())){ ;  
			L.Remove();
		}	
		else{ L.MoveDown(); } 
		
	}

	if(L.Size()==0) return 1;
	return 0;
}


bool BranchAndPrune::Contract(list<VECTOR<interval> > &L){
////////////////////////////////////////////////////////////////////////
//	Contract each box in the list L 
//	and if a box becomes empty remove it from L.
//	
//	Output:	1 if all boxes contract to empty.
//				0 otherwise.
////////////////////////////////////////////////////////////////////////

list<VECTOR<interval> >::iterator it( L.begin() );
while(it !=L.end())
{

	if( F.LoopSweepForwardBackward( *it ) ) it = L.erase(it);	
	else it++; 

}

if(L.empty()) return 1;
return 0;
}

void BranchAndPrune::BAP1_initialize()
{

Iter	=	0; 

// scale 

setScale();

// CC

CC_firstTime	=	1;
CC_prevCover.Clean();

// IOVR

IOVR_prev = 0.0;
IOVR_iter_check = IOVR_iter_check_period;

// CON

RangeList.MoveTop(); 
CON_set_ranges(RangeList.Data());

}
	

bool BranchAndPrune::Prune(){

	if( !PRUNE_enable ) return 0;
	return (*pruner)(F);
}

void BranchAndPrune::getNextBox(){

	F.V = RangeList.pop();
}

void BranchAndPrune::splitBox(){

	S.SplitAndPush(F.V, RangeList);
}


void BranchAndPrune::BAP1_printInfo(){

}

bool BranchAndPrune::BAP1(){ 
////////////////////////////////////////////////////////////////////////
//
//	BAP1, Branch And Prune algorithm 1, 
//	operates on boxes present in RangeList. 
//
//	pruning function: BranchAndPrune::pruner
//
//	Termination Conditions:
//			-- !CC_isLarge
//			-- IOVR_isLarge
//			-- RangeList empty 
//			-- Rangelist size exceeding RangeListMaxSize 
//			-- Iter exceeding MaxIter
//	
//	Return: isInconsistent()
//			
//
////////////////////////////////////////////////////////////////////////
	
BAP1_initialize();
		
while(	(!RangeList.empty())						&& 
			(RangeList.Size()<RangeListMaxSize)&& 
			(Iter<MaxIter)){
		
		BAP1_printInfo();	
		Iter++;
		if( !CC_isLarge() )			{continue;}
		if( IOVR_isLarge() )			{continue;}
		getNextBox();
		if( Prune() )					{continue;}
		setScale();
		if( SMALL() )					{continue;}            
		if( CON_consistent() )		{continue;}
		splitBox();
}

return isInconsistent(); 
} 
	

//***********************************************
//
//	CON
//
//***********************************************


double BranchAndPrune::CON_softExclusionDegree(interval &X, interval &Y){
	
	//safe
	
	//if (isEmpty(X&Y)) return FP::PosInf;
	
	double d1,d2;
	
	if (inf(X) != inf(Y)) d1=2.0*( inf(Y)-inf(X) )  / (  fabs( inf(Y) )+fabs( inf(X) )  );
	else d1=0.0;
	
	if (sup(X) != sup(Y)) d2=2.0*( sup(X)-sup(Y) )  /  (  fabs( sup(Y) )+fabs( sup(X) )  );
	else d2=0.0;
	
	if (d1<d2) return d2;
	return d1;
}


void BranchAndPrune::CON_set_ranges(const VECTOR<interval> &X){
////////////////////////////////////////////////////////////////////////
//	set consistency range
////////////////////////////////////////////////////////////////////////

	if(!CON_enable) return;

	for(int i=1; i<=CON_components.Size(); i++){ 
		CON_ranges(i) = X( CON_components(i) );
	}
}


bool BranchAndPrune::CON_consistent(){		
////////////////////////////////////////////////////////////////////////
//	Check if F.V is CON_ranges consistent.
//
//	if consistent:
//			*	push it on the ConsistencyList and 
//			*	return 1
//
//	else: *	return 0. 
//
////////////////////////////////////////////////////////////////////////
//	1)
//	If there are interval valued constants 
//	they might have contracted. We therefore ensure
//	they have their initial values:
//
////////////////////////////////////////////////////////////////////////

if(!CON_enable) return 0;

VECTOR<interval> X(F.V);
//1)
SetConstants(*F.G, X); 
F.SweepForwardIO(X);



//INCONSISTENCY TEST
//HARD --only hard inconsistency available
if(CON_inconsistencyTest==1){
	for(int i=1; i<=CON_components.Size(); i++){
		if( isEmpty( X(CON_components(i)) & CON_ranges(i) )   ){
		  //cout<<"\nPOSITIVE INCONSISTENCY TEST";
			return 1;
		}
	}
}


//CONSISTENCY TEST

if(CON_softDegree>0.0){ 
//SOFT	
	for(int i=1; i<=CON_components.Size(); i++){
	
		if( CON_softExclusionDegree(X(CON_components(i)),CON_ranges(i) )>CON_softDegree)  return 0;
	}
}  

else{
//HARD
	for(int i=1; i<=CON_components.Size(); i++){
	//cout<<"\nCON_ranges(i)= "<<CON_ranges(i);
		if(   !(   X(CON_components(i)) <<=  CON_ranges(i)   )      ) return 0;
		
	}
}



ConsistencyList.push(F.V); 
return 1;
}



//***********************************************
//	IOVR
//***********************************************




void BranchAndPrune::IOVR_checking(double IOVR_curr)
{
		
	if(IOVR_curr <= 1.1*IOVR_prev && IOVR_curr<0.9*IOVR)
	{
		IOVR_iter_check = Iter+ IOVR_iter_check_period;
	} 
	else
	{
		IOVR_iter_check = Iter + IOVR_iter_check_period / 10;
	}
	
	IOVR_prev = IOVR_curr;
}


bool BranchAndPrune::IOVR_isLarge(){
////////////////////////////////////////////////////////////////////////
//  Ouput: 	1 if innerVol/outerVol > InnerOuterVolumeRatio.
//				0 otherwise
////////////////////////////////////////////////////////////////////////

if(!IOVR_enable) return 0;

if (Iter<IOVR_iter_check) return 0;

double innerVol, outerVol;

innerVol = Measure::Volume(ConsistencyList, S.scale, S.Select);

outerVol = innerVol;
outerVol += Measure::Volume(RangeList, S.scale, S.Select);
outerVol += Measure::Volume(IndefiniteList, S.scale, S.Select);

double ratio = innerVol/outerVol;
IOVR_checking(ratio);

bool isLarge = (ratio  > IOVR);


if (isLarge){
	Push(IndefiniteList, RangeList);
	RangeList.Clean();
}


return isLarge;
}



//***********************************************
//	Cover Contraction
//***********************************************


bool BranchAndPrune::CC_isLoosingCoverVolume(){

	double currVol, prevVol;
	
	currVol	=	Measure::Volume(RangeList,	S.scale, S.Select);
	currVol	+=Measure::Volume(ConsistencyList,	S.scale, S.Select);
	currVol	+=Measure::Volume(IndefiniteList,	S.scale, S.Select);
	
	prevVol	=	Measure::Volume(CC_prevCover,	S.scale, S.Select);
	
	
return  ( currVol < CC_contractionRatio* prevVol ) ;  
		
	
}

bool BranchAndPrune::CC_isLarge(){
////////////////////////////////////////////////////////////////////////
// If the covering refinement ( RangeList.Size() ) 
// grows much
// faster than the gain of information about solution
// structure then we STOP, that is,
// we move the whole RangeList to the IndefiniteList. 
//
// Control parameters :
// CC_prevCover:
//				previous cover
//
// CC_minCoverSize: 
//	
//					Threshold, only if the size of the 
//					current cover 
// 					is larger than this threshold do we 
//					compute contraction ratio. 
//					Setting this threshold to a large value
//					, say, 99999999, disables the Contraction detection. 
//	
// CC_contractionRatio: a number in (0,1),
//	only if currVol>CC_ratio*prevVol do we return 0  
//

if(!CC_enable) return 1;

bool aboveThreshold = (CoverSize() > CC_minCoverSize );

// If above threshold for the first time then set CC_prevCover
if(CC_firstTime && aboveThreshold) 
	{ 
		CC_prevCover.Clean();
		PushAllLists(CC_prevCover);
		CC_firstTime=0;
	}


//	If current list size is above threshold and 
//	is twice as large as the previous list size, 
//	then check if there were any contraction.
if( aboveThreshold && 
	CoverSize() > CC_sizeRatio*CC_prevCover.Size() )
	
	{
	
	if( !CC_isLoosingCoverVolume() ){
	
			//-------STOP---------
			Push(IndefiniteList, RangeList); 
			RangeList.Clean();
			return 0; 
	}
			
	CC_prevCover.Clean();
	PushAllLists(CC_prevCover);
	
	}


return 1;
}


bool BranchAndPrune::ExpansionRate_upAdjustment(){

if (E.rateAdjuster>0.0 &&
	E.Count == BAP1_Expand_nRejections){ 
	
	E.rateAdjustment("up");
	E.Count = 0;
	return 1;
}  

return 0;
}


bool BranchAndPrune::ExpansionRate_downAdjustment(){
////////////////////////////////////////////////////////////////////////
// Reject a high expansion rate.
//	If E.Count is zero then the current box was 
//	BAP1-accepted without expansion, 
//	and we return 0;
////////////////////////////////////////////////////////////////////////

if(	E.rateAdjuster>0.0	&&
	E.rateMin<E.rate	&&
	(0<E.Count)		 
	){ 
		
		E.rateAdjustment("down"); 
		CleanAllLists();		
		RangeList.push(E.X0); 
		return 1;
}

return 0;
}


bool BranchAndPrune::Expand(){
	
	if(E.isPossibleX1())
	{
		
		ExpansionRate_upAdjustment();
		
 		E.X0 = E.X1;
 		E.Expand(E.X1);
		RangeList.push(E.X1);
		E.Count++;	
		return 1;
	}
	else
	{
	  return 0; 
	}


}

void BranchAndPrune::BAP1_Expand_initialize(){

RangeList.MoveTop(); 
E.X0 = RangeList.Data(); 
E.X1 = E.X0;			
E.Count=0;
	
}



bool BranchAndPrune::BAP1_Expand(){
	
	

////////////////////////////////////////////////////////////////////////
 /*
 *		---	Inconsistency detector: BAP1().
		---	Expander function: E.Expand()

 *	Define: accepted = non-rejected

 *	Input:	One box B in RangeList. 
				B expands on BAP1-rejection.
				Expansion continues until B is BAP1-accepted.  

 *	Output: 

	--  E.Count	: total number of expansion operations.   
	--	E.X0		: Largest BAP1-rejected box (if E.count == 0 )
	--	E.X1  	: Smallest BAP1-accepted box.
		
	--	The collection of boxes in RangeList, 
		ConsistencyList and IndefiniteList computed 
		by BAP1 on the first accepted box E.X1.
		Their hull is included in E.X1, making E.X1 
		obselete for the output. 
		
	

 *	Remark: 

	The largest inconsistent box E.X0 is here only 
	used in the (optional) expansion-rate rejection 
	subroutine. Other usage of E.X0 (perhaps pruning)
	have not been implemented.
*/	
////////////////////////////////////////////////////////////////////////	




E.Count=0;
if(RangeList.Size() == 0){return 1;}
RangeList.MoveTop(); 
E.X0 = RangeList.Data(); 
E.X1 = E.X0;			

while( BAP1() )
{
	
	if(E.isPossibleX1())
	{
		
		ExpansionRate_upAdjustment();
 		E.X0 = E.X1;
 		E.Expand(E.X1);
		E.Count++; 
		RangeList.push(E.X1);	
	}
	else
	{
		return 1; 
	}
	
}

//////////////////////////////////
// Here: BAP1-accepted
/////////////////////////////////


if ( ExpansionRate_downAdjustment() ) BAP1_Expand();

return 0;
}







////////////////////////////////////////////////////////////////////////
// Globals outputs of BAP1_Expand_Symmetric()

VECTOR<interval> 				gFirstHit;
VECTOR<interval>				gCFirstHit;
LIST<VECTOR<interval> >	glist_CFirstHit;
VECTOR<interval>				gSymExp;

VECTOR<interval>				gpoint;
VECTOR<interval>				gimage;

////////////////////////////////////////////////////////////////////////


bool BranchAndPrune::BAP1_Expand_Symmetric(){

////////////////////////////////////////////////////////////////////////
//	The computations are, in sequence:
//  1. BAP1_Expand
//	2. Symmetric Expansion
//	3. BAP1
////////////////////////////////////////////////////////////////////////

// Save initial box X_Init.
RangeList.MoveTop();
VECTOR<interval> X_Init(RangeList.Data());

////////////////////////////////////////////////////////////////////////
// BAP1_Expand

if( BAP1_Expand() ){ 
  //cout<<"\nINCONCISTENCY by BAP.BAP1_Expand()		\
  //		in BAP1_Expand_Symmetric()"; 
return 1; 
}

LIST<VECTOR<interval> >	HitList;
VECTOR<interval>				hull_HitList(F.V.Size()); 

PushAllLists(HitList);
setEmpty(hull_HitList);
Hull(hull_HitList, HitList);

////////////////////////////////////////////////////////////////////////
//Save in globals
gFirstHit 			= E.X1;
glist_CFirstHit 	= HitList;
gCFirstHit 			= hull_HitList;
		
////////////////////////////////////////////////////////////////////////
// Symmetric expansion
	
VECTOR<interval> symBoxInit(X_Init);//begin with the initial box again
	
for(int k=1; k<= E.NodeNumber.Size(); k++){ 
	symBoxInit(E.NodeNumber(k)) = hull_HitList(E.NodeNumber(k)); 
}	
E.SymmetricExpand(symBoxInit, X_Init);

// Save Symmetrized Expansion
gSymExp = symBoxInit;
////////////////////////////////////////////////////////////////////////

CleanAllLists();
RangeList.push(symBoxInit);
if( BAP1() ){ 
	cout<<"\n INCONSISTENCY by BAP.BAP1() \
				in BAP1_Expand_Symmetric() after\
				symmetric expansion"; 
	return 1;
}
	
return 0;
}




void BranchAndPrune::get_point_image(VECTOR<interval> &point,
														VECTOR<interval> &image,
														vNODE &p1, NODE &p2, NODE &p3){

////////////////////////////////////////////////////////////////////////  
//		This subfunction of 
//		BAP1_Expand_Symmetric(vNODE &,NODE &,NODE&)
//
//		chooses the midpoint of the (p1,p2,p3) 
//		subvector of the first box in the contracted 
//		first hit list and sweep it forward to get the 
//		image.
////////////////////////////////////////////////////////////////////////

 
	//
	//chose a point
	//
	
glist_CFirstHit.MoveTop();
F.V=glist_CFirstHit.Data();
	
int k;
for(int i=1; i<=p1.Size(); i++){
	k=p1(i).NodePtr()->num;
	point(i)	= interval(mid(F.V(k)));
	F.V(k) 	= point(i);
}

k=p2.NodePtr()->num;
point(p1.Size()+1)	= interval(mid(F.V(k)));
F.V(k) = point(p1.Size()+1);

k=p3.NodePtr()->num;
point(p1.Size()+2)	= interval(mid(F.V(k)));
F.V(k) = point(p1.Size()+2);
	
	//
	//compute its image 
	//
	
F.SweepForwardIO();

for(int i=1; i<=E.NodeNumber.Size(); i++){	
	image(i) = F.V( E.NodeNumber(i) );
}

}


bool BranchAndPrune::BAP1_Expand_Symmetric(vNODE &p1, 
																		NODE &p2, 
																		NODE &p3){


////////////////////////////////////////////////////////////////////////
//	The computations are, in sequence:
//  1. BAP1_Expand
//	2. Symmetric Expansion
//	3. BAP1
////////////////////////////////////////////////////////////////////////

// Save initial box X_Init.
RangeList.MoveTop();
VECTOR<interval> X_Init(RangeList.Data());

////////////////////////////////////////////////////////////////////////
// BAP1_Expand

if( BAP1_Expand() ){ 
  //cout<<"\nINCONCISTENCY by BAP.BAP1_Expand()		\
  //			in BAP1_Expand_Symmetric()"; 
return 1; 
}
LIST<VECTOR<interval> >	HitList;
VECTOR<interval>				hull_HitList(F.V.Size()); 

PushAllLists(HitList);
setEmpty(hull_HitList);
Hull(hull_HitList, HitList);


////////////////////////////////////////////////////////////////////////
//Save in globals
gFirstHit 			= E.X1;
glist_CFirstHit 	= HitList;
gCFirstHit 			= hull_HitList;
		
////////////////////////////////////////////////////////////////////////
// Symmetric expansion

VECTOR<interval> symBoxInit(X_Init); 

VECTOR<interval> point(p1.Size()+2); //(p1,p2,p3)
VECTOR<interval> image(E.NodeNumber.Size());

get_point_image(point, image, p1, p2, p3); 
//save in globals
gpoint = point;
gimage = image;

//
//data as endpoints(Minimal expansion)
//

for(int i=1; i<=E.NodeNumber.Size(); i++){	
	symBoxInit( E.NodeNumber(i) ) = image(i);
}
E.SymmetricExpand(symBoxInit, X_Init);

double rate_saved = E.rate;
E.rate = reExpansionRate;
E.Expand(symBoxInit);
E.rate = rate_saved;


// Save Symmetrized Expansion
gSymExp = symBoxInit;
////////////////////////////////////////////////////////////////////////

CleanAllLists();
RangeList.push(symBoxInit);
if( BAP1() ){ 
	return 1;
}
	
return 0;
}


void BranchAndPrune::setScale(){
////////////////////////////////////////////////////////////////////////
// set S.scale 
////////////////////////////////////////////////////////////////////////


//if iter = 0 or a multiple of scale_update_period
//then set scale.
if (Iter%scale_update_period != 0) return; 


VECTOR<interval> V(F.V);  
HullAllLists(V);  
VECTOR<double> s(S.Select.Size());

for(int i=1; i<=S.Select.Size(); i++){
		s(i) = width(V(S.Select(i)));
}

S.scale = s; 


for(int i=1; i<=V.Size(); i++){
		scale(i) = width(V(i));
}

}



void BranchAndPrune::Info(std::ostream &o){

F.Info(o);
S.Info(o);
E.Info(o);
o<<"\n#BAP::Info()";
o<<"\nBAP::RangeList.Size "			<<RangeList.Size();
o<<"\nBAP::ConsistencyList.Size "	<<ConsistencyList.Size();
o<<"\nBAP::IndefiniteList.Size() "		<<IndefiniteList.Size();
o<<"\nBAP::Iter "							<<Iter;
o<<"\nBAP::MaxIter "						<<MaxIter;
o<<"\nBAP::MinIter "						<<MinIter;
o<<"\nBAP::RangeListMaxSize "		<<RangeListMaxSize;
o<<"\nBAP::RangeListMinSize "		<<RangeListMinSize;
o<<"\nBAP::SMALL_width "				<<SMALL_width;
o<<"\nBAP::SMALL_abswidth "			<<SMALL_abswidth;
o<<"\nBAP::SMALL_abswidth_count "<<SMALL_abswidth_count;
o<<"\nBAP::SMALL_width_measure "<<SMALL_width_measure;
o<<"\nBAP::scale_update_period "	<<scale_update_period;
o<<"\nBAP::CC_enable "					<<CC_enable;
o<<"\nBAP::IOVR_enable "				<<IOVR_enable;
o<<"\nBAP::SMALL_enable "			<<SMALL_enable;
o<<"\nBAP::CON_enable "				<<CON_enable;
o<<"\nBAP::PRUNE_enable "			<<PRUNE_enable;
	
o<<"\nBAP::CC_contractionRatio "	<<CC_contractionRatio;
o<<"\nBAP::CC_sizeRatio "				<<CC_sizeRatio;
o<<"\nBAP::CC_minCoverSize "		<<CC_minCoverSize;
o<<"\nBAP::CON_softDegree "		<<CON_softDegree;
o<<"\nBAP::IOVR = "						<<IOVR;
o<<"\nBAP::IOVR_iter_check_period "<<IOVR_iter_check_period;
o<<"\nBAP1_Expand_nRejections"	<<BAP1_Expand_nRejections;
}




void Multi_Split(
		 VECTOR< VECTOR<interval> > &U,
		 const VECTOR<interval> &V,
		 int i,
		 int n_subbox){
		 	
////////////////////////////////////////////////////////////////////////
//	Partition the box V into n_subbox subboxes by 
//	partitioning coordinate i into n_subbox equally 
//	sized subintervals.
//
// Remark:  V(i) must be bounded.
//
// Needs improvement: split it into pieces with 
// with geometric length.
////////////////////////////////////////////////////////////////////////
  
double t = inf(V(i));
double r = width(V(i))/((double)n_subbox);

for(int k=1; k<=n_subbox; k++){
		
		U(k) = V;
		U(k)(i) = interval(t , t+r);
		t += r;  
}

U(n_subbox)(i) = interval( inf( U(n_subbox)(i) ), sup( V(i)) );
}



void MakeUnions(LIST<VECTOR<VECTOR<interval> > > &ListOfUnion,
		const VECTOR<int> &component,
		int n_subbox,
		const VECTOR<interval> &V){

	VECTOR< VECTOR<interval> > U(n_subbox);
	for(int k=1; k<=component.Size(); k++){
		Multi_Split(U, V, component(k), n_subbox);
		ListOfUnion.push(U);
	}
}





