#include "DAG/DAGRewrite.h"
#include "DAG/DAG.h"
#include "DAG/NODE.h"
#include "DAG/DAGUtilities.h"

#include "Structs/LIST.h"



DAGRewrite::DAGRewrite():stdOutput(0){}

bool DAGRewrite::isMulPow_xa_Pow_xb (NODE *n, NODE *x, NODE *a, NODE *b){





if (    Symbol(n)                == "*" && 
	Symbol(n->pred1())	 == "^" &&
	Symbol(n->pred2())	 == "^" &&
	n->pred1()-> pred1() == n->pred2()-> pred1() //same base 
   )
   
   {
  	x = n->pred1()->pred1();
  	a = n->pred1()->pred2();
  	b = n->pred2()->pred2();
  	
  	return 1;
   }

return 0;
}

NODE* DAGRewrite::Pow_xa_Mul_Pow_xb(DAG &G, const string &nodesym, LIST<NODE*> *predList){


if(nodesym == "*"){
	
	NODE* p1 = (*predList)[0];
	NODE* p2 = (*predList)[1];
	
	// (x^a) * (x^b) -> x^(a+b)
	if(	Symbol(p1)		 == "^" &&
		Symbol(p2)		 == "^" &&
		p1->pred1() == p2->pred1() //same base 
   	)
   	{
  		NODE *x = p1->pred1();
  		NODE *a = p1->pred2();
  		NODE *b = p2->pred2();
  	
  		return &( (*x)^( *a + *b) );
   	}
	// x*(x^b) -> x^(1+b)   
	if(	Symbol(p2)	 == "^" &&
		p1		 == p2->pred1() //same base 
   	)
   	{	
  	return &( (*p1)^( 1.0 + *p2->pred2()) );
   	}
	// (x^b)*x -> x^(1+b)
	if(	Symbol(p1)	 == "^" &&
		p2		 == p1->pred1() //same base 
   	)
   	{	
  	return &( (*p2)^( 1.0 + *p1->pred2()) );
   	}
}

return 0;
}
	
NODE* DAGRewrite::FactorSmallestExp(DAG &G, const string &nodesym, LIST<NODE*> *predList){


}


NODE* DAGRewrite::RightAssoc(DAG &G, const string &nodesym, LIST<NODE*> *predList){

	/*

	(a+b)+c -> a+(b+c)
	(a*b)*c -> a*(b*c)

	Result: Left predecessor of a addition nodes are \emph{not} addition nodes.


	*/
	
	

	if (nodesym == "+" && Symbol( (*predList)[0] ) == "+"){ 

		NODE *a = (*predList)[0] -> pred1();
		NODE *b = (*predList)[0] -> pred2();
		NODE *c = (*predList)[1] ;
		return &( *a + (*b + *c) );

 	}
	if (nodesym == "*" && Symbol( (*predList)[0] ) == "*"){ 

		NODE *a = (*predList)[0] -> pred1();
		NODE *b = (*predList)[0] -> pred2();
		NODE *c = (*predList)[1] ;
		return &( *a * (*b * *c) );

 	}
 	
 	
 	
return 0;
}





NODE* DAGRewrite::Constant(DAG &G, const string &nodesym, LIST<NODE*> *predList){
// Binary operators
if(	(predList->Size() == 2) && 
	(isConstant((*predList)[0]) || isConstant((*predList)[1]))  ){
      
    	// 0+n -> n
    	// n+0 -> n
    	if(nodesym == "+"){
   
      		if( isConstantAndZero((*predList)[0]) ){ return (*predList)[1];  }
      		if( isConstantAndZero((*predList)[1]) ){ return (*predList)[0];  }
	}
    	// 0-n -> -n
    	// n-0 ->  n   
    	else if(nodesym == "-"){
   
        	if( isConstantAndZero((*predList)[1]) ){ return (*predList)[0];  }
        	if( isConstantAndZero((*predList)[0]) ){ return &(- *(*predList)[1] ); }
	}
	// 1*n     -> n
	// n*1     -> n
	// (-1)*n  -> -n  -OPTIONAL
        // n*(-1)  -> -n  -OPTIONAL	
	else if(nodesym == "*"){
      
      		if( isConstantAndOne((*predList)[0]) ){ return (*predList)[1];  }
      		if( isConstantAndOne((*predList)[1]) ){ return (*predList)[0];  }
       		if( isConstantAndMinusOne((*predList)[0]) ){ NODE* n = (*predList)[1]; return &(-(*n));  }
      		if( isConstantAndMinusOne((*predList)[1]) ){ NODE* n = (*predList)[0]; return &(-(*n));  }
	}
    	// n/1    ->  n
    	// n/(-1) -> -n  -OPTIONAL 
    	else if (nodesym == "/"){
  
      		if( isConstantAndOne((*predList)[1]) ){ return (*predList)[0];  }
      		if( isConstantAndMinusOne((*predList)[1]) ){ NODE* n = (*predList)[0]; return &(-(*n));  }
	}
    	// n^1    ->  n
    	// 1^n	  ->  1
	else if (nodesym == "^"){
   
      		if( isConstantAndOne((*predList)[1]) ){ return (*predList)[0];  }
      		if( isConstantAndOne((*predList)[0]) ){ return (*predList)[0];  }
      	}
}

return 0;
}

 
NODE* DAGRewrite::Constant1(DAG &G, const string &nodesym, LIST<NODE*> *predList){
	
	


	bool bin_op = 0;
	NODE *p1, *p2;  
	
	if ( predList->Size() == 2 ){
		bin_op = 1; 
		p1 = (*predList)[0];
		p2 = (*predList)[1]; 
	}
   
   
   
if (  bin_op && ( isConstant(p1) || isConstant(p2) )  )   {
    
	if (nodesym == "*"){
  	//
  	// n*0 -> 0
  	// 0*n -> 0
  	//
  	// Not true if the range of n is unbounded
		if( isConstantAndZero(p1) ) return p1;
		if( isConstantAndZero(p2) ) return p2;
	  }

	if(nodesym == "/"){
 	//
 	// 0/n -> 0
 	//
 	// Not true if the range of n encloses zero
  	if( isConstantAndZero(p1) ){ return p1; }
  	//return {G.ConstructNODE( interval(0.0) ); }

 	}
 


 	if(nodesym == "^"){
	//	
   	// 0^n -> 0  
  	//
  	// Not true if the range of n encloses zero
  	if( isConstantAndZero(p1) ) {return p1;}
  	//{ return G.ConstructNODE( interval(0.0) ); }
   	//
   	// n^0 -> 1
   	//
   	// Not true if the range of n is unbounded
  	if( isConstantAndZero(p2) ){ return G.ConstructNODE( interval(1.0) ); }
  	
  	}

}

return 0; 
 
}



NODE* DAGRewrite::Subtraction(DAG &G, const string &nodesym, LIST<NODE*> *predList){


	// -(-n) -> n
  
	if (nodesym == "u-"){   
		if( Symbol( (*predList)[0] ) == "u-" )
		{ return (*predList)[0] -> pred1(); }
    	}
   

	// x+(-y) -> x-y
	// (-x)+y -> y-x
    
	if (nodesym == "+"){
	
		if( Symbol( (*predList)[1] ) == "u-"){ 
			NODE* x =(*predList)[0];
			NODE* y =(*predList)[1]->pred1();
			return &( *x - *y  ); 
      		}
		if( Symbol( (*predList)[0] ) == "u-"){ 
			NODE* y =(*predList)[1];
			NODE* x =(*predList)[0]->pred1();
			return &( *y - *x  ); 
		}
	} 

   
   	
	// x-(-y) -> x+y
	
	if (nodesym =="-"){
   
		if( Symbol( (*predList)[1] )=="u-"){ 
			NODE* x =(*predList)[0];
			NODE* y =(*predList)[1]->pred1();
			return &( *x + *y  ); 
		}
      	} 
    
   
   
	// (-x)-y -> -(x+y)
    
	if (nodesym =="-"){
   
		if( Symbol( (*predList)[0] )=="u-"){ 
			NODE* y =(*predList)[1];
			NODE* x =(*predList)[0]->pred1();
			return &( -(*x + *y)  ); 
		}
	} 
    

	// (-x)*y, x*(-y) -> -(x*y)
    
	if (nodesym =="*"){
   
		if( Symbol( (*predList)[0] )=="u-") { 	
			NODE* x =(*predList)[0]->pred1();
			NODE* y =(*predList)[1];
			return &( -(*x * *y)  ); 
		}
      		if( Symbol( (*predList)[1] )=="u-") { 
			NODE* x =(*predList)[1]->pred1();
			NODE* y =(*predList)[0];
			return &( -(*x * *y)  ); 
		}
      
      }


    
	// (-x)/y, x/(-y) -> -(x/y)
	
	if (nodesym =="/"){
   
		if( Symbol( (*predList)[0] )=="u-") { 
			NODE* x =(*predList)[0]->pred1();
			NODE* y =(*predList)[1];
			return &( -(*x / *y)  ); 
      		}
      		if( Symbol( (*predList)[1] )=="u-") { 
			NODE* y =(*predList)[1]->pred1();
			NODE* x =(*predList)[0];
			return &( -(*x / *y)  ); 
     		}
      
      
    	}
    

return 0;
}



NODE* DAGRewrite::Identities(DAG &G, const string &nodesym, LIST<NODE*> *predList){

	// exp(log(n)) -> n
	// log(exp(n)) -> n
	if (nodesym=="exp" && Symbol((*predList)[0])== "log")		
		{return ((*predList)[0])->pred1();}  	

	if (nodesym=="log" && Symbol((*predList)[0])=="exp")	
		{return  ((*predList)[0]) -> pred1(); }
  	
  	// y/exp(x) -> y*exp(-x)
	if (nodesym == "/" && Symbol((*predList)[1])=="exp" ){
		return &( *(*predList)[0] * exp( - *( (*predList)[1] -> pred1() ) ) );
	}



	// exp(x+y) -> exp(x)*exp(y);
	if (nodesym == "exp" && Symbol( (*predList)[0] ) == "+" ){

		NODE *x = (*predList)[0] -> pred1();
		NODE *y = (*predList)[0] -> pred2();

		return &( exp(*x) * exp(*y) );
	}


	// exp(x)^y -> exp(x*y)

	if (nodesym == "^" && Symbol( (*predList)[0] ) == "exp" ){
		return &( exp(  *( (*predList)[0]->pred1() ) *  *( (*predList)[1]  ) ) );
	}




if (    nodesym                  == "*" && 
	Symbol( (*predList)[0] ) == "^" &&
	Symbol( (*predList)[1] ) == "^" &&
	(*predList)[0] -> pred1() == (*predList)[1] -> pred1()    ){
	
	NODE *p1 =(*predList)[0];
	NODE *p2 =(*predList)[1];		
  
  	return &(  ( *(p1->pred1())  ) ^ 
		(    *(p1 -> pred2())  +  *(p2 -> pred2())  )  ); 
	
	}
if (    nodesym                  == "/" && 
	Symbol( (*predList)[0] ) == "^" &&
	Symbol( (*predList)[1] ) == "^" &&
	(*predList)[0] -> pred1() == (*predList)[1] -> pred1()
   ){
	NODE *p1 =(*predList)[0];
	NODE *p2 =(*predList)[1];		
  
  	return &(  ( *(p1->pred1())  ) ^ 
          	(    *(p1 -> pred2())  -  *(p2 -> pred2())  )  ); 
    	}


	// (x^a)^b -> x^(a*b)

	if(	nodesym=="^" && 
		Symbol((*predList)[0])=="^"){

		NODE* n = (*predList)[0];
		return & ( ( *n->pred1() ) ^ ( *n->pred2() *  *(*predList)[1])  );
	}



	//a*(b*c) -> (a*b)*c if a and b are constants
	//	  -> (a*c)*b if a and c are constants
	if(	nodesym=="*" && 
		Symbol((*predList)[1])=="*"){
	
		NODE *a = (*predList)[0];
		NODE *b = (*predList)[1]->pred1();
		NODE *c = (*predList)[1]->pred2();
	
		if(isConstant(a)){		
			if(isConstant(b)){
				return &(    ((*a)*(*b))  *  (*c)   ); 
			}
			else if(isConstant(c)){
				return &(   ((*a)*(*c))  *  (*b)   ); 
			}
		}  
}




return 0;
}


NODE* DAGRewrite::FindCommon(LIST<NODE*> &L1, LIST<NODE*> &L2){

	for(int i=0; i<=L1.Size()-1; i++){
		for(int j=0; j<=L2.Size()-1; j++){
   
      			if ( L1[i]==L2[j] ){ return L1[i]; }
    		}
  	}
	return 0;
}


NODE* DAGRewrite::The_Other_Predecessor(NODE *p, NODE*c){
  
	// p = binary node
	// c = one of the predecessors
	// return the other predecessor 
 
	if (c == p->pred1()) 
		return p->pred2(); 
	else 
		return p->pred1();
	}


NODE* DAGRewrite::FirstLevelFactorization(DAG &G, const string &nodesym, LIST<NODE*> *predList){
  
  
  //a*x + b*x  -> (a+b)*x
  //a*x - b*x  -> (a-b)*x
    
	if (nodesym =="+" || nodesym =="-"){
   		if( Symbol((*predList)[0])=="*" &&  Symbol((*predList)[1])=="*"){ 
     
      			NODE* m1 = (*predList)[0]; 
      			NODE* m2 = (*predList)[1]; 
      
      			//if m1 and m2 has a common factor x
      			NODE *x = FindCommon(*(m1->Pred()),*(m2->Pred()));
			if (x != 0){ 
	  			NODE *a = The_Other_Predecessor(m1,x);
	  			NODE *b = The_Other_Predecessor(m2,x);
	      
	  			if (nodesym == "+") 
	    				return &( (*a + *b) * *x  );
	  			else 
	    				return &( (*a - *b) * *x  );
			} 
    		}
    
  	}

   
  
  //a/x + b/x  -> (a+b)/x
  //a/x - b/x  -> (a-b)/x
  
	if (nodesym =="+" || nodesym =="-"){      
    		if( Symbol((*predList)[0])=="/" &&  Symbol((*predList)[1])=="/"){ 
     
      			NODE* m1 = (*predList)[0]; 
      			NODE* m2 = (*predList)[1]; 
      			//if m1 and m2 has a common divisor x
      			NODE *x = FindCommon(*(m1->Pred()),*(m2->Pred()));
      
			if (x != 0){ 
	  			NODE *a = The_Other_Predecessor(m1,x);
	  			NODE *b = The_Other_Predecessor(m2,x);
	      
	  			if (nodesym == "+") 
	    				return &( (*a + *b) / *x  );
	  			else 
	    				return &( (*a - *b) / *x  );
			} 
    		}
    
  	}
cout<<endl<<"end";
return 0;
}


NODE* DAGRewrite::SinglePred(DAG &G, const string &nodesym, LIST<NODE*> *predList){

  	/*
  	A binary operator with one predecessor is rewritten to canonical form.
  
  	n+n  -> 2*n  
        n-n  -> 0
        n*n  -> n^2
        n/n ->  1
   	*/
 
	if (  predList->Size()==2  &&  (*predList)[0]==(*predList)[1] ){
		NODE *n = (*predList)[0];
    		if      (nodesym == "+"){ return &( 2.0 * (*n) );  } 
    		else if (nodesym == "-"){ return G.ConstructNODE( interval(0.0) );  }
    		else if (nodesym == "*"){ return &( ( *n )^2.0 );  }
    		else if (nodesym == "/"){ return G.ConstructNODE( interval(1.0) );  }
    	}
return 0; //IE, MAKE NEW NODE
}
 


NODE* DAGRewrite::ConstFold(DAG &G, const string &nodesym, LIST<NODE*> *predList){
  
	interval r;
	bool Folded = 0;

	// Unary operators 
	if (  predList->Size()==1 && isConstant((*predList)[0])  ){

	NODE* p1 = (*predList)[0];
	if      (nodesym == "u-") 	{ r = -(p1->range);   Folded = 1;}
	else if (nodesym == "exp")	{ r = exp(p1->range); Folded = 1;}
	else if (nodesym == "log")	{ r = log(p1->range); Folded = 1;}
	}
  
	
	//Binary operators 

	else if ( predList->Size() == 2 && 
	  	isConstant( (*predList)[0] ) && 
	  	isConstant( (*predList)[1] )  
	  	){

		NODE* p1 = (*predList)[0];
  		NODE* p2 = (*predList)[1];
  		if      (nodesym == "+"){ r = (p1->range)+(p2->range); Folded = 1;}
  		else if (nodesym == "-"){ r = (p1->range)-(p2->range); Folded = 1;}
  		else if (nodesym == "*"){ r = (p1->range)*(p2->range); Folded = 1;}
  		else if (nodesym == "/"){ r = (p1->range)/(p2->range); Folded = 1;}
  		else if (nodesym == "^"){ r = (p1->range)^(p2->range); Folded = 1;}
  
  
 	}
	if ( Folded ){ return G.ConstructNODE(r); }

return 0;
}






NODE* DAGRewrite::Simplify(DAG &G, const string &nodesym, LIST<NODE*> *predList){
  

  NODE *s;
  s = Constant(G, nodesym,predList);
  if (s!=0){ cout<<endl<<"DAGRewrite::Constant"; return s;}

  s = SinglePred(G, nodesym, predList);
  if(s!=0){ cout<<endl<<"DAGRewrite::SinglePred"; return s; }
  
  //s = ConstFold(G, nodesym, predList);
  if(s!=0){ cout<<endl<<"DAGRewrite::ConstFold"; return s; }

  s = Constant1(G, nodesym, predList);
  if(s!=0){ cout<<endl<<"DAGRewrite::Constant1";return s;}

  s= Subtraction(G, nodesym, predList);
  if(s!=0){cout<<endl<<"DAGRewrite::Subtraction";return s;}

  s = FirstLevelFactorization(G, nodesym, predList);
  if(s!=0){cout<<endl<<"DAGRewrite::FirstLevelFactorization"; return s;}
  
  //s = Identities(G, nodesym, predList);
  if(s!=0){ cout<<endl<<"DAGRewrite::Identities"; return s; }
  
  s = RightAssoc(G, nodesym, predList);
  if(s!=0){ cout<<endl<<"DAGRewrite::RightAssoc"; return s;}
  
return 0;
}


void DAGRewrite::Info(){}







































bool isMulPow_xa_Pow_xb (NODE *n, NODE *x, NODE *a, NODE *b){


// if n = (x^a) * (x^b), return x,a,b.



if (    Symbol(n)                == "*" && 
	Symbol(n->pred1())	 == "^" &&
	Symbol(n->pred2())	 == "^" &&
	n->pred1()-> pred1() == n->pred2()-> pred1()
   )
   
   {
  	x = n->pred1()->pred1();
  	a = n->pred1()->pred2();
  	b = n->pred2()->pred2();
  	
  	return 1;
   }


return 0;
}

NODE* Pow_xa_Mul_Pow_xb(DAG &G, const string &nodesym, LIST<NODE*> *predList){


if(nodesym == "*"){
	
	NODE* p1 = (*predList)[0];
	NODE* p2 = (*predList)[1];
	
	// (x^a) * (x^b) -> x^(a+b)
	if(	Symbol(p1)		 == "^" &&
		Symbol(p2)		 == "^" &&
		p1->pred1() == p2->pred1() //same base 
   	)
   	{
  		NODE *x = p1->pred1();
  		NODE *a = p1->pred2();
  		NODE *b = p2->pred2();
  	
  		return &( (*x)^( *a + *b) );
   	}
	// x*(x^b) -> x^(1+b)   
	if(	Symbol(p2)	 == "^" &&
		p1		 == p2->pred1() //same base 
   	)
   	{	
  	return &( (*p1)^( 1.0 + *p2->pred2()) );
   	}
	// (x^b)*x -> x^(1+b)
	if(	Symbol(p1)	 == "^" &&
		p2		 == p1->pred1() //same base 
   	)
   	{	
  	return &( (*p2)^( 1.0 + *p1->pred2()) );
   	}
}

return 0;
}



//NODE* RewriteFactorSmallestExp(DAG &G, sym nodesym, LIST<NODE*> *predList){
NODE* RewriteFactorSmallestExp(DAG &G, const string &nodesym, LIST<NODE*> *predList){

//cout<<endl<<"RewriteFactorSmallestExp";

// (x^a) + (x^b) -> (x^a)(1+x^(b-a))  if a=<b
//		 -> (x^b)(x^(a-b) +1) if a>b 

}



NODE* RewriteRightAssoc(DAG &G, const string &nodesym, LIST<NODE*> *predList){

cout<<"9";
	/*

	(a+b)+c -> a+(b+c)
	(a*b)*c -> a*(b*c)

	Result: Left predecessor of a addition nodes are \emph{not} addition nodes.


	*/
	
	

	if (nodesym == "+" && Symbol( (*predList)[0] ) == "+"){ 

		NODE *a = (*predList)[0] -> pred1();
		NODE *b = (*predList)[0] -> pred2();
		NODE *c = (*predList)[1] ;
		return &( *a + (*b + *c) );

 	}
	if (nodesym == "*" && Symbol( (*predList)[0] ) == "*"){ 

		NODE *a = (*predList)[0] -> pred1();
		NODE *b = (*predList)[0] -> pred2();
		NODE *c = (*predList)[1] ;
		return &( *a * (*b * *c) );

 	}
return 0;
}








NODE* RewriteConstant(DAG &G, const string &nodesym, LIST<NODE*> *predList){


  
  if (  predList->Size() == 1 && ( isConstant( (*predList)[0] ) ) )
  {
  
	if (nodesym == "exp"   &&   isConstantAndZero( (*predList)[0] ) )
			
		{ return G.ConstructNODE( interval(1.0) ); }
  	
  	if (nodesym == "log"   &&   isConstantAndOne( (*predList)[0] ) )
			
		{ return G.ConstructNODE( interval(0.0) ); }
  	
  }
  


  // Binary operators
  
  if (  predList->Size() == 2 && 
  	( isConstant( (*predList)[0] ) || isConstant( (*predList)[1] ) ) )
  	
  	
  {
  	
  	    
      
    	// 0+n -> n
    	// n+0 -> n
    
    	if (nodesym == "+"){
   
      		if( isConstantAndZero( (*predList)[0] ) ){ return (*predList)[1];  }
      		if( isConstantAndZero( (*predList)[1] ) ){ return (*predList)[0];  }
    
    	}
    
    	// 0-n -> -n
    	// n-0 ->  n
    
    	else if (nodesym == "-"){
   
        	if( isConstantAndZero((*predList)[1]) ){ return (*predList)[0];  }
        	if( isConstantAndZero((*predList)[0]) ){ return &(- *(*predList)[1] );}
	        
 
	}

	
   
	// 1*n     -> n
	// n*1     -> n
	// (-1)*n  -> -n  -OPTIONAL
        // n*(-1)  -> -n  -OPTIONAL
	
	else if (nodesym == "*"){
      
      		if( isConstantAndOne((*predList)[0]) ){ return (*predList)[1];  }
      		if( isConstantAndOne((*predList)[1]) ){ return (*predList)[0];  }

       		if( isConstantAndMinusOne((*predList)[0]) ){ NODE* n = (*predList)[1]; return &(-(*n));  }
      		if( isConstantAndMinusOne((*predList)[1]) ){ NODE* n = (*predList)[0]; return &(-(*n));  }
      
      
         
    	}
   
    
    	// n/1    ->  n
    	// n/(-1) -> -n  -OPTIONAL 
    
    	else if (nodesym == "/"){
   
      		if( isConstantAndOne((*predList)[1]) ){ return (*predList)[0];  }
      		if( isConstantAndMinusOne((*predList)[1]) ){ NODE* n = (*predList)[0]; return &(-(*n));  }
    

    	}
    	
    	
    	// n^1    ->  n
    	// 1^n	  ->  1
    		
    	else if (nodesym == "^"){
   
      		if( isConstantAndOne((*predList)[1]) ){ return (*predList)[0];  }
      		if( isConstantAndOne((*predList)[0]) ){ return (*predList)[0];  }
      	
    	}
  
  
  }
  
//cout<<" NA";
return 0; 
}

 
//NODE* RewriteConstant1(DAG &G, sym nodesym, LIST<NODE*> *predList){
NODE* RewriteConstant1(DAG &G, const string &nodesym, LIST<NODE*> *predList){


	


	bool bin_op = 0;
	NODE *p1, *p2;  
	
	if ( predList->Size() == 2 ){
		bin_op = 1; 
		p1 = (*predList)[0];
		p2 = (*predList)[1]; 
	}
   
   
   
if (  bin_op && ( isConstant(p1) || isConstant(p2) )  )   {
    
	if (nodesym == "*"){
  	//
  	// n*0 -> 0
  	// 0*n -> 0
  	//
  	// Not true if the range of n is unbounded
		if( isConstantAndZero(p1) ) return p1;
		if( isConstantAndZero(p2) ) return p2;
	  }

	if(nodesym == "/"){
 	//
 	// 0/n -> 0
 	//
 	// Not true if the range of n encloses zero
  	if( isConstantAndZero(p1) ){ return p1; }
  	//return {G.ConstructNODE( interval(0.0) ); }

 	}
 


 	if(nodesym == "^"){
	//	
   	// 0^n -> 0  
  	//
  	// Not true if the range of n encloses zero
  	if( isConstantAndZero(p1) ) {return p1;}
  	//{ return G.ConstructNODE( interval(0.0) ); }
   	//
   	// n^0 -> 1
   	//
   	// Not true if the range of n is unbounded
  	if( isConstantAndZero(p2) ){ return G.ConstructNODE( interval(1.0) ); }
  	
  	}

}
//cout<<" NA";
return 0; 
 
 }




//NODE* RewriteSubtraction(DAG &G, sym nodesym, LIST<NODE*> *predList){
NODE* RewriteSubtraction(DAG &G, const string &nodesym, LIST<NODE*> *predList){

//cout<<endl<<"Rewrite_Subtraction";
  



	// -(-n) -> n
  
	if (nodesym == "u-"){   
		if( Symbol( (*predList)[0] ) == "u-" )
		{ return (*predList)[0] -> pred1(); }
    	}
   

	// x+(-y) -> x-y
	// (-x)+y -> y-x
    
	if (nodesym == "+"){
	
		if( Symbol( (*predList)[1] ) == "u-"){ 
			NODE* x =(*predList)[0];
			NODE* y =(*predList)[1]->pred1();
			return &( *x - *y  ); 
      		}
		if( Symbol( (*predList)[0] ) == "u-"){ 
			NODE* y =(*predList)[1];
			NODE* x =(*predList)[0]->pred1();
			return &( *y - *x  ); 
		}
	} 

   
   	
	// x-(-y) -> x+y
	
	if (nodesym =="-"){
   
		if( Symbol( (*predList)[1] )=="u-"){ 
			NODE* x =(*predList)[0];
			NODE* y =(*predList)[1]->pred1();
			return &( *x + *y  ); 
		}
      	} 
    
   
   
	// (-x)-y -> -(x+y)
    
	if (nodesym =="-"){
   
		if( Symbol( (*predList)[0] )=="u-"){ 
			NODE* y =(*predList)[1];
			NODE* x =(*predList)[0]->pred1();
			return &( -(*x + *y)  ); 
		}
	} 
    

	// (-x)*y, x*(-y) -> -(x*y)
    
	if (nodesym =="*"){
   
		if( Symbol( (*predList)[0] )=="u-") { 	
			NODE* x =(*predList)[0]->pred1();
			NODE* y =(*predList)[1];
			return &( -(*x * *y)  ); 
		}
      		if( Symbol( (*predList)[1] )=="u-") { 
			NODE* x =(*predList)[1]->pred1();
			NODE* y =(*predList)[0];
			return &( -(*x * *y)  ); 
		}
      
      }


    
	// (-x)/y, x/(-y) -> -(x/y)
	
	if (nodesym =="/"){
   
		if( Symbol( (*predList)[0] )=="u-") { 
			NODE* x =(*predList)[0]->pred1();
			NODE* y =(*predList)[1];
			return &( -(*x / *y)  ); 
      		}
      		if( Symbol( (*predList)[1] )=="u-") { 
			NODE* y =(*predList)[1]->pred1();
			NODE* x =(*predList)[0];
			return &( -(*x / *y)  ); 
     		}
      
      
    	}
    






//cout<<" NA";  
return 0;
}



NODE* RewriteIdentities(DAG &G, const string &nodesym, LIST<NODE*> *predList){


if (nodesym == "exp"   &&   Symbol( (*predList)[0] ) == "log")
			
		{ return ((*predList)[0]) -> pred1(); }
  	


if (nodesym == "log"   &&   Symbol( (*predList)[0] ) == "exp" )
			
		{ return  ((*predList)[0]) -> pred1(); }
  	


// y/exp(x) -> y*exp(-x)


if (nodesym == "/" && Symbol( (*predList)[1] ) == "exp" ){

return &( *(*predList)[0] * exp( - *( (*predList)[1] -> pred1() ) ) );

}


// exp(x+y) -> exp(x)*exp(y);


if (nodesym == "exp" && Symbol( (*predList)[0] ) == "+" ){

NODE *x = (*predList)[0] -> pred1();
NODE *y = (*predList)[0] -> pred2();

return &( exp(*x) * exp(*y) );


}




if (nodesym == "^" && Symbol( (*predList)[0] ) == "exp" ){

return &( exp(  *( (*predList)[0]->pred1() ) *  *( (*predList)[1]  ) ) );

}

if (    nodesym                  == "*" && 
	Symbol( (*predList)[0] ) == "^" &&
	Symbol( (*predList)[1] ) == "^" &&
	(*predList)[0] -> pred1() == (*predList)[1] -> pred1()
   )
   	
	{
	
	
	NODE *p1 =(*predList)[0];
	NODE *p2 =(*predList)[1];		
  
  	return &(  ( *(p1->pred1())  ) ^ 
          (    *(p1 -> pred2())  +  *(p2 -> pred2())  )  ); 
	
	}
if (    nodesym                  == "/" && 
	Symbol( (*predList)[0] ) == "^" &&
	Symbol( (*predList)[1] ) == "^" &&
	(*predList)[0] -> pred1() == (*predList)[1] -> pred1()
   )
   	
	{
	
	
	NODE *p1 =(*predList)[0];
	NODE *p2 =(*predList)[1];		
  
  	return &(  ( *(p1->pred1())  ) ^ 
          (    *(p1 -> pred2())  -  *(p2 -> pred2())  )  ); 
	
	}

// (x^a)^b -> x^(a*b)

if (    nodesym                  == "^" && 
	Symbol( (*predList)[0] ) == "^" )
{

NODE* n = (*predList)[0];

return & ( ( *n->pred1() ) ^ ( *n->pred2() *  *(*predList)[1])  );

}





// THE FOLLOWING ONLY FOR z^z






//a*(b*c) -> (a*b)*c if a and b are constants
//	  -> (a*c)*b if a and c are constants
if(	nodesym=="*" && 
	Symbol((*predList)[1])=="*"){
	
	NODE *a = (*predList)[0];
	NODE *b = (*predList)[1]->pred1();
	NODE *c = (*predList)[1]->pred2();
	
	if(isConstant(a)){
		
		if(isConstant(b)){
			return &(    ((*a)*(*b))  *  (*c)   ); 
		}
		else if(isConstant(c)){
			return &(   ((*a)*(*c))  *  (*b)   ); 
		}
	}  
}


if(nodesym == "+"){
	NODE *p1 = (*predList)[0];
	NODE *p2 = (*predList)[1];

	//(a-c1)+c2 -> a+(c2-c1) if c1,c2 are constants
	if( Symbol(p1)=="-" ){
		if(isConstant(p1->pred2()) && isConstant(p2)){ return &( *(p1->pred1()) + (*p2 - *(p1->pred2())) ); }
	
	} 
}
	
if(nodesym == "-"){
	NODE *p1 = (*predList)[0];
	NODE *p2 = (*predList)[1];

	//(a-c1)-c2 -> a-(c1+c2) if c1,c2 are constants
	if( Symbol(p1)=="-" ){
		if(isConstant(p1->pred2()) && isConstant(p2)){ return &( *(p1->pred1()) - (*p2 + *(p1->pred2())) ); }
	
	} 
}


return 0;
}




NODE* FindCommon(LIST<NODE*> &L1, LIST<NODE*> &L2){

  for(int i=0; i<=L1.Size()-1; i++){
    for(int j=0; j<=L2.Size()-1; j++){
   
      if ( L1[i]==L2[j] ){ return L1[i]; }

    }
  }

  return 0;
}


NODE* The_Other_Predecessor(NODE *p, NODE*c){
  
  // p = binary node
  // c = one of the predecessors
  // return the other predecessor 
 
  if (c == p->pred1()) 
    return p->pred2(); 
  else 
    return p->pred1();
}



NODE* Rewrite_FirstLevelFactorization(DAG &G, const string &nodesym, LIST<NODE*> *predList){

//cout<<endl<<"Rewrite_FirstLevelFactorization";
  
  //-----------------------
  //a*x + b*x  -> (a+b)*x
  //a*x - b*x  -> (a-b)*x
  //-----------------------
  cout<<"0";
  if (nodesym =="+" || nodesym =="-"){
   
	NODE* m1 = (*predList)[0]; 
	NODE* m2 = (*predList)[1]; 
      
	if( Symbol(m1)=="*" ){
		if (Symbol(m2)=="*"){ 
  			//a*x + b*x  -> (a+b)*x
  			//a*x - b*x  -> (a-b)*x
  			cout<<"1";
     			NODE *x = FindCommon(*(m1->Pred()),*(m2->Pred()));
      			cout<<"2";
      			if (x != 0){
      			cout<<"3"; 
	  			NODE *a = The_Other_Predecessor(m1,x);
	  			NODE *b = The_Other_Predecessor(m2,x);
	      cout<<"4";
	  			if (nodesym == "+") 
	    				return &( (*a + *b) * *x  );
	  			else 
	    				return &( (*a - *b) * *x  );
			} 
    		} 
  		else if( m1->pred1()==m2 ){
  			//x*a+x -> x(1+a)
  			if(nodesym=="+")
  				return &( *m2 *(1.0 + *(m1->pred2()) ) );
  			else
  				return &( *m2 *(1.0 - *(m1->pred2()) ) );
  		}
  		else if( m1->pred2()==m2){
  		//a*x+x->x(1+a)
  			
  			if(nodesym=="+")
  				return &( *m2 *( 1.0 + *(m1->pred1())) );
  			else
				return &( *m2 *( 1.0 - *(m1->pred1())) );
  			
  		}
  	}
  	else if(Symbol(m2)=="*"){
  
  		if( m1==m2->pred1()){
  			//x+x*b -> x(1+b)
  			if(nodesym=="+")
  				return &( *m1 *(1.0 + *(m2->pred2())) );
  			else
  				return &( *m1 *(1.0 - *(m2->pred2())) );
  		}
  		else if( m1==m2->pred2()){
  		//x+b*x ->x(1+b)
  			if(nodesym=="+")
  				return &(*m1 *( 1.0 + *(m2->pred1())));
  			else
  				return &(*m1 *( 1.0 - *(m2->pred1())));
  		}
  	}
  }

   
   
   
   
  //-----------------------
  //a/x + b/x  -> (a+b)/x
  //a/x - b/x  -> (a-b)/x
  //-----------------------
  if (nodesym =="+" || nodesym =="-"){
   
      
    if( Symbol( (*predList)[0] )=="/" &&  Symbol( (*predList)[1] )=="/"){ 
     
      NODE* m1 = (*predList)[0]; 
      NODE* m2 = (*predList)[1]; 
      
      //if m1 and m2 has a common divisor x
      NODE *x = FindCommon(*(m1->Pred()),*(m2->Pred()));
      
	if (x != 0){ 
	cout<<"7";
	  NODE *a = The_Other_Predecessor(m1,x);
	  NODE *b = The_Other_Predecessor(m2,x);
	      
	  if (nodesym == "+") 
	    return &( (*a + *b) / *x  );
	  else 
	    return &( (*a - *b) / *x  );
	} 
    }
    
  }





//cout<<" NA"; 
return 0;
}



//NODE* RewriteSinglePred(DAG &G, sym nodesym, LIST<NODE*> *predList){
  NODE* RewriteSinglePred(DAG &G, const string &nodesym, LIST<NODE*> *predList){
  
//cout<<endl<<"RewriteSinglePred";


  /*
  
  A binary operator with one predecessor is rewritten to canonical form.
  
  	n+n  -> 2*n  
        n-n  -> 0
        n*n  -> n^2
        n/n ->  1
   
  
*/
 
  
  
  
  if (  predList->Size() == 2  &&  (*predList)[0] == (*predList)[1] ){
    
    NODE *n = (*predList)[0];
    
    if      (nodesym == "+"){ return &( 2.0 * (*n) );  } 
    else if (nodesym == "-"){ return G.ConstructNODE( interval(0.0) );  }
    else if (nodesym == "*"){ return &( ( *n )^2.0 );  }
    else if (nodesym == "/"){ return G.ConstructNODE( interval(1.0) );  }
  
  
  }

//cout<<" NA";
return 0; //IE, MAKE NEW NODE
}
 



//NODE* RewriteConstFold(DAG &G, sym nodesym, LIST<NODE*> *predList){
  NODE* RewriteConstFold(DAG &G, const string &nodesym, LIST<NODE*> *predList){
  
  
  //cout << endl << "RewriteConstFold";


  interval r;
  bool Folded = 0;



// Unary operators 


 if (  predList->Size() == 1 && isConstant( (*predList)[0] )   )

{

  // COMPUTE NEW CONSTANT RANGE

  NODE* p1 = (*predList)[0];
  if      (nodesym == "u-") 	{ r = -(p1->range);   Folded = 1;}
  else if (nodesym == "exp")    { r = exp(p1->range); Folded = 1;}
  else if (nodesym == "log")    { r = log(p1->range); Folded = 1;}
  
  
}
  



//Binary operators 



else if ( predList->Size() == 2 && 
	  isConstant( (*predList)[0] ) && 
	  isConstant( (*predList)[1] )  
	  )

{

  // COMPUTE NEW CONSTANT RANGE

  NODE* p1 = (*predList)[0];
  NODE* p2 = (*predList)[1];
  if      (nodesym == "+"){ r = (p1->range)+(p2->range); Folded = 1;}
  else if (nodesym == "-"){ r = (p1->range)-(p2->range); Folded = 1;}
  else if (nodesym == "*"){ r = (p1->range)*(p2->range); Folded = 1;}
  else if (nodesym == "/"){ r = (p1->range)/(p2->range); Folded = 1;}
  else if (nodesym == "^"){ r = (p1->range)^(p2->range); Folded = 1;}
  
  
 }


if ( Folded ){ return G.ConstructNODE(r); }

//cout<<" NA";
return 0;
}








//NODE* Simplify(DAG &G, sym nodesym, LIST<NODE*> *predList){
  NODE* Simplify(DAG &G, const string &nodesym, LIST<NODE*> *predList){
  

  NODE *s;

  s = Pow_xa_Mul_Pow_xb(G, nodesym, predList);
  if (s!=0){ cout<<endl<<"RewritePow";return s;}

  s = RewriteConstant(G, nodesym,predList);
  if (s!=0){ cout<<endl<<"RewriteConstant";return s;}

  s = RewriteSinglePred(G, nodesym, predList);
  if(s!=0){ cout<<endl<<"RewriteSinglePred";return s; }
  
  s = RewriteConstFold(G, nodesym, predList);
  if(s!=0){cout<<endl<<"RewriteConstFold";return s;}

  s= RewriteConstant1(G, nodesym, predList);
  if(s!=0){cout<<endl<<"RewriteConstant1";return s;}

  s= RewriteSubtraction(G, nodesym, predList);
  if(s!=0){cout<<endl<<"RewriteSubtraction";return s;}

  s = Rewrite_FirstLevelFactorization(G, nodesym, predList);
  if(s!=0){cout<<endl<<"Rewrite_FirstLevelFactorization";return s;}
  
  s = RewriteIdentities(G, nodesym, predList);
  if(s!=0){cout<<endl<<"RewriteIdentities";return s;}
  
  s = RewriteRightAssoc(G, nodesym, predList);
  if(s!=0){cout<<endl<<"RewriteRightAssoc";return s;}
  

return 0;
}




void StandardReuseRewrite(){

	DAG::sEnableReuse = 1;
	DAG::sEnableRewrite = 1;
	VECTOR<DAG::RewriteFunction>RW(4);
	RW(1) = DAGRewrite::Constant;
	RW(2) = DAGRewrite::Constant1;
	RW(3) = DAGRewrite::SinglePred;
	RW(4) = DAGRewrite::ConstFold;
	DAG::sCurrDAG->RewriteSystem = RW;

}





