#include "DAG/DAGUtilities.h"
#include "Structs/Sort.h"
#include "Structs/Tools.h"
#include "DAG/NODE.h"
#include "Structs/symUtilities.h"





///////////////////////////////////////////////////////////////////////
// Initialization
////////////////////////////////////////////////////////////////////////

void Initialize(Filter &F,
		vNODE &t,		const VECTOR<interval> &r_t,
		vNODE &p1,	const VECTOR<interval> &r_p1,
		NODE &p2,	const interval& r_p2,
		NODE &p3,	const interval& r_p3){    
		
SetConstants(*F.G, F.V);
F.Write(t,r_t);
F.Write(p1,r_p1);
F.Write(p2,r_p2);
F.Write(p3,r_p3);
F.SweepForwardIO();
}


Derivative::Derivative(){}
Derivative::Derivative(DAG &aDAG){

	G = &aDAG;
	MakeConstant(zero, aDAG, interval(0.0), "0");
	MakeConstant(one, aDAG, interval(1.0), "1");	
	SetUsed(zero);SetUsed(one);
	
}


void Derivative::SetUsed(NODE *v){

		if( !In<NODE*>(v,Used) ){ Used.push(v); }
}
void Derivative::SetUsed(VECTOR<NODE*> &N){

	for(int i=1; i<=N.Size(); i++){

		if( !In<NODE*>(N(i),Used) ){ Used.push(N(i)); }
	}
}
void Derivative::RemoveUnused(){

		RemoveUnusedNodes(*G, Used);
		G->NodeEnumeration();
}


void Derivative::cInit(NODE* n){	
	
	cD = VECTOR<NODE>(G->Size());
	for(int i=1; i<=cD.Size(); i++){ cD(i) = *zero; }
	cD( n->NodePtr()->num ) = *one;
}




void Derivative::nInit(NODE* n){
	
	nD = VECTOR<interval>(G->Size());
	for(int i=1; i<=nD.Size(); i++){ nD(i) = interval(0.0); }
	nD( n->NodePtr()->num ) = interval(1.0);
}
void Derivative::cBackward(NODE *y){
  
  	cInit(y);
 	PATH = Path(*G,y);
 	for(int i=1; i<=PATH.Size(); i++){  PATH(i)->cBPD(cD);  }
}
void Derivative::cBackward(NODE *y, NODE *x){
  
  	
  	cInit(y);
 	PATH = Path(*G,x,y); Reverse<NODE*>(PATH);
	for(int i = 1; i<=PATH.Size(); i++){  PATH(i)->cBPD(cD);  }
}
void Derivative::cForward(NODE *x){
  	
  	cInit(x);
  	PATH = VECTOR<NODE*>(G->Size());
	ForwardOrdering(*G, PATH);
	for(int i = 1; i<=PATH.Size(); i++){  PATH(i)->cFPD(cD);  }
}
void Derivative::cForward(NODE *y, NODE *x){
 
	
  	cInit(x);
  	PATH = Path(*G,x,y);
	for(int i = 1; i<=PATH.Size(); i++){  PATH(i)->cFPD(cD);  }
}





void Derivative::nBackward(NODE *y){
  
  	//Numerical backward computation of dydp for all predecessors p of y
  	//Requires that nV is precomputed.
  	nInit(y);
 	PATH = Path(*G,y);
 	for(int i=1; i<=PATH.Size(); i++){  PATH(i)->BPD(nV,nD);  }
}
void Derivative::nBackward(NODE *y, NODE *x){
  
  	//Numerical backward computation of dydx 
  	nInit(y);
 	PATH = Path(*G,x,y); Reverse<NODE*>(PATH);
	for(int i = 1; i<=PATH.Size(); i++){  PATH(i)->BPD(nV,nD);  }
}
void Derivative::nForward(NODE *x){
  	
  	//Numerical forward computation of dydx for all successors y of x
  	nInit(x);
  	PATH = VECTOR<NODE*>(G->Size());
	ForwardOrdering(*G, PATH);
	for(int i = 1; i<=PATH.Size(); i++){  PATH(i)->FPD(nV,nD);  }
}
void Derivative::nForward(NODE *y, NODE *x){
 
	//Numerical forward computation of dydx
  	nInit(x);
  	PATH = Path(*G,x,y);
	for(int i = 1; i<=PATH.Size(); i++){  PATH(i)->FPD(nV,nD);  }
}



void Derivative::cForwardMixed(VECTOR<NODE*> &df,  NODE *f, const VECTOR<NODE*> &x){

 	for(int i=1; i<=x.Size(); i++){
 
		cForward(f,x(i));
		f = cD(f->num).NodePtr();
		df(i) = f;
 	}
}

void Derivative::cBackwardMixed(VECTOR<NODE*> &df,  NODE *f, const VECTOR<NODE*> &x){

 
 	for(int i=1; i<=x.Size(); i++){
 
		cBackward(f,x(i));
		f = cD(x(i)->num).NodePtr();
		df(i) = f;
 	}
}
void Derivative::cForwardDiff(VECTOR<NODE*> &z, NODE *y, NODE *x){
        	
         	z(1)=y;
        	for(int i=1; i<=z.Size()-1; i++){
        		cForward(z(i),x);
                	z(i+1) = cD(z(i)->num).NodePtr();
		}
		
}

void Derivative::cBackwardDiff(VECTOR<NODE*> &z, NODE *y, NODE *x){
        	

        	z(1)=y;
        	for(int i=1; i<=z.Size()-1; i++){
        		cBackward(z(i),x);
                	z(i+1) = cD(x->num).NodePtr();
		}
		
}



void Derivative::cProgressiveExpansion(	NODE &y,
					VECTOR<NODE> &t,
					VECTOR<NODE> &c,
					VECTOR<NODE> &x,
					NODE&(*g)(VECTOR<NODE>&) ){ 

	VECTOR<NODE> u(x.Size());				
	for(int i=1; i<=x.Size(); i++){
		
		u(i) = c(i) + t(i)*( x(i)-c(i) );				
	}	
	VECTOR<NODE> z(x.Size());
	for(int i=1; i<=x.Size(); i++){
		
		z(i)=x(i);
	}
	NODE v0,v1,v2,*e;
	v0 = g(x);
	for(int i=1; i<=x.Size(); i++){
		
		z(i)=c(i);
		v1=g(z);
		z(i)=u(i);
		v2=g(z);
		cBackward(v2.NodePtr(), u(i).NodePtr());
		v2 = v1 + cD(u(i).NodePtr()->num) * (x(i)-c(i));
		e=InsertEqualityNode(*G);
		MakeEqual(v0.NodePtr(),e);
		MakeEqual(v2.NodePtr(),e);
		if(i==1){y=*e;}
		z(i)=c(i);
		v0=v1;	
	}
}


bool isIn(NODE *n, VECTOR<NODE*> X){

// is n in X

	for(int i=1; i<=X.Size(); i++){ 
		if (X(i)==n) return 1; 
	}
	return 0;
}


bool isSubDAG(VECTOR<NODE*> &X, DAG &G){

return 0;
}

void getSubDAGSource(LIST<NODE*> &S, VECTOR<NODE*> &X, DAG &G){

}
void getSubDAGSink(LIST<NODE*> &S, VECTOR<NODE*> &X, DAG &G){

}


bool SearchAndIntersect(DAG &G, VECTOR<NODE*> &E, VECTOR<NODE*> &F);
 
bool isSource(NODE *n){
	return (  n!=0 && n->Pred()->empty()  );  
}
bool   isSink(NODE *n){
	return ( n!= 0 && n->Succ()->empty() );
}


bool   isIntermediate(NODE *n){
	return (   n!=0 && !(n->Pred()->empty()) && !( n->Succ()->empty() )   ); 
}

bool isConstant(NODE *n){
	NODE_con c ; 
	return ( typeid(*n) == typeid(c) ); 
}
bool isVariable(NODE *n){  
	NODE_var v; 
	return ( typeid(*n) == typeid(v) );  
}
  



bool isConstantAndZero(NODE *n)		{  return isConstant(n) && n->range == interval(0.0); } 
bool isConstantAndOne(NODE *n)		{  return isConstant(n) &&  n->range == interval(1.0) ; } 
bool isConstantAndMinusOne(NODE *n)	{  return isConstant(n) &&  n->range == interval(-1.0) ; } 
bool containsZero(NODE *n)		{  return (0.0 <= n->range); }
bool isThinConstant(NODE *n)		{  return isConstant(n) && inf(n->range) == sup(n->range) ; }




// Symbolic identifiers:

bool isUSub(NODE *n)		{return Symbol(n) == "u-";}
bool isAdd(NODE *n)		{return Symbol(n) == "+";}
bool isSub(NODE *n)		{return Symbol(n) == "-";}
bool isMul(NODE *n)		{return Symbol(n) == "*";}
bool isDiv(NODE *n)		{return Symbol(n) == "/";}
bool isExp(NODE *n)		{return Symbol(n) == "exp";}
bool isLog(NODE *n)		{return Symbol(n) == "log";}
bool isPow(NODE *n)		{return n!=0 && Symbol(n) == "^";}
bool isSum(NODE *n)		{return Symbol(n) == "sum";}
bool isProd(NODE *n)		{return Symbol(n) == "prod";}
bool isEquality(NODE *n)	{return Symbol(n) == "=";}
bool isInclusion(NODE *n)	{return Symbol(n) == "in";}
bool isIntersection(NODE *n)	{return Symbol(n) == "&";}

bool isLessOrEqual(NODE *n)	{return Symbol(n) == "<=";}

bool isAddition(NODE *n)	{return isAdd(n) || isSum(n);}
bool isMultiplication(NODE *n)  {return isMul(n) || isProd(n);}


bool isUnary(NODE *n)           {return (n->Pred()->Size()) == 1; }
bool isBinary(NODE *n)          {return (n->Pred()->Size()) == 2; } 
bool isk_ary(NODE *n)           {return (n->Pred()->Size()) >= 3; }


bool isUnaryOp(NODE *n)         {return isUSub(n) || 
					isExp(n)  ||
					isLog(n);
					}

bool isBinaryOp(NODE *n)         {return isAdd(n) || 
					isSub(n)  ||
					isMul(n)  || 
					isDiv(n)  ||
					isPow(n)  ||
					isLessOrEqual(n);
					}

bool isk_aryOp(NODE *n)		{return isSum(n)      ||
					isProd(n)     || 
					isEquality(n) ||
					isInclusion(n) || 
					isIntersection(n); 
					}





bool isArithmetic (NODE *n){
                   
	return (  n!=0 && 
		(Symbol(n) == "+" 	 ||
		 Symbol(n) == "-" 	 ||
                 Symbol(n) == "*" 	 ||
                 Symbol(n) == "/" 	 ||
                 Symbol(n) == "sum" ||
                 Symbol(n) == "prod" )   
               );
}


bool isCommutative(NODE *n){
                  
                  return ( n!=0 && 
                   	    (Symbol(n) == "+"    || 
                   	     Symbol(n) == "*"    ||
                   	     Symbol(n) == "sum"  ||
                   	     Symbol(n) == "prod" ||
                   	     Symbol(n) == "&"    ||
                   	     Symbol(n) == "=") 
                   	 );
}






bool areSymbolicallyEquivalent(NODE *n1, NODE* n2){

   



  
  
  if (n1 == 0 || n2 == 0) return 0;
  
  if ( !( Symbol(n1) == Symbol(n2) ) ) return 0;

  if ( !isCommutative(n1) ){
  
       
	LIST<NODE*> *L1 = n1->Pred();
	LIST<NODE*> *L2 = n2->Pred();
	if (L1->Size() != L2->Size()) return 0;
	L1->MoveTop();
	L2->MoveTop();
	while (!L1->eol())
		{ 
		if (  !(Symbol((L1->Data())) == Symbol((L2->Data())) ) ) return 0;
		L1->MoveDown();
		L2->MoveDown();
		 }
  }
  else{
	 // commutative
	
        LIST<string> *S1 ;
	LIST<string> *S2 ;
	
	LIST<NODE*> *L1 = n1->Pred();
	LIST<NODE*> *L2 = n2->Pred();
	if (L1->Size() != L2->Size()) return 0;
	
	L1->MoveTop();
	L2->MoveTop();
	while (!L1->eol())
		{ 
		S1->push( Symbol((L1->Data()))  );
		S2->push( Symbol((L2->Data()))  );
		L1->MoveDown();
		L2->MoveDown();
		 }
    	 return areEqual(S1, S2); 
    
  }
  
}











bool isMulAddition(NODE *n){ 
	
	return (  isMul(n) &&
		 (isAddition(n->pred1()) || isAddition(n->pred2()) )
	 	);
}
		 
bool isMulSub(NODE *n){ 
	
	return (  isMul(n) &&
		 (isSub(n->pred1()) || isSub(n->pred2()) )
	 	);
}




bool isInverted(NODE*n, NODE* &x){
	
	/*
	
	Identification of the multiplicative inverse:
	1) n = 1/x
	2) n = x^(-1)
	If n is any of these operators,
	then return node x.
	
	*/

if( isDiv(n) && isConstantAndOne(n->pred1()) )	
	{ x = n->pred2(); return 1;} 
if( isPow(n) && isConstantAndMinusOne(n->pred2()) )
	{ x = n->pred1(); return 1;}

return 0;
}





bool isNegated(NODE* n, NODE* &x){


	/*
	
	Identification of the three standard negation operators:
	1)	n = -x
	21)	n = (-1)*x  
	22)	n = x*(-1)
	3)	n = x/(-1)

	If n is any of these operators,
	then return the negated node x.
	
	*/



	if 	(isUSub(n))
		{       
			x = n->pred1(); return 1;
		}	
		
	else if	(isMul(n))
		{
		
		
		if 	( isConstantAndMinusOne( n->pred1() ))
			{ 
				x = n->pred2(); return 1;
			}
		else if ( isConstantAndMinusOne( n->pred2() ) )
			{ 
				x = n->pred1(); return 1;
			}
		
		}
	
	else if ( isDiv(n) && isConstantAndMinusOne( n->pred2() ) )
		{
			x = n->pred1(); return 1;
		}
	
	else return 0;
 
}



bool isNegatedAddition(NODE *n, NODE* &x){

	// If n is a negated addition,
	// then return the negated addition node x.

        
	return ( isNegated(n,x) && isAddition(x) );
}



bool isNegatedSubtraction(NODE *n, NODE* &x){

	// If n is a negated subtraction
	// then return the negated subtraction node x.

	return isNegated(n,x) && isSub(x);
}






NODE* Negation(NODE *n){ return &( -(*n) ); }

void NegateAndPush(LIST<NODE*> *A,  LIST<NODE*> *B){

	// Push on A all negations of nodes in B.
	
	B->MoveTop();
	while(!B->eol()){ 
		A->push( Negation( B->Data() ) ); 
		B->MoveDown();
	} 
}



void DeFactorization(LIST<NODE*> *S, NODE *v){

	NODE *x = 0;
	LIST<NODE*> *T = new LIST<NODE*>;
	T -> push(v);
	
	
	while ( !T->empty() ){
		
		v = T -> pop();
		
		if 	( isAddition(v) )  
			{ 
				Push<NODE*>(T, v->Pred()); 
			}
		
		else if ( isSub(v) )  
			{            
		  		T -> push( v->pred1() ); 
		  		T -> push( Negation(v->pred2()) );
			}
		
		else if ( isNegatedAddition(v,x) ) 
		        {
		        	NegateAndPush(T, x->Pred());
		        }
		
		else if ( isNegatedSubtraction(v,x) ) 
			{ 
		  		T -> push( Negation(x->pred1()) );
		  		T -> push( x->pred2() ); 
			}
		
		else S -> push(v);
	
	}
	
	delete T;

}









void GetSISNodes(DAG &G,
		LIST<NODE*> &SOURCE,
		LIST<NODE*> &INTERMEDIATE,
		LIST<NODE*> &SINK){
	
	
	LIST<NODE*> *L;
	G.GetNodeList(L);
	
	
 	int i = 1;
 	L->MoveTop();
 	while ( !L->eol() ){
                 
        	if 	( isSource(L->Data() ) )  	SOURCE.push(L->Data()) ;
        	else if ( isIntermediate(L->Data() ) )  INTERMEDIATE.push(L->Data()) ;
         	else if ( isSink(L->Data() ) )  	SINK.push(L->Data()) ;
         	
         	L->MoveDown();
 	}
 	
 
}



void GetSourceNodes(LIST<NODE*> &SOURCE, DAG &G){
	
	LIST<NODE*> *L;
	G.GetNodeList(L);
	
 	int i = 1;
 	L->MoveTop();
 	while ( !L->eol() ){
                 
        	if ( isSource(L->Data() ) )  SOURCE.push(L->Data()) ;
         	L->MoveDown();
 	}
}

void GetSourceNodes(LIST<NODE*> &SOURCE, NODE *x){

	
	NODE *n;
	LIST<NODE*> L, S;
	
	L.push(x); 
	while(!L.empty()){
		n = L.pop();
		if (isSource(n)) S.push(n);
		else Push<NODE*>(&L, n->Pred());

	}
	
	RemoveMultiples<NODE*>(SOURCE, S);
}

void GetSinkNodes(LIST<NODE*> &SINK, DAG &G){
	
	LIST<NODE*> *L;
	G.GetNodeList(L);
	
 	int i = 1;
 	L->MoveTop();
 	while ( !L->eol() ){
                 
        	if ( isSink(L->Data() ) )  SINK.push(L->Data()) ;
         	L->MoveDown();
 	}
}

void GetSymbolNodes(LIST<NODE*> &SYM, const string &NodeSymbol, DAG &G){

	LIST<NODE*> *L;
	G.GetNodeList(L);
	
 	int i = 1;
 	L->MoveTop();
 	while ( !L->eol() ){
                 
        	if ( Symbol(L->Data()) == NodeSymbol)  SYM.push(L->Data()) ;
         	L->MoveDown();
 	}
}

void NumberToPointer(VECTOR<NODE*> &N, DAG &G){


	LIST<NODE*> *L;
	G.GetNodeList(L);

	NODE *n;
	L->MoveTop();
	while(!L->eol()){
		n=L->Data();
		N(n->num)=n;
		L->MoveDown();
	}

}

void GetNodeNumbers(VECTOR<int> &ind, VECTOR<NODE> &n){
	
	for(int i=1; i<=n.Size(); i++){
		ind(i)=n(i).NodePtr()->num;
	}
}

void GetNodeNumbers(VECTOR<int> &ind, VECTOR<NODE*> &n){
	
	for(int i=1; i<=n.Size(); i++){
		ind(i)=n(i)->num;
	}
}

void GetNodeNumbers(VECTOR<int> &ind, vNODE &n){
	
	for(int i=1; i<=n.Size(); i++){
		ind(i)= n(i).NodePtr()->num;
	}
}



bool isSubGraphMatch( VECTOR<NODE*> &X, DAG &G, NODE *n, NODE *e){

/*
        
        A simple subgraph matching test. 
        
          	
	Input:
	G - DAG 
	n - node in G. 
	e - node from some other DAG
	The e subdag is tested for match with the n subdag.

	X - vector of nodes with size equal or larger to the size of subdag e, 
	X(i) = 0 all i.
	
	Output:
	Return value 1  - iff match :  Means that e is a subdag in n. 
	Return value 0  - otherwise :  Means that e does not match exactly to n in the 
			 		  strict graph matching sense (not in mathematical sense). 
			 		  Fox, the commutativity and associativity of + and * is
			 		  not taken into account. 
	 		  
	X -  The nonzero elemets of X contain the nodes of the n subdag
	     that correspond in a 1-1 way to the nodes of subdag e.
	
*/


LIST<NODE*> L;
L.push(e); 
L.push(n);
L.MoveTop();

while(!L.empty()){ 

e = L.pop();
n = L.pop();


	if( X(e->num) == 0 ){
	
		if(isIn(n,X)) return 0; // X must be injective 		
		X(e->num) = n;
		
	}
	else if ( X(e->num) != n ) { return 0; }// X must be single valued


	if( !isSource(e) ){
	
		if( (!(Symbol(e)==Symbol(n))) || isSource(n)  ) {return 0;}
	
	
	
                if (isUnaryOp(e)){
                	L.push(e->pred1());
                	L.push(n->pred1());
                }
                else if(isBinaryOp(e)){
                        L.push(e->pred1());
                	L.push(n->pred1());
                        L.push(e->pred2());
                	L.push(n->pred2());
                }	
	}
	// if (e is a source, do nothing)
	
}
return 1;
}






bool CopyDAG(VECTOR<NODE*> &X, DAG &H){

	/* 
	Input:
	X  - a vector of NODE*, a "complete" subDAG 
	     forwardly ordered: For each i, Pred(X(i)) IN { X(j): j<i }, 
	     the nodes X(i) are enumerated from 1 to X.Size(),
	
	Output:
	H - copy DAG of X 
	
	*/

	
        VECTOR<NODE*> C( X.Size() );
        setZero<NODE*>(C); 
	
	NODE *n;
	
	DAG::sCurrDAG = &H;
	DAG::sEnableReuse = 0;
	DAG::sEnableRewrite = 0;
	
	for(int i=1; i<=X.Size(); i++){

		if( isSource(X(i)) ){ 
		
			if (isConstant(X(i)) ){
			
				   C(i) = H.ConstructNODE( X(i)->range ); 
			}
			else{ 
				C(i) = new NODE_var();  
				H.InsertNode( C(i) ); 
			}	
 			
 			C(i)->num = X(i)->num; 
 			Symbol(C(i)) = Symbol(X(i));
 			 
 			
		}
		
		else if( isUnaryOp(X(i)) ){
			
				
			if( isUSub(X(i)) ){
				
				C(i) = &-( *C( X(i)->NodePtr()->pred1()->num) );
			}
			else if( isExp(X(i)) ){
				
				C(i) = &exp( *C( X(i)->NodePtr()->pred1()->num) );
			}
			else if( isLog(X(i)) ){
				
				C(i) = &log( *C( X(i)->NodePtr()->pred1()->num) );
			}
		
			C(i)->num = X(i)->num;  
			
						
		}
		
		else if ( isBinaryOp(X(i)) ){ 
		
		        if( isAdd(X(i)) ){
				
				C(i) = &( *C( X(i)->NodePtr()->pred1()->num ) + *C( X(i)->NodePtr()->pred2()->num ));
			}
			else if( isSub(X(i)) ){
			
				C(i) = &( *C( X(i)->NodePtr()->pred1()->num ) - *C( X(i)->NodePtr()->pred2()->num ));
				
			}
			else if( isMul(X(i)) ){
	
				C(i) = &( *C( X(i)->NodePtr()->pred1()->num ) * *C( X(i)->NodePtr()->pred2()->num ));
			}
			else if( isDiv(X(i)) ){
	
				C(i) = &( *C( X(i)->NodePtr()->pred1()->num ) / *C( X(i)->NodePtr()->pred2()->num ));
			}
			else if( isPow(X(i)) ){

				C(i) = &( *C( X(i)->NodePtr()->pred1()->num ) ^ *C( X(i)->NodePtr()->pred2()->num ));	
				
			}
			
			C(i)->num = X(i)->num; 
			
			
			
		}
		else if ( isk_aryOp(X(i)) ){ 
		
		 cout<<" NOT IMPLEMENTED";
		        if( isSum(X(i)) ){               		
			}
			else if( isProd(X(i)) ){
			}
			else if( isEquality(X(i)) ){
			}
			else if( isIntersection(X(i)) ){
			}
		}
		
	}
	return 0;
}




bool FindExpression(NODE *e, DAG &G, LIST<NODE*> &EXPR){

/*

 Finds every occurence of the expression e in the DAG G and pushes 
 them on EXPR.
 
 Implementation: Traverse G (in arbitrary order). 
 At each visited node n we test isSubGraphMatch(X,G,n,e). If true, 
 push n on EXPR. Continue to next node.
 
 
 */
 
 
 
 
NODE *n;
VECTOR<NODE*> Z( G.Size() ); 
			
LIST<NODE*> *L;
G.GetNodeList(L);

	L->MoveTop();
	while(!L->eol()){
		n = L->Data();
 		setZero<NODE*>(Z);
 		if ( isSubGraphMatch(Z,G,n,e) ) { EXPR.push(n); }
 		L->MoveDown(); 
	}

if ( EXPR.empty() ) return 0;
return 1;
}

bool FindExpression(NODE *e, DAG &G, LIST< VECTOR<NODE*> > &EXPR){

/*

 Traverse G (in arbitrary order). 
 At each visited node n we test isSubGraphMatch(X,G,n,e). If true, 
 push X on EXPR. Continue to next node.
 
 
 */
 
 
 
 
NODE *n;
VECTOR<NODE*> Z( G.Size() ); 
LIST<NODE*> *L;
G.GetNodeList(L);

	L->MoveTop();
	while(!L->eol()){
		n = L->Data();
 		setZero<NODE*>(Z);
 		if ( isSubGraphMatch(Z,G,n,e) ) { EXPR.push(Z); }
 		L->MoveDown(); 
	}

if ( EXPR.empty() ) return 0;
return 1;
}




int N_GPred(DAG &G, NODE *n){

	//Number of predecessors of n, including n.
		
	VECTOR<NODE*> PRED(G.Size());
	GPred(PRED,n);
	int count = 0;
	for(int i=1; i<=PRED.Size(); i++){ if(PRED(i)) count++; }
	return count;

}

void GPred(VECTOR<NODE*> &PRED, NODE *n){

	/*
	Global Predecessor set
	All predecessors of n, and all their predecessors...
	Initially,
	VECTOR<NODE*> PRED(G.Size());
	OUTPUT : PRED(i) !=0 -> PRED(i) is a global predecessor of n
	*/

	setZero<NODE*>(PRED);
	
	LIST<NODE*> L;
	L.push(n);
	while (!L.empty()) {
		n = L.pop();
		PRED(n->num) = n;
		if ( !isSource(n) ){ Push(&L, n->Pred()); }
	}


}

void GPred(VECTOR<NODE*> &PRED, LIST<NODE*> &L){

	/*
	All predecessors of all nodes in L, and all their predecessors...
	Warning: Destructive for L
	*/
	
	setZero<NODE*>(PRED);
	
	NODE *n;
	while (!L.empty()) {
		n = L.pop();
		PRED(n->num) = n;
		if ( !isSource(n) ){ Push(&L, n->Pred()); }
	}
}



void MakePredecessor(NODE*n, NODE*m){

	//Make m predecessor of n
	n->Pred()->push(m);
	m->Succ()->push(n);
}

void MakePredecessors(NODE* n, LIST<NODE*> *L){

        NODE *m;
	L -> MoveTop();
	while( !L->eol() ){
	        m = L -> Data(); 
		n -> Pred() -> push(m); 
		m -> Succ() -> push(n);
		L -> MoveDown();
	}

}
		     




bool RemoveEdge(NODE *m, NODE *n){

	/*
	
	Remove the edge (m,n).
	The order of the nodes is unimportant, we search both.
	(m TO n is checked first)
	
	Output: 
	1 if the edge is removed, 
	0 if the edge does not exist.
		
	*/	
	
	
	
	if ( Remove<NODE*>(m,*(n->Pred()) ) ) { 
		
		if (Remove<NODE*>(n, *(m->Succ()))){ return 1; }
		else {cout<<"Error in RemoveEdge 1"<<endl;;exit(1);}
	}
	
	else if ( Remove<NODE*>(n,*(m->Pred())) ){
		
		if (Remove<NODE*>(m, *(n->Succ()))){ return 1; }
		else {cout<<"Error in RemoveEdge 2"<<endl;exit(1);}
	}
	
	else { return 0; } // The edge does not exist
	
}



void InsertPredEdges(NODE* o){

	LIST<NODE*> *P = o->Pred();
	P->MoveTop();
	while(!P->eol()){
	
		P->Data()->Succ()->push(o);
		P->MoveDown();

	}
}

void InsertSuccEdges(NODE* o){

	LIST<NODE*> *S = o->Succ();
	S->MoveTop();
	while(!S->eol()){
	
		S->Data()->Pred()->push(o);
		S->MoveDown();

	}
}

void RemovePredEdges(NODE* o){

	LIST<NODE*> *P = o->Pred();
	P->MoveTop();
	while(!P->eol()){
	
		Remove<NODE*>(o, *(P->Data()->Succ()));
		P->MoveDown();

	}
}

void RemoveSuccEdges(NODE* o){

	LIST<NODE*> *S = o->Succ();
	S->MoveTop();
	while(!S->eol()){
	
		Remove<NODE*>(o, *(S->Data()->Pred()));
		S->MoveDown();
	}
}

void ReplaceSuccEdges(NODE *o, NODE *r){
	
	LIST<NODE*> *S = o->Succ();
	S ->MoveTop();
	while(!S->eol()){
	
		Replace<NODE*>(o,r, *(S->Data()->Pred()));	
		S->MoveDown();
	}
}

void ReplacePredEdges(NODE* o, NODE *r){

	LIST<NODE*> *P = o->Pred();
	P->MoveTop();
	while(!P->eol()){
	
		Replace<NODE*>(o,r, *(P->Data()->Succ()));
		P->MoveDown();

	}
}

void DAGNodeInsert(DAG *G, NODE *o){ 

	/*
	Insert node o in graph G
	*/
	
        InsertPredEdges(o);
	InsertSuccEdges(o);
	G->InsertNode(o); 
}



void DAGNodeRemove(DAG *G, NODE *o){ 

	/*
	Remove node o from graph G
	*/
	
        RemovePredEdges(o);
	RemoveSuccEdges(o);
	G->RemoveNode(o); 
}

void DAGNodeReplace(DAG *G, NODE *o, NODE *r){


	RemovePredEdges(o);		//1.
	ReplaceSuccEdges(o,r);		//2.
	Push(r->Succ(), o->Succ());	//3.
	G->RemoveNode(o);		//4.
}



void RemoveUnusedNodes(DAG &G, LIST<NODE*> &Used){

	/*
	1) Label the predecessor set of Used.
	2) Remove every node that is not labeled. 
	
	Assumes the nodes in G are enumerated from 1 to G.Size().
		Does not make a renumbering itself!
	*/

	
	VECTOR<NODE*> MARK(G.Size());
	GPred(MARK, Used);

	VECTOR<NODE*> N(G.Size());
	NumberToPointer(N, G);

	for (int i=1; i<=MARK.Size(); i++){

		if( !MARK(i) ){ DAGNodeRemove(&G, N(i)); } 
	}

	
}



////////////////////////////////////////////////////////////////////////
//SpecialFunction_PGs
////////////////////////////////////////////////////////////////////////

NODE* SpecialFunction_OEDmodel2s(NODE* t, NODE *p1, NODE *p2){
////////////////////////////////////////////////////////////////////////
//	No check if there already is a NODE_SF_OEDmodel2s with these 
//	arguments.
////////////////////////////////////////////////////////////////////////

NODE* n = new NODE_SF_OEDmodel2s("SF_OEDmodel2s");
DAG::sCurrDAG->InsertNode(n);

n->Pred()->push(t);
n->Pred()->push(p1);
n->Pred()->push(p2);		

return n;
}


//--------------
// LESSOREQUAL
//--------------


void leq(NODE *a, NODE *b){
ConstructLessOrEqualNode(DAG::sCurrDAG, a, b);
}

void geq(NODE *a, NODE *b){
ConstructLessOrEqualNode(DAG::sCurrDAG, b, a);
}

bool hasLEQSucc(NODE* &leq, NODE *a, NODE *b){
	
	
	LIST<NODE*> *s = a->Succ();
	s->MoveTop();
	while(!s->eol()){
		
		if( isLessOrEqual(s->Data()) && s->Data()->pred1()==a && s->Data()->pred2()==b ){ 
		
			leq = s->Data(); 
			return 1; 
		
		}
		s->MoveDown();
	}
	return 0;
}

void ConstructLessOrEqualNode(DAG *G, NODE *a, NODE *b){

//-----------------------------------
// Construction of a binary NODE_leq
//-----------------------------------
// 1. If a==b return.
// 2. If a and b already have a 
//    inequality successor a<=b then return.
// 3. Otherwise construct a new a<=b node.
//----------------------------------- 
 	
 	NODE *leq;
 
	if( a==b || hasLEQSucc(leq,a,b) ) return;
	
	else{ 

		leq = new NODE_leq("<=");
		G->InsertNode(leq);
		
		leq->Pred()->push(a);
		a->Succ()->push(leq);
		
		leq->Pred()->push(b);
		b->Succ()->push(leq);
	}
	
	
}

//---------------
// INCLUSION
//---------------

void in(NODE *a, NODE *b){

	ConstructInclusionNode(DAG::sCurrDAG, a, b);
}
void in(NODE *a, const interval &b){

	NODE *n = DAG::sCurrDAG->ConstructNODE(b);
	ConstructInclusionNode(DAG::sCurrDAG, a, n);
}
void ConstructInclusionNode(DAG *G, NODE *x, NODE *y){

	if(x==y) return;
	NODE *in;
	if(hasInclusionSucc(in, y)){ 
	
		LIST<NODE*> *s = x->Succ();
		s->MoveTop();
		while(!s->eol()){
			
			if(s->Data()==in) return;
			s->MoveDown();
		}
		
		MakePredecessor(in,x); }
	
	else{
		InsertInclusionNode(G,x,y);
	}
}

bool hasInclusionSucc(NODE*& in, NODE* y){

	// if y has a inclusion successor of which y is the first predecessor then 
	// it is returned in in and the function returns true(1).
	
	LIST<NODE*> *s = y->Succ();
	s->MoveTop();
	while(!s->eol()){
	
		if( isInclusion( s->Data() ) ){ 
			
			s->Data()->Pred()->MoveTop(); 
			if(s->Data()->Pred()->Data()==y) { in = s->Data(); return 1;}
		}
		s->MoveDown();
	}
return 0;
}

void InsertInclusionNode(DAG *G, NODE *x, NODE *y){
  
 	// Construct a NODE_in in G
 	// and push y and then x to the predecessor list.
 	// y then becomes the first node in the predecessor list.
 	// 
	// Unconditionally, ie, no test if there exist a NODE_in
	// which has y as the first predecessor.
	
	NODE_in *in = new NODE_in("in");
	G->InsertNode(in); 
	MakePredecessor(in,y); 
	MakePredecessor(in,x); 
 
}



//----------------
// EQUALITY
//----------------

void equal(NODE *m, NODE *n){
	
	ConstructEqualityNode(DAG::sCurrDAG, m, n);
}

void equal(NODE &m, NODE &n){
	
	ConstructEqualityNode(DAG::sCurrDAG, m.NodePtr(), n.NodePtr());
}

void ConstructEqualityNode(DAG *G, NODE *m, NODE *n){

//1. If m==n return.
//2. If Both m and n have equality successors, move it to the equality node of n and delete the equality node of m
//3. If Only m has a equality successor, make n a predecessor of it
//4. If Only n has a equality successor, make m a predecessor of it.
//5. If Nobody has a equality successor, make a new equality node and set m,n as its predecessors.


	if(m==n) return;
	
	
	
	NODE *eq_m, *eq_n;
	eq_m=0;
	eq_n=0;
	hasEqSucc(eq_m, m); 
	hasEqSucc(eq_n,n);
	
	if( eq_m ){ 
		
		if ( eq_n  &&  (eq_m != eq_n) ){ //BOTH	

						
				ReplacePredEdges(eq_m, eq_n); // Predeges to eq_m are replaced by prededges to eq_n
				Push<NODE*>( eq_n->Pred(), eq_m->Pred() ); //Push on eq_n->Pred()
				eq_m->Pred()->Clean();
				G->RemoveNode(eq_m);
				
			
		}
		else{ //eq_m only

			eq_m->Pred()->push(n);
			n->Succ()->push(eq_m);			
		}
	
	}
	else if( eq_n ){//eq_n only

		eq_n->Pred()->push(m);
		m->Succ()->push(eq_n);
	
	}
	else{ //nobody

		NODE_equal *eq = new NODE_equal("=");
		G->InsertNode(eq);
		
		eq->Pred()->push(m);
		m->Succ()->push(eq);
		
		eq->Pred()->push(n);
		n->Succ()->push(eq);
	}
}
 	
NODE_equal* InsertEqualityNode(DAG &G){
  
 	// Construct a NODE_equal in G 
	// Unconditionally
	
	NODE_equal *n = new NODE_equal("=");
	G.InsertNode(n);  
  	return n;
}


bool hasEqSucc(NODE*& eq, NODE* n){

	// if n has a equality successor then 
	// it is returned in eq and the function returns true(1).
	
	LIST<NODE*> *s = n->Succ();
	s->MoveTop();
	while(!s->eol()){
	
		if( isEquality( s->Data() ) ){ eq = s->Data(); return 1;}
		s->MoveDown();
	}
return 0;
}


void MakeEqual(NODE *x, NODE *eq){



if( !isEquality(eq) ){cout<<endl<<"Node not NODE_equal in void MakeEqual(NODE*,NODE*)"; exit(1);}

//1. 
	eq->Pred()->push(x);
//2.
	x->Succ()->push(eq);



}

void MakeEqual(LIST<NODE*> &X, NODE *eq){

	X.MoveTop();
	while(!X.eol()){
		
		MakeEqual(X.Data(), eq);
		X.MoveDown();
	}
}

void SearchAndEquate(DAG &G, NODE *s, VECTOR<NODE> &arg, NODE& (*F)(VECTOR<NODE> &arg) ){


	/*
	
	Search for expression s with arguments arg.
	If found equate with F(arg)

	Should I replace s and arg with NODE (*S)(VECTOR<NODE> &arg)
	and then generate the expression in some dag H.
	Maybe this is better. But then I need my DAG destructor again.
	
	
	*/


	NODE *eq;
	VECTOR<NODE*> X;
	LIST< VECTOR<NODE*> > EXPR;

	NODE T;
	
	if ( FindExpression(s,G,EXPR) ){  

		EXPR.MoveTop();
		while(!EXPR.eol()){
		
			X = EXPR.Data();
			
			if ( !hasEqSucc(eq, X(s->num)) ) {
			 	eq = new NODE_equal("="); 
			 	G.InsertNode(eq);
			} 
			MakeEqual(X(s->num), eq); 
			
			
			VECTOR<NODE> A(arg.Size());
			for(int i=1; i<=A.Size(); i++){
			A(i) = *X( arg(i).NodePtr()->num );
			}
			DAG::sCurrDAG=&G;
			T = F(A);
			
			
			MakeEqual(T.NodePtr(), eq);
			
			EXPR.MoveDown();
		}
	}
}







void SetConstants(DAG &G, VECTOR<interval> &V){
  
	// The range of a constant node (NODE_con) n in graph G 
	// is written to component n.num of vector V.

	NODE* n;
	G.nodelist->MoveTop();
	while (!G.nodelist->eol() ){
		n = G.nodelist->Data();
		if ( isConstant( n ) ){ V( n->num ) =  n->range  ; }
    		G.nodelist->MoveDown();
  	}
}




	
	/*

	Levels and Paths

	*/


void SuccLevel(NODE* n, VECTOR<int> &LEVEL){
  
  /*
  
  Computes, recursively, a vector, LEVEL, of 
  "levels" of all successor nodes s of n.
  
  A successor s of a current node n is assigned a 
  level strictly larger than n.
 
 */
 
	NODE* s;
	n->Succ()->MoveTop();
	while( !( n->Succ()->eol() ) ){
         
		s = n->Succ()->Data(); 
        	if( LEVEL(s->num)<LEVEL(n->num)+1 ){  
        	
       			LEVEL(s->num) = LEVEL(n->num)+1;  
        		SuccLevel(s, LEVEL);  
        	}
        	n->Succ()->MoveDown();
    	}
}
  

void PredLevel(NODE *n, VECTOR<int> &LEVEL){
   

  	/*
  	A predecessor node is given a LEVEL 
  	strictly larger than a current node
	*/


	NODE *p;
	n->Pred()->MoveTop();
  
	while( !(n->Pred()->eol()) ){
         
        	p = n->Pred()->Data(); 
		if ( LEVEL(p->num)< (LEVEL(n->num)+1) ){ 
			LEVEL(p->num) = LEVEL(n->num)+1;  
			PredLevel(p, LEVEL);  
		}
        	n->Pred()->MoveDown();
	}
}




void ComputeSuccLevels(DAG &G, VECTOR<int> &LEVEL){


	/* 	
 	Calls SuccLevel for every source node in G 
  
	i is a predecessor of j  IMPLIES   LEVEL(i) < LEVEL(j);

	Input:  LEVEL, an int vector with size of the graph G
	Output: LEVEL, LEVEL(i) = SuccLevel of node i	
	*/


  // GET ALL SOURCE NODES
	LIST<NODE*> SourceList; 
   
	GetSourceNodes(SourceList,G);


  // INITIALIZE THE LEVELS OF EACH NODE
	for(int i=1; i<=LEVEL.Size();i++){ LEVEL(i) = 0; }
  

   // CALL SuccLEVEL FOR EACH SOURCE NODE
	SourceList.MoveTop();
	while ( !SourceList.eol() ){
     		SuccLevel( SourceList.Data(), LEVEL );
          	SourceList.MoveDown();
	}
}

VECTOR<NODE*> PathF(DAG &G, NODE *start){//ToDo
	
	/*
	Topological Range:
	Computes a forward ordering of the subdag defined by the start 
	node and all its successor nodes.
	All successors of the start, and only these, are in the path.
	The first node in the vector returned is the start node. 
	The second node is a immediate successor.
	....
	NOTE: This gives a topological ordering for FORWARD walks.
	
	Example of Use:
		1) Use a forward(or backward) derivative computation on this path to
		  compute dfdstart for all sink nodes f.
	*/

}
VECTOR<NODE*> Path(DAG &G, NODE *target){
  
	/*
	
	Computes a backward ordering of the subdag defined by the target 
	node and all its predecessor nodes.
	All predecessors of the target, and only these, are in the path.
	The first node in the vector returned is the target. 
	The second node is a immediate predecessor
	....
	NOTE: This gives a topological ordering for BACKWARD walks.

	Example of Use: 
		1)Use a backward derivative computation on this path to
		 compute the gradient of the target node in all source nodes.
		2) Reverse this path and use a forward derivative computation to compute
		the gradient of the target node in all source nodes.
	*/



// COMPUTE PRE-LEVELS

	VECTOR<int> PLEVEL ( G.Size() );
	for(int i=1; i<=PLEVEL.Size(); i++){  PLEVEL(i) = 0; }
	PLEVEL(target->num) = 99;//any positive integer works!
	PredLevel(target, PLEVEL);

// THEN SORT THE NODEPTRS IN INCREASING PLEVEL ORDER
// NODEPTR WITH LOWEST PLEVEL IS FIRST IN SORTED

	VECTOR<NODE*> SORTED(PLEVEL.Size());
	VECTOR<NODE*> NTP(PLEVEL.Size());
	G.NumberToPointer(NTP);
	SORT<NODE*>::SelectionSort(SORTED, PLEVEL, NTP);
 
 
// AND REMOVE ZERO PLEVEL NODES

	int NumOfNodes = 0; 
	for(int i=1;i<=PLEVEL.Size();i++){  if ( PLEVEL(i) != 0 ) NumOfNodes++; }
 
	VECTOR<NODE*> PATH(NumOfNodes);
	for(int i=1; i<=PATH.Size(); i++){
		PATH(i) = SORTED( PLEVEL.Size()-PATH.Size() + i );
	}
return PATH;
}


VECTOR<NODE*> Path(DAG &G, NODE *n1, NODE *n2 ){
  
	/*
	Computes [n1,n2], i.e. a path from n1 to n2
	The first node in the returned vector is n1. 
	The second node is a immediate successor.
	....
	The last node is n2.
	
	NOTE: [n,n]=n, [m,n]=/emptyset if m \in Succ n
	
	Primary use: To compute dn1dn2 we do it on the path [n2,n1].
	*/

  
	// COMPUTE PRE- and SUCC-LEVELS

	VECTOR<int> PLEVEL(G.Size()), SLEVEL(G.Size());
	for(int i=1; i<=PLEVEL.Size(); i++){PLEVEL(i)=0; SLEVEL(i)=0;}
	PLEVEL(n2->num) = 1;  
  	SLEVEL(n1->num) = 1;
  	PredLevel(n2, PLEVEL);
  	SuccLevel(n1, SLEVEL);
  
	// AND INTERSECT ( and store in SLEVEL )

	for(int i=1; i<=G.Size(); i++){  if ( PLEVEL(i)==0 ) SLEVEL(i)=0; }

	// THEN SORT THE NODEPTRS IN INCREASING SLEVEL ORDER
	// NODEPTR WITH LOWEST SLEVEL IS FIRST IN SORTED

	VECTOR<NODE*> SORTED(SLEVEL.Size());
	VECTOR<NODE*> NTP(SLEVEL.Size());
	G.NumberToPointer(NTP);
	SORT<NODE*>::SelectionSort(SORTED, SLEVEL, NTP);
 
   
	// AND REMOVE ZERO SLEVEL NODES

	int NumOfNodes = 0; 
	for(int i=1;i<=SLEVEL.Size();i++){  if ( SLEVEL(i) != 0 ) NumOfNodes++; }
 	VECTOR<NODE*> PATH(NumOfNodes);
	for(int i=1; i<=PATH.Size(); i++){
		PATH(i) = SORTED( SLEVEL.Size()-PATH.Size() + i );	
	}
 	return PATH;
}




void ForwardByLevel(DAG &G, VECTOR<NODE*> &FORWARD, VECTOR<int> &LEVEL){

  	/*
  	
  	Given a LEVEL vector, we return a vector FORWARD of 
  	nodepointers which is a topological ordering
  	for forward walk according to LEVEL.
	
	*/
	
	VECTOR<NODE*> NTP( G.Size() );
	G.NumberToPointer ( NTP );
	VECTOR<NODE*> SORTED( LEVEL.Size() );
	SORT<NODE*>::SelectionSort(FORWARD, LEVEL, NTP);
}



VECTOR<NODE*> ForwardOrderingByLevel(DAG &G){
	
	VECTOR<int> LEVEL(G.Size());
	ComputeSuccLevels(G, LEVEL);
	VECTOR<NODE*> aFORWARD(G.Size());
	ForwardByLevel(G, aFORWARD, LEVEL);
	
	return aFORWARD;
}

VECTOR<NODE*> BackwardByReverseForward(const VECTOR<NODE*> &FORWARD){

	VECTOR<NODE*> BACKWARD(FORWARD.Size());
	for(int i=1; i<=FORWARD.Size(); i++){
		BACKWARD(i) = FORWARD(FORWARD.Size()-i+1);
	}
	
	return BACKWARD;
}
	

void ForwardOrdering(DAG &G, VECTOR<NODE*> &FORWARD){
  


	int i = 1;  

	G.nodelist->MoveTop();
	while ( !G.nodelist->eol() ){
    		FORWARD(i) = G.nodelist->Data();
    		i++;
    		G.nodelist->MoveDown();
  	}
}
void BackwardOrdering(DAG &G, VECTOR<NODE*> &BACKWARD){
  

  
	int s = BACKWARD.Size(); 
	int i = 1;  

	G.nodelist->MoveTop();
	while ( !G.nodelist->eol() ){
		BACKWARD(s-i+1) = G.nodelist->Data();
		i++;
  		 G.nodelist->MoveDown();
 	}
 }



VECTOR<int> nofSuccessors(DAG &G){

   // V(i) = number of successors of node number i 
  
  

  NODE *n;
  VECTOR<int> V( G.nodelist->Size() );
  G.nodelist->MoveTop();
  while(!G.nodelist->eol()){
    n = G.nodelist->Data();
    V( n->num ) = n->Succ()->Size();      
    G.nodelist->MoveDown();
    
  }
  return V;
}         




void NumberOfSuccessors(DAG &G, 
			VECTOR<NODE*> &SORTED_NODEVECTOR, 
			VECTOR<int> &NUM_OF_SUCC){

  // Sorts the nodes in increasing order wrt successor size 


  
NODE *n;
  VECTOR<int> nSucc( G.nodelist->Size() );
  G.nodelist->MoveTop();
  while(!G.nodelist->eol()){
    n = G.nodelist->Data();
    nSucc( n->num ) = n->Succ()->Size();          
    G.nodelist->MoveDown();
    
  }

 
 VECTOR<NODE*> NTP( G.nodelist->Size() );
 G.NumberToPointer ( NTP );
 
 SORT<NODE*>::SelectionSort(SORTED_NODEVECTOR, nSucc, NTP);
 
 for(int i=1; i<=NUM_OF_SUCC.Size(); i++){
   NUM_OF_SUCC(i) = SORTED_NODEVECTOR(i)->Succ()->Size();
 }

}





VECTOR<int> MaxSuccessor(DAG &G){
  
	
	// i \in MaxSuccessor(G) iff node number i has  
	// the maximum number of successors

	
	
	
VECTOR<int> nofSucc( G.Size() );
nofSucc = nofSuccessors(G);
 
VECTOR<NODE*> SORTED(G.Size());
VECTOR<int> NOS(G.Size());
 
NumberOfSuccessors(G, SORTED, NOS);
 
 
 int MAX = NOS(NOS.Size()), numOfMax = 0; 
 for (int i = NOS.Size(); i>0; i--) {
   if ( NOS(i) == MAX ) numOfMax++;
 }
 
 
 VECTOR<int> MAXSUCC(numOfMax); int j = 1;
 for (int i=1; i<=NOS.Size(); i++){
   
   if ( NOS(i) == MAX ) { MAXSUCC(j) = SORTED(i)->num;  j++; }
 }
 
 return MAXSUCC;
}


void MaxSuccessor(LIST<NODE*> &L, DAG &G){
  
	
	// node in L iff node has  
	// the maximum number of successors

VECTOR<int> nofSucc( G.Size() );
nofSucc = nofSuccessors(G);
 
VECTOR<NODE*> SORTED(G.Size());
VECTOR<int> NOS(G.Size());
 
NumberOfSuccessors(G, SORTED, NOS);
 
int MAX = NOS(NOS.Size()); 
for (int i = NOS.Size(); i>0; i--) {
   if ( NOS(i) == MAX ) L.push(SORTED(i));
 }

}



void Print(VECTOR<NODE> &n, const VECTOR<interval> &V, ostream& out){


  for(int i=1; i<= n.Size();i++){
    out<<"\n"<<V(n(i).NodePtr()->num);
  }
}


void Print(vNODE &n, const VECTOR<interval> &V, ostream& out){

  for(int i=1; i<= n.Size();i++){
    out<<"\n"<<V(n(i).NodePtr()->num);
  }
}

void Print(const VECTOR<int> &n, const VECTOR<interval> &V, ostream& out){

  for(int i=1; i<= n.Size();i++){
    out<<"\n" << V(n(i));
  }
}



void PE(NODE *n){

  // Initial call PE(n)

  if (n!=0){
          cout << Symbol(n);
          if ( !isSource(n) ){
              cout << "(";
              int s = n->Pred()->Size(); 
	      n->Pred()->MoveTop();
	      for (int i=0; i<s-1; i++){
		PE( n->Pred()->Data() );
		cout << ",";
	        n->Pred()->MoveDown();
	      }
	      PE( n->Pred()->Data() );
              
	      cout<< ")";
	  }
  }
}



void PE1(NODE *n, NODE *s, ostream &out){


 
  
  if (n){
          
          if      ( isSource(n) )
          		
          		{ 
          		out << Symbol(n);
          		 }
          		 
	  else if ( isArithmetic(n) || isPow(n) )
	  		
	  		{
                        out << "(";  
		    	
		       		n->Pred()->MoveTop();  PE1(n->Pred()->Data(),n, out);
		       		out << Symbol(n);
		       		n->Pred()->MoveDown(); PE1(n->Pred()->Data(),n, out);
		       
	                out << ")";
	  		}
	  
	  else{

	      out << Symbol(n) << "{";
	      		
	      		int s = n->Pred()->Size();
	      		n->Pred()->MoveTop();
	      		for (int i=0; i<s-1; i++){
		
		  		PE1( n->Pred()->Data(),n, out);
		  		out << ",";
	          		n->Pred()->MoveDown();
	      
	      		}		
	     		PE1( n->Pred()->Data(),n, out);
	      
	      out << "}";
	  }
  }
}



void PrintExpression(NODE *n, ostream& out){  
	PE1(n, 0, out);  
}
void PrintExpression(NODE &n, ostream& out){ 

	PE1(n.NodePtr(),0,out); 
}


void PrintExpression(VECTOR<NODE> &n, ostream& out){

  for(int i=1; i<= n.Size();i++){
    out << Symbol(&n(i)) << " : ";
    PE1(n(i).NodePtr(),0, out);
    out << endl;
  }
}

void PrintExpression(const VECTOR<NODE*> &n, ostream& out){

	for(int i=1; i<= n.Size();i++){
		out<<endl;PrintExpression(n(i),out);
  	}
}


void PrintExpression(LIST<NODE*> &L, ostream& out){
	L.MoveTop();
	while(!L.eol()){
		out << endl;
		PrintExpression(L.Data(),out);
		L.MoveDown();
	}

}
void PrintExpression(VECTOR<int> &NodeNumbers, DAG &G, ostream& out){

	VECTOR<NODE*> NTP(G.Size());
	NumberToPointer(NTP,G);
	for(int i=1; i<=NodeNumbers.Size(); i++){
		PrintExpression(NTP(NodeNumbers(i)),out);
	}
}
void PrintSinks(DAG &G, ostream& out){

	LIST<NODE*> SinkList;
	GetSinkNodes(SinkList,G);
	PrintExpression(SinkList, out);
}

void PrintCodeList(DAG &G, ostream& out){


	G.NodeEnumeration();
	VECTOR<int> LEVEL(G.Size());
	ComputeSuccLevels(G, LEVEL);
	VECTOR<NODE*> FORWARD(G.Size());
	ForwardByLevel(G, FORWARD, LEVEL);

	LIST<NODE*> *p;
	for(int i=1; i<=FORWARD.Size();i++){

		out<<endl<<endl<<( "v" + ToString<int>(FORWARD(i)->num)  )<<"  "<<Symbol(FORWARD(i));
		
		out<<endl<<"pred:";
	
		p = FORWARD(i)->Pred();
		p->MoveTop();
		while(!p->eol()){
		

			out<<( "v" + ToString<int>( p->Data()->num)  ) <<",";	
			
			p->MoveDown();
		}
	}

}



void Operation(NODE *n, ostream &out){ 

	n->Pred()->MoveTop();
	if( isConstant(n)| isVariable(n)){ out << Symbol(n);}
	else{
	out << "(";
    		while( !(n->Pred()->eol()) ){ 
      		out <<  n->Pred()->Data()->num;
      		n->Pred()->MoveDown();
      		if (!( n->Pred()->eol() )){ out << ",";}
    		}
    
    	out<<")" << Symbol(n);
    
  }
}




void DAGInformation(DAG &G, ostream &out){

LIST<NODE*> SOURCE;
LIST<NODE*> INTERMEDIATE;
LIST<NODE*> SINK;
GetSISNodes(G, SOURCE, INTERMEDIATE, SINK);
		
  
out << endl << "Function expression: " << endl << endl;
  
SINK.MoveTop();
while(!SINK.eol()){
	PE1( SINK.Data(),0,out); 
	out << endl;
	SINK.MoveDown();
}
  
out << endl << "*******************************************************" << endl;
  
  

out << endl << "NODE*, number, level, Node Operation, Total Operation" << endl;
 
LIST<NODE*> *L = new LIST<NODE*>;
G.GetNodeList(L);
   
L->MoveTop();
while ( !L->eol() ){

	out << endl << L->Data() << "  " << L->Data()->num <<"    "; 
	Operation( L->Data(), out);   
	out <<"     "; PE1( L->Data() , 0 ,out);
	L->MoveDown();
}
  
  
out << endl << "*******************************************************" << endl;

SINK.MoveTop();
     
out << endl << endl << "sources:" << endl << SOURCE;
out << endl << endl << "intermediates:" << endl << INTERMEDIATE;
out << endl << endl << "sinks:" << endl << SINK; 

out << endl << endl << "*******************************************************" << endl;
out << G;








}


void DAGInformation(DAG &G, const VECTOR<interval> &RANGE, ostream &out){
  
  //Prints the range vector V

  out << endl << "value" << '\t' << "number"<< '\t' << "Node Operation" << '\t' << "Total Operation" << endl;
 
  LIST<NODE*> *L = new LIST<NODE*>;
  G.GetNodeList(L);
   
  L->MoveTop();
  while ( !L->eol() ){
    
    out << endl << RANGE( Number(L->Data() ) ) << '\t' << L->Data()->num << '\t'; Operation( L->Data(), out);   
    
    out <<'\t'<<'\t'; PE1( L->Data() , 0 ,out);
    L->MoveDown();
  }
  
 
}
















NODE* MakeVar(DAG *G, const string &name){
  
 // Construct a NODE_var in G   

  NODE *n = new NODE_var(name);
  G->InsertNode(n);  // Insert into G.nodelist
  return n;

}




NODE& MakeSource(DAG &G, const string &name){
  
 // Construct a NODE_var in G   

  NODE *n = new NODE_var(name);
  G.InsertNode(n);  // Insert into G.nodelist
  return *n;

}

NODE& MakeConstant(DAG &G, interval c, const string &name){
  
  // Construct/Reuse a NODE_con in G
  
  NODE* n   = G.ConstructNODE(c);
  Symbol(n) = name;    
  return *n;
       
}
void MakeConstant(NODE* &n, DAG &G, const interval &c, const string &name){
  
  // Construct/Reuse a NODE_con in G
  
  n  =  G.ConstructNODE(c);
  Symbol(n) = name;    
}

NODE& MakeConstant(DAG &G, interval c){
 
 // Construct/Reuse a NODE_con in G
  
  NODE* n   = G.ConstructNODE(c);
  return *n;
}


NODE& MakeNODE(DAG &G, const string &name){
 
  //Construct a NODE in G
 
  NODE *n = new NODE(name);
  G.InsertNode(n);  // Insert into G.nodelist
  return *n;

}

void MakeSource(NODE &v, const string &name, DAG &G){

// Construct a NODE_var in G

        NODE *p = new NODE_var();
        v = *p; 
	G.InsertNode( v.NodePtr() );
	Symbol( v.NodePtr() )	= name;					
			
}




void MakeSource(VECTOR<NODE> &V, const string &name, DAG &G){

// Construct NODE_var's in G
 
	NODE *p;
	for(int i=1; i<=V.Size(); i++){

			G.InsertNode( &V(i) );
			Symbol( &V(i) ) = name + ToString<int>(i);		
			
	}
}

vNODE& MakeSource(DAG &G, const string &name, int dim){

  // Construct a vNODE in G

   vNODE *V = new vNODE(dim);                        
   for(int i=1; i<=dim; i++){  
     
     G.InsertNode( &(*V)(i)  );  
     Symbol( &(*V)(i)  ) = name + ToString<int>(i); 
     
   } 

   return *V;
}


















