#include "DAG/NODE.h"
#include "DAG/DAG.h"

//#include "DAG/SpecialFunctions.h"


// NODE construction, destruction ----------------------------------------

 NODE::NODE(){             
 
              
      nodeptr     = this;
      ns          = "ND"; //Not Defined
      pred        = new LIST<NODE*>;
      succ        = new LIST<NODE*>;
}
   

NODE::NODE(const string &s){  //MODIFIED                  

   
  	nodeptr = this;//nodeptr         
  	ns          = s;
	pred        = new LIST<NODE*>;
	succ        = new LIST<NODE*>;
    
} 

 

NODE::~NODE(){
  
nodeptr = 0; 
pred->MoveTop();
while ( !pred->eol() ) { pred->Data()=0; pred->MoveDown(); }
delete pred;

succ->MoveTop();
while ( ! succ->eol() ) { succ->Data()=0; succ->MoveDown(); }
delete succ;
}

// 2009June9
NODE::NODE(NODE &n){

// COPY constructor
nodeptr = n.nodeptr;
ns 		= n.ns;
pred 	= n.pred;
succ		= n.succ; 
}




LIST<NODE*>*  NODE::Pred()   { return pred; }
LIST<NODE*>*  NODE::Succ()   { return succ; }

NODE* NODE::pred1(){
		pred->MoveTop();
		return pred->Data();
}
NODE* NODE::pred2(){
                 pred->MoveTop();
                 pred->MoveDown();
                 return pred->Data();
}


 


ostream & operator<<(ostream& stream, const NODE &n){

  stream << endl << "This:"       << endl << (NODE*)&n;
  stream << endl << "nodeptr:"    << endl << n.nodeptr;
  stream << endl << "nodenumber:" << endl << n.nodeptr->num;
  stream << endl << "symbol: "    << endl << Symbol(n.nodeptr); 
  stream << endl << "Pred:"       << endl << *n.nodeptr->Pred();
 stream << endl << "Succ:"       << endl << *n.nodeptr->Succ();
     

  return stream; 
}



/*

FP() 
Forward Propagate Range
Constrain range of current node using range of the predecessor nodes


FPIO() 
Forward Propagate Range Intersection Off 
Evaluate range of current node using range of the predecessor nodes

BP() 
Backward Propagate Range
Constrain range of each predecessor node of the current node using the range 
of the current node and the other predecessor nodes.
FPD()
Forward Propagate Derivative

cFPD()
construct Forward Propagate Derivative

BPD()
Backward Propagate Derivative

cBPD()
construct Backwar Propagate Derivative


*/


//Special Function
// OEDmodel2s *********************


bool NODE_SF_OEDmodel2s::FP(VECTOR<interval> &V){   

//pred1=T
//pred2=P1
//pred3=P2
 
pred->MoveTop();
interval T = V(pred->Data()->num);
pred->MoveDown();
interval P1 = V(pred->Data()->num);
pred->MoveDown();
interval P2 = V(pred->Data()->num);

interval range;
range =SpecialFunction_OEDmodel2s(T,P1,P2);
V(num) &= range;
return isEmpty(V(num));
}

void NODE_SF_OEDmodel2s::FPIO(VECTOR<interval> &V){	

pred->MoveTop();
interval T = V(pred->Data()->num);
pred->MoveDown();
interval P1 = V(pred->Data()->num);
pred->MoveDown();
interval P2 = V(pred->Data()->num);

interval range;
range = SpecialFunction_OEDmodel2s(T,P1,P2);
V(num) = range;
}

bool NODE_SF_OEDmodel2s::BP(VECTOR<interval> &V){
return 0;
}
void NODE_SF_OEDmodel2s::FPD(VECTOR<interval> &V,VECTOR<interval> &D){
}
void NODE_SF_OEDmodel2s::cFPD(VECTOR<NODE> &D){  
}
void NODE_SF_OEDmodel2s::BPD(VECTOR<interval> &V,VECTOR<interval> &D){
}
void NODE_SF_OEDmodel2s::cBPD(VECTOR<NODE> &D){ 
}




// leq *********************


bool NODE_leq::FP(VECTOR<interval> &V){   
 
 	//a<=b 
 
 	pred->MoveTop();
 	interval a = V(pred->Data()->num);
 	pred->MoveDown();
 	interval b = V(pred->Data()->num);
 	
 	if( sup(a)>sup(b) ) a.sethi(sup(b));
 	if( inf(b)<inf(a) ) b.setlo(inf(a));
	if( isEmpty(a) || isEmpty(b) ) return 1;
	return 0;
}
void NODE_leq::FPIO(VECTOR<interval> &V){	
}
bool NODE_leq::BP(VECTOR<interval> &V){
   
 	pred->MoveTop();
 	interval a = V(pred->Data()->num);
 	pred->MoveDown();
 	interval b = V(pred->Data()->num);
 	
 	if( sup(a)>sup(b) ) a.sethi(sup(b));
 	if( inf(b)<inf(a) ) b.setlo(inf(a));
	if( isEmpty(a) || isEmpty(b) ) return 1;
	return 0;
}
void NODE_leq::FPD(VECTOR<interval> &V,VECTOR<interval> &D){
}
void NODE_leq::cFPD(VECTOR<NODE> &D){  
}
void NODE_leq::BPD(VECTOR<interval> &V,VECTOR<interval> &D){
}
void NODE_leq::cBPD(VECTOR<NODE> &D){ 
}





// in *********************


bool NODE_in::FP(VECTOR<interval> &V){   
 
 	pred->MoveTop();
 	interval Y = V(pred->Data()->num);
 	pred->MoveDown();
	while(!pred->eol()){
		
		V(pred->Data()->num) &= Y;
		if(isEmpty( V(pred->Data()->num))) return 1;
		pred->MoveDown();
	}
	return 0;
}
void NODE_in::FPIO(VECTOR<interval> &V){	
}
bool NODE_in::BP(VECTOR<interval> &V){
   
 	pred->MoveTop();
 	interval Y = V(pred->Data()->num);
 	pred->MoveDown();
	while(!pred->eol()){
		
		V(pred->Data()->num) &= Y;
		if(isEmpty( V(pred->Data()->num))) return 1;
		pred->MoveDown();
	}
	return 0;
}
void NODE_in::FPD(VECTOR<interval> &V,VECTOR<interval> &D){
}
void NODE_in::cFPD(VECTOR<NODE> &D){  
}
void NODE_in::BPD(VECTOR<interval> &V,VECTOR<interval> &D){
}
void NODE_in::cBPD(VECTOR<NODE> &D){ 
}




// Helper-function




void RangeSum(interval &S, const VECTOR<interval> &V,  LIST<NODE*> *L){
      	
      	// Adds the sum of all V(L -> Data()-> num) to S
      	 
      	L -> MoveTop();
 	while ( !L -> eol() ){
 		S += V( L -> Data()-> num );
 		L -> MoveDown();
 	}
}



bool NODE_sum::FP(VECTOR<interval> &V){   
 
 interval S(0);
 RangeSum(S, V, pred); 
 V(num) &= S;
 return isEmpty( V(num) );
}


void NODE_sum::FPIO(VECTOR<interval> &V){
 
 interval S(0);
 RangeSum(S,V,pred);
 V(num) = S;
}

bool NODE_sum::BP(VECTOR<interval> &V){
   
 interval S(0);
 int k;
 
 RangeSum(S,V,pred);
 pred -> MoveTop();
 while ( !pred -> eol() ){
        k = pred -> Data()-> num;  
 	V(k) &= ( V(num) - DSUB(S, V(k)) );
 	if ( isEmpty( V(k) ) ) return 1;
 	pred -> MoveDown();
 }
 
 return 0;
 }
 
void NODE_sum::FPD(VECTOR<interval> &V, VECTOR<interval> &D)

	{cout<<"NODE_sum::FPD not implemented";}
void NODE_sum::cFPD (VECTOR<NODE> &D)

	{cout<<"NODE_sum::cFPD not implemented";}

void NODE_sum::BPD(VECTOR<interval> &V,VECTOR<interval> &D){
  
  pred -> MoveTop();
  while(!pred -> eol()){ 
  	D(pred -> Data() -> num) += D(num); //f_pred_i=1 for all i 
  	pred -> MoveDown();
  }
      
}
void NODE_sum::cBPD(VECTOR<NODE> &D){ 

pred -> MoveTop();
  while(!pred -> eol()){ 
  	D(pred -> Data() -> num) += D(num); //f_pred_i=1 for all i 
  	pred -> MoveDown();
  }
  
}








// Helper-function

void RangeProd(interval &P, const VECTOR<interval> &V,  LIST<NODE*> *L){
      	
      	// Multiplies the product of all V(L -> Data()-> num) to P
      	 
      	
      	L -> MoveTop();
 	while ( !L -> eol() ){
 		P *= V( L -> Data()-> num );
 		L -> MoveDown();
 	}
}



bool NODE_prod::FP(VECTOR<interval> &V){   
 
 interval P(0);
 RangeProd(P, V, pred); 
 V(num) &= P;
 return isEmpty( V(num) );
}


void NODE_prod::FPIO(VECTOR<interval> &V){
 
 interval P(0);
 RangeSum(P,V,pred);
 V(num) = P;
}

bool NODE_prod::BP(VECTOR<interval> &V){
  
 cout<<"NODE_prod::BP not implemented";
 return 0;
 }
 
void NODE_prod::FPD(VECTOR<interval> &V, VECTOR<interval> &D)

	{cout<<"NODE_prod::FPD not implemented";}

void NODE_prod::cFPD (VECTOR<NODE> &D)

	{cout<<"NODE_prod::cFPD not implemented";}

void NODE_prod::BPD(VECTOR<interval> &V,VECTOR<interval> &D)
 	
 	{cout<<"NODE_prod::BPD not implemented";}
 
     
void NODE_prod::cBPD(VECTOR<NODE> &D)
	
 	{cout<<"NODE_prod::cBPD not implemented";}
 
  






// equal *********************


// Help-functions
void RangeIntersect(interval &r, const VECTOR<interval> &V, LIST<NODE*> *L){
 
 	// Intersect r with all V(L -> Data()-> num).
      	 
      	L -> MoveTop();
 	while ( !L -> eol() ){
 		r &= V( L -> Data()-> num );
 		L -> MoveDown();
 	}
}
void RangeAssign(const interval &r, VECTOR<interval> &V, LIST<NODE*> *L){
 
 	// Assign range r to all V(L -> Data()-> num).
      	 
      	L -> MoveTop();
 	while ( !L -> eol() ){
 		V( L -> Data()-> num ) = r;
 		L -> MoveDown();
 	}
}


bool NODE_equal::FP(VECTOR<interval> &V){   
 
 	// pred_i <-- Intersect( pred_j: all j }
	
	RangeIntersect(V(num), V, pred);
	if ( isEmpty(V(num)) ) return 1;
	RangeAssign(V(num),V,pred);
	return 0;
	
	//return isEmpty( V(num) );

}
void NODE_equal::FPIO(VECTOR<interval> &V){
	interval r(FP::NegInf, FP::PosInf);
	RangeIntersect(r,V,pred);
	V(num)=r;
}
bool NODE_equal::BP(VECTOR<interval> &V){
   
	RangeIntersect(V(num), V, pred);
	if ( isEmpty(V(num)) ) return 1;
	RangeAssign(V(num),V,pred);
	return 0;
}
void NODE_equal::FPD(VECTOR<interval> &V,VECTOR<interval> &D){


  
	pred->MoveTop();
	D(num) = D(pred->Data()->num);
}
void NODE_equal::cFPD(VECTOR<NODE> &D){ 

  
	pred->MoveTop();
	D(num) = D(pred->Data()->num);  
}
void NODE_equal::BPD(VECTOR<interval> &V,VECTOR<interval> &D){


  
	pred->MoveTop();
	D(pred->Data()->num) += D(num);
}
void NODE_equal::cBPD(VECTOR<NODE> &D){ 



	pred -> MoveTop();
	D(pred->Data()->num) += D(num) ;  
}






bool NODE_intersect::FP(VECTOR<interval> &V){   
 
	RangeIntersect(V(num), V, pred);
	return isEmpty( V(num) );

}

void NODE_intersect::FPIO(VECTOR<interval> &V){
	
	interval r(interval::Entire);
	RangeIntersect(r,V,pred);
	V(num) = r;
}

bool NODE_intersect::BP(VECTOR<interval> &V){
   
   RangeIntersect(V(num), V, pred);
   if ( isEmpty(V(num)) ) return 1;
   RangeAssign(V(num),V,pred);
   return 0;
}


void NODE_intersect::FPD(VECTOR<interval> &V,VECTOR<interval> &D){

  cout<<endl<<"NODE_intersect::FPD  not implemented";
}

void NODE_intersect::cFPD(VECTOR<NODE> &D){ 

cout<<endl<<"NODE_intersect::cFPD  not implemented";
}
  
void NODE_intersect::BPD(VECTOR<interval> &V,VECTOR<interval> &D){

cout<<endl<<"NODE_intersect::BPD  not implemented";

}

void NODE_intersect::cBPD(VECTOR<NODE> &D){ 

cout<<endl<<"NODE_intersect::BPD  not implemented";

}









bool NODE::FP    (VECTOR<interval> &V){ return 0; }
void NODE::FPIO  (VECTOR<interval> &V){}
bool NODE::BP    (VECTOR<interval> &V){ return 0; }
void NODE::FPD   (VECTOR<interval> &V, VECTOR<interval> &D){}
void NODE::cFPD  (VECTOR<NODE> &V){}
void NODE::BPD   (VECTOR<interval> &V, VECTOR<interval> &D){}
void NODE::cBPD  (VECTOR<NODE> &D){}

//**************

bool NODE_con::FP   (VECTOR<interval> &V){ return 0; }
void NODE_con::FPIO (VECTOR<interval> &V){}
bool NODE_con::BP   (VECTOR<interval> &V){ return 0; }
void NODE_con::FPD  (VECTOR<interval> &V, VECTOR<interval> &D){}
void NODE_con::cFPD (VECTOR<NODE> &D){}
void NODE_con::BPD  (VECTOR<interval> &V, VECTOR<interval> &D){}
void NODE_con::cBPD (VECTOR<NODE> &D){}

//***************

bool NODE_var::FP   (VECTOR<interval> &V){ return 0; }
void NODE_var::FPIO (VECTOR<interval> &V){}
bool NODE_var::BP   (VECTOR<interval> &V){ return 0; }
void NODE_var::FPD  (VECTOR<interval> &V, VECTOR<interval> &D){}
void NODE_var::cFPD (VECTOR<NODE> &D){}
void NODE_var::BPD  (VECTOR<interval> &V, VECTOR<interval> &D){}
void NODE_var::cBPD (VECTOR<NODE> &D){}

//**************



// for the template version remember to have implemented &=,+=,....



bool NODE_add::FP(VECTOR<interval> &V){   
	V(num) &= (V(pred1()->num) + V(pred2()->num ));
	return isEmpty( V(num) );
}
void NODE_add::FPIO(VECTOR<interval> &V){
	V(num) = V(pred1()->num) + V(pred2()->num);
}
bool NODE_add::BP(VECTOR<interval> &V){
   	V(pred1()->num) &= (  V(num) - V(pred2()->num)  ) ;
	V(pred2()->num) &= (  V(num) - V(pred1()->num)  ) ;
	return(  isEmpty( V(pred1()->num) ) | isEmpty(  V(pred2()->num)   )   );  
}
//f_pred1=1, f_pred2=1
void NODE_add::FPD(VECTOR<interval> &V,VECTOR<interval> &D){
 	D(num) = D(pred1()->num) + D(pred2()->num);   
}
void NODE_add::cFPD(VECTOR<NODE> &D){
	D(num) = D(pred1()->num) + D(pred2()->num);   
}
void NODE_add::BPD(VECTOR<interval> &V,VECTOR<interval> &D){
	D(pred1()->num) +=  D(num); 
	D(pred2()->num) +=  D(num); 
}
void NODE_add::cBPD(VECTOR<NODE> &D){ 
	D(pred1()->num) =  D(pred1()->num) + D(num) ;  
	D(pred2()->num) =  D(pred2()->num) + D(num) ;  
}



bool NODE_sub::FP(VECTOR<interval> &V){
	V(num) &= ( V(pred1()->num) - V(pred2()->num) ) ;
	return isEmpty( V(num) );
}
void NODE_sub::FPIO(VECTOR<interval> &V){
 	V(num) = V(pred1()->num) - V(pred2()->num)  ;
}
bool NODE_sub::BP(VECTOR<interval> &V){
	V(pred1()->num) &= ( V(num) + V(pred2()->num)  ) ;
	V(pred2()->num) &= ( V(pred1()->num) - V(num)  ) ;
 	return(  isEmpty( V(pred1()->num) ) | isEmpty(  V(pred2()->num)   )   );
}
//f_pred1=1,f_pred2=-1
void NODE_sub::FPD(VECTOR<interval> &V, VECTOR<interval> &D){
 	D(num) = D(pred1()->num) - D(pred2()->num);   
}
void NODE_sub::cFPD(VECTOR<NODE> &D){ 
	D(num) = D(pred1()->num) - D(pred2()->num);   
}
void NODE_sub::BPD(VECTOR<interval> &V,VECTOR<interval> &D){
 	D(pred1()->num) += D(num);      	
 	D(pred2()->num) += -D(num);  
}
void NODE_sub::cBPD(VECTOR<NODE> &D){ 

	D(pred1()->num) =  D(pred1()->num) + D(num) ;  //f_pred1 = 1 
	D(pred2()->num) =  D(pred2()->num) - D(num) ;  //f_pred2 =-1
}


bool NODE_usub::FP(VECTOR<interval> &V){
	V(num) &=  -V(pred1()->num);
	return isEmpty( V(num) );
}
void NODE_usub::FPIO(VECTOR<interval> &V){
	V(num) = -V(pred1()->num);
}
bool NODE_usub::BP(VECTOR<interval> &V){
	V(pred1()->num) &= -V(num);
	return (  isEmpty( V(pred1()->num) )   );
}
//f_pred1=-1
void NODE_usub::FPD(VECTOR<interval> &V,VECTOR<interval> &D){
 	D(num) = -D(pred1()-> num);   
}
void NODE_usub::cFPD(VECTOR<NODE> &D){  
	D(num) = -D(pred1()-> num);   
}
void NODE_usub::BPD(VECTOR<interval> &V,VECTOR<interval> &D){ 
	D(pred1()->num) -= D(num);      
}
void NODE_usub::cBPD(VECTOR<NODE> &D){ 
	D(pred1()->num) =  D(pred1()->num) - D(num) ; 
	//can be replaced by V(pred1()->num) -=  V(num) ; after implementing -= for NODE  
}




bool NODE_mul::FP(VECTOR<interval> &V){
   
  V(num) &= ( V(pred1()->num) * V(pred2()->num)  ) ;
  return isEmpty( V(num) );
}
void NODE_mul::FPIO(VECTOR<interval> &V){
   
  V(num) = V(pred1()->num) * V(pred2()->num)  ;
}
bool NODE_mul::BP(VECTOR<interval> &V){
  

  //Range(pred1())  &= (  Range(this) / Range(pred2())  ) ; 
  //Range(pred2())  &= (  Range(this) / Range(pred1())  ) ;

  DivideAndHull( V(pred1()->num), V(num), V(pred2()->num) );
  DivideAndHull( V(pred2()->num), V(num), V(pred1()->num ));
  return (  isEmpty( V(pred1()->num) ) | isEmpty(  V(pred2()->num)   )   );
}
void NODE_mul::FPD(VECTOR<interval> &V,VECTOR<interval> &D){
  
    int p1 = pred1()-> num;
    int p2 = pred2()-> num;
    
    D(num) = D(p1) * V(p2) +  V(p1) * D(p2);    
    //f_pred1 = pred2, f_pred2 = pred1
}

void NODE_mul::cFPD(VECTOR<NODE> &D){

    int p1 = pred1()-> num;
    int p2 = pred2()-> num;
    
    D(num) = D(p1) *  *pred2() +  *pred1() * D(p2);    
    //f_pred1 = pred2, f_pred2 = pred1
}

void NODE_mul::BPD(VECTOR<interval> &V,VECTOR<interval> &D){
  
    D(pred1()->num) += D(num) * V(pred2()->num) ;//f_pred1=pred2
    D(pred2()->num) += D(num) * V(pred1()->num) ;//f_pred2=pred1
}

void NODE_mul::cBPD(VECTOR<NODE> &D){

  D(pred1()->num) = D(pred1()->num)   +  D(num) * *pred2()  ;
  D(pred2()->num) = D(pred2()->num)   +  D(num) * *pred1()  ;



//cD(pred1()->num) = &(  *cD(pred1()->num)   +   *cD(num) * *pred2()  );
//cD(pred2()->num) = &(  *cD(pred2()->num)   +   *cD(num) * *pred1()  );

}



	//div
	
	
	
	
	
bool NODE_div::FP(VECTOR<interval> &V){

	DivideAndHull( V(num), V(pred1()->num), V(pred2()->num) );
	return isEmpty( V(num) );
  
}
void NODE_div::FPIO(VECTOR<interval> &V){
                    
	//DivideAndHullIO( V(num), V(pred1()->num), V(pred2()->num) );                   
 	V(num) = V( pred1()->num )/V( pred2()->num ) ;                   
 
	// Vi kan inte visa inklusion om vi har division med 0, eller ?  Om division med 0, så icke Konsistent.
}

bool NODE_div::BP(VECTOR<interval> &V){
  
	V(pred1()->num) &= ( V(num) * V(pred2()->num)  ) ;
	DivideAndHull( V(pred2()->num), V(pred1()->num), V(num) );
	return (  isEmpty( V(pred1()->num) ) | isEmpty(  V(pred2()->num)   )   );

}

void NODE_div::FPD(VECTOR<interval> &V,VECTOR<interval> &D){
 
 // z = x/y;
 // ddx(z) = 1/y
 // ddy(z) = -x/y^2 = (-1/y)z
 
 
int x = pred1() -> num;
int y = pred2() -> num;
int z = num;

//D(z) = ( 1/V(y) )*D(x) + ( -V(x) / (V(y)^2) )*D(y);


	interval term1(interval::Entire);
 	DivideAndHull(term1, D(x), V(y) );
 	
 	interval m(interval::Entire);
 	DivideAndHull(m, V(x), V(y)^2 );
 	
 	
D(z) = term1 - m*D(y); 	
}

void NODE_div::cFPD(VECTOR<NODE> &D){ 
 
 // z = x/y;
 // ddx(z) = 1/y
 // ddy(z) = -x/y^2 = (-1/y)z
 
D(num) = ( 1/ (*pred2()) ) * D(pred1()->num) + ( (- *pred1())  / ( (*pred2())^2) ) * D(pred2()->num);

}


void NODE_div::BPD(VECTOR<interval> &V,VECTOR<interval> &D){
 
 
  //RangeD(pred1()) +=  RangeD(this)/Range(pred2());                                //f_pred1=1/pred2
  //RangeD(pred2()) +=  RangeD(this)*(  -Range(pred1()) / ((Range(pred2()))^2)  ) ;//f_pred2=-pred1/(pred2)^2


 interval tmp(interval::Entire);
 DivideAndHull(tmp, D(num),V(pred2()->num) );
 D(pred1()->num) += tmp;

 tmp = interval(interval::Entire);
 DivideAndHull(tmp, -V(pred1()->num),  ( ( V( pred2()->num ) )^2 )   );
 D(pred2()->num) += D(num)*tmp;


}

void NODE_div::cBPD(VECTOR<NODE> &D){ 
  
  D(pred1()->num) = D(pred1()->num)   +   D(num) / *pred2()  ;
  D(pred2()->num) = D(pred2()->num)   -   ( D(num) / *pred2() ) * *this ;  //vet ej vilket man ska vælja?

  //  cD(pred1()->num) = &(  *cD(pred1()->num)   +   *cD(num) / *pred2()  );
  ////cD(pred2()->num) = &(  *cD(pred2()->num)   -   *cD(num) * *this / *pred2() );
  //cD(pred2()->num) = &(  *cD(pred2()->num)   -   (*cD(num) / *pred2() ) * *this  );  //vet ej vilket man ska vælja?
}



interval FPPower(interval z, interval x, interval y){
	
	// Power z=x^y
	
	if	 (inf(x)>=0.0){ return z & exp(y*log(x)); }
	else if  (sup(x)<=0.0){ 
	 
 			interval a = abs(x); 
 			if (isEven(y)){ return FPPower (z,a,y); }
 			//if (isEven(y)){ return z & exp( y*log(a)); }
 			if (isOdd(y)){ return -FPPower(-z,a,y);} 
 			//if (isOdd(y)){ return -(-z & exp( y*log(a)));} 
 			//return  -FPPower(-z,a,y) | FPPower(z,a,y); 
			return  -(-z & exp( y*log(a))) | ( z & exp( y*log(a)) ); 
	}
 	else if  (inf(y)<0.0 && 0.0<sup(y)){ return z;}
	else {
		return FPPower(z, interval(inf(x),0.0),y) | FPPower(z,interval(0.0, sup(x)),y) ; //(1)
		//return -FPPower(-z, interval(0.0, -inf(x)), y) | FPPower(z, abs(x), y);//(2)

	}
}




void BPPower(const interval &z, interval &x, interval &y){
	
	/*
	Inverting z=x^y
	Reduction to the case x>=0. 
	No test if exponent is even/odd integer	
	*/
	
	
	if(0<=inf(x)){
		
		interval zpos = z & (x^y);	
				
		//if( !(y==0.0) ){ x= z^(1/y) & x; }      
		//if( !(x==0.0) ) {y=log(z)/log(x) & y;}
		//cout<<endl<<"1) "<<endl<<zpos<<endl<<x<<endl<<y;
		interval D,E;		
		if( !(y==0.0) ){
		
			xdiv(interval(1.0), y, D, E);
		//	cout<<endl<<"2) "<<endl<<D<<endl<<E;
			
			x = ((zpos^D) & x) | ((zpos^E) & x);
		//	cout<<endl<<"3) "<<endl<<z<<endl<<x;
		}
		if( !(x==0.0) ){
			
			xdiv(log(zpos), log(x), D, E);
			y = (D & y) | (E & y);
		
		}
	}
	
	else if(sup(x)<=0){
	
		interval a = abs(x);
        	interval z_negPart = abs(z) & (a^y);

        	BPPower(z_negPart, a, y);
        	x = -a & x;
	}
	
	else{
		interval xneg = interval(inf(x),0.0);
		interval xpos = interval(0.0,sup(x));
		interval y1(y), y2(y);
		BPPower(z, xpos, y1);
		BPPower(z, xneg, y2);
		x = xneg | xpos;
		y = y1 | y2; 			
	}

}




bool NODE_power::FP(VECTOR<interval> &V){
   
	
	
	//
	//V(num) &= ( V(pred1()->num) ^ V(pred2()->num)  ) ;
	
	V(num) = FPPower(V(num), V(pred1()->num), V(pred2()->num)  );

	return isEmpty( V(num) );  
}

void NODE_power::FPIO(VECTOR<interval> &V){
   
 	V(num) = V(pred1()->num) ^ V(pred2()->num);  	
}


bool NODE_power::BP(VECTOR<interval> &V){
  
//


             
	if( isEven(V(pred2()->num)) ){ 
	
		if( !(V(pred2()->num)==0.0) ){
	   		interval b =  V(num)^( 1.0/V(pred2()->num));
	   		V(pred1()->num) = ( V(pred1()->num) & -b ) | ( V(pred1()->num) & b )  ;
  		 }
	}
	
	else if (isOdd(V(pred2()->num))){ 
	
		interval a = ( interval(fabs(inf(V(num)))) )^(1.0/V(pred2()->num)); if(inf(V(num))<0.0) a=-a;
		interval b = ( interval(fabs(sup(V(num)))) )^(1.0/V(pred2()->num)); if(sup(V(num))<0.0) b=-b;
		V(pred1()->num) &= (a|b); 
	} 
    
    
  
    
   
   BPPower(V(num), V(pred1()->num), V(pred2()->num));
	
   
   

return( isEmpty(V(pred1()->num)) || isEmpty(V(pred2()->num)) );
}

void NODE_power::FPD(VECTOR<interval> &V, VECTOR<interval> &D){
  
  //  z    = x^y
  //  dzdx = y * x^(y-1)
  //  dzdy = ln(x)*z
   
  int x = pred1()->num;
  int y = pred2()->num;
  
  D(num) = V(y) * ( V(x)^(V(y)-1) ) * D(x)  +  log(V(x)) * V(num) * D(y);
}


void NODE_power::cFPD(VECTOR<NODE> &D){
  
  //  z    = x^y
  //  dzdx = y * x^(y-1)
  //  dzdy = ln(x)*z
   
  int x = pred1()->num;
  int y = pred2()->num;
  
  D(num) = (*pred2()) * ( (*pred1())^((*pred2())-1) ) * D(x)  +  log((*pred1())) * (*this) * D(y);

   
}

void NODE_power::BPD(VECTOR<interval> &V, VECTOR<interval> &D){
  
  //  z    = x^y = exp(ylogx)
  //  dzdx = (y/x)z = y * x^(y-1).  
  //  dzdy = log(x)*z

	int num1 = pred1()->num;
	int num2 = pred2()->num;
	/*
	D(num1) += D(num) * ( (V(num2)/V(num1))  *  V(num) );
	//D(num1) += D(num) * (  V(num2) * (  V(num1) ^ (V(num2)-1.0) ) );
	D(num2) += D(num) * ( log( V(num1) ) * V(num) );
	*/
	
	interval t(interval::Entire);		
	DivideAndHull(t, V(num2), V(num1));
  	D(num1) += D(num) * ( t *  V(num) );
	
	D(num2) += D(num) * ( log( V(num1) ) * V(num) );
	
	
}


void NODE_power::cBPD(VECTOR<NODE> &D){

   
	int num1 = pred1()->num;
	int num2 = pred2()->num;

	//D(num1) += D(num) * ( ( (*pred2()) /(*pred1()))  * *this );
	D(num2) += D(num) * ( log( *pred1() ) * *this );
	
	D(num1) += D(num) * (  (*pred2()) * ( ( *pred1() )^( (*pred2()) - 1.0  )   )  ); 
	





}



	
	
	
bool NODE_exp::FP(VECTOR<interval> &V){
  
  // Range(this) &= exp(Range(pred1()))  ;
    
    V(num) &= exp(V(pred1()->num))  ;
    return isEmpty( V(num) );

}
void NODE_exp::FPIO(VECTOR<interval> &V){
  
  // Range(this) &= exp(Range(pred1()))  ;
    
    V(num) = exp(V(pred1()->num))  ;
}
bool NODE_exp::BP(VECTOR<interval> &V){                                                                                               
 
 V(pred1()->num)  &= log(V(num)) ; 
 return isEmpty( V(pred1()->num) );
}

void NODE_exp::FPD(VECTOR<interval> &V,VECTOR<interval> &D){
 
  D(num) = V(num) * D(pred1()->num);   
}

void NODE_exp::cFPD(VECTOR<NODE> &D){
 
   D(num) = (*this) * D(pred1()->num);   

}

void NODE_exp::BPD(VECTOR<interval> &V,VECTOR<interval> &D){

  D(pred1()->num) +=  D(num) * V(num);  
}

void NODE_exp::cBPD(VECTOR<NODE> &D){
 
//D(pred1()->num) =  D(pred1()->num)  +  D(num) * *this ;
D(pred1()->num) +=  D(num) * *this ;

}


	
	//log




bool NODE_log::FP(VECTOR<interval> &V){
    
    V(num) &= log(V(pred1()->num))  ;
    return isEmpty( V(num) );

}

void NODE_log::FPIO(VECTOR<interval> &V){
  
    V(num) = log(V(pred1()->num))  ;
}

bool NODE_log::BP(VECTOR<interval> &V){                                                                                               
 V(pred1()->num)  &= exp(V(num)) ; 
 return isEmpty( V(pred1()->num) );
}

void NODE_log::FPD(VECTOR<interval> &V,VECTOR<interval> &D){
  //D(num) = ( 1.0 / ( V(pred1()->num) ) ) * D(pred1()->num);  
  
  	interval t(interval::Entire);		//
	DivideAndHull(t, interval(1.0), V(pred1()->num));
  	D(num) = t * D(pred1()->num);  
}

void NODE_log::cFPD(VECTOR<NODE> &D){
  
  D(num) = ( 1.0 / ( *(pred1()) ) ) * D(pred1()->num);  
}

void NODE_log::BPD(VECTOR<interval> &V,VECTOR<interval> &D){
int num1 = pred1()->num;

  //D(num1) +=  D(num) * ( 1.0 / ( V(num1) ) );  
  
        interval t(interval::Entire);		//
	DivideAndHull(t, D(num), V(num1) );
  	D(num1) += t;  
}

void NODE_log::cBPD(VECTOR<NODE> &D){
 
  int num1 = pred1()->num;
  D(num1) =  D(num1)  +  D(num) * ( 1.0 / ( *pred1() ) ) ;


}






NODE& NODE::operator&(const NODE &b)const{

  LIST<NODE*> *predList = new LIST<NODE*>;
  predList->push(nodeptr);
  predList->push(b.nodeptr);
  //sym NodeSymbol("&");
  //return *(DAG::sCurrDAG->ConstructNODE("&", predList)  );
  return *(DAG::sCurrDAG->ConstructNODE("&", predList)  );
  
}






		


NODE& NODE::operator+(const NODE &b)const{

	LIST<NODE*> *predList = new LIST<NODE*>;
	predList->push(nodeptr);
	predList->push(b.nodeptr);
	return *(DAG::sCurrDAG->ConstructNODE("+", predList)  );
}
NODE& NODE::operator+(const double &s)const{   
	if (s==0.0){ return *(NODE*)this; }    
	LIST<NODE*> *predList = new LIST<NODE*>;
	predList->push(nodeptr);
        NODE* n = DAG::sCurrDAG->ConstructNODE(interval(s));
  	predList->push(n);
	return *( DAG::sCurrDAG->ConstructNODE("+", predList)  );    
}
NODE& operator+(const double &s, const NODE &a){

	if (s==0.0){ return (NODE&)a; }                                         
	LIST<NODE*> *predList = new LIST<NODE*>;
	predList->push(a.nodeptr);
	NODE* n = DAG::sCurrDAG->ConstructNODE(interval(s));
  	predList->push(n);
	return *( DAG::sCurrDAG->ConstructNODE("+", predList)  );
}
NODE& NODE::operator+(const interval &s)const{   
	if (s==0.0){ return *(NODE*)this; }    
	LIST<NODE*> *predList = new LIST<NODE*>;
	predList->push(nodeptr);
        NODE* n = DAG::sCurrDAG->ConstructNODE(s);
  	predList->push(n);
	return *( DAG::sCurrDAG->ConstructNODE("+", predList)  );    
}
NODE& operator+(const interval &s, const NODE &a){

	if (s==0.0){ return (NODE&)a; }                                         
	LIST<NODE*> *predList = new LIST<NODE*>;
	predList->push(a.nodeptr);
	NODE* n = DAG::sCurrDAG->ConstructNODE(s);
  	predList->push(n);
	return *( DAG::sCurrDAG->ConstructNODE("+", predList)  );
}
NODE& NODE::operator+()const{        
	return *(NODE*)this;
}





NODE& NODE::operator-(const NODE &b)const{
  
	LIST<NODE*> *predList = new LIST<NODE*>;
	predList->push(nodeptr);
	predList->push(b.nodeptr);
  	return *( DAG::sCurrDAG->ConstructNODE("-", predList)  ); 
}
NODE& NODE::operator-(const double &s)const{   
  
	if (s==0.0){ return *(NODE*)this; }  //11MARS  
	LIST<NODE*> *predList = new LIST<NODE*>;
	predList->push((NODE*)this);
	NODE* n = DAG::sCurrDAG->ConstructNODE(interval(s));
	predList->push(n);
	return *( DAG::sCurrDAG->ConstructNODE("-", predList)  );    
}
NODE& operator-(const double &s, const NODE &a){

	if (s==0.0){ return (NODE&)a; }                                         
	LIST<NODE*> *predList = new LIST<NODE*>;
	NODE* n = DAG::sCurrDAG->ConstructNODE(interval(s));
	predList->push(n);
  	predList->push(a.nodeptr);  
   	return *( DAG::sCurrDAG->ConstructNODE("-", predList)  );   
}
NODE& NODE::operator-(const interval &s)const{   
  
	if (s==0.0){ return *(NODE*)this; }  //11MARS  
	LIST<NODE*> *predList = new LIST<NODE*>;
	predList->push((NODE*)this);
	NODE* n = DAG::sCurrDAG->ConstructNODE(s);
	predList->push(n);
	return *( DAG::sCurrDAG->ConstructNODE("-", predList)  );    
}
NODE& operator-(const interval &s, const NODE &a){

	if (s==0.0){ return (NODE&)a; }                                         
	LIST<NODE*> *predList = new LIST<NODE*>;
	NODE* n = DAG::sCurrDAG->ConstructNODE(s);
	predList->push(n);
  	predList->push(a.nodeptr);  
   	return *( DAG::sCurrDAG->ConstructNODE("-", predList)  );   
}
NODE& NODE::operator-()const{              
 
	LIST<NODE*> *predList = new LIST<NODE*>;
	predList->push(nodeptr);
	return *( DAG::sCurrDAG->ConstructNODE("u-", predList)  );
}



NODE& NODE::operator*(const NODE &b)const{
  
	LIST<NODE*> *predList = new LIST<NODE*>;
	predList->push(nodeptr);
  	predList->push(b.nodeptr);
	return *( DAG::sCurrDAG->ConstructNODE("*", predList)  );
}
NODE& NODE::operator*(const double &s)const{   
  
	if (s==1.0){ return *(NODE*)this; }    
	LIST<NODE*> *predList = new LIST<NODE*>;
	predList->push(nodeptr);
	NODE* n = DAG::sCurrDAG->ConstructNODE(interval(s));
	predList->push(n);
	return *(DAG::sCurrDAG->ConstructNODE("*", predList)  );  
}
NODE& operator*(const double &s, const NODE &a){

	if (s==1.0){ return (NODE&)a; }                                         
	LIST<NODE*> *predList = new LIST<NODE*>;
	predList->push(a.nodeptr);
	NODE* n = DAG::sCurrDAG->ConstructNODE(interval(s));
	predList->push(n);
	return *( DAG::sCurrDAG->ConstructNODE("*", predList)  );  
}
NODE& NODE::operator*(const interval &s)const{   
  
  if (s==1.0){ return *(NODE*)this; }    
  LIST<NODE*> *predList = new LIST<NODE*>;
  predList->push(nodeptr);
  NODE* n = DAG::sCurrDAG->ConstructNODE(s);
  predList->push(n);
  return *(DAG::sCurrDAG->ConstructNODE("*", predList)  );
    
}
NODE& operator*(const interval &s, const NODE &a){

  if (s==1.0){ return (NODE&)a; }                                         
  LIST<NODE*> *predList = new LIST<NODE*>;
  predList->push(a.nodeptr);
  NODE* n = DAG::sCurrDAG->ConstructNODE(s);
  predList->push(n);
  return *( DAG::sCurrDAG->ConstructNODE("*", predList)  );
  
}






NODE& NODE::operator/(const NODE &b)const{
  
  LIST<NODE*> *predList = new LIST<NODE*>;
  predList->push(nodeptr);
  predList->push(b.nodeptr);
  return *( DAG::sCurrDAG->ConstructNODE("/", predList)  );
 
}
NODE& NODE::operator/(const double &s)const{   
  
  if (s==1.0){ return *(NODE*)this; }                                  
  LIST<NODE*> *predList = new LIST<NODE*>;
  predList->push(nodeptr);          
  NODE* n = DAG::sCurrDAG->ConstructNODE(interval(s));
  predList->push(n);
  return *( DAG::sCurrDAG->ConstructNODE("/", predList)  );  
}
NODE& operator/(const double &s, const NODE &a){
  LIST<NODE*> *predList = new LIST<NODE*>;
  NODE* n = DAG::sCurrDAG->ConstructNODE(interval(s));
  predList->push(n);
  predList->push(a.nodeptr);  
  return *( DAG::sCurrDAG->ConstructNODE("/", predList)  );
}
NODE& NODE::operator/(const interval &s)const{   
  
  if (s==1.0){ return *(NODE*)this; }                                  
  LIST<NODE*> *predList = new LIST<NODE*>;
  predList->push(nodeptr);          
  NODE* n = DAG::sCurrDAG->ConstructNODE(s);
  predList->push(n);
  return *( DAG::sCurrDAG->ConstructNODE("/", predList)  );  
}
NODE& operator/(const interval &s, const NODE &a){
  LIST<NODE*> *predList = new LIST<NODE*>;
  NODE* n = DAG::sCurrDAG->ConstructNODE(s);
  predList->push(n);
  predList->push(a.nodeptr);  
  return *( DAG::sCurrDAG->ConstructNODE("/", predList)  );
}




NODE& NODE::operator^(const NODE &b)const{
  
  LIST<NODE*> *predList = new LIST<NODE*>;
  predList->push(nodeptr);
  predList->push(b.nodeptr);
  return *( DAG::sCurrDAG->ConstructNODE("^", predList)  );
}
NODE& NODE::operator^(const double &s)const{   
  
  if (s==1.0){ return *(NODE*)this; }    
  LIST<NODE*> *predList = new LIST<NODE*>;
  predList->push(nodeptr);                                               
  NODE* n = DAG::sCurrDAG->ConstructNODE(interval(s));
  predList->push(n);
  return *( DAG::sCurrDAG->ConstructNODE("^", predList)  ); 
}
NODE& operator^(const double &s, const NODE &a){
                                
  LIST<NODE*> *predList = new LIST<NODE*>;
  NODE* n = DAG::sCurrDAG->ConstructNODE(interval(s));
  predList->push(n);
  predList->push(a.nodeptr);  
  return *( DAG::sCurrDAG->ConstructNODE("^", predList)  );
}
NODE& NODE::operator^(const interval &s)const{   
  
  if (s==1.0){ return *(NODE*)this; }    
  LIST<NODE*> *predList = new LIST<NODE*>;
  predList->push(nodeptr);                                               
  NODE* n = DAG::sCurrDAG->ConstructNODE(s);
  predList->push(n);
  return *( DAG::sCurrDAG->ConstructNODE("^", predList)  ); 
}
NODE& operator^(const interval &s, const NODE &a){
                                
  LIST<NODE*> *predList = new LIST<NODE*>;
  NODE* n = DAG::sCurrDAG->ConstructNODE(s);
  predList->push(n);
  predList->push(a.nodeptr);  
  return *( DAG::sCurrDAG->ConstructNODE("^", predList)  );
}







void NODE::operator+=(const NODE &a){
	*this = *this + a;
}
void NODE::operator+=(const double &a){
	*this = *this + a;
}
void NODE::operator-=(const NODE &a){
	*this = *this - a;
}
void NODE::operator-=(const double &a){
	*this = *this - a;
}

void NODE::operator*=(const NODE &a){
	*this = *this * a;
}
void NODE::operator*=(const double &a){
	*this = *this * a;
}



NODE& exp(const NODE &b){
  
  LIST<NODE*> *predList = new LIST<NODE*>;
  predList->push(b.nodeptr);
  return *( DAG::sCurrDAG->ConstructNODE("exp", predList)  );
}

NODE& log(const NODE &b){

  LIST<NODE*> *predList = new LIST<NODE*>;
  predList->push(b.nodeptr);
  return *( DAG::sCurrDAG->ConstructNODE("log", predList)  );
}







vNODE&  vNODE::operator=(vNODE &x){
  //
  //

  if(size != x.size){cout<<endl<<"dimensions disagree in   vNODE&  vNODE::operator=(vNODE &x)"<<endl;exit(1);}
  for (int i=0; i<size; i++){ vec[i] = x.vec[i];}//just copying of nodeptr
  return (*this);
}   



vNODE::vNODE(){ vec=0; size=0; }
 

vNODE::vNODE(int dim, char *name){
           
               vec = new NODE_var[dim];                        
		       size = dim;
		       //for(int i=0; i<size; i++){  
		          
			 //DAG::sCurrDAG->InsertNode(vec[i].nodeptr );  
			 //Symbol( &vec[i] ) = sym(name)&int2sym(i+1);
			 
		       //} 
  
}

vNODE::vNODE(int dim){
           
               vec = new NODE_var[dim];                        
		       size = dim;
		       
}


vNODE::~vNODE(){


  if (vec != 0){ delete []vec; vec = 0; size = 0;}

}



vNODE::vNODE(const vNODE &x){

    // COPY constructor
     

   size = x.size;
    if (x.vec == 0){ vec = 0;}                                  
    else {  vec = new NODE[size]; 
      for (int i=0; i<size; i++){ vec[i] = x.vec[i]; } //
    }
 }






NODE& vNODE::operator()(int index){                //non-constant

  // index range 1,..., size
     if (index<1 || index>size ){ cout << "\n\n\nINDEX NOT IN RANGE, NODE& vNODE::operator()(int index)"; exit(1);}
   return vec[index-1];
}
const NODE& vNODE::operator()(int index)const{     //constant

  // index range 1...,size
  
  if (index<1 || index>size ){ cout << "\n\n\nINDEX NOT IN INDEX RANGE, const NODE& vNODE::operator()(int index)const"; exit(1);}
  return vec[index-1];
}



ostream & operator<<(ostream& stream, const vNODE &V){
  
  for(int i=1; i<=V.size; i++){
    stream<<endl<<V(i);
  }
  return stream; 
}




















































