using namespace std;
#include "Structs/interval.h"
#include "Structs/round.h"



//SKROTA
double fp_pow(double x, int e){ 
  double p = 1.0;
  while (e != 0){
         if (e & 1){p = p*x;}
         e = e >> 1;
         x = x*x;
  }
  return p;
}




const int FP::_Max[] = {0xffffffff, 0x7fefffff};
const double FP::Max = *(double*)&_Max;

const int FP::_Min[] = {0xffffffff, 0xffefffff};
const double FP::Min = *(double*)&_Min;

const int FP::_PosInf [] = {0x00000000, 0x7ff00000};
const double FP::PosInf  = *(double*)&_PosInf;

const int FP::_NegInf [] = {0x00000000, 0xfff00000};
const double FP::NegInf  = *(double*)&_NegInf;

const int FP::_aNaN [] = {0x00000001, 0x7ff00000};
const double FP::aNaN  = *(double*)&_aNaN;

const int FP::_Ita [] = {0x00000001, 0x00000000};
const double FP::Ita  = *(double*)&_Ita;

const int FP::_Epsilon [] = {0x00000000, 0x3cb00000};
const double FP::Epsilon  = *(double*)&_Epsilon;


void FP::PrintHex(double w){  
//  Print double in hexadecimal base    
printf("%.8x%.8x" ,((int *)&w)[1], ((int *)&w)[0]);
}

double FP::Shift(double d, int n){

  // Shift n floating point spaces upwards (n>0) or downwards (n<0)

  
  if(n>0){
  
  	  if (d==NegInf){ d=Min; n-=1; } 
          RoundUp();
          for(int i=1; i<=n; i++){ d = d + Ita ; }
  }
  if(n<0){
  
  	  if(d==PosInf){ d=Max; n+=1;}
  
          RoundDown();
          for(int i=1; i<=-n; i++){ d = d - Ita ; }
  }
  
  RoundNear();
  return d;

}


bool FP::isInf(double x){

	return  (x==PosInf || x== NegInf) ;
}   

bool FP::isNaN(double x){

	int m_mask [2] = {0xffffffff, 0x000fffff};
	int e_mask [2] = {0x00000000, 0x7ff00000};
	int * fp = (int*)&x;
 
	return( ( (e_mask[1]&fp[1]) == e_mask[1] ) && 
		( (m_mask[0]&fp[0]) != 0  || ( m_mask[1]&fp[1]) != 0 ) );

}

double FP::RoundMode(){

	double s = 1.0e-20;
	if      (1.0-s==1.0+s){return 0.5;}          //round to nearest
	else if (1.0+s==1.0 && -1.0+s==-1.0){return -1.0;} //round down
	else if (1.0-s==1.0){return 1.0;}              //round up
	else {return 0.0;}                         //round to zero

}

void FP::RoundDown()	{ setRound(ROUND_DOWN); }
void FP::RoundUp()	{ setRound(ROUND_UP);   } 
void FP::RoundNear()	{ setRound(ROUND_NEAR); } 
void FP::RoundToZero()	{ setRound(ROUND_ZERO); }	


double FP::minimum(double x, double y){return ( (x<y) ? x : y );}
double FP::maximum(double x, double y){return ( (x<y) ? y : x );}





const interval interval::Empty(FP::PosInf,FP::NegInf); 
//const interval interval::Empty(FP::aNaN,FP::aNaN);   //Standard
const interval interval::Entire(FP::NegInf,FP::PosInf);
const interval interval::Finite(FP::Min,FP::Max);
const interval interval::NonNegative(0.0,FP::PosInf);
const interval interval::NonPositive(FP::NegInf,0.0);
const interval interval::PositiveTail(FP::Max,FP::PosInf);
const interval interval::NegativeTail(FP::NegInf,FP::Min);

double	inf(const interval &a){

	if(a.e)return FP::PosInf;
	return(a.lo);
}

double	sup(const interval &a){
	
	if(a.e)return FP::NegInf;  
	return(a.hi);
}



interval iinf(const interval &a){
	
	if(a.e)return interval::Empty;
	if(FP::NegInf<a.lo){return interval(a.lo);}
	return interval::NegativeTail;  
}

interval isup(const interval &a){
	
	if(a.e)return interval::Empty;
	if(a.hi<FP::PosInf){return interval(a.lo);}
	return interval::PositiveTail;  
}

bool	empty(const interval &a){
return(a.e || a.lo>a.hi);
}
bool	isEmpty(const interval &a){  
return(a.e || a.lo>a.hi);
}
bool	isThin(const interval &a){
if(a.lo==a.hi || a.e) return 1;
return 0;
}
bool	isAtom(const interval &a){
return ( isThin(a) || FP::Shift(a.lo, +1) == a.hi);
}
bool	isFinite(const interval &a){
if (a.e || (FP::NegInf<a.lo && a.hi<FP::PosInf) ) return 1;
return 0;
}



double width(const interval &x){

	if(x.e)return 0.0;
	if (x.lo<x.hi){ 
			if (!isinf(x.lo) && !isinf(x.hi) ) { 
			  return x.hi-x.lo;
			} 
			else return FP::PosInf;
	}
	else return 0.0;  
}
              

double mid(const interval &x){
	
	if (x.e) return FP::aNaN;	
	if ( x.lo < x.hi ) {
	
		if(FP::NegInf<x.lo){  
			
			//both finite
			if (x.hi<FP::PosInf) return (x.lo*0.5 + x.hi*0.5); //27APR
			//xlo finite, xhi infinite
			else return FP::Max;				
		}
		
		//xlo infinite, xhi finite
		else { if(x.hi<FP::PosInf) return FP::Min;	
		
		//both infinite
		else return 0.0;					}
	}
	else return x.lo;
}


// Interval valued width and mid. 
// Only to be applied on bounded arguments.

interval iwidth(const interval &x){ 
	
	if(x.e)return interval::Empty;	
	return( interval(x.hi) - interval(x.lo) );
}
interval imid (const interval &x){

	if(x.e)return interval::Empty;
	return( ( interval(x.lo) + x.hi )/2.0 ); 
}                                 

double mig(const interval &a){

	if (a.e) return FP::aNaN;
	
	if			(a.lo>=0.0)	return a.lo; 
	else if	(a.hi<=0.0)	return -a.hi; 
	else							return 0.0;	 
}  

double mag(const interval &a){

	if (a.e) return FP::aNaN;
	
	if			(a.lo>=0.0)	return a.hi; 
	else if	(a.hi<=0.0)	return -a.lo; 
	else if	(-a.lo<a.hi)	return a.hi;
	else							return -a.lo;
}  

interval abs(const interval &a){

	if (a.e) return interval::Empty;
	
	if			(a.lo>=0.0)	return interval(a.lo,a.hi); 
	else if	(a.hi<=0.0)	return interval(-a.hi,-a.lo); 
	else if	(-a.lo<a.hi)	return interval(0.0,a.hi);
	else							return interval(0.0,-a.lo);
	
	
}

double HausdorffDistance(const interval &a, const interval &b){

 // Upper bound on Hausdorff distance.
  
if(a.e||b.e) return(FP::PosInf);

double l(0.0),u(0.0);

FP::RoundUp();
		
	if(a.lo<b.lo){	
		l=b.lo-a.lo; 
	}
	else if(b.lo<a.lo){
		l=a.lo-b.lo;
	}
	 

	if(a.hi<b.hi){	
		u=b.hi-a.hi;
	}
	else if(b.hi<a.hi){
		u=a.hi-b.hi;
	}
	

FP::RoundNear(); 

	if(l<u) return u;
	return l;
}  
  
 

interval gap(const interval &a, const interval &b){

	//returns the gap "between" a and b
	
	if (a.e||b.e)return interval::Empty;
	
	interval g;
	
	if      (a.lo>b.hi) { g.lo=b.hi; g.hi=a.lo;}
  	else if (b.lo>a.hi) { g.lo=a.hi; g.hi=b.lo; }
  	else                { g=interval::Empty; }
  	return g;	
}

double SeparationDistance(const interval &a, const interval &b){
	
	if(a.e||b.e)return FP::PosInf;
	double d;
	FP::RoundUp();
  	if(a.lo>b.hi) 		{ d = a.lo - b.hi; }
  	else if(b.lo>a.hi) 	{ d = b.lo - a.hi; }
  	else                	{ d = 0.0; }
  	FP::RoundNear();
  	return d;


}

interval max(const interval &X, const interval &Y){

	// max(X,Y)	= { max(x,y) : x in X, y in Y } 
	//		= [ max(X.lo,Y.lo), max(X.hi, Y.hi) ]

	double l,u; 
	if (X.lo<=Y.lo) l = Y.lo;
	else		l = X.lo;
	if (X.hi<=Y.hi) u = Y.hi;
	else  		u = X.hi;

	return interval(l,u);
}

interval maxInv(const interval &Z, const interval &X){

	// maxInv(Z,X) = { y: max(x,y) in Z for some x in X } 

	if	(Z.hi<X.lo)	return interval::Empty;
	else if (X.hi>=Z.lo)	return interval(FP::NegInf,Z.hi);
	else 			return Z;	
}



interval min(const interval &X, const interval &Y){


	// min(X,Y)	= { min(x,y) : x in X, y in Y } 
	//		= [ min(X.lo,Y.lo), min(X.hi, Y.hi) ]


	double l,u; 
	if (X.lo<=Y.lo) l = X.lo;
	else		l = Y.lo;
	if (X.hi<=Y.hi) u = X.hi;
	else  		u = Y.hi;

	return interval(l,u);
}

interval minInv(const interval &Z, const interval &X){

	// minInv(Z,X) = { y: min(x,y) in Z for some x in X } 

	if	(Z.hi<X.lo)	return Z;
	else if (X.hi>=Z.lo)	return interval(Z.lo, FP::PosInf); //Case of Nonempty intersection
	else 			return interval::Empty;	
}


interval hull(const interval &X, const interval &Y)
{

 	if(X.e)
 	{
    	if(Y.e)	return interval::Empty; 
    	else		return Y;
    }
  	else{ if (Y.e) return X; }
  
	double l(X.lo), u(X.hi);
	if(Y.lo<l)	l=Y.lo;
	if(u<Y.hi)	u=Y.hi;
	
	return interval(l,u);
}

interval hullInv(const interval &Z, const interval &X){

	if			(Z.hi<X.lo)		return interval(FP::NegInf,Z.hi);
	else if	(X.hi<Z.lo)		return interval(Z.lo,FP::PosInf);
	else 						return interval::Entire;
}




void interval::set(double a, double b){lo = a; hi = b; e = (hi<lo);}  
void interval::setlo(double a){lo = a; e = (hi<lo);}
void interval::sethi(double a){hi = a; e = (hi<lo);}

interval integer(const interval &X)
{
	return interval(ceil(X.lo), floor(X.hi));
}


double sign2(const interval &X)
{
	if(isEmpty(X))return FP::aNaN;
	
	if(X.lo<0.0 && X.hi<=0.0) return -1.0;
	if(X.lo>=0.0 && X.hi>0.0) return 1.0; 
}


interval& interval::operator=(const interval &a){           
  
    if (this == &a) { return *this; }
    lo = a.lo;
    hi = a.hi;
    e  = a.e;
    return *this;
}

interval interval::operator+(const interval & a)const{                                         
if(e||a.e)return Empty;
  
	interval r;
				
	if(FP::NegInf<lo && FP::NegInf<a.lo){FP::RoundDown(); r.lo = lo + a.lo;}
	else					{r.lo=FP::NegInf;}
	
	if(hi<FP::PosInf && a.hi<FP::PosInf){FP::RoundUp(); r.hi = hi + a.hi;}
	else					{r.hi=FP::PosInf;}
    
	return r;
    
}
interval interval::operator+(const double s)const{                                         

	if(e)return Empty;
  
	interval r;
	if(FP::NegInf<lo && FP::NegInf<s){FP::RoundDown();r.lo=lo+s;}
	else				{r.lo=FP::NegInf;}
	if(hi<FP::PosInf && s<FP::PosInf){FP::RoundUp();r.hi=hi+s;}
	else				{r.hi=FP::PosInf;}
	
	return r;
	  
}

interval operator+(const double s, const interval & a){                                        

	return a+s;
}

interval interval::operator+()const{    //unary plus    

	return *this;
}





interval interval::operator-(const interval &a)const{

	if(e||a.e)return Empty;
 
 	interval r;					
	
	if(FP::NegInf<lo && a.hi<FP::PosInf){FP::RoundDown(); r.lo=lo-a.hi;}
	else					{r.lo=FP::NegInf;}
	
	if(hi<FP::PosInf && a.lo>FP::NegInf){FP::RoundUp(); r.hi=hi-a.lo;}
	else					{r.hi=FP::PosInf;}
	
	return r;
}

interval interval::operator-(const double s)const{     

	interval I(s);
	return *this-I;
}
interval operator-(const double s, const interval &a){   

	return -(a-s);
}

interval interval::operator-()const{
if(e)return Empty;
  
    //unary minus

    interval result;
    result.lo = -hi;
    result.hi = -lo;
    return result;
}



interval DSUB(const interval &a, const interval &b){
  if(a.e||b.e)return interval::Empty;
  interval x(-b.lo, -b.hi);
  return(a + x);  
}
interval DADD(const interval &a, const interval &b){
  if(a.e||b.e)return interval::Empty;
  interval x(b.hi, b.lo);
  return(a + x);  
}


void interval::operator+=(const interval & a){

    // Incremental addition. 

    if(e||a.e){*this = Empty;}
    else{ *this = *this + a; }
    
}
void interval::operator-=(const interval & a){

    // Incremental subtraction. 

    if(e||a.e){*this = Empty;}
    else{ *this = *this - a; }
    
}

//x*0=0 Multiplication
interval interval::operator*(const interval &y)const{


	if(e||y.e)return Empty;
  
	interval p;
  
	if ( 0 <= lo ){
		if       ( 0 <= y.lo ){
				if(lo==0.0 || y.lo==0.0){p.lo=0.0;}
                           	else{ FP::RoundDown(); p.lo = lo*y.lo; }
                           	if(hi==0.0 || y.hi==0.0){p.hi=0.0;}
                           	else{ FP::RoundUp();   p.hi = hi*y.hi; }
		}
		else if  ( 0 <= y.hi ){
				if(hi==0.0||y.lo==0.0){p.lo=0.0;}
				else{FP::RoundDown(); p.lo = hi*y.lo;}
				if (hi==0.0||y.hi==0.0){p.hi=0.0;}
				else{FP::RoundUp(); p.hi = hi*y.hi;}
		}
		else{		if(hi==0.0||y.lo==0.0){p.lo=0.0;}
				else{FP::RoundDown(); p.lo = hi*y.lo;}
				if(lo==0.0||y.hi==0.0){p.hi=0.0;}
				else{FP::RoundUp(); p.hi = lo*y.hi;}
		}
	}
	else if ( 0 <= hi ){// lo < 0 
		if  ( 0 <= y.lo ){
				if(lo==0.0||y.hi==0.0){p.lo=0.0;}
				else{FP::RoundDown(); p.lo = lo*y.hi;}
				if(hi==0.0||y.hi==0.0){p.hi=0.0;}
				else{FP::RoundUp(); p.hi = hi*y.hi;}
	 	}
	 	else if ( 0 <= y.hi ){
	                  //p.lo = FP::minimum(hi*y.lo, lo*y.hi);
				FP::RoundDown();
			  
				if(hi==0.0||y.lo==0.0){p.lo=0.0;}
				else{p.lo = hi*y.lo;}
			  
				double r;
			 
			 	if(lo==0.0||y.hi==0.0){r=0.0;}
			  	else{r = lo*y.hi;}
			  
			  	if (p.lo>r){p.lo=r;}

	                  //p.hi = FP::maximum(lo*y.lo, hi*y.hi);
	                  	FP::RoundUp();
			  
			  	if(lo==0.0||y.lo==0.0){p.hi=0.0;}
			  	else{p.hi = lo*y.lo;}
			
				if(hi==0.0||y.hi==0.0){r=0.0;}
				else{r = hi*y.hi;}
				
				if (p.hi<r){p.hi = r;}
	 	}
		else{
	                  	if(hi==0.0||y.lo==0.0){p.lo=0.0;}
	                  	else{FP::RoundDown(); p.lo = hi*y.lo;}
				if(lo==0.0||y.lo==0.0){p.hi=0.0;}
				else{FP::RoundUp(); p.hi = lo*y.lo;}
		}	
	}
	else{//hi < 0  
		if (0 <= y.lo){
				if(lo==0.0||y.hi==0.0){p.lo=0.0;}
               			else{FP::RoundDown(); p.lo = lo*y.hi;}
				if(hi==0.0||y.lo==0.0){p.hi=0.0;}
			  	else{FP::RoundUp(); p.hi = hi*y.lo;}
     		}
     		else if ( 0 <= y.hi ){
				if(lo==0.0||y.hi==0.0){p.lo=0.0;}
				else{FP::RoundDown(); p.lo = lo*y.hi;}
				if(lo==0.0||y.lo==0.0){p.hi=0.0;}
				else{FP::RoundUp(); p.hi = lo*y.lo;}
     		}
     		else{
				if(hi==0.0||y.hi==0.0){p.lo=0.0;}
				else{FP::RoundDown(); p.lo = hi*y.hi;}
                          	if(lo==0.0||y.lo==0.0){p.hi=0.0;}
                          	else{FP::RoundUp();p.hi = lo*y.lo;}
		}
	}
	
FP::RoundNear();
return p;
}


interval interval::operator* (const double s)const{

  	// Right multiplication by double.

	if(e)return Empty;
	return *this *interval(s);
	
}
  
interval operator*(const double s, const interval &a){  
return a*s;
}

interval interval::DMUL(const interval & y){

 // Dependent MULtiplication
 // Does not function with infinities

if(e||y.e)return Empty;
  
 interval d;
 
 if( 0<y.lo ){
   if(lo>=0){
     FP::RoundDown();     d.lo = lo*y.hi;
     FP::RoundUp()  ;     d.hi = hi*y.lo;
     FP::RoundNear();
     }
   else if(hi>=0){
     FP::RoundDown();     d.lo = lo*y.lo;
     FP::RoundUp()  ;     d.hi = hi*y.lo;
     FP::RoundNear();
   }
   else{
     FP::RoundDown();     d.lo = lo*y.lo;
     FP::RoundUp()  ;     d.hi = hi*y.hi;
     FP::RoundNear();
   }
 }
 else if( y.hi<0 ){ 
        if(hi<=0){
	  FP::RoundDown();    d.lo = hi*y.lo;
	  FP::RoundUp()  ;    d.hi = lo*y.hi;
	  FP::RoundNear();
	}
        else if( lo<=0 ){
	  FP::RoundDown();    d.lo = hi*y.hi;
	  FP::RoundUp()  ;    d.hi = lo*y.hi;
	  FP::RoundNear();
	}
        else{
	  FP::RoundDown();    d.lo = hi*y.hi;
	  FP::RoundUp()  ;    d.hi = lo*y.lo;
	  FP::RoundNear();
	}
 }
 

return d;
}

interval interval::operator/(const interval &y)const{

	if(e||y.e)return Empty;	
	interval d;
 
	if( 0<y.lo ){
	
		if(lo>=0){
    			if(y.hi==FP::PosInf){d.lo=0.0;}
			else{FP::RoundDown(); d.lo = lo/y.hi;}
			if(hi==FP::PosInf){d.hi=FP::PosInf;}
     			else{FP::RoundUp(); d.hi = hi/y.lo;}
     			FP::RoundNear();
		}
		else if(hi>=0){
			if(lo==FP::NegInf){d.lo = FP::NegInf;}
     			else{FP::RoundDown(); d.lo = lo/y.lo;}
     			if(hi==FP::PosInf){d.hi=FP::PosInf;}
     			else{FP::RoundUp(); d.hi = hi/y.lo;}
     			FP::RoundNear();
		}
   		else{
   			if(lo==FP::NegInf){d.lo=FP::NegInf;}
     			else{FP::RoundDown(); d.lo = lo/y.lo;}
     			if(y.hi==FP::PosInf){d.hi=0.0;}
     			else{FP::RoundUp(); d.hi = hi/y.hi;}
     			FP::RoundNear();
   		}
 	}
 	else if( y.hi<0 ){
   
		d = -( *this/(-y) );
 	}
 	else{ //0<<=y
 		interval D,E;
 		xdiv(*this,y,D,E);
 		return (D|E);
 	} 

return d;
}

interval interval::operator/(const double s)const{

	if(e)return Empty;
	return(*this/interval(s));
}

interval operator/(const double s, const interval &a){

  if (a.e) return interval::Empty;
  return(interval(s)/a);
}





void xdiv(const interval & x, const interval & y, interval & D, interval & E){
 
 E = interval::Empty;
 
 if        ( !(0.0<<=y) ){ D = x/y; }
 else if   (  0.0<<=x   ){ D = interval(FP::NegInf, FP::PosInf); }
 else if   (  y==0.0    ){ D = interval::Empty; }
 else if   (  x<0.0     ){
                        if(y.hi == 0.0){ 
                        	if(x.hi==FP::NegInf && y.lo==FP::NegInf){D.lo=0.0;}
                        	else{FP::RoundDown(); D.lo = x.hi/y.lo;}
				D.hi = FP::PosInf;
				FP::RoundNear();
			}
                        else if (y.lo == 0.0){         
                        	D.lo = FP::NegInf;
                        	if(x.hi==FP::NegInf && y.hi==FP::PosInf){D.hi=0.0;}
			        else{FP::RoundUp(); D.hi = x.hi/y.hi;}
				FP::RoundNear();
			}
                        else{ 
                        	D.lo = FP::NegInf;
                        	if(x.hi==FP::NegInf && y.hi==FP::PosInf){D.hi=0.0;}
			        else{FP::RoundUp(); D.hi = x.hi/y.hi;}
			        
			        if(x.hi==FP::NegInf && y.lo==FP::NegInf){E.lo=0.0;}
			        else{FP::RoundDown(); E.lo = x.hi/y.lo;}
				E.hi = FP::PosInf;
				E.e = 0;
				FP::RoundNear();
			}
 }
 else{//x>0
			if(y.hi == 0.0){ 
				D.lo = FP::NegInf;
				if(x.lo==FP::PosInf && y.lo==FP::NegInf){D.hi=0.0;}
	                     	else{FP::RoundUp(); D.hi = x.lo/y.lo;}
			     	FP::RoundNear();
			}
			else if(y.lo == 0.0){ 
				if(x.lo==FP::PosInf && y.hi==FP::PosInf){D.lo=0.0;}
				else{FP::RoundDown(); D.lo = x.lo/y.hi;}
	                	D.hi = FP::PosInf;
	   			FP::RoundNear(); 
			}
			else{
	     			D.lo = FP::NegInf;
	     			if(x.lo==FP::PosInf && y.lo==FP::NegInf){D.hi=0.0;}
	     			else{FP::RoundUp(); D.hi = x.lo/y.lo;}
	     			
	     			if(x.lo==FP::PosInf && y.hi==FP::PosInf){E.lo=0.0;}
	     			else{FP::RoundDown();E.lo = x.lo/y.hi;}
	     			E.hi = FP::PosInf;
	               		E.e = 0;
				FP::RoundNear();
	}
 }


}


void interval::operator*=(const interval & a){

    if(e||a.e){*this = Empty;}
    else{ *this = *this * a; }
}
void interval::operator*=(const double &a){

    if(e){*this = Empty;}
    else{ *this = *this * a; }
}
void interval::operator/=(const interval &a){

    if(e||a.e){*this = Empty;}
    else{ *this = *this / a; }
}
void interval::operator/=(const double &a){

    if(e){*this = Empty;}
    else{ *this = *this / a; }
}
void DivideAndHull(interval &result, const interval &num, const interval &denum){

	interval D,E;
	xdiv(num, denum, D, E);
	result = (result & D) | (result & E);
}
void DivideAndHullIO(interval &result, const interval &num, const interval &denum){

	interval D,E;
	xdiv(num,denum, D, E);
 	result = ( D|E );
}
 
interval sqrt(const interval &a){

  // SquareRoot
  // With intersection of the argument to the domain [0,FP::PosInf] of sqrt.
  
	if(a.e)return interval::Empty;
  
	interval r;
	if( a.hi<=0){ r = interval(0,a.hi); }
  	else{
    		double t = a.lo;
    		if(a.lo<0){ t=0; }
    
    		FP::RoundDown(); r.lo = sqrt(t);
    		FP::RoundUp();   r.hi = sqrt(a.hi);
    		FP::RoundNear();
  	}
  	return r;
}

bool isEven(const interval &x){
	//inf evaluates to 0
	if(x.lo==x.hi){ int s=(int)x.lo; if(x.lo==s && s%2==0){return 1;} }
	return 0;
}
bool isOdd(const interval &x){
	//inf evaluates to 0
	if(x.lo==x.hi){ int s=(int)x.lo; if(x.lo==s && !(s%2==0)){return 1;} }
	return 0;
}

interval interval::operator^(const interval &y)const{  
 
	// Power
 	
	if	 (lo>=0){ return exp( y*log(*this)); }
	else if  (hi<=0){ 
 			interval a = abs(*this); 
 			if (isEven(y)){return a^y;}
 			if (isOdd(y)){return -(a^y);} 
 			return  -( a^y ) | ( a^y ); 
	}
 	else if  (y.lo<0 && 0<y.hi){ return interval(FP::NegInf, FP::PosInf); }
	else {return interval(lo,0.0)^y | interval(0.0, hi)^y ;}
}

interval interval::operator^(int n)const{       


	if(e)return Empty;
  
	interval y(1.0);
 
	if (n>=0){

  
		if ( n == 0 ) return y;   //we define 0^0=1
		if ( (n&1) == 0 ){

   			// n is even
  	
    			FP::RoundDown(); y.lo = fp_pow(mig(*this), n);
    			FP::RoundUp();   y.hi = fp_pow(mag(*this), n);
    			FP::RoundNear();
    			return y;
		}
		else {
       
       			// n is odd
  
       			if (lo>=0){
	   			FP::RoundDown(); y.lo = fp_pow(lo,n);
           			FP::RoundUp();   y.hi = fp_pow(hi,n);
       			}
       			else{
	    			FP::RoundUp(); y.lo = -fp_pow(-lo,n);
            			if (hi>=0){ y.hi = fp_pow(hi,n); }
	    			else{
	        	 		FP::RoundDown(); y.hi = -fp_pow(-hi,n);
	    			}
       			}	
       			FP::RoundNear();
       			return y;
		}
	}

	else{ // n<0
      		return( 1/(*this) )^(-n); 
	}
}


interval exp(const interval & x){


  
	if (x.e){ return interval::Empty; }
	interval p;
  
  	// Thickened floating point computation.
	
	if (x.lo == FP::NegInf){p.lo = 0.0;}
	else{p.lo = FP::Shift(exp(x.lo), -3);}
	if (x.hi == FP::PosInf){ p.hi = FP::PosInf; }
	else{p.hi = FP::Shift(exp(x.hi), +3);  }
	
	FP::RoundNear();

	if(p.lo>p.hi){ cout<<"\n exp produced empty interval,   interval exp(const interval & x)"; exit(1);}   
	if(p.lo>=p.hi && x.lo<x.hi){ cout<<"\n exp FAILED TO ENCLOSE THE RANGE, interval exp(const interval & x)"; exit(1);}   
	
return p;
}




interval log (const interval & x){

	
	interval y = x & interval(0, FP::PosInf);
  	if (y.e){ return interval::Empty; }
  	interval l;
  	
  	// Thickened  

  	if   (y.lo == 0)          { l.lo = FP::NegInf; }
  	else                      { l.lo = FP::Shift( log(y.lo), -3 ); }
  	if   (y.hi == FP::PosInf)    { l.hi = FP::PosInf; }
  	else                      { l.hi = FP::Shift( log(y.hi), +3 ); }



	if(l.lo > l.hi){ cout<< "log produced empty interval,   interval log (const interval & x)";exit(1); }
  	if(l.lo == l.hi && l.lo<l.hi){ cout<< "log produced empty interval,   interval log (const interval & x)";exit(1); }
  	
return l;
}

bool interval::operator<=(const interval & b)const{
return(hi <= b.lo);
}
bool interval::operator<=(const double b)const{
return(hi <= b);
}
bool operator<=(const double a, const interval & b){
return(a <= b.lo);
}  
bool interval::operator<(const interval & b)const{
return(hi < b.lo);
}
bool interval::operator<(const double b)const{
return(hi < b);
}
bool operator<(const double a, const interval & b){
return(a < b.lo);
}
bool interval::operator>=(const interval & b)const{
return(lo >= b.hi);
}
bool interval::operator>=(const double b)const{
return(lo >= b);
}
bool operator>=(const double a, const interval & b){
return(a >= b.hi);
}  
bool interval::operator>(const interval & b)const{
return(lo > b.hi);
}
bool interval::operator>(const double b)const{
return(lo > b);
}
bool operator>(const double a, const interval & b){
return(a > b.hi);
}


  
bool interval::operator!=(const interval & b)const{
if (e||b.e)return 1;
  
  return(hi < b.lo || b.hi < lo);
}
bool interval::operator==(const interval & b)const{

  if(e){
        if(b.e){return 1;}
        else{return 0;}
  }
  else{
       if(b.e){ return 0; }     
  }
  return(lo == b.lo && hi == b.hi);
}
bool interval::operator==(const double b)const{

if(e){return 0;}
else return(lo==b && hi==b);

}


// & INTERSECTION ------------------------------------------------------------

interval interval::operator&(const interval & a)const{

  if(e||a.e)return Empty;

  interval y;
  
  y.lo = FP::maximum(lo, a.lo);
  y.hi = FP::minimum(hi, a.hi);
  if(y.lo>y.hi)y.e=1;
  return y;
}

// | HULL  ------------------------------------------------------------------

interval interval::operator|(const interval & b)const{

  if(e){
    if(b.e){return Empty; }
    else{return b;}
    }
  else{ if(b.e)return *this; }
  
  return( interval(FP::minimum(lo, b.lo), FP::maximum(hi, b.hi)));
}

// --------------------------------------------------------------------------

void interval::operator&=(const interval & a){
  if(e||a.e){*this = Empty;}
  else{  
       lo = FP::maximum(lo, a.lo);
       hi = FP::minimum(hi, a.hi);
       if (lo>hi){e=1;}   //24FEB
  }
}
void interval::operator|=(const interval & a){
  if(e){*this = a;}
  else{
      if(!a.e){
               lo = FP::minimum(lo, a.lo);
               hi = FP::maximum(hi, a.hi);
      }
  }
}


// <<=, << INCLUSION --------------------------------------------------------
//    
//  INCLUSION: a <<= b means a in b 
//  STRICT INCLUSION: a >>= b means a covers b
//
//    
bool interval::operator<<=(const interval & b)const{
  if(e){return 1;}
  else {if(b.e)return 0;}
return( b.lo <= lo && hi <= b.hi);
}
bool interval::operator<<=(double s)const{
 if(e)return 1;
return( s == lo && s == hi);
}
bool operator<<=(double s, const interval & b){
 if(b.e)return 0;
return( b.lo <= s && s <= b.hi );
} 
bool interval::operator>>=(const interval & b)const{
  if(b.e){return 1;}
  else {if(e)return 0;}
  return(lo <= b.lo && b.hi <= hi);
}
bool interval::operator>>=(double s)const{
 if(e)return 0;
return( lo<=s && s<=hi);
}
bool operator>>=(double s, const interval & b){
 if(b.e)return 1;
return( s == b.lo && s == b.hi );
}


bool interval::operator<<(const interval & b)const{
 if(e){return 1;}
  else {if(b.e)return 0;}  
return( b.lo < lo && hi < b.hi);
}
bool interval::operator<<(double s)const{
 if(e)return 1;
return(lo>s && hi<s); //silly
}
bool operator<<(double s, const interval & b){
 if(b.e)return 0;
return( b.lo < s && s < b.hi );
} 
bool interval::operator>>(const interval & b)const{
 if(b.e){return 1;}
  else {if(e)return 0;}  
return(lo < b.lo && b.hi < hi);
}
bool interval::operator>>(double s)const{
 if(e)return 0;  
return( lo<s && s<hi);
}
bool operator>>(double s, const interval & b){
 if(b.e)return 1;  
return (b.lo>s && b.hi<s);//silly
}



// Output -------------------------------------------------------------------

void PrintHex(const interval &x){

  // Prints interval in hexadecimal base

 if(x.e){ cout << "[empty]";}  

 else{  
       int * plo = (int *) &(x.lo);
       int * phi = (int *) &(x.hi);
       printf("[%.8x%.8x,%.8x%.8x]" ,plo[1],plo[0],phi[1],phi[0]);
 }
}


bool ParseEndPoint(istream &stream, double &c){

		if('0'<=stream.peek() && stream.peek()<='9'){ 
				stream>>c;
				return 1;
		}
		else if(stream.peek()=='i'){ 
				stream.get();stream.get();stream.get(); //inf
				c=FP::PosInf;
				return 1; 
		}
		else if(stream.peek()=='m'){
				stream.get();stream.get();stream.get(); //max
				c=FP::Max;
				return 1; 
		}	
		else if (stream.peek()=='N'){
				stream.get();stream.get();stream.get(); //NaN
				c=FP::aNaN;
				return 1;
		}
		return 0;
}


istream & operator>>(istream &stream, interval &a){


	
/*

[	->	Mark of beginning of left endpoint
]	->	Mark of end of right endpoint
,	->	Mark of end of left endpoint and beginning of right endpoint 

inf 	->	FP::PosInf
+inf	->	FP::PosInf
-inf	->	FP::NegInf
max	->	FP::Max
+max	->	FP::Max
-max	->	FP::Min	
NaN	->	FP::aNaN
+NaN	->	FP::aNaN
-NaN	->	FP::aNaN


NO SPACES ALLOWED: Ex) [2.1, 3.2] contains space after the comma, this is not allowed.

*/


// Parsing left bracket


	char c = stream.peek(); if (c=='['){stream.get(c);} 
	while( c!='[' ){stream.get(c);}

		
// Parsing lower endpoint
		
		
		if	(ParseEndPoint(stream, a.lo)){}
		else if (stream.peek()=='-'){ 
			stream.get();
			if(ParseEndPoint(stream, a.lo)){a.lo = -a.lo;}
		}
		else if (stream.peek()=='+'){ 
			stream.get();
			if(ParseEndPoint(stream, a.lo)){}
		}
		else if (stream.peek()=='e'){ 
			 a=interval::Empty; return stream; 
		}
		else{cout<<endl<<"Cannot parse, UNKNOWN LOWER BOUND"; exit(1);}
		
		
		
		stream.get(); //comma
	        
// Parsing upper endpoint
	        
		if	(ParseEndPoint(stream, a.hi)){}
		else if (stream.peek()=='-'){ 
			stream.get();
			if(ParseEndPoint(stream, a.hi)){a.hi = -a.hi;}
		}
		else if (stream.peek()=='+'){ 
			stream.get();
			if(ParseEndPoint(stream, a.hi)){}
		}
		else{cout<<endl<<"Cannot parse, UNKNOWN UPPER BOUND"; exit(1);}
		
		
		stream.get(); //"]"
	
	return stream;	
}


ostream & operator<<(ostream &stream, const interval &a){

	if(a.e){ stream << "[empty]";}  
	else{
     
		stream << "[";
	 
		
		if		(a.lo == FP::NegInf){stream << "-inf";}
		else if	(a.lo == FP::Min)	{stream << "-max";}
	   	else if	(a.lo == FP::Max)	{stream << "max";}
	   	else if	(a.lo == FP::PosInf){stream << "+inf";}
	   	else{stream << a.lo;}
       
		stream << ",";
	
		
		if		(a.hi == FP::NegInf){stream << "-inf";}
	  	else if	(a.hi == FP::Min)	{stream << "-max";}
	   	else if	(a.hi == FP::Max)	{stream << "max";}
	   	else if	(a.hi == FP::PosInf){stream << "+inf";}
	   	else						{stream << a.hi; }

       	stream << "]";
	}
       
return stream;
}





interval FPpower(const interval &z, const interval &x, const interval &y){
	
	// Power z=x^y
	
	if	 (inf(x)>=0.0){ return z & exp(y*log(x)); }
	else if  (sup(x)<=0.0){ 
	 
 			interval a = abs(x); 
 			if (isEven(y)){ return FPpower (z,a,y); }
 			if (isOdd(y)){ return -FPpower(-z,a,y);} 
 			return  -(-z & exp( y*log(a))) | ( z & exp( y*log(a)) ); 
	}
 	else if  (inf(y)<0.0 && 0.0<sup(y)){ return z;}
	else {
		return FPpower(z, interval(inf(x),0.0),y) | FPpower(z,interval(0.0, sup(x)),y) ; //(1)
	}
}



void BPpower(const interval &z, interval &x, interval &y){
	
		
	if(0<=inf(x)){
		
		interval zpos = z & (x^y);	
				
		interval D,E;		
		if( !(y==0.0) ){
		
			xdiv(interval(1.0), y, D, E);
			x = ((zpos^D) & x) | ((zpos^E) & x);
				}
		if( !(x==0.0) ){
			
			xdiv(log(zpos), log(x), D, E);
			y = (D & y) | (E & y);
		}
	}
	
	else if(sup(x)<=0){
	
		interval a = abs(x);
        	interval z_negPart = abs(z) & (a^y);

        	BPpower(z_negPart, a, y);
        	x = -a & x;
	}
	
	else{
		interval xneg = interval(inf(x),0.0);
		interval xpos = interval(0.0,sup(x));
		interval y1(y), y2(y);
		BPpower(z, xpos, y1);
		BPpower(z, xneg, y2);
		x = xneg | xpos;
		y = y1 | y2; 			
	}

}


bool Intersect(interval &h, const interval &u){
h &= u;
if(isEmpty(h)) return 1;
return 0;
}
bool FPaddition(interval &h, const interval &u, const interval &v){
h &= u+v;
return isEmpty(h);
}
bool BPaddition(const interval &h, interval &u, interval &v){
u &= h-v;
v &= h-u;
return (isEmpty(u)||isEmpty(v));
}
bool FPsubtraction(interval &h, const interval &u, const interval &v){
h &= u-v;
return isEmpty(h);
}
bool BPsubtraction(const interval &h, interval &u, interval &v){
v &= u-h;
u &= h+v;
return (isEmpty(u)||isEmpty(v));
}
bool FPusubtraction(interval &h, const interval &u){
h &= -u;
return isEmpty(h);
}
bool BPusubtraction(const interval &h, interval &u){
u &= -h;
return isEmpty(u);
}

bool FPmultiplication(interval &h, const interval &u, const interval &v){
h &= u*v;
return isEmpty(h);
}
bool BPmultiplication(const interval &h, interval &u, interval &v){

DivideAndHull(u, h, v);
DivideAndHull(v, h, u);
return (isEmpty(u)||isEmpty(v));
}
bool FPdivision(interval &h, const interval &u, const interval &v){

DivideAndHull(h, u, v);
return isEmpty(h);
}
bool BPdivision(const interval &h, interval &u, interval &v){
u &= h*v;
DivideAndHull(v, u, h) ; 
return (isEmpty(u)||isEmpty(v));
}
bool FPexp(interval &h, const interval &u){
h &= exp(u);
return isEmpty(h);
}
bool BPexp(const interval &h, interval &u){
u &= log(h);
return isEmpty(u);
}
bool FPlog(interval &h, const interval &u){
h &= log(u);
return isEmpty(h);
}
bool BPlog(const interval &h, interval &u){
u &= exp(h);
return isEmpty(u);
}
bool FPpow(interval &h, const interval &u, const interval &v){
h=FPpower(h,u,v); //h &= exp(v*log(u));
return isEmpty(h);
}
bool BPpow(const interval &h, interval &u, interval &v){
BPpower(h,u,v);
return (isEmpty(u)||isEmpty(v));
}

bool FPabs(interval &h, const interval &u)
{
	h &= abs(u);
	return isEmpty(h);
}

bool BPabs(const interval &h, interval &u)
{
	interval hpos(h);
	if(hpos<0.0) return 1;
	if(h.lo<0.0) hpos.lo=0.0;
	
	u=hull(u&hpos, u&(-hpos));
	return isEmpty(u);
}

