#include "Structs/intervalUtilities.h"
#include "Structs/Tools.h"
#include <fstream>
#include <iostream>

using namespace std;



////////////////////////////////////////////////////////////////////////
// Basic Help functions
////////////////////////////////////////////////////////////////////////


VECTOR<interval> Hull(list<VECTOR<interval> > &L){

list<VECTOR<interval> >::iterator it(L.begin());
VECTOR<interval> H(*it);
it++;

while(it != L.end()){

	for(int i=1; i<= H.Size(); i++){
			H(i) = H(i) | (*it)(i);
	}
	it++;
}

return H;
}




interval hull(double a, double b){

if(a<b) return interval(a,b);
return interval(b,a);

}

double relativeDistance(interval &X, interval &Y){
	
	//safe
	
	double d1,d2;
	
	if (inf(X) != inf(Y)) d1=2.0*fabs( inf(Y)-inf(X) )  / (  fabs( inf(Y) )+fabs( inf(X) )  );
	else d1=0.0;
	
	if (sup(X) != sup(Y)) d2=2.0*fabs( sup(Y)-sup(X) )  / (  fabs( sup(Y) )+fabs( sup(X) )  );
	else d2=0.0;
	
	if (d1<d2) return d2;
	return d1;
}




void DoubleToInterval(VECTOR<interval> &Ix, const VECTOR<double> &x){
	for(int i=1;i<=x.Size(); i++){
		Ix(i)=interval(x(i));
	}
}
void setEmpty(interval &x){
	x = interval::Empty;
}
void setEmpty(VECTOR<interval> &x){
	for(int i=1; i<=x.Size(); i++){ x(i) = interval::Empty;}
}

bool isEmpty(const VECTOR<interval> &H){
	
	for(int i=1; i<=H.Size(); i++){ if(isEmpty(H(i))) return 1; }
	return 0;
}

bool isFinite(const VECTOR<interval> &X){

	for(int i=1; i<=X.Size(); i++){ 
		if( FP::isInf(inf(X(i))) || FP::isInf(sup(X(i)))  ){ return 0; }
	}
	return 1;
}

double SumOfWidths(const VECTOR<interval> & X){
	
	double s = width(X(1));
	for(int i=2; i<=X.Size(); i++){ s += width( X(i) ) ; }
	return s;
}

double Volume(const VECTOR<interval> & X){
	
	double s = width(X(1));
	for(int i=2; i<=X.Size(); i++){ s *= width( X(i) ) ; }
	return s;
}


void Intersection(VECTOR<interval> &I,
	  	  const VECTOR<interval> &X){

	// Intersection(I,X) stores the intersection of I and X in I
	
	for(int i=1; i<= I.Size(); i++){
		I(i) = I(i)& X(i);
	}
}

void Intersection(LIST<VECTOR<interval> > &L, const VECTOR<interval> &x){

	//Intersect each element of L with x and store in L

	L.MoveTop();
	while(!L.eol()){ 
		Intersection(L.Data(), x); 
		L.MoveDown();
	}
} 
void Intersection(VECTOR< VECTOR<interval> > &U,
	  	  const VECTOR<interval> &X){

	//Store the intersection of U and X in U

	for(int i=1; i<= U.Size(); i++)	{	
		Intersection(U(i),X);
	}
}

bool IntersectBoth(VECTOR< VECTOR<interval> > &U,
	  	   VECTOR<interval> &X){

	Intersection(U,X);
	Hull(X,U);
	if(isEmpty(X)) return 1;
	return 0;  	  
}

void Hull(interval &x, LIST<interval> &L){

	L.MoveTop();
	while(!L.eol()){
		x|=L.Data();
		L.MoveDown();
	}
}


void Hull(VECTOR<interval> &x, LIST<VECTOR<interval> > &L){

	L.MoveTop();
	while(!L.eol()){
		Hull(x,L.Data());
		L.MoveDown();
	}
}

void Hull(VECTOR<interval> &H,
	  const VECTOR<interval> &X1,
	  const VECTOR<interval> &X2){

	//Hull(H,X1,X2) stores the hull of X1 and X2 in H

	for(int i=1; i<= H.Size(); i++){
      		H(i) = X1(i)| X2(i);
    	}
}


void Hull(VECTOR<interval> &H,
	  const VECTOR<interval> &X){

	//Hull(H,X) stores the hull of H and X in H

	for(int i=1; i<= H.Size(); i++){
		H(i) = H(i) | X(i);
	}
}

void Hull(VECTOR<interval> &H, const VECTOR< VECTOR<interval> > &U){

  	// Hull(H,U) stores the hull of the subboxes of U in H

      H = U(1);
      for(int j=2; j<=U.Size(); j++){ Hull(H, U(j) ); }    
}

void mid(VECTOR<double> &m, VECTOR<interval> &V){

	for(int i=1; i<=m.Size(); i++){ 
		m(i) = mid(V(i));
	}
}


////////////////////////////////////////////////////////////////////////
// leq, geq
////////////////////////////////////////////////////////////////////////

interval leq(double x){ return interval(FP::NegInf,x);}
interval geq(double x){ return interval(x,FP::PosInf);}

////////////////////////////////////////////////////////////////////////
// Vertex
////////////////////////////////////////////////////////////////////////


void Vertex(int i, const VECTOR<interval> &x, VECTOR<double> &z, LIST< VECTOR<double> > &L){

	// Store vertices of the box x in the list L.
	// Call: Vertex(0,x,z,L)

	if (i==x.Size()){ L.push(z); }
	else {
		z(i+1)=inf(x(i+1)); Vertex(i+1,x,z, L);
		z(i+1)=sup(x(i+1)); Vertex(i+1,x,z, L);
	}
}


////////////////////////////////////////////////////////////////////////
// Set Difference
////////////////////////////////////////////////////////////////////////

interval LComplement(const interval &x){

	//LowerComplement
	
	if(isEmpty(x)) 		return interval::Entire;
	if(inf(x)==FP::NegInf)	return interval::Empty;
	
	return interval(FP::NegInf,inf(x)); 
}

interval UComplement(const interval &x){

	//UpperComplement
	
	if(isEmpty(x)) 		return interval::Entire;
	if(sup(x)==FP::PosInf)	return interval::Empty;
	
	return interval(sup(x),FP::PosInf); 
}

void SetDifference(LIST<interval> &L, const interval &y, const interval &x){

	
	// Push cl(y\x) on L
	
	if(isEmpty(y)){return;}
	if(isEmpty(x)){L.push(y); return;}
	
	//nonoverlapping
	if ( inf(x) >= sup(y) || sup(x) <= inf(y) ){
		L.push(y);
	}
	//both endpoints of x cut int(y)
	else if( inf(x) > inf(y) && sup(x)<sup(y) ){ 
		
		L.push(interval(inf(y),inf(x))); 
		L.push(interval(sup(x),sup(y))); 
	}
	//inf(x) cuts int(y)
       	else if( inf(x) > inf(y)){ 
       		L.push(interval(inf(y),inf(x))); 
       	}
       	//sup(x) cuts int(y)
       	else if( sup(x) < sup(y)){
       		L.push(interval(sup(x),sup(y)));
       	}
       	// no endpoint of x cuts int(y)
       	else{}
}



void SetDifference(LIST<VECTOR<interval> > &L, VECTOR<interval> y, VECTOR<interval> x){

////////////////////////////////////////////////////////////////////////
// Push cl(y/x) on the list L.
// No push if y<<=x.
////////////////////////////////////////////////////////////////////////
 

	int n = x.Size();
	VECTOR<int> B(n);
	VECTOR<interval> z(n);

	if(In<interval>(y,x)) return;
  
	for (int i=1; i<=n; i++){
    
		if ( inf(x(i)) >= sup(y(i)) || sup(x(i)) <= inf(y(i)) ){ L.push(y); return; }
       		else if( inf(x(i)) > inf(y(i)) && sup(x(i)) < sup(y(i))){ B(i) = 2; }
       		else if( inf(x(i)) > inf(y(i))){ B(i) = 0; }
       		else if( sup(x(i)) < sup(y(i))){ B(i) = 1; }
       		else                           { B(i) = -1;}
  	}

  

  
	for (int i=1; i<=n; i++){
    
	    if      ( B(i) == 0){
	                      z = y;
	                      z(i) = interval( inf(y(i)), inf(x(i)) ); 
	                      y(i) = interval( inf(x(i)), sup(y(i)) );
			      L.push(z);
	    }	      
	    else if ( B(i) == 1 ){
	                      z = y;
	                      z(i) = interval( sup(x(i)), sup(y(i)) ); 
	                      y(i) = interval( inf(y(i)), sup(x(i)) );
			      L.push(z);
	    }
	    else if ( B(i) == 2 ){
	                      z = y;
	                      z(i) = interval( inf(y(i)), inf(x(i)) ); 
	                      y(i) = interval( inf(x(i)), sup(y(i)) );
			      L.push(z);

			      z = y;
	                      z(i) = interval( sup(x(i)), sup(y(i)) ); 
	                      y(i) = interval( inf(y(i)), sup(x(i)) );
			      L.push(z);
	    }
    else {}
    
  }
}




void SetDifference(LIST<VECTOR<interval> > &R, const VECTOR<interval> &x, LIST<VECTOR<interval> > &L){

////////////////////////////////////////////////////////////////////////
//       R := x\L
////////////////////////////////////////////////////////////////////////

  
	VECTOR<interval> r; 

	if(L.empty()){ R.push(x); return; }
	else{
	    	L.MoveTop();
            	SetDifference(R, x, L.Data());
		L.MoveDown();
	}
	while( !L.eol() && !R.empty() ){

	      	R.init();
		while( R.FINT() ){               //while( R.FINT() && !R.empty() )// behövs inte 
                     r = R.pop();
	             SetDifference(R, r, L.Data());
        	      	L.MoveDown();
		}
	}
}



bool isIncluded(const VECTOR<interval> &x, LIST<VECTOR<interval> > &L){

	LIST<VECTOR<interval> > R;
	SetDifference(R,x,L);
	return R.empty();
}


void DisjointUnion(LIST<VECTOR<interval> > &L, VECTOR<interval> x, VECTOR<interval> y){

	// Push x\y and y on the list L.

	if ( In<interval>(x,y) ){ L.push(y); return; }
	if ( In<interval>(y,x) ){ L.push(x); return; }
	SetDifference(L,x,y);
	L.push(y);
}


void DisjointUnion(LIST<VECTOR<interval> > &L){

////////////////////////////////////////////////////////////////////////
// Given a list L of boxes, we return a list L that 
//	preserves the union but all its boxes are
// nonempty and pairwise nonoverlapping 
//	( their interiors have empty intersection ).
////////////////////////////////////////////////////////////////////////


  
	VECTOR<interval> x,y;
	bool flag;
	LIST<VECTOR<interval> > R;
  
	while(!L.empty()){
	        x = L.pop(); 
        	if ( isEmpty(x) ){ continue; } 	 
        	R.init(); // <-> The next box that is pushed on R becomes the first_in.
		flag = 1;

		while ( R.FINT() && flag ){
	 
	       		y = R.pop();
	      		 if( In<interval>(x,y) ){ R.push(y); flag = 0; }
	       		else{ SetDifference(R,y,x);     }
		}
		if (flag == 1){ R.push(x); }
   	}
  	L = R;
}



void DisjointUnion(LIST<VECTOR<interval> > & L1, LIST<VECTOR<interval> > & L2){

  // Returns the union of L1 and L2 in L1.
  // L2 is deleted.

  while(!L2.empty()){ L1.push( L2.pop() ); } 
  DisjointUnion(L1);
}


void Cube(VECTOR<interval> &x, double side){

	for(int i=1; i<=x.Size();i++){
		x(i)=interval(-side/2.0,side/2.0);	
	}
}


void old_ConnectedComponents(LIST<VECTOR<interval> >  L, 
											LIST< LIST< VECTOR<interval > > > &COMP)
{
////////////////////////////////////////////////////////////////////////
// Returns a list of lists, where each list contains a
// connected component of L.
// Computationally EXPENSIVE.
////////////////////////////////////////////////////////////////////////
  
  LIST<VECTOR<interval> > A,B;
  VECTOR<interval> a;
  
 
  while ( !(L.empty()) )
  {
        a = L.pop(); 
        A.Clean(); 
        A.push(a);
		COMP.MoveTop();
		while( (!COMP.empty()) && (!COMP.eol()) )
		{
	      B = COMP.Data();
	      if   ( old_Intersection(A,B) )
	      	{	
	      		//A = Union(A,B);
	      		Push(A,B); 
	      		COMP.Remove(); 
	      		COMP.MoveTop(); 
	      	}
	      else
	      {
	      	 COMP.MoveDown(); 
	       }
		}
        COMP.push(A);
  }


}


VECTOR<interval> old_Hull(LIST<VECTOR<interval> > & L){

  // Hulls all elements of L
  
  VECTOR<interval> x;
  if( !L.empty() )
  {
	x = L.pop(); 	
	while(!L.empty()){ Hull(x,L.pop()); }
	return x;
  }
  
}


bool old_Intersection(const VECTOR<interval> &a, 
								const VECTOR<interval> &b){
for(int i=1;i<=a.Size();i++){
	if( sup(a(i))<inf(b(i)) || sup(b(i))<inf(a(i)) ) return 0;
}
return 1;
}

bool old_Intersection (const VECTOR<interval> &x, 
								LIST<VECTOR<interval> > &L)
{
  // Tests if x intersects nonempty with an element of L.
  
if( L.empty() ) return 0;
L.MoveTop();
while(!L.eol())
{ 
  	if ( old_Intersection( x,L.Data() ) ) return 1; 
  	L.MoveDown(); 
}
return 0;
}

bool old_Intersection (LIST<VECTOR<interval> > &A,
								 LIST<VECTOR<interval> > &B){

  // Tests if the lists A, B intersect nonempty.
  
	if( A.empty() || B.empty() ) return 0;
	A.MoveTop();
	while(!A.eol())
	{
        B.MoveTop();
        while(!B.eol())
        {
	  	  if( old_Intersection( A.Data(), B.Data() ) ) return 1;
	  	  B.MoveDown();
		}
		A.MoveDown();
 	}
  return 0;
}


////////////////////////////
// STORAGE
////////////////////////////



bool WriteBoxesVertexFormat(LIST<VECTOR<interval> > L, const std::string &filename){

	//Save boxes in vertex format

	int n;  
	if ( L.empty() ){ cout<<endl<<"WriteBoxesVertexFormat: List is Empty"; return 0; }
	else{ L.MoveTop(); n = L.Data().Size(); }

	if (n==3){
  
		ofstream o(filename.c_str());
		if(!o){cout<<endl<<"Cannot write to file"; return 0;}
		LIST<VECTOR<double> > K;
		VECTOR<interval> x(3);
		VECTOR<double> z(3);
 		
 		while ( !L.empty() ){
     			x = L.pop();
     			if(!isEmpty(x)){

     				K.Clean(); Vertex(0,x,z,K);
     				K.MoveTop(); int k=1;
     				while(k<=8){
     					o<<K.Data()(1)<<" "<<K.Data()(2)<<" "<<K.Data()(3)<<endl;
     					K.MoveDown();
     					k++;
  				}
     			}
 		}
        o.close();
	}

	else if (n==2){
 
		ofstream o(filename.c_str());
		VECTOR<interval> x(2);
		
		while ( !L.empty() ){

        		x=L.pop();
	 			o << inf(x(1)) << " " << inf(x(2)) << endl;
        		o << sup(x(1)) << " " << inf(x(2)) << endl;
        		o << sup(x(1)) << " " << sup(x(2)) << endl;
        		o << inf(x(1)) << " " << sup(x(2)) << endl;
	 	
	 	}
	o.close();
	}
return 1;	
}


bool WriteBoxesInlineFormat(LIST<VECTOR<interval> > &L, const std::string &filename){
  

  //***************************************************************
  //
  // Writes m=L.Size() VECTOR<interval>'s of dimension n into a textfile 
  // using the format:
  //
  //
  //   m             n     
  // 
  //   inf( x1(1) )  sup (x1(1) ) ...  inf( x1(n) ) sup( x1(n) )
  //   ...
  //   ...
  //   inf( xm(1) )  sup( xm(1) ) ... inf( xm(n) )  sup( xm(n) )
  //
  //**************************************************************




  if (L.empty()){ cout<<endl<<"WriteBoxesInlineFormat: List is empty"; return 0; }

  ofstream utobj(filename.c_str());
  cout << endl << "Storing Boxes  in " << filename << endl;

   
  
  L.MoveTop(); 
  int n = L.Data().Size(); 

  int m = L.Size();
  utobj << m << " " << n << endl;
  

  VECTOR<interval> x(n);
  while ( !L.eol() ){
    x = L.Data();
    for(int i=1; i<=n; i++){ utobj << inf( x(i) ) << " " << sup( x(i) ) << " "; }
    utobj << endl;
    L.MoveDown();
  
  }
 
  utobj.close();

  return 1;
}


bool ReadBoxesInlineFormat(LIST<VECTOR<interval> > &L, const std::string filename){
  

  //***************************************************************
  //
  // Reads boxes from a textfile "filename" that has 
  // m boxes, each of dimension n, and the boxes have a line-up 
  // representation:
  //
  //
  //   m             n            
  // 
  //   inf( x1(1) )  sup (x1(1) ) ...  inf( x1(n) ) sup( x1(n) )
  //   ...
  //   ...
  //   inf( xm(1) )  sup( xm(1) ) ... inf( xm(n) )  sup( xm(n) )
  //
  // ... and store the boxes as VECTOR<interval>'s in L.
  //
  //**************************************************************




  
  ifstream inobj(filename.c_str());
  cout << endl << "Reading Boxes in " << filename << endl;
 
   
  int m,n;
  inobj >> m;
  inobj >> n;

  VECTOR<interval> x(n);
  double lo,hi;
  for (int i = 1; i <= m; i++){ 
    for (int j = 1; j <= n; j++){
      
      inobj >> lo; 
      inobj >> hi;
      x(j) = interval(lo,hi);  
    }
    L.push(x);
  }
  cout << endl << "Reading Boxes Completed" <<endl;

  return 1;
}



bool WriteBoxes(LIST<VECTOR<interval> > &L, const std::string &filename){

////////////////////////////////////////////////////////////////////////
//	L.Size() VECTOR<interval>'s of dimension n into 
//	a textfile using the format
// 
//   inf( x1(1) )  sup (x1(1) ) ...  inf( x1(n) ) sup( x1(n) )
//   ...
//   ...
//   inf( xm(1) )  sup( xm(1) ) ... inf( xm(n) )  sup( xm(n) )
//
////////////////////////////////////////////////////////////////////////

if (L.empty()){ cout<<endl<<"WriteBoxes: List is empty"; return 0; }

ofstream o(filename.c_str());
  //cout<<"\nWriting Boxes to "<<filename<<"\n";

L.MoveTop(); 
int n = L.Data().Size(); 

VECTOR<interval> x(n);
while ( !L.eol() ){
	x = L.Data();
    for(int i=1; i<=n; i++){ o<< inf( x(i) ) << " " << sup( x(i) ) << " "; }
    o<<endl;
    L.MoveDown();
}
  // cout<<"\nWriting Boxes to "<<filename<<" completed \n";

return 1;
}


bool ReadBoxes(LIST<VECTOR<interval> > &L, const std::string filename, int dim){

////////////////////////////////////////////////////////////////////////
// Reads dim dimensional boxes from a textfile 
// "filename" with boxes that have a line  
// representation:
// 
//   inf( x1(1) )  sup (x1(1) ) ...  inf( x1(n) ) sup( x1(n) )
//   ...
//   ...
//   inf( xm(1) )  sup( xm(1) ) ... inf( xm(n) )  sup( xm(n) )
//
// ... and pushes the boxes to L.
//
////////////////////////////////////////////////////////////////////////


ifstream file(filename.c_str());
cout<<"\nReading Boxes in " << filename <<"\n";

VECTOR<interval> x(dim);
double lo,hi;

while(1){

	for (int j=1; j<=dim; j++){
      
		file >> lo; 
		file >> hi;
		x(j)	=	interval(lo,hi);  
    
	}
	if( file.eof() ) break;
	L.push(x);

}
cout<<"\nReading Boxes in "<<filename <<" completed\n";
return 1;
}




bool WriteRealVectorsInlineFormat(LIST<VECTOR<double> > &L, const std::string filename){
  
  
  //***************************************
  //
  // Stores the boxes of L in the file 
  // "filename" using the format 
  // 
  //   x1(1) x1(2) x1(3) ..x1(n)
  //   ...
  //   ...
  //   xm(1) xm(2) xm(3).. xm(n)
  //
  //**************************************



  if (L.empty()){ cout << endl << "WriteRealVectors: List is empty"; return 0; }

  ofstream utobj(filename.c_str());
  cout << endl << "Storing Real Vectors In  " << filename << endl;



  L.MoveTop();
  int n = L.Data().Size();
  VECTOR<double> x(n);
  while ( !L.eol() ){
   
    x = L.Data();
    for(int i=1; i<=n; i++){ utobj << x(i) << " "; }
    utobj << endl;
    L.MoveDown();

  }
 
  utobj.close();

  return 1;
}


//Merge


void MergeAdjacent(LIST<interval> &M){
   
    // Merge adjacent intervals in M
    // If SeparationDistance = 0 Then Merge
    // Only merge of intervals that are positioned consecutively in the list
  
	interval a,b;
  
	if ( !M.empty() ){
		a = M.pop();
		M.init();         //the next element that will be pushed on M is the first_in
		while ( M.FINT() ){
			b = M.pop();
			if (SeparationDistance(a,b)==0.0){ a = a|b; }
			else{ M.push(a); a = b;}
		}
		M.push(a);
	}
}



bool Merge(VECTOR<interval> &a, VECTOR<interval> &b){
  
	int flag = 0;
	int iHull;
  
	for(int i=1; i<=a.Size(); i++){
 
		if ( !(a(i) == b(i)) ){
			if ( flag || ( inf(a(i)) != sup(b(i)) && sup(a(i)) != inf(b(i)) )   ){ return 0; }
			else { flag = 1; iHull=i; }
		}
  	}
  	if(flag==1){ a(iHull) |= b(iHull); }
  	return 1;
}


void MergeAdjacent( LIST<VECTOR<interval> > &L){
	
	int dim;
	if( !L.empty() ){ L.MoveTop(); dim = L.Data().Size(); }
	else {return;}
	VECTOR<interval> a(dim),b(dim);
	
	int Lsize = L.Size() + 1;
	while( Lsize>L.Size() ){
      		Lsize = L.Size();
  
       		if ( !(L.empty()) ){  
            		a = L.pop();
            		b = a;
            		L.init();         // The next element that will be pushed on L is the "first_in" element. 
            		while ( L.FINT() ){
                   		b = L.pop();
		   		if (!Merge(a,b)){ L.push(a); a = b; }
	    		}
	    		L.push(a);
       		}
	}
}

 
bool StoreIntervalVECTOR(const VECTOR<interval> &v, const string &filename){
 
	ofstream o(filename.c_str());
	if(!o){cout<<endl<<"Cannot write to file"; return 0;}
		
	for(int i=1; i<=v.Size(); i++){ 
		o<< inf(v(i)) << " "<< sup(v(i)) << endl; 
	}
	o.close();
	
  return 1;
}







