// Returns a uniform random variable from [min, max].

double uniformRnd(double min, double max)
{
  return ((double) rand()/RAND_MAX)*(max - min) + min;
}



// Polar (Box-Mueller) method; See Knuth v2, 3rd ed, p122.
// Return a gaussian random variable from N(0, sigma).


double gaussianRnd(const double sigma)
{
  double x, y, r;
  do {                                 // Pick x,y uniformly from [-1,-1]2.
    x = uniformRnd(-1, +1);
    y = uniformRnd(-1, +1);
    r = x*x + y*y;
  }
  while ( (r > 1.0) || (r == 0) );     // See if it is in the unit circle.

  return sigma*y*sqrt(-2.0*log(r)/r);  // Box-Muller transform.
} 

void Generate_NormalSamples(VECTOR<double> &x,
			    double mu,
			    double sigma){
			   			
	for(int i=1; i<=x.Size(); i++){
		x(i) = mu + gaussianRnd(sigma);
	}
}


void Generate_LogNormalSamples(VECTOR<double> &x,
			    double mu,
			    double sigma){
			   			
	for(int i=1; i<=x.Size(); i++){
		x(i) = exp(mu + gaussianRnd(sigma));
	}
}

double LogNormalSample(double mu, double sigma){
	
	return exp(mu + gaussianRnd(sigma));

}


void Mean(double &m, const VECTOR<double> &X){

	int n = X.Size();
	m = 0.0;
	for (int i=1; i<=n; i++){
		m+=X(i);
	}
	m/=(double)(n);
}

void MeanOfMean(double &m, const VECTOR<interval> &X){

	int n = X.Size();
	m = 0.0;
	for (int i=1; i<=n; i++){
		m+=mid(X(i));
	}
	m/=(double)(n);
}




interval RandIv(){

  double a = 0.0001*( rand()%10000   );
  double b = 0.0001*( rand()%10000  );
  if (a<b)  return interval(a,b); 
  else      return interval(b,a);
}



