using namespace std;

#include <iostream>
#include <iomanip>
#include <fstream>
#include <math.h> 
#include <string>
#include <stdio.h>
#include <stdlib.h>
#include <cmath>
#include "Structs/interval.h"
#include "Structs/ParameterReader.h"
#include "Structs/Tools.h"
#include "Structs/symUtilities.h"
#include "DAG/GraphViz.h"

  
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
//  Constants
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
const int gi_max=10000; // max number of individuals per group
const int g_max=12; // max number of groups
const int npts_max=4; // max number of time-points per time-series
const int nParameter_max=6; // # max number of parameters in model
const int pListSize_max=50000;


// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
//  Model and design variables
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
int model=1; // # model 1 or 2
int action=1; // # 1: Plot model, 2: Parameter estimation, 3: optimal design
int subaction=1; // 1 is standard value, >1 for alternatives, see README file
int nParameter=2; // # parameters in model
int groups=1; // number of groups
double T_MAX=24.0; // max time to sample from, min time is always 0
int ITER=1; // Iterations for each individual for a given design 
double epsilon=0.0; // noise added
double alpha=0.0; // uncertainty in the true parameters beta_pop
double sampleWindow; // sample window
double doseWindow=0.0; // dose window
double lambda; // lambda, constant for the impact of bias in error function
int nTimePointsInSearchDomain=10; // discrete time-points in design domain
int npts[g_max]; // time-points per time-series in each group
int gi[g_max]; // number of individuals per group
int doplot;
double timeLB[g_max][npts_max]; // group, time lower bound
double timeUB[g_max][npts_max]; // group, time upper bound
double psharp[nParameter_max]; // Standard parameter values
interval psharp_iv[nParameter_max]; // Standard parameter values as interval
interval pstar[nParameter_max]; // Correct parameter for an individual, used to caluclate bias in error term.
double dpsharp[nParameter_max]; // Standard parameter values + random value from d
double d[nParameter_max]; // Standard d-values
double psearch[nParameter_max][2]; // Search domain
double dose[g_max]; // dose
interval p[gi_max][nParameter_max]; // p found
interval pexp[gi_max][g_max][nParameter_max]; // p found for one full experiment
interval pexptem[gi_max][g_max][nParameter_max]; // p found for one full experiment
double pest[nParameter_max]; // p found estimated mean
double pestsigma2[nParameter_max]; // p found estimated sigma2=variance
double errorf[g_max][gi_max]; // errorf found for group and individual
int pListSize;
interval pList[nParameter_max][pListSize_max]; // list of all sub p found
int pListLabel[pListSize_max]; // initially zero <-> unassigned
interval pListHull[nParameter_max][pListSize_max]; // Hull of each label
int numberOfLabels;// number of used labels in pListLabel
string stringcounter[pListSize_max]; // used in plot3d









// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
//  Function definitions
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
void writeProblem(int,int);
void readOutput(int,int);
void plot(string);
void plothist(string);
void plot3d(string,int,int);
double paraest(int,bool,int,int);
double uniformRnd(double,double);
void plotModel(int,string);
void meanandvar(int);
void labelPlist();









// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
//  Main program
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
int main(int argc, char *argv[]) {



  // ---------------------------------------------------------------------------
  //
  // Initial echo
  //  
  // ---------------------------------------------------------------------------
  cout << "Run program " << argv[0] << " with arguments " << endl;
  for (int i = 1; i < argc; i++) {
    cout << argv[i] << " ";
  }
  cout << endl;



  // ---------------------------------------------------------------------------
  //
  // Read input file
  //  
  // ---------------------------------------------------------------------------
  ParameterReader pr;
  string s=argv[1];
  pr.readFile(s);
  action = parseString<int>( pr.find("Action") );// # 1: Plot model, 2: Parameter estimation, 3: optimal design
  subaction = parseString<int>( pr.find("SubAction") );// 
  model = parseString<int>( pr.find("Model") );// # model 1 or 2
  nParameter = parseString<int>( pr.find("nParameter") );// # data-points per time-series
  groups = parseString<int>( pr.find("nGroups") );// number of groups
  for (int i=0;i< groups;i++){
    string st="nInGroup"+ToString(i);
    gi[i] = parseString<int>( pr.find(st) );// individuals in group i  
    st="nptsInGroup"+ToString(i);
    npts[i] = parseString<int>( pr.find(st));// # data-points per time-series in group i  
  }
  T_MAX = parseString<double>( pr.find("T_MAX") );// max time to sample from, min time is always 0
  nTimePointsInSearchDomain = parseString<int>( pr.find("nTimePointsInSearchDomain") );// sample window
  sampleWindow = parseString<double>( pr.find("sampleWindow") );// sample window
  string sss=pr.find("doseWindow");
  if (sss.length()>0)
    doseWindow = parseString<double>( pr.find("doseWindow") );// dose window
  else
    doseWindow = 0.0;
  ITER = parseString<int>( pr.find("ITER") );// Iterations for each individual for a given design 
  epsilon = parseString<double>( pr.find("epsilon") ); // noise added
  alpha = parseString<double>( pr.find("alpha") );// uncertainty in the true parameters beta_pop
  lambda = parseString<double>( pr.find("lambda") ); // constant for bias impact in error function
  int debug = parseString<int>( pr.find("debug") );// debug
  doplot= parseString<int>( pr.find("doplot") );// plot when running (no 0 or yes 1)?
  int randomSeed = parseString<int>( pr.find("randomSeed") );// randomSeed (0: take from clock)




  // ---------------------------------------------------------------------------
  //
  // Echo essential variables to standard output
  //  
  // ---------------------------------------------------------------------------
  cout<<"Action "<<action<< endl;
  cout<<"SubAction "<<subaction<< endl;
  cout<<"Model "<<model<< endl;
  cout<<"nParameter "<<nParameter<< endl;
  cout<<"epsilon "<<epsilon<< endl;
  cout<<"groups "<<groups<< endl; 
  for (int i=0;i< groups;i++){
    cout<<"Individuals in group "<<i<<"="<<gi[i]<< endl;
    cout<<"Data-points per time-series in group "<<i<<"="<<npts[i]<< endl;
  }
  cout<<"T_MAX "<<T_MAX<< endl;
  cout<<"nTimePointsInSearchDomain "<<nTimePointsInSearchDomain<< endl;
  cout<<"sampleWindow "<<sampleWindow<< endl;
  cout<<"doseWindow "<<doseWindow<< endl;
  cout<<"ITER "<<ITER<< endl;
  cout<<"alpha "<<alpha<< endl;  
  cout<<"lambda "<<lambda<< endl;
  cout<<"debug "<<debug<< endl;
  cout<<"doplot "<<doplot<< endl;
  cout<<"randomSeed "<<randomSeed<< endl;
  if (randomSeed==0) {
    randomSeed=time(0);
    cout<<"new randomSeed "<<randomSeed<< endl;
  }
  srand(randomSeed);

  string filename="action_";
  if (action==1) filename+="1";
  else if (action==2) filename+="2";
  else if (action==3) filename+="3";
  else {cout<<"action non standard"; exit(1);}
  if (subaction==1) filename+="_1";
  else if (subaction==2) filename+="_2";
  else if (subaction==3) filename+="_3";
  else if (subaction==4) filename+="_4";
  else if (subaction==5) filename+="_5";
  else if (subaction==6) filename+="_6";
  else if (subaction==7) filename+="_7";
  else if (subaction==8) filename+="_8";
  else if (subaction==9) filename+="_9";
  else if (subaction==10) filename+="_10";
  else {cout<<"subaction non standard"; exit(1);}
  if (model==1) filename+="_model_1";
  if (model==2) filename+="_model_2";
  if (model==3) filename+="_model_3";
  if (action>1) {
    if (epsilon>0.0) filename+="_noisyData";
    else filename+="_prfctData";
    if (alpha==0.0) filename+="_alpha_00";
    else if (alpha==0.15) filename+="_alpha_15";
    else if (alpha==0.30) filename+="_alpha_30";
    else {cout<<"alpha non standard";filename+="_alpha_?";}
    if (action>2) {
      if (lambda>0.0) filename+="_biasInErrFunc";
      else filename+="_standErrFunc";
    }
  }
  cout<<"Constructed filename to save images in: "<<filename<<endl;






  // ---------------------------------------------------------------------------
  //
  // Default setings Model I
  //  
  // ---------------------------------------------------------------------------
  if (model==1) {
    psharp[0]=3.0;
    psharp[1]=30.0;
    psharp[2]=0.0;
    psharp[3]=0.0;
    d[0]=0.0; 
    d[1]=0.0; 
    d[2]=0.25;
    d[3]=25;
    psearch[0][0]=3.0;
    psearch[0][1]=3.0;
    psearch[1][0]=30.0;
    psearch[1][1]=30.0;

    psearch[2][0]=-2;
    psearch[2][1]=2;
    psearch[3][0]=-20;
    psearch[3][1]=20;

    dose[0]= 450;
    for (int i=0; i<groups; i++) {
      dose[i]= dose[0];
      if (gi[i]>gi_max) {
	cout<<"Increase gi_max!"<<endl;
	return 1;
      }
    }
    psharp_iv[0]=interval(psharp[0]-d[2]*alpha,psharp[0]+d[2]*alpha); 
    psharp_iv[1]=interval(psharp[1]-d[3]*alpha,psharp[1]+d[3]*alpha); 
    psharp_iv[2]=interval(psharp[2],psharp[2]); 
    psharp_iv[3]=interval(psharp[3],psharp[3]);

    if (alpha>0.0) {
      psearch[0][0]=1;
      psearch[0][1]=5;
      psearch[1][0]=10;
      psearch[1][1]=50;
    }
  }
 



  
  // ---------------------------------------------------------------------------
  //
  // Default setings Model II
  //  
  // ---------------------------------------------------------------------------
  if (model==2) {
    psharp[0]=2.71;
    psharp[1]=0.0763;
    psharp[2]=0.0373;
    psharp[3]=0.0;
    psharp[4]=0.0;
    psharp[5]=0.0;

    d[0]=0.0; 
    d[1]=0.0;
    d[2]=0.0; 
    d[3]=0.7840;
    d[4]=0.0185;
    d[5]=0.0238;

    psearch[0][0]=2.71;
    psearch[0][1]=2.71;
    psearch[1][0]=0.0763;
    psearch[1][1]=0.0763;
    psearch[2][0]=0.0373;
    psearch[2][1]=0.0373;
    psearch[3][0]=-5.0;
    psearch[3][1]=5.0;
    psearch[4][0]=-3.0;
    psearch[4][1]=3.0;
    psearch[5][0]=-3.0;
    psearch[5][1]=3.0;

    if (action==2 && subaction==1){
      psearch[3][0]=-8.0;
      psearch[3][1]=8.0;
      psearch[4][0]=-5.0;
      psearch[4][1]=5.0;
      psearch[5][0]=-5.0;
      psearch[5][1]=5.0;
    }

    dose[0]= 4.02;
    dose[1]= 4.40;
    dose[2]= 4.53;
    dose[3]= 4.40;
    dose[4]= 5.86;
    dose[5]= 4.00;
    dose[6]= 4.95;
    dose[7]= 4.53;
    dose[8]= 3.10;
    dose[9]= 5.50;
    dose[10]=4.92;
    dose[11]=5.30;

    if (subaction==2){
      dose[0]= 5.86;
      dose[1]= 4.6;
      dose[2]= 3.1;
    }

    for (int i=0; i<groups; i++) {
      if (gi[i]>gi_max) {
	cout<<"Increase gi_max!"<<endl;
	return 1;
      }
    }
    psharp_iv[0]=interval(psharp[0]*exp(-d[3]*alpha),psharp[0]*exp(d[3]*alpha)); 
    psharp_iv[1]=interval(psharp[1]*exp(-d[4]*alpha),psharp[1]*exp(d[4]*alpha)); 
    psharp_iv[2]=interval(psharp[2]*exp(-d[5]*alpha),psharp[2]*exp(d[5]*alpha)); 
    psharp_iv[3]=interval(psharp[3],psharp[3]); 
    psharp_iv[4]=interval(psharp[4],psharp[4]); 
    psharp_iv[5]=interval(psharp[5],psharp[5]);

    if (alpha>0.0) {
      psearch[0][0]=0.0785;    
      psearch[0][1]=93.5690;
      psearch[1][0]=0.0443;
      psearch[1][1]=0.1315;
      psearch[2][0]=0.0201;
      psearch[2][1]=0.0691;
    }

  }






  // ---------------------------------------------------------------------------
  //
  // Default sampling times
  //  
  // ---------------------------------------------------------------------------
  for (int igroup=0;igroup< groups;igroup++){
    timeLB[igroup][0]=0.5+igroup*0.25;
    timeUB[igroup][0]=timeLB[igroup][0]+sampleWindow;
    timeLB[igroup][1]=3+igroup*0.25;
    timeUB[igroup][1]=timeLB[igroup][1]+sampleWindow;
    timeLB[igroup][2]=15+igroup*0.25;
    timeUB[igroup][2]=timeLB[igroup][2]+sampleWindow;
  }
  


   
  // ---------------------------------------------------------------------------
  //
  // Initiate string counter
  //  
  // ---------------------------------------------------------------------------
  for (int i=0;i<pListSize_max;i++) stringcounter[i]=ToString(i);





  // ---------------------------------------------------------------------------
  //
  // action == 1 
  // Plot
  //  
  // ---------------------------------------------------------------------------
  if (action==1) {
    plot(filename); // gnuplot function for various doses
    return 1;
  }
   

   
   
  // ---------------------------------------------------------------------------
  //
  // action == 2 
  // Parameter estimation
  //  
  // ---------------------------------------------------------------------------
  if (action==2){
    // Use ITER=1
    // Choose desired gi[igroup], e.g., 500
    for (int igroup=0;igroup< groups;igroup++){
       
      if (model==1) {
	timeLB[igroup][0]=0;
	timeUB[igroup][0]=timeLB[igroup][0]+sampleWindow;
	timeLB[igroup][1]=24;
	timeUB[igroup][1]=timeLB[igroup][1]+sampleWindow;
	if (npts[igroup]==3) {
	  timeLB[igroup][0]=0;
	  timeUB[igroup][0]=timeLB[igroup][0]+sampleWindow;
	  timeLB[igroup][1]=0;
	  timeUB[igroup][1]=timeLB[igroup][1]+sampleWindow;
	  timeLB[igroup][2]=24;
	  timeUB[igroup][2]=timeLB[igroup][2]+sampleWindow;
	}
      }
      if (model==2) {
	timeLB[igroup][0]=0.2;
	timeUB[igroup][0]=timeLB[igroup][0]+sampleWindow;
	timeLB[igroup][1]=3;
	timeUB[igroup][1]=timeLB[igroup][1]+sampleWindow;
	timeLB[igroup][2]=15;
	timeUB[igroup][2]=timeLB[igroup][2]+sampleWindow;
      }

      // PopED optimal design of problem 1 for model 2
      if (subaction==8){
	timeLB[igroup][0]=0.31;
	timeUB[igroup][0]=timeLB[igroup][0]+sampleWindow;
	timeLB[igroup][1]=1.2;
	timeUB[igroup][1]=timeLB[igroup][1]+sampleWindow;
	timeLB[igroup][2]=15;
	timeUB[igroup][2]=timeLB[igroup][2]+sampleWindow;
	if (igroup==4){
	  timeLB[igroup][0]=1.2;
	  timeUB[igroup][0]=timeLB[igroup][0]+sampleWindow;
	  timeLB[igroup][1]=2.8;
	  timeUB[igroup][1]=timeLB[igroup][1]+sampleWindow;
	}
	if (igroup==8){
	  timeLB[igroup][2]=2.8;
	  timeUB[igroup][2]=timeLB[igroup][2]+sampleWindow;
	}
      }


      // My optimal design of problem 1 for model 2
      if (subaction==9){
	if (igroup==0){
	  timeLB[igroup][0]=0.31;
	  timeUB[igroup][0]=timeLB[igroup][0]+sampleWindow;
	  timeLB[igroup][1]=2.8;
	  timeUB[igroup][1]=timeLB[igroup][1]+sampleWindow;
	  timeLB[igroup][2]=11;
	  timeUB[igroup][2]=timeLB[igroup][2]+sampleWindow;
	}
	if (igroup==1){
	  timeLB[igroup][0]=0.31;
	  timeUB[igroup][0]=timeLB[igroup][0]+sampleWindow;
	  timeLB[igroup][1]=2.8;
	  timeUB[igroup][1]=timeLB[igroup][1]+sampleWindow;
	  timeLB[igroup][2]=15;
	  timeUB[igroup][2]=timeLB[igroup][2]+sampleWindow;
	}
	if (igroup==2){
	  timeLB[igroup][0]=0.31;
	  timeUB[igroup][0]=timeLB[igroup][0]+sampleWindow;
	  timeLB[igroup][1]=7.7;
	  timeUB[igroup][1]=timeLB[igroup][1]+sampleWindow;
	  timeLB[igroup][2]=20;
	  timeUB[igroup][2]=timeLB[igroup][2]+sampleWindow;
	}
	if (igroup==3){
	  timeLB[igroup][0]=0.31;
	  timeUB[igroup][0]=timeLB[igroup][0]+sampleWindow;
	  timeLB[igroup][1]=2.8;
	  timeUB[igroup][1]=timeLB[igroup][1]+sampleWindow;
	  timeLB[igroup][2]=11;
	  timeUB[igroup][2]=timeLB[igroup][2]+sampleWindow;
	}
	if (igroup==4){
	  timeLB[igroup][0]=0.31;
	  timeUB[igroup][0]=timeLB[igroup][0]+sampleWindow;
	  timeLB[igroup][1]=2.8;
	  timeUB[igroup][1]=timeLB[igroup][1]+sampleWindow;
	  timeLB[igroup][2]=15;
	  timeUB[igroup][2]=timeLB[igroup][2]+sampleWindow;
	}
	if (igroup==5){
	  timeLB[igroup][0]=0.31;
	  timeUB[igroup][0]=timeLB[igroup][0]+sampleWindow;
	  timeLB[igroup][1]=7.7;
	  timeUB[igroup][1]=timeLB[igroup][1]+sampleWindow;
	  timeLB[igroup][2]=25;
	  timeUB[igroup][2]=timeLB[igroup][2]+sampleWindow;
	}
	if (igroup==6){
	  timeLB[igroup][0]=0.31;
	  timeUB[igroup][0]=timeLB[igroup][0]+sampleWindow;
	  timeLB[igroup][1]=2.8;
	  timeUB[igroup][1]=timeLB[igroup][1]+sampleWindow;
	  timeLB[igroup][2]=25;
	  timeUB[igroup][2]=timeLB[igroup][2]+sampleWindow;
	}
	if (igroup==7){
	  timeLB[igroup][0]=0.31;
	  timeUB[igroup][0]=timeLB[igroup][0]+sampleWindow;
	  timeLB[igroup][1]=2.8;
	  timeUB[igroup][1]=timeLB[igroup][1]+sampleWindow;
	  timeLB[igroup][2]=7.7;
	  timeUB[igroup][2]=timeLB[igroup][2]+sampleWindow;
	}
	if (igroup==8){
	  timeLB[igroup][0]=0.31;
	  timeUB[igroup][0]=timeLB[igroup][0]+sampleWindow;
	  timeLB[igroup][1]=2.8;
	  timeUB[igroup][1]=timeLB[igroup][1]+sampleWindow;
	  timeLB[igroup][2]=11;
	  timeUB[igroup][2]=timeLB[igroup][2]+sampleWindow;
	}
	if (igroup==9){
	  timeLB[igroup][0]=0.31;
	  timeUB[igroup][0]=timeLB[igroup][0]+sampleWindow;
	  timeLB[igroup][1]=2.8;
	  timeUB[igroup][1]=timeLB[igroup][1]+sampleWindow;
	  timeLB[igroup][2]=11;
	  timeUB[igroup][2]=timeLB[igroup][2]+sampleWindow;
	}
	if (igroup==10){
	  timeLB[igroup][0]=0.31;
	  timeUB[igroup][0]=timeLB[igroup][0]+sampleWindow;
	  timeLB[igroup][1]=2.8;
	  timeUB[igroup][1]=timeLB[igroup][1]+sampleWindow;
	  timeLB[igroup][2]=15;
	  timeUB[igroup][2]=timeLB[igroup][2]+sampleWindow;
	}
	if (igroup==11){
	  timeLB[igroup][0]=0.31;
	  timeUB[igroup][0]=timeLB[igroup][0]+sampleWindow;
	  timeLB[igroup][1]=1.2;
	  timeUB[igroup][1]=timeLB[igroup][1]+sampleWindow;
	  timeLB[igroup][2]=20;
	  timeUB[igroup][2]=timeLB[igroup][2]+sampleWindow;
	}

      }

    }

    // Subaction 1, plot histogram and time-series
    if (subaction==1) {	


      if (model==1111) { // Do all individuals simultanouosly

	cout<<endl<<"Call paraest";
	int seed=(int)uniformRnd(0,100000);
	double w = paraest(0,false,seed,debug);// igroup=0
	cout<<endl<<"Sampling times =";
	for (int igroup=0;igroup< groups;igroup++){
	  for (int i=0;i<npts[igroup];i++) {
	    cout<<" ["<<timeLB[igroup][i]<<","<<timeUB[igroup][i]<<"]";
	  }
	}
	cout<<" w = "<<w<<endl;
	for (int igroup=0;igroup< groups;igroup++){
	  for (int ind=0;ind<gi[igroup];ind++){
	    for (int ip=0;ip<nParameter;ip++) cout<<pexp[ind][igroup][ip]<<" ";
	    cout<<"\n";
	  }
	}
	cout<<endl;
	if (doplot>0) {
	  plothist(filename);
	  plotModel(0,filename);
	  cout<<endl<<"Figures saved in SavedPlots/ as";
	  cout<<endl<<filename<<"_[i parameter].eps";
	  cout<<endl<<filename<<"_model.eps"<<endl;
	}
      }
      if (model==1 || model==2) { // Do one individual at a time
	int seed=(int)uniformRnd(0,100000);
	double w=0.0;
	int ORIGINAL_NIND=gi[0];
	gi[0]=1;
	for (int i=ORIGINAL_NIND-1;i>=0;i--){
	  if (i%50==0) cout<<"countdown "<<i<<endl;
	  w = paraest(0,false,seed+i,debug-1);
	  int igroup=0;
	  for (int ip=0;ip<nParameter;ip++) {
	    pexp[i][igroup][ip]=pexp[0][igroup][ip];
	  }
	}
	gi[0]=ORIGINAL_NIND;
	for (int igroup=0;igroup< groups;igroup++){
	  for (int ind=0;ind<gi[igroup];ind++){
	    for (int ip=0;ip<nParameter;ip++) cout<<pexp[ind][igroup][ip]<<" ";
	    cout<<"\n";
	  }
	}
	cout<<endl;
      }
    }


    // Subaction 8, test of Poped optimal design for mdoel 2 problem 1
    // Subaction 9, test of my optimal design for model 2 problem 1
    // Subaction 9, test of my optimal design for model 2 problem 1 BIAS
    if (subaction>=8 &&  subaction<=10) {	

      if (model==2) {
	int seed=(int)uniformRnd(0,100000);
	double w=0.0;
	int accum=0;
	for (int igroup=0;igroup< groups;igroup++){
	  int ORIGINAL_NIND=gi[igroup];
	  gi[igroup]=1;
	  for (int i=ORIGINAL_NIND-1;i>=0;i--){
	    if (i%100==0) cout<<"countdown "<<i<<endl;
	    w = paraest(igroup,false,seed+accum,debug-1);
	    accum++;
	    for (int ip=0;ip<nParameter;ip++) {
	      pexp[i][igroup][ip]=pexp[0][igroup][ip];
	    }
	  }
	  gi[igroup]=ORIGINAL_NIND;
	  for (int ind=0;ind<gi[igroup];ind++){
	    for (int ip=0;ip<nParameter;ip++) cout<<pexp[ind][igroup][ip]<<" ";
	    cout<<"\n";
	  }
	}
	cout<<endl;
	 
      }
    }
     
    // Subaction 2, plot 3d
    if (subaction==2 && doplot) {
      if (model==2) {
	for (int i=0;i<pListSize_max;i++) stringcounter[i]=ToString(i);
	int seed=93283;//example of failure
	double w = paraest(0,false,seed,2);//example of success
	string tempfilename1=filename+"_good";
	plot3d(tempfilename1,0,1);
	seed=16918;//example of failure
	w = paraest(0,false,seed,2);
	string tempfilename2=filename+"_bad";
	plot3d(tempfilename2,0,2);
	cout<<endl<<"Figures saved in SavedPlots/ as";
	cout<<endl<<tempfilename1<<".eps";
	cout<<endl<<tempfilename2<<".eps";
      }
    }

    // Subaction 3, plot  -1.91752, 0.0967215,-0.10458 -1.91752, 0.0967215,-0.10458 3d but without singularity
    if (subaction>=3 && doplot) {
      if (model==2) {
	for (int i=0;i<pListSize_max;i++) stringcounter[i]=ToString(i);
	int seed=93283;//example of failure
	double w = paraest(0,false,seed,2);//example of success
	string tempfilename1=filename+"_good";
	plot3d(tempfilename1,0,1);
	seed=16918;//example of failure
	w = paraest(0,false,seed,2);
	string tempfilename2=filename+"_bad";
	plot3d(tempfilename2,0,2);
	cout<<endl<<"Figures saved in SavedPlots/ as";
	cout<<endl<<tempfilename1<<".eps";
	cout<<endl<<tempfilename2<<".eps";
      }
    }

    return 0;
  }




  // ---------------------------------------------------------------------------
  //
  // action == 3
  // Optimal design
  //  
  // ---------------------------------------------------------------------------

  // ---------------------------------------------------------------------------
  // Global search, individually for each group, ITER iterations for each individual.
  // ---------------------------------------------------------------------------
  if (action==3){
    for (int igroup=0;igroup<groups;igroup++){
      double wbest=999999;
      int i1b=0;
      int i2b=0;
      int i3b=0;
      int i4b=0;
      double MYEXP=2.0;
      for (int i1=1;i1<=nTimePointsInSearchDomain;i1++){
	for (int i2=i1;i2<=nTimePointsInSearchDomain;i2++){
	  for (int i3=i2;i3<=nTimePointsInSearchDomain;i3++){
	    for (int i4=i3;i4<=nTimePointsInSearchDomain;i4++){
	      timeLB[igroup][0]=pow(i1-1,MYEXP)/pow(nTimePointsInSearchDomain-1,MYEXP)*T_MAX;
	      timeUB[igroup][0]=timeLB[igroup][0]+sampleWindow;
	      timeLB[igroup][1]=pow(i2-1,MYEXP)/pow(nTimePointsInSearchDomain-1,MYEXP)*T_MAX;
	      timeUB[igroup][1]=timeLB[igroup][1]+sampleWindow;
	      if (npts[igroup]>=3) {
		timeLB[igroup][2]=pow(i3-1,MYEXP)/pow(nTimePointsInSearchDomain-1,MYEXP)*T_MAX;
		timeUB[igroup][2]=timeLB[igroup][2]+sampleWindow;
	      }
	      if (npts[igroup]>=4) {
		timeLB[igroup][3]=pow(i4-1,MYEXP)/pow(nTimePointsInSearchDomain-1,MYEXP)*T_MAX;
		timeUB[igroup][3]=timeLB[igroup][3]+sampleWindow;
	      }
	      bool is2=npts[igroup]==2 && 
		timeUB[igroup][0]<timeLB[igroup][1] && 
		i3==nTimePointsInSearchDomain &&
		i4==nTimePointsInSearchDomain;
	      bool is3=npts[igroup]==3 && 
		timeUB[igroup][0]<timeLB[igroup][1] && 
		timeUB[igroup][1]<timeLB[igroup][2] && 
		i4==nTimePointsInSearchDomain;
	      bool is4=npts[igroup]==4 && 
		timeUB[igroup][0]<timeLB[igroup][1] && 
		timeUB[igroup][1]<timeLB[igroup][2] && 
		timeUB[igroup][2]<timeLB[igroup][3];

	      if (debug>1) cout<<endl<<timeLB[igroup][0]<<" "<<timeUB[igroup][0]<<" "<<timeLB[igroup][1]<<" "<<timeUB[igroup][1];

	      if (is2 || is3 || is4){
		int seed=(int)uniformRnd(0,100000);
		if (debug>1){
		  cout<<endl;
		  for (int i=0;i<npts[igroup];i++) {
		    cout<<"Call with times "<<timeLB[igroup][i]<<" ";
		  }
		}
		double w;
		if (!(model==1 && subaction==1)){
		  int ITERorg=ITER;
		  ITER=ITER/10+1;
		  double wsn = paraest(igroup,false,seed,0);	
		  if (debug>0){
		    cout<<endl;
		    for (int i=0;i<npts[igroup];i++) {
		      cout<<timeLB[igroup][i]<<" ";
		    }
		    cout<<" wsn="<<wsn<< " (wbest*3="<<wbest*3<<")";
		  }
		  ITER=ITERorg;
		  if (wsn<wbest*3) {
		    w = paraest(igroup,false,seed,0);
		    if (debug>0){
		      cout<<" w="<<w;
		    }
		  }
		  else w=wsn;
		}

		if (model==1 && subaction==1){
		  int ORIGINAL_NIND=gi[0];
		  gi[0]=1;
		  if (debug>0){
		    cout<<endl;
		    for (int i=0;i<npts[igroup];i++) {
		      cout<<timeLB[igroup][i]<<" ";
		    }
		  }
		  w=0;
		  int FACTOR=1;
		  double wlocal[FACTOR*ORIGINAL_NIND];
		  for (int i=FACTOR*ORIGINAL_NIND-1;i>=0;i--){
		    seed=(int)uniformRnd(0,100000);
		    wlocal[i]=paraest(0,true,seed,debug-1); 		     
		    w+=wlocal[i];
		    int igroup=0;
		    for (int ip=0;ip<nParameter;ip++) {
		      pexp[i][igroup][ip]=pexp[0][igroup][ip];
		    }
		  }
		  int i;
		  double key;
		  for(int j=0;j<FACTOR*ORIGINAL_NIND;j++) {
		    key=wlocal[j];
		    i=j-1;
		    while(wlocal[i]>key && i>=0) {
		      wlocal[i+1]=wlocal[i];
		      i--;
		    }
		    wlocal[i+1]=key;
		  }
		  cout<<wlocal[FACTOR*ORIGINAL_NIND/2]; 
		  gi[0]=ORIGINAL_NIND;
		}
		if (w<wbest) {
		  wbest=w;
		  if (subaction>1) cout<<endl<<" new wbest="<<wbest;
		  i1b=i1;
		  i2b=i2;
		  i3b=i3;
		  i4b=i4;
		}
	      } 
	      else {
		// dummy
	      }
	    }
	  }
	} 
      }
      int i1=i1b;
      int i2=i2b;
      int i3=i3b;
      int i4=i4b;
      timeLB[igroup][0]=pow(i1-1,MYEXP)/pow(nTimePointsInSearchDomain-1,MYEXP)*T_MAX;
      timeUB[igroup][0]=timeLB[igroup][0]+sampleWindow;
      timeLB[igroup][1]=pow(i2-1,MYEXP)/pow(nTimePointsInSearchDomain-1,MYEXP)*T_MAX;
      timeUB[igroup][1]=timeLB[igroup][1]+sampleWindow;
      if (npts[igroup]>=3) {
	timeLB[igroup][2]=pow(i3-1,MYEXP)/pow(nTimePointsInSearchDomain-1,MYEXP)*T_MAX;
	timeUB[igroup][2]=timeLB[igroup][2]+sampleWindow;
      }

      if (npts[igroup]>=4) {
	timeLB[igroup][3]=pow(i4-1,MYEXP)/pow(nTimePointsInSearchDomain-1,MYEXP)*T_MAX;
	timeUB[igroup][3]=timeLB[igroup][3]+sampleWindow;
      }
      if (debug>0){
	cout<<endl;
	for (int i=0;i<npts[igroup];i++) {
	  cout<<"["<<timeLB[igroup][i]<<" "<<timeUB[igroup][i]<<"] ";
	}
	cout<<" "<<wbest<<" (igroup="<<igroup<<")";
	cout<<endl<<"{";
	for (int i=0;i<npts[igroup];i++) {
	  if (i==npts[igroup]-1) cout<<timeLB[igroup][i];
	  else cout<<timeLB[igroup][i]<<",";
	}
	cout<<"}";
	cout<<endl;
      }
    }
  }
 

  // ---------------------------------------------------------------------------
  // Local search 
  // ---------------------------------------------------------------------------
  if (!(model==1 && subaction==1)){
    int seed=(int)uniformRnd(0,100000);
    double wbest = paraest(-1,false,seed,0);
    if (debug>0){
      cout<<endl;
      for (int igroup=0;igroup< groups;igroup++){
	cout<<"group="<<igroup<<" sampling times =";
	for (int i=0;i<npts[igroup];i++) {
	  cout<<" ["<<timeLB[igroup][i]<<","<<timeUB[igroup][i]<<"]";
	}
	cout<<endl;
      }
      cout<<"wbest "<<wbest<<endl;
    }
    
    int ntp=0;
    int getGroup[g_max];
    int getTime[npts_max];
    for (int igroup=0;igroup<groups;igroup++){
      for (int i=0;i<npts[igroup];i++){
	getGroup[ntp]=igroup;
	getTime[ntp]=i;
	cout<<" getGroup["<<ntp<<"]="<<getGroup[ntp]<<endl;
	cout<<" getTime["<<ntp<<"]="<<getTime[ntp]<<endl;
	ntp++;
      }
    }								     
    cout<<" Fine-tune, no of time-points to vary = ntp = "<<ntp<<endl;
    bool success;
    do { 
      success=false;
      for (int i=0;i<ntp;i++){
	cout<<"test time-point  "<<i<<endl;
	for (int up=0;up<2;up++){ 
	  double MYEXP=2.0;
	  int timeindex=(int)round(sqrt(timeLB[getGroup[i]][getTime[i]]*pow(nTimePointsInSearchDomain-1,MYEXP)/T_MAX)+1);
	  if (timeindex<nTimePointsInSearchDomain && up==1){
	    timeindex++;
	  }
	  else if (timeindex>1 && up==0){
	    timeindex--;
	  }
	  else {
	    continue;
	  }
	  timeLB[getGroup[i]][getTime[i]]=pow(timeindex-1,MYEXP)/pow(nTimePointsInSearchDomain-1,MYEXP)*T_MAX;
	  timeUB[getGroup[i]][getTime[i]]=timeLB[getGroup[i]][getTime[i]]+sampleWindow;
	  bool ok=true;
	  for (int ip=0;ip<npts[getGroup[i]]-1;ip++){ 
	    if (timeLB[getGroup[i]][ip]>timeLB[getGroup[i]][ip+1]) ok=false;
	  }
	  if (!ok){ 
	    if (up==1) timeindex--;
	    else if (up==0) timeindex++;
	    timeLB[getGroup[i]][getTime[i]]=pow(timeindex-1,MYEXP)/pow(nTimePointsInSearchDomain-1,MYEXP)*T_MAX;
	    timeUB[getGroup[i]][getTime[i]]=timeLB[getGroup[i]][getTime[i]]+sampleWindow;
	    continue;
	  }	  
	  int seed=(int)uniformRnd(0,100000);
	  double w = paraest(-1,false,seed,0);
	  if (debug>0){
	    for (int igroup=0;igroup< groups;igroup++){
	      cout<<"group="<<igroup<<" sampling times =";
	      for (int i=0;i<npts[igroup];i++) {
		cout<<" ["<<timeLB[igroup][i]<<","<<timeUB[igroup][i]<<"]";
	      }	
	      cout<<endl;
	    }
	    cout<<"w "<<w<<endl;
	  }
	  if (w<wbest){
	    wbest=w;
	    cout<<" new wbest "<<wbest<<endl;
	    success=true;
	    if (up==0) break; 
	  }
	  else{ 
	    if (up==1) timeindex--;
	    else if (up==0) timeindex++;
	    timeLB[getGroup[i]][getTime[i]]=pow(timeindex-1,MYEXP)/pow(nTimePointsInSearchDomain-1,MYEXP)*T_MAX;
	    timeUB[getGroup[i]][getTime[i]]=timeLB[getGroup[i]][getTime[i]]+sampleWindow;
	  }
	  
	}
      } 
    } while(success); 

  }
  cout<<"Finally, do the best one"<<endl;
  int seed=(int)uniformRnd(0,100000);
  double w = paraest(-1,true,seed,1);
  for (int igroup=0;igroup< groups;igroup++){
    cout<<endl<<"group="<<igroup<<" sampling times =";
    for (int i=0;i<npts[igroup];i++) {
      cout<<" ["<<timeLB[igroup][i]<<","<<timeUB[igroup][i]<<"]";
    }	
    }											     
  cout<<" w = "<<w<<endl;
  cout<<"Finished."<<endl;  
  for (int igroup=0;igroup< groups;igroup++){
    cout<<endl<<"group="<<igroup<<"\n";
      for (int i=0;i<npts[igroup];i++) {
	cout<<timeLB[igroup][i]<<" ";
      }	
  }
  cout<<endl;
  return 0;
}










// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
//  Parameter estimation 
//
//  Estimate all parameters in all groups if igroup<0, 
//  otherwise estimate in igroup only
//  onlyonce true if only one iteration
//           false if ITER iterations
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
double paraest(int igroup,
	       bool onlyonce,
	       int seed,
	       int debug){
  double errorfavg=0.0;
  int count=0;
     
  for (int jj=0;jj<ITER;jj++) {
    if (debug>1 && jj%50==0) cout << endl<<"paraestAll iteration = "<<jj;
    writeProblem(igroup,seed+jj); 
    int out=system("./Project/OptExpDes/OptExpDes");
    readOutput(igroup,debug);
       
    if (igroup<0){
      for (int ig=0;ig<groups;ig++) {
	for (int ind=0;ind<gi[ig];ind++) {
	  errorfavg+=errorf[ig][ind];
	}
      }
    }
    else {
      for (int ind=0;ind<gi[igroup];ind++) {
	errorfavg+=errorf[igroup][ind];
      }
    }
    count++;
    if (onlyonce) break;
  }
  if (debug>0) cout<<endl<<"Final errorfavg = "<<errorfavg;
  return errorfavg/count;
}


















// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// 
// Write paraest problem for all groups to file if igroup<0, 
// otherwise igroup only 
//
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
void writeProblem(int igroup, int seed){

  ofstream myfile;
  myfile.open ("Input/Input.txt");
  if (action==2 && subaction==2 && model==2) myfile << "nModel 22\n";
  else myfile << "nModel "<<model<<"\n";
  myfile << "\n";
  myfile << "#----------------------------------------------------------------------\n";
  myfile << "#OED parameters\n";
  myfile << "#----------------------------------------------------------------------\n";
  myfile << "\n";
  myfile << "OED_nParameter "<<nParameter<<"\n";
  if (igroup<0) myfile << "OED_nGroups "<<groups<<"\n";
  else  myfile << "OED_nGroups 1\n";
  myfile << "\n";
  myfile << "# Number of time-points in each group\n";
    
  if (igroup<0) {
    for (int i=0;i<groups;i++) {
      myfile << "OED_group"<<(i+1)<<"_nTime "<<npts[i]<<"\n";
    }
  }
  else {
    myfile << "OED_group1_nTime "<<npts[igroup]<<"\n";
  }
  myfile << "\n";
  myfile << "# Number of individuals in the groups\n";
    
  if (igroup<0) {
    for (int i=0;i<groups;i++) {
      myfile << "OED_group"<<(i+1)<<"_nInd "<<gi[i]<< endl;
    }
  }
  else {
    myfile << "OED_group1_nInd "<<gi[igroup]<< endl;
  }
  myfile << "\n";
  myfile << "# Dose for each group\n";
    
  if (igroup<0) {
    for (int i=0;i<groups;i++) {
      myfile << "OED_group"<<(i+1)<<"_A ["<<dose[i]-doseWindow*dose[i]/2<<","<<dose[i]+doseWindow*dose[i]/2<<"]\n";
    }
  }
  else {
    myfile << "OED_group1_A ["<<dose[igroup]-doseWindow*dose[igroup]/2<<","<<dose[igroup]+doseWindow*dose[igroup]/2<<"]\n";
  }

    
  if (igroup<0) {
    for (int igr=0;igr<groups;igr++) {
      myfile <<"\n";
      myfile << "# Time-points for group "<<(igr+1)<<"\n";
      for (int i=0;i<npts[igr];i++) {
	myfile <<std::fixed <<std::setprecision(15)<< "OED_group"<<(igr+1)<<"_time"<<(i+1)<<" ["<<timeLB[igr][i]<<","<<timeUB[igr][i]<<"]\n";
      }
    }
  }    
  else {
    myfile <<"\n";
    myfile << "# Time-points for group 1\n";
    for (int i=0;i<npts[igroup];i++) {
      myfile <<std::fixed <<std::setprecision(15)<< "OED_group1_time"<<(i+1)<<" ["<<timeLB[igroup][i]<<","<<timeUB[igroup][i]<<"]\n";
    }
  }

  int OED_numberOfCommonTimes=0;
  if (igroup<0) {
    for (int igr=0;igr<groups;igr++) {
      for (int i=0;i<npts[igr];i++) {
	for (int igr2=igr+1;igr2<groups;igr2++) {
	  for (int i2=0;i2<npts[igr2];i2++) {
	    if (timeLB[igr][i]==timeLB[igr2][i2] && 
		timeUB[igr][i]==timeUB[igr2][i2]){
	      OED_numberOfCommonTimes++;
	    }
	  }
	}
      }
    }
  }
  myfile <<"\n";
  myfile << "OED_numberOfCommonTimes "<<OED_numberOfCommonTimes<<"\n";
    
    
  if (igroup<0) {
    int count=0;
    for (int igr=0;igr<groups;igr++) {
      for (int i=0;i<npts[igr];i++) {
	for (int igr2=igr+1;igr2<groups;igr2++) {
	  for (int i2=0;i2<npts[igr2];i2++) {
	    if (timeLB[igr][i]==timeLB[igr2][i2] && 
		timeUB[igr][i]==timeUB[igr2][i2]){
	      myfile <<"\n";
	      myfile <<"# Group "<<(igr+1)<<" time-point "<<(i+1)<<" in common with group "<<(igr2+1)<<" time-point "<<(i2+1)<<"\n"; 
	      myfile <<"OED_common_time"<<++count<<" "<<(igr+1)<<" "<<(i+1)<<" "<<(igr2+1)<<" "<<(i2+1)<<"\n"; 
	    }
	  }
	}
      }
    }
  }

  myfile <<"\n";
  for (int i=0;i<nParameter/2;i++) {
    myfile << "OED_p"<<(i+1)<<"_isCommon 1\n";
  }
  for (int i=nParameter/2;i<nParameter;i++) {
    myfile << "OED_p"<<(i+1)<<"_isCommon 0\n";
  }


  myfile <<"\n";
  for (int i=0;i<nParameter;i++) {
    myfile << "OED_p"<<(i+1)<<"sharp "<<psharp_iv[i]<<"\n";
  }
  
  myfile <<"\n";
  for (int i=0;i<nParameter;i++) {
    myfile << "OED_p"<<(i+1)<<"search ["<<psearch[i][0]<<","<<psearch[i][1]<<"]\n";
  }
  myfile <<"\n";
  myfile <<"\n";
  myfile << "#----------------------------------------------------------------------\n";
  myfile << "#	RNGmode: fixed, variable\n";
  myfile << "#----------------------------------------------------------------------\n";
  myfile << "RNGmode fixed\n";
  myfile << "RNGseed "<<seed<<"\n";
  myfile <<"\n";
  myfile <<"#----------------------------------------------------------------------\n";
  myfile <<"#	ParameterPerturbation:  \n";  
  myfile <<"#	ParameterPerturbation_Distribution: \n";
  myfile <<"						uniform, normal, lognormal\n";
  myfile <<"#	ParameterPerturbation_Mode: absolute,relative\n";
  myfile <<"#----------------------------------------------------------------------\n";
  myfile <<"\n";
  myfile <<"ParameterPerturbation_Enable 1\n";
  myfile <<"\n";
  for (int i=nParameter/2;i<nParameter;i++) {
    myfile <<"ParameterPerturbation_Distribution"<<(i+1)<<" normal\n";
  }
  myfile <<"\n";
  for (int i=nParameter/2;i<nParameter;i++) {
    if (model==1) myfile <<"ParameterPerturbation_Mode"<<(i+1)<<" absolute\n";
    if (model==2) myfile <<"ParameterPerturbation_Mode"<<(i+1)<<" absolute\n";
  }
  myfile <<"\n";
  for (int i=nParameter/2;i<nParameter;i++) {
    myfile <<"ParameterPerturbation_epsilon"<<(i+1)<<" 0.0\n";
  }
  myfile <<"\n";
  for (int i=nParameter/2;i<nParameter;i++) {
    myfile <<"ParameterPerturbation_mean"<<(i+1)<<" 0.0\n";
  }
  myfile <<"\n";
  for (int i=nParameter/2;i<nParameter;i++) {
    myfile <<"ParameterPerturbation_std"<<(i+1)<<" "<<sqrt(d[i])<<"\n";
  }
  myfile <<"\n";
  myfile <<"#----------------------------------------------------------------------\n";
  myfile <<"#	DataPerturbation:\n";
  myfile <<"#	DataPerturbation_Distribution: uniform, normal\n";
  myfile <<"#	DataPerturbation_Mode: absolute,relative\n";
  myfile <<"#----------------------------------------------------------------------\n";
  myfile <<"\n";
  myfile <<"DataPerturbation_Enable 1\n";
  myfile <<"\n";
  myfile <<"DataPerturbation_Distribution normal\n";
  if (model==1) myfile <<"DataPerturbation_Mode relative\n";
  if (model==2) myfile <<"DataPerturbation_Mode absolute\n";
  myfile <<"DataPerturbation_epsilon 0.0\n";
  myfile <<"DataPerturbation_mean 0.0\n";
  myfile <<"DataPerturbation_std "<<epsilon<<"\n";
  myfile <<"\n";
  myfile <<"#----------------------------------------------------------------------\n";
  myfile <<"#	Filter: \n";
  myfile <<"#----------------------------------------------------------------------\n";
  myfile <<"\n";

  myfile <<"Filter_ContractionRatio 0.9999\n";
  myfile <<"Filter_MinWidth 0.0001\n";
  myfile <<"Filter_MaxIter 999\n";
  myfile <<"Filter_MinIter 1\n";
  myfile <<"\n";
  myfile <<"#----------------------------------------------------------------------\n";
  myfile <<"#	Splitter: \n";
  myfile <<"#\n";
  myfile <<"#	Split_CUT_cutFunction: safeGM, safeAM\n";
  myfile <<"#	Split_COMP_componentSelector:\n";
  myfile <<"#	max_width, max_rel_width, max_scaled_width\n";
  myfile <<"#	bestBranch\n";
  myfile <<"#----------------------------------------------------------------------\n";
  myfile <<"\n";
  myfile <<"Split_CUT_cutFunction safeAM\n";	
  myfile <<"Split_CUT_safeGM_r 0.50\n";
  myfile <<"Split_COMP_componentSelector max_scaled_width\n";
  myfile <<"\n";
  myfile <<"#----------------------------------------------------------------------\n";
  myfile <<"#	BAP:\n";
  myfile <<"#----------------------------------------------------------------------\n";
  myfile <<"\n";
  if (action==2) {
    myfile <<"BAP_RangeListMaxSize 1000\n";
    myfile <<"BAP_MaxIter 1000\n";
  }
  else {
    myfile <<"BAP_RangeListMaxSize 100\n";
    myfile <<"BAP_MaxIter 100\n";
  }
  myfile <<"BAP_SMALL_width 0.1\n";
  myfile <<"BAP_SMALL_width_measure absolute\n";
  myfile <<"BAP_SMALL_abswidth 0.01\n";
  myfile <<"BAP_scale_update_period 12\n";
  myfile <<"\n";
  myfile <<"BAP_CC_enable 0\n";
  myfile <<"BAP_IOVR_enable 0\n";
  myfile <<"BAP_CON_enable 0\n";
  myfile <<"BAP_SMALL_enable 1\n";
  myfile <<"BAP_PRUNE_enable 1\n";
  myfile <<"\n";
  myfile <<"BAP_CC_minCoverSize 100\n";
  myfile <<"BAP_CC_contractionRatio 0.7\n";
  myfile <<"BAP_CC_sizeRatio 8\n";
  myfile <<"BAP_IOVR 0.1\n";
  myfile <<"BAP_IOVR_iter_check_period 10000\n";
  myfile <<"BAP_CON_softDegree 0.0\n";
  myfile <<"BAP_CON_inconsistencyTest 0\n";
  myfile <<"BAP_reExpansionRate 0.30\n";
  myfile <<"BAP_pruner LoopSweepForwardBackward\n";
  myfile <<"\n";
  myfile <<"#----------------------------------------------------------------------\n";
  myfile <<"#	Expander:\n";
  myfile <<"#	Expander_Mode: absolute, relative\n";
  myfile <<"#----------------------------------------------------------------------\n";
  myfile <<"\n";
  if (model==1) {
    myfile <<"Expander_Mode relative\n";
  }
  else if (model==2) {
    myfile <<"Expander_Mode absolute\n";
  }
  else{
    cout<<"unknown model when setting expander mode\n";
    exit(1);
  }
  myfile <<"Expander_Rate 0.1\n";
  myfile <<"Expander_absTerm 0.001\n";
  myfile <<"Expander_RateAdjuster 0.1\n";
  myfile <<"Expander_RateMin 0.00001\n";
  myfile <<"Expander_MinCount 10 \n";
  myfile <<"Expander_MaxWidth 99999999999.0\n";
  myfile <<"\n";
  myfile <<"#----------------------------------------------------------------------\n";
  myfile.close();  
}












// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
//
// Read output for all groups if igroup<0, ow igroup only 
//
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
void readOutput(int igroup,int debug){
  for (int igr=0;igr<groups;igr++) {
    if (igroup>=0 && igr!=igroup) continue;
    for (int ind=0;ind<gi[igr];ind++) {
      string ssfile="Output/Output.txt";
      string ssn="pstar_g"+stringcounter[igr+1]+"i"+stringcounter[ind+1];
      if (igroup>=0){
	ssn="pstar_g"+stringcounter[0+1]+"i"+stringcounter[ind+1];
      }
      ifstream inFile;
      inFile.open(ssfile.c_str());
      if (!inFile) {
	cout << "Unable to open "<<ssfile;
	exit(1);
      }
      string temp;
      while (inFile >> temp){
	if (temp==ssn){
	  for (int k=0;k<nParameter;k++) {
	    if (inFile >> temp){
	      pstar[k] = parseString<interval>(temp);
	    }
	  }
	  break;
	}
      }
      for (int kk=0;kk<nParameter;kk++) pexp[ind][igr][kk]=interval(1.0,0.0);
      if (debug>2) cout << endl<<"group  " << igr+1<< "ind "<<ind+1<<"\n";
      string ss="Output/p_boxes_g"+stringcounter[igr+1]+"i"+stringcounter[ind+1]+".txt";
      if (igroup>=0){
	ss="Output/p_boxes_g"+stringcounter[0+1]+"i"+stringcounter[ind+1]+".txt";
      }
      ifstream inFile2;
      inFile2.open(ss.c_str());
      if (!inFile2) {
	cout << "Unable to open "<<ss;
	exit(1); 
      }
      pListSize=0;
      double temp1,temp2;
      bool cont = true;
      while ( cont != false ){
	if (debug>2) cout <<endl<<"pListSize " << pListSize;
	if (pListSize>=pListSize_max){
	  cout << endl<<"pListSize " << pListSize;
	  cout << endl<<"pListSize_max " << pListSize_max<<endl;
	  exit (1);
	}
	for (int kk=0;kk<nParameter;kk++) {
	  if (inFile2 >> temp1 && inFile2 >> temp2){
	    pList[kk][pListSize]=interval(temp1,temp2);
	    if (debug>2) cout << endl<<"p"<<kk+1<<"_"<<pListSize<<"="<<pList[kk][pListSize];
	    pexp[ind][igr][kk]=pexp[ind][igr][kk] | pList[kk][pListSize];
	    if (debug>2) cout << endl<<"pexp igr=" << igr+1<< " ind="<<ind+1<<" ipara="<<kk+1<<"="<<pexp[ind][igr][kk];
	  }
	  else {
	    cont=false;
	    break;
	  }
	}
	if (cont) pListSize++;
      }
      inFile2.close();
      if (action==2 && subaction==2 && model==2) {
	labelPlist(); 
	for (int iLabel=1;iLabel<=numberOfLabels;iLabel++){
	  if (debug>0) cout<<endl<<"Hull "<<pListHull[0][iLabel]<<" "<<pListHull[1][iLabel]<<" "<<pListHull[2][iLabel]<<pListHull[3][iLabel]<<" "<<pListHull[4][iLabel]<<" "<<pListHull[5][iLabel]<<" label="<<iLabel;
	}
	if (debug>0) cout<<endl;
      }
      else if (model==2) {
	if (igroup<0 && debug>2) cout << endl<<"pexp igr=" << igr+1<< " ind="<<ind+1;
	labelPlist(); 
	if (model==2){ 	  
	  int bestLabel=1; 
	  for (int ipara=0;ipara<nParameter;ipara++) {
	    p[ind][ipara]=pListHull[ipara][bestLabel];
	  }
	  if (debug>0) cout<<endl;
	  for (int iLabel=1;iLabel<=numberOfLabels;iLabel++){
	    if (debug>0) cout<<endl<<"Hull "<<pListHull[0][iLabel]<<" "<<pListHull[1][iLabel]<<" "<<pListHull[2][iLabel]<<pListHull[3][iLabel]<<" "<<pListHull[4][iLabel]<<" "<<pListHull[5][iLabel]<<" label="<<iLabel;
	    if (pListHull[0][iLabel]*exp(pListHull[3][iLabel])>pListHull[1][iLabel]*exp(pListHull[4][iLabel])){
	      for (int ipara=0;ipara<nParameter;ipara++) {
		p[ind][ipara]=pListHull[ipara][iLabel];
	      }
	      bestLabel=iLabel;      
	    }
	  }
	  for (int ipara=0;ipara<nParameter;ipara++)
	    pexp[ind][igr][ipara]=p[ind][ipara];
	}
      }
      int ERRORMEASURE=1;//   0 = VOLUME,  1 = WIDTH
      errorf[igr][ind]=0.0;
      double bias[nParameter_max];
      for (int k=0;k<nParameter;k++) bias[k]=0.0;
      for (int i=0;i<pListSize;i++){
	double vol=1.0; 
	for (int k=0;k<nParameter;k++) {
	  if (ERRORMEASURE==1){ 
	    if (psearch[k][0]==psearch[k][1]){
	      // nothing
	    }
	    else if (abs(psearch[k][0])!=abs(psearch[k][1])) {
	      errorf[igr][ind]+=abs(width(pList[k][i])/mid(pList[k][i]));
	    }
	    else{
	      errorf[igr][ind]+=abs(width(pList[k][i]))/abs(psearch[k][0]);
	    }
	  }
	  if (ERRORMEASURE==0){
	    double len=0.0;
	    if (psearch[k][0]==psearch[k][1]){
	      len=1.0;
	    }
	    else if (abs(psearch[k][0])!=abs(psearch[k][1])) {
	      len=abs(width(pList[k][i])/mid(pList[k][i]));
	    }
	    else{
	      len=abs(width(pList[k][i]))/abs(psearch[k][0]);
	    }
	    if (len<1E-14) len=1E-14;
	    vol*=len;
	  }
	  if (psearch[k][0]==psearch[k][1]){
	    // nothing
	  }
	  else if (psearch[k][0]<0.0 && psearch[k][1]>0.0) {
	    bias[k]+=abs(mid(pstar[k])-mid(pList[k][i]))/abs(psearch[k][0]);
	  }
	  else{ 
	    bias[k]+=abs(mid(pstar[k])-mid(pList[k][i])/mid(pstar[k]));
	  }
	}
	if (ERRORMEASURE==0){
	  errorf[igr][ind]+=vol;
	}
      }
      errorf[igr][ind]=errorf[igr][ind]/nParameter;
      double biastot=0.0;
      for (int k=0;k<nParameter;k++) {
	bias[k]=bias[k]/pListSize; 
	biastot+=bias[k];
      }
      biastot=biastot/nParameter;
      errorf[igr][ind]+=lambda*biastot;
      if (debug>4) cout<<endl<<"errorf["<<igr<<","<<ind<<"] = "<<errorf[igr][ind];
    } 
  }
  
}














// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
//  Gnuplot function for various doses
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
void plot(string filename){
  ofstream mydatfile;
  int N=5;
  double doseloc[N]; 
  if (model==1){
    doseloc[0]= 250;
    doseloc[1]= 350;
    doseloc[2]= 450;
    doseloc[3]= 550;
    doseloc[4]= 650;
  }
  else if (model==2){
    doseloc[0]= 3.0;
    doseloc[1]= 4.0;
    doseloc[2]= 5.0;
    doseloc[3]= 6.0;
    doseloc[4]= 7.0;
  }

  mydatfile.open ("myData.data");
  for (int i=0;i<101;i++){
    double t=(i*T_MAX)/100;
    mydatfile<<t<<" ";
    for (int jj=0;jj<N;jj++){
      double f=0.0;
      if (model==1){
	f=doseloc[jj]/psharp[1]*exp(-psharp[0]/psharp[1]*t); 
      }
      else if (model==2){
	f=doseloc[jj]*psharp[0]*psharp[1]/ ( psharp[2]*(psharp[0]-psharp[1]) ) * 
	  (exp(-psharp[1]*t) -  exp(-psharp[0]*t) );
      }
      mydatfile<<f<<" ";
    }
    mydatfile<<endl;
  }
  mydatfile.close();
  
  ofstream myplotfile;
  myplotfile.open ("myPlot.gp");
  myplotfile<<endl<<"set xlabel \"Time (h)\"";
  myplotfile<<endl<<"set ylabel \"Concentration\"";
  myplotfile<<endl<<"set grid";
  myplotfile<<endl<<"set style line 1 lt 1 lw 5";
  myplotfile<<endl<<"set style line 2 lt 2 lw 5";
  myplotfile<<endl<<"set style line 3 lt 3 lw 5";
  myplotfile<<endl<<"set style line 4 lt 4 lw 5";
  myplotfile<<endl<<"set style line 5 lt 5 lw 5";
  if (doplot){
    myplotfile<<endl<<"plot ";
    for (int jj=0;jj<N;jj++){
      myplotfile<<"\"myData.data\" using 1:"<<jj+2<<" title 'a="<<doseloc[jj]<<"' with lines ls "<<jj+1;
      if (jj<N-1) myplotfile<<", \\\n";
    }
    myplotfile<<endl;
  }
  // FOR SAVING TO DISK AS .EPS-FILE
  myplotfile<<endl<<"set term postscript eps color blacktext \"Helvetica\" 24";
  if (model==1) myplotfile<<endl<<"set output 'SavedPlots/"<<filename<<".eps'";
  if (model==2) myplotfile<<endl<<"set output 'SavedPlots/"<<filename<<".eps'";
  myplotfile<<endl<<"plot ";
  for (int jj=0;jj<N;jj++){
    myplotfile<<"\"myData.data\" using 1:"<<jj+2<<" title 'a="<<doseloc[jj]<<"' with lines ls "<<jj+1;
    if (jj<N-1) myplotfile<<", \\\n";
  }
  myplotfile<<endl<<"set output";
  myplotfile.close(); 
  int out=system("gnuplot -persist < myPlot.gp");
  cout<<endl<<"saved as "<<filename<<".eps"<<endl;
}












// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
//  Gnuplot of time-series for certain parameters
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
void plotModel(int igroup,string figname){

  ofstream mydatfile;
  mydatfile.open ("myModel.data");
  for (int i=0;i<101;i++){
    double t=(i*T_MAX)/100;
    mydatfile<<t<<" ";
    double f=0.0;
    if (model==1) {
      double p0=psharp[0]+psharp[2];
      double p1=psharp[1]+psharp[3];
      f=dose[igroup]/p1*exp(-p0/p1*t); 
    }
    if (model==2){
      double p0=psharp[0]*exp(psharp[3]);
      double p1=psharp[1]*exp(psharp[4]);
      double p2=psharp[2]*exp(psharp[5]);
      f=dose[igroup]*p0*p1/ ( p2*(p0-p1) )*(exp(-p1*t) -  exp(-p0*t) );
    }
    mydatfile<<f<<" ";
    if (model==1) {
      double p0=pest[0]+pest[2];
      double p1=pest[1]+pest[3];
      f=dose[igroup]/p1*exp(-p0/p1*t); 
    }
    if (model==2){
      double p0=pest[0]*exp(pest[3]);
      double p1=pest[1]*exp(pest[4]);
      double p2=pest[2]*exp(pest[5]);
      f=dose[igroup]*p0*p1/ ( p2*(p0-p1) )*(exp(-p1*t) -  exp(-p0*t) );
    }
    mydatfile<<f<<" "; 
    mydatfile<<endl;
  }
  mydatfile.close();

  ofstream myplotfile;
  myplotfile.open ("myModel.gp");
  myplotfile<<endl<<"set xlabel \"Time (h)\"";
  myplotfile<<endl<<"set ylabel \"Concentration\"";
  myplotfile<<endl<<"set grid";
  myplotfile<<endl<<"set style line 1 lt 1 lw 5";
  myplotfile<<endl<<"set style line 2 lt 3 lw 5";
  if (doplot){
    myplotfile<<endl<<"plot ";
    myplotfile<<"\"myModel.data\" using 1:2 title 'True parameters' w l ls 2";
    myplotfile<<", \\\n";
    myplotfile<<"\"myModel.data\" using 1:3 title 'Estimated parameters' w l ls 1";
    myplotfile<<endl;
  }
  myplotfile<<endl<<"set term postscript eps color blacktext \"Helvetica\" 24";
  myplotfile<<endl<<"set output 'SavedPlots/"<<figname<<"_model.eps'";
  myplotfile<<endl<<"plot ";
  myplotfile<<"\"myModel.data\" using 1:2 title 'True parameters' w l ls 2";
  myplotfile<<", \\\n";
  myplotfile<<"\"myModel.data\" using 1:3 title 'Estimated parameters' w l ls 1";
  myplotfile<<endl<<"set output";
  myplotfile.close(); 
  int out=system("gnuplot -persist < myModel.gp");
}

























// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
//   Gnuplot histogram of several repetitiosn given one design.
//   Note that empty intervals are already considered in the called
//   functions inf and sup in the class interval 
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
void plothist(string figname){
  int debug=0;
  ofstream mydatfile;
  const int BINS=500;
  double cmax[nParameter];  
  int N=0; // number of intervals to consider
  for (int igroup=0;igroup< groups;igroup++){
    N+=gi[igroup];
  }
  int occur[N];
  for (int ipara=nParameter/2;ipara<nParameter;ipara++){
    cmax[ipara]=0.0;
    if (ipara==0) mydatfile.open ("myDataHist1.data");
    if (ipara==1) mydatfile.open ("myDataHist2.data");
    if (ipara==2) mydatfile.open ("myDataHist3.data");
    if (ipara==3) mydatfile.open ("myDataHist4.data");
    if (ipara==4) mydatfile.open ("myDataHist5.data");
    if (ipara==5) mydatfile.open ("myDataHist6.data");
    double binlength=(psearch[ipara][1]-psearch[ipara][0])/BINS;
    double ctot=0;
    double lnsum=0; 
    double lnsumMID=0;
    for (int jj=0;jj<N;jj++) occur[jj]=0;
    for (int i=0;i<BINS;i++){
      double temval_lo=psearch[ipara][0]+i*binlength;
      double temval_mi=psearch[ipara][0]+(i+0.5)*binlength;
      double temval_hi=psearch[ipara][0]+(i+1)*binlength;
      int count=0;
      for (int igroup=0;igroup<groups;igroup++){
	for (int jj=0;jj<gi[igroup];jj++){
	  if (inf(pexp[jj][igroup][ipara])<temval_hi && sup(pexp[jj][igroup][ipara])>=temval_lo) {
	    occur[count]++;
	  }
	  count++;
	}
      }
    }
    for (int i=0;i<BINS;i++){
      double temval_lo=psearch[ipara][0]+i*binlength;
      double temval_mi=psearch[ipara][0]+(i+0.5)*binlength;
      double temval_hi=psearch[ipara][0]+(i+1)*binlength;
      double c=0.0;
      int count=0;
      for (int igroup=0;igroup< groups;igroup++){
	for (int jj=0;jj<gi[igroup];jj++){
	  if (i==0) lnsumMID+=mid(pexp[jj][igroup][ipara]);
	 	  
	  if (inf(pexp[jj][igroup][ipara])<temval_hi && sup(pexp[jj][igroup][ipara])>=temval_lo) {
	    double impact=1.0/occur[count]; // Each interval has equal weight.
	    c+=impact;
	    ctot+=impact;
	    lnsum+=impact*temval_mi;
	    
	  }
	  count++;
	}
      }
    }
    if (debug>0) cout<<endl<<"ctot is "<<ctot;
    double plottedvol=0.0;
    double volume=ctot*binlength;
    if (debug>0) cout<<endl<<"binlength is  "<<binlength;
    if (debug>0) cout<<endl<<"volume is  "<<volume;
    for (int i=0;i<BINS;i++){
      double temval_lo=psearch[ipara][0]+i*binlength;
      double temval_mi=psearch[ipara][0]+(i+0.5)*binlength;
      double temval_hi=psearch[ipara][0]+(i+1)*binlength;
      double c=0.0;
      int count=0;
      for (int igroup=0;igroup< groups;igroup++){
	for (int jj=0;jj<gi[igroup];jj++){
	  if (inf(pexp[jj][igroup][ipara])<temval_hi && sup(pexp[jj][igroup][ipara])>=temval_lo) {
	    double impact=1.0/occur[count]; // Each interval has equal weight.
	    c+=impact;
	  }
	  count++;
	}
      }
      mydatfile<<temval_mi<<" "<<c/volume<<endl;
      plottedvol+=c/volume*binlength;
      if (c/volume>cmax[ipara]) cmax[ipara]=c/volume;
    }
    if (debug>0) cout<<endl<<"plotted volume "<<plottedvol;
    double mu=lnsum/ctot;
    double muMID=lnsumMID/N;
    mydatfile.close();
    double sum=0;
    for (int i=0;i<BINS;i++){
      double temval_lo=psearch[ipara][0]+i*binlength;
      double temval_mi=psearch[ipara][0]+(i+0.5)*binlength;
      double temval_hi=psearch[ipara][0]+(i+1)*binlength;
      int count=0;
      for (int igroup=0;igroup< groups;igroup++){
	for (int jj=0;jj<gi[igroup];jj++){
	  if (inf(pexp[jj][igroup][ipara])<temval_hi && sup(pexp[jj][igroup][ipara])>=temval_lo) {
	    double impact=1.0/occur[count];
	    sum+=impact*pow(temval_mi-mu,2);
	  }
	  count++;
	}
      }
    }
    if (debug>0) cout<<endl<<"sum is  "<<sum;
    if (debug>0) cout<<endl<<"ctot is  "<<ctot;
    double sigma2=sum/(ctot-1);
    double root=0.0,lb=0.0,ub=0.0,lb2=0.0,ub2=0.0;
    double z=1.96;
    double worst[N];
    double best[N];
    double worstmean=0;
    double bestmean=0;
    double worstvar=0;
    double bestvar=0;
    int jj=0;
    for (int igroup=0;igroup<groups;igroup++){
      for (int j=0;j<gi[igroup];j++){
	double low=abs(inf(pexp[j][igroup][ipara])-mu);
	double high=abs(sup(pexp[j][igroup][ipara])-mu);
	if (low<high){
	  best[jj]=inf(pexp[j][igroup][ipara]);
	  worst[jj]=sup(pexp[j][igroup][ipara]);
	}
	else{
	  best[jj]=sup(pexp[j][igroup][ipara]);
	  worst[jj]=inf(pexp[j][igroup][ipara]);
	}
	bestmean+=best[jj];
	worstmean+=worst[jj];
	jj++;
      }
    }
    bestmean=bestmean/N;
    worstmean=worstmean/N;
    for (int jj=0;jj<N;jj++){
      bestvar+=pow(best[jj]-bestmean,2);
      worstvar+=pow(worst[jj]-worstmean,2);
    }
    bestvar/=(N-1);
    worstvar/=(N-1);
    cout<<"\nBest variance for parameter "<<ipara<<" is "<<bestvar<<"\n";
    cout<<"\nWorst variance for parameter "<<ipara<<" is "<<worstvar<<"\n";
    if (ctot==1) z=12.706;
    else if (ctot==2) z=4.303;
    else if (ctot==3) z=3.182;
    else if (ctot==4) z=2.776;
    else if (ctot==5) z=2.571;
    else if (ctot==6) z=2.447;
    else if (ctot==7) z=2.365;
    else if (ctot==8) z=2.306;
    else if (ctot==9) z=2.262;
    else if (ctot==10) z=2.228;
    else if (ctot<=15) z=2.131;
    else if (ctot<=20) z=2.086;		
    else if (ctot<=30) z=2.042;
    else if (ctot<=60) z=2.000;
    else if (ctot<=120) z=1.980;
    else z=1.96;
    root=sqrt(pow(sigma2,2)/ctot);
    lb=mu-z*root; 
    ub=mu+z*root; 
    pest[ipara]=mu;
    pestsigma2[ipara]=sigma2;
    double meanest=mu;
    double meantrue=psharp[ipara];
    if (model==1) {
      if (ipara==2 || ipara==3) {
	cout<<std::fixed <<std::setprecision(4)<<endl<<"\\(\\ \\beta_{pop"<<ipara-1<<"} \\) & "<<psharp[ipara-2]<<" & "<<psharp[ipara-2]+pest[ipara]<<"& ["<<psharp[ipara-2]+lb<<","<<psharp[ipara-2]+ub<<"] \\\\";
	cout<<std::fixed <<std::setprecision(4)<<endl<<"\\(d_"<<ipara-1<<"\\) & "<<d[ipara]<<" & "<<pestsigma2[ipara]<<"& \\\\";
      }
    }
    if (model==2) {  
      if (ipara==3 || ipara==4 || ipara==5) {
	cout<<std::fixed <<std::setprecision(4)<<endl<<"\\(\\ \\beta_{pop"<<ipara-2<<"} \\) & "<<psharp[ipara-3]<<" & "<<psharp[ipara-3]*exp(pest[ipara])<<"& ["<<psharp[ipara-3]*exp(lb)<<","<<psharp[ipara-3]*exp(ub)<<"] \\\\";
	cout<<std::fixed <<std::setprecision(4)<<endl<<"\\(d_"<<ipara-2<<"\\) & "<<d[ipara]<<" & "<<pestsigma2[ipara]<<"& \\\\";
      }
    }
    ofstream myplotfile;
    if (ipara==0) myplotfile.open ("myPlotHist1.gp");
    if (ipara==1) myplotfile.open ("myPlotHist2.gp");
    if (ipara==2) myplotfile.open ("myPlotHist3.gp");
    if (ipara==3) myplotfile.open ("myPlotHist4.gp");
    if (ipara==4) myplotfile.open ("myPlotHist5.gp");
    if (ipara==5) myplotfile.open ("myPlotHist6.gp");
    if (model==1) {
      if (ipara==0) myplotfile<<endl<<"set xlabel \"beta_pop1\"";
      if (ipara==1) myplotfile<<endl<<"set xlabel \"beta_pop2\"";
      if (ipara==2) myplotfile<<endl<<"set xlabel \"b_1\"";
      if (ipara==3) myplotfile<<endl<<"set xlabel \"b_2\"";
    }
    if (model==2) {
      if (ipara==0) myplotfile<<endl<<"set xlabel \"beta_pop1\"";
      if (ipara==1) myplotfile<<endl<<"set xlabel \"beta_pop2\"";
      if (ipara==2) myplotfile<<endl<<"set xlabel \"beta_pop3\"";
      if (ipara==3) myplotfile<<endl<<"set xlabel \"b_1\"";
      if (ipara==4) myplotfile<<endl<<"set xlabel \"b_2\"";
      if (ipara==5) myplotfile<<endl<<"set xlabel \"b_3\"";
    }
    myplotfile<<endl<<"set ylabel \"Frequency\"";
    myplotfile<<endl<<"set yrange [0:"<<1.1*cmax[ipara]<<"]";
    myplotfile<<endl<<"set grid";
    myplotfile<<endl<<"beta="<<psharp[ipara];
    myplotfile<<endl<<"betaest="<<pest[ipara];
    myplotfile<<endl<<"sigma = sqrt("<<d[ipara]<<")";
    myplotfile<<endl<<"sigmaest = sqrt("<<pestsigma2[ipara]<<")";
    myplotfile<<endl<<"mean = beta";
    myplotfile<<endl<<"set style line 1 lt 1 lw 1";
    myplotfile<<endl<<"set style line 2 lt 3 lw 5";
    myplotfile<<endl<<"set style line 3 lt 2 lw 5";
    if (action==3 && subaction==1){
      myplotfile<<endl<<"myf(x)=exp(-(x-mean)*(x-mean)/sigma/sigma/2.0)/sqrt(2*pi)/sigma";
      myplotfile<<endl<<"myfest(x)=exp(-(x-mean)*(x-mean)/sigmaest/sigmaest/2.0)/sqrt(2*pi)/sigmaest";
    }
    else{
      myplotfile<<endl<<"myf(x)=exp(-(x-mean)*(x-mean)/sigma/sigma/2.0)/sqrt(2*pi)/sigma";
      myplotfile<<endl<<"myfest(x)=exp(-(x-mean)*(x-mean)/sigmaest/sigmaest/2.0)/sqrt(2*pi)/sigmaest";
    }
    if (doplot) {
      myplotfile<<endl<<"plot \"myDataHist"<<(ipara+1)<<".data\"  title 'Coverage of intervals' with boxes ls 1,\\";
      myplotfile<<endl<<"myf(x) title 'True distribution' w l ls 2,\\";
      myplotfile<<endl<<"myfest(x) title 'Estimated distribution' w l ls 3"<<endl;
    }
    myplotfile<<endl<<"set term postscript eps solid color blacktext \"Helvetica\" 24";
    if (ipara==0)   myplotfile<<endl<<"set output 'SavedPlots/"<<figname<<"_1.eps'";
    if (ipara==1)   myplotfile<<endl<<"set output 'SavedPlots/"<<figname<<"_2.eps'";
    if (ipara==2)   myplotfile<<endl<<"set output 'SavedPlots/"<<figname<<"_3.eps'";
    if (ipara==3)   myplotfile<<endl<<"set output 'SavedPlots/"<<figname<<"_4.eps'";
    if (ipara==4)   myplotfile<<endl<<"set output 'SavedPlots/"<<figname<<"_5.eps'";
    if (ipara==5)   myplotfile<<endl<<"set output 'SavedPlots/"<<figname<<"_6.eps'";
    if (model==1) {
      if (ipara==0) myplotfile<<endl<<"set xlabel \"beta_pop1\"";
      if (ipara==1) myplotfile<<endl<<"set xlabel \"beta_pop2\"";
      if (ipara==2) myplotfile<<endl<<"set xlabel \"b_1\"";
      if (ipara==3) myplotfile<<endl<<"set xlabel \"b_2\"";
    }
    if (model==2) {
      if (ipara==0) myplotfile<<endl<<"set xlabel \"beta_pop1\"";
      if (ipara==1) myplotfile<<endl<<"set xlabel \"beta_pop2\"";
      if (ipara==2) myplotfile<<endl<<"set xlabel \"beta_pop3\"";
      if (ipara==3) myplotfile<<endl<<"set xlabel \"b_1\"";
      if (ipara==4) myplotfile<<endl<<"set xlabel \"b_2\"";
      if (ipara==5) myplotfile<<endl<<"set xlabel \"b_3\"";
    }
    myplotfile<<endl<<"set ylabel \"Frequency\"";
    myplotfile<<endl<<"set grid";

    myplotfile<<endl<<"plot \"myDataHist"<<(ipara+1)<<".data\"  title 'Coverage of intervals' with boxes ls 1,\\";
    if (doplot) {
      myplotfile<<endl<<"myf(x) title 'True distribution' w l ls 2,\\";
      myplotfile<<endl<<"myfest(x) title 'Estimated distribution' w l ls 3"<<endl;
    }
    myplotfile<<endl<<"set output";
    myplotfile.close(); 
    if (doplot && ipara==0)  int out=system("gnuplot -persist < myPlotHist1.gp");
    if (doplot && ipara==1)  int out=system("gnuplot -persist < myPlotHist2.gp");
    if (doplot && ipara==2)  int out=system("gnuplot -persist < myPlotHist3.gp");
    if (doplot && ipara==3)  int out=system("gnuplot -persist < myPlotHist4.gp");
    if (doplot && ipara==4)  int out=system("gnuplot -persist < myPlotHist5.gp");
    if (doplot && ipara==5)  int out=system("gnuplot -persist < myPlotHist6.gp");
  }
  cout<<endl;

  for (int ipara=0;ipara<nParameter/2;ipara++){
    if (model==1) pest[ipara]=psharp[ipara];
    if (model==2) pest[ipara]=psharp[ipara];
  }


}




















// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// Gnuplot splot of pList data 
// arrowtype 0- no, 1-arrow to best region, 2-arrow to hardcoded position
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
void plot3d(string figname,int igroup,int arrowtype){
  labelPlist();
  for (int i=0;i<pListSize;i++){
    ofstream mydatfile;
    string ss="myData3d_"+stringcounter[i]+".data";
    mydatfile.open(ss.c_str());
    double a1=inf(pList[3][i]);
    double a2=sup(pList[3][i]);
    double b1=inf(pList[4][i]);
    double b2=sup(pList[4][i]);
    double c1=inf(pList[5][i]);
    double c2=sup(pList[5][i]);
    double con=0.05;
    if (abs(a1-a2)<abs(a1)*0.1) {a1-=a1*con;a2+=a2*con;}
    if (abs(b1-b2)<abs(b1)*0.1) {b1-=b1*con;b2+=b2*con;}
    if (abs(c1-c2)<abs(c1)*0.1) {c1-=c1*con;c2+=c2*con;}
    if (a1<psearch[3][0]) a1=psearch[3][0];
    if (b1<psearch[4][0]) b1=psearch[4][0];
    if (c1<psearch[5][0]) c1=psearch[5][0];
    if (a2>psearch[3][1]) a2=psearch[3][1];
    if (b2>psearch[4][1]) b2=psearch[4][1];
    if (c2>psearch[5][1]) c2=psearch[5][1];
    mydatfile<<endl<<a1<<" "<<b1<<" "<<c1;
    mydatfile<<endl<<a2<<" "<<b1<<" "<<c1;
    mydatfile<<endl<<a2<<" "<<b2<<" "<<c1;
    mydatfile<<endl<<a1<<" "<<b2<<" "<<c1;
    mydatfile<<endl<<a1<<" "<<b1<<" "<<c1;
    mydatfile<<endl<<a1<<" "<<b1<<" "<<c2;
    mydatfile<<endl<<a1<<" "<<b2<<" "<<c2;
    mydatfile<<endl<<a2<<" "<<b2<<" "<<c2;
    mydatfile<<endl<<a2<<" "<<b1<<" "<<c2;
    mydatfile<<endl<<a1<<" "<<b1<<" "<<c2;
    mydatfile<<endl<<a2<<" "<<b1<<" "<<c2;
    mydatfile<<endl<<a2<<" "<<b1<<" "<<c1;
    mydatfile<<endl<<a2<<" "<<b2<<" "<<c1;
    mydatfile<<endl<<a2<<" "<<b2<<" "<<c2;
    mydatfile<<endl<<a1<<" "<<b2<<" "<<c2;
    mydatfile<<endl<<a1<<" "<<b2<<" "<<c1;
    mydatfile<<endl; 
    mydatfile.close();
  }
  ofstream myplotfile;
  myplotfile.open ("myPlot3d.gp");
  myplotfile<<endl<<"set xlabel \"b_1\"";
  myplotfile<<endl<<"set ylabel \"b_2\"";
  myplotfile<<endl<<"set zlabel \"b_3\"";
  myplotfile<<endl<<"set nokey";
  if (arrowtype==2){
    myplotfile<<endl<<"set xtics -5,0.5,5";
    myplotfile<<endl<<"set ytics -3, 0.5, 3.0";
    myplotfile<<endl<<"set ztics -3, 0.5, 3.0";
    myplotfile<<endl<<"set xr [-5.0:5.0]";
    myplotfile<<endl<<"set yr [-3.0:3.0]";
    myplotfile<<endl<<"set zr [-3.0:3.0]";
    myplotfile<<endl<<"set view 30,325";
  }
  else{
    myplotfile<<endl<<"set xtics -5,0.5,5";
    myplotfile<<endl<<"set ytics -3, 0.5, 3.0";
    myplotfile<<endl<<"set ztics -3, 0.5, 3.0";
    myplotfile<<endl<<"set xr [-5.0:5.0]";
    myplotfile<<endl<<"set yr [-3.0:3.0]";
    myplotfile<<endl<<"set zr [-3.0:3.0]";
    myplotfile<<endl<<"set view 30,325";
  }
  myplotfile<<endl<<"set ticslevel 0";
  myplotfile<<endl<<"set pointsize 3";
  myplotfile<<endl<<"set style line 1 lt 1 lw 2";
  if (arrowtype==1) {
    myplotfile<<endl<<"set arrow from -0.341924+2, 0.209908,  2 to -0.341924+0.05, 0.209908,  -0.193059 ls 1";
  }
  if (arrowtype==2) {
    myplotfile<<endl<<"set arrow from -1.91752+2, 0.0967215,2  to  -1.91752+0.05, 0.0967215,-0.10458 ls 1";
    myplotfile<<endl<<"set arrow from  -3.45-1.5, 1.6, 2 to -3.45-0.05, 1.6, -0.1 ls 1";
  }
  myplotfile<<endl<<"splot ";
  for (int i=0;i<pListSize;i++){
    if (i<pListSize-1) {
      if (pListLabel[i]==1) myplotfile<<"\"myData3d_"<<i<<".data\"  u 1:2:3 with lines 3 4,\\"<<endl;
      else if (pListLabel[i]==2) myplotfile<<"\"myData3d_"<<i<<".data\"  u 1:2:3 with lines 3 4,\\"<<endl;
      else myplotfile<<"\"myData3d_"<<i<<".data\"  u 1:2:3 with lines 3 4,\\"<<endl;
    }
    else{
      if (pListLabel[i]==1) myplotfile<<"\"myData3d_"<<i<<".data\"  u 1:2:3 with lines 3 4"<<endl;
      else if (pListLabel[i]==2) myplotfile<<"\"myData3d_"<<i<<".data\"  u 1:2:3 with lines 3 4"<<endl;
      else  myplotfile<<"\"myData3d_"<<i<<".data\"  u 1:2:3 with lines 3 4"<<endl;
    }
  }
  myplotfile<<endl<<"set term postscript eps solid color blacktext \"Helvetica\" 16";
  myplotfile<<endl<<"set output 'SavedPlots/"<<figname<<".eps'";
  myplotfile<<endl<<"splot ";
  for (int i=0;i<pListSize;i++){
    if (i<pListSize-1) {
      if (pListLabel[i]==1) myplotfile<<"\"myData3d_"<<i<<".data\"  u 1:2:3 with lines 3 4,\\"<<endl;
      else if (pListLabel[i]==2) myplotfile<<"\"myData3d_"<<i<<".data\"  u 1:2:3 with lines 3 4,\\"<<endl;
      else myplotfile<<"\"myData3d_"<<i<<".data\"  u 1:2:3 with lines 3 4,\\"<<endl;
    }
    else{
      if (pListLabel[i]==1) myplotfile<<"\"myData3d_"<<i<<".data\"  u 1:2:3 with lines 3 4"<<endl;
      else if (pListLabel[i]==2) myplotfile<<"\"myData3d_"<<i<<".data\"  u 1:2:3 with lines 3 4"<<endl;
      else  myplotfile<<"\"myData3d_"<<i<<".data\"  u 1:2:3 with lines 3 4"<<endl;
    }
  }
  myplotfile<<endl<<"set output";
  myplotfile.close(); 
  int out=system("gnuplot -persist myPlot3d.gp");
}















// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// label items in pListSize.
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
void labelPlist(){
  int currentLabel=0;
  bool done=false;
  double FACTOR=1.0;
  for (int i=0;i<pListSize_max;i++) pListLabel[i]=0;
  for (;;) {
    currentLabel++; // Update currentLabel and assign to first 
    for (int i=0;i<pListSize;i++) {
      if (pListLabel[i]==0) {
	pListLabel[i]=currentLabel;
	break;
      }
    }
    bool succ;
    do {
      succ=false;
      for (int i=0;i<pListSize;i++){
	for (int j=0;j<pListSize;j++){
	  if (pListLabel[i]==currentLabel && pListLabel[j]==0){
	    bool connected=true; 
	    for (int ip=0;ip<nParameter;ip++) {
	      double a1=inf(pList[ip][i]);
	      double a2=sup(pList[ip][i]);
	      double b1=inf(pList[ip][j]);
	      double b2=sup(pList[ip][j]);
	      if (! (
		     (b1>=a1*FACTOR  && b1*FACTOR <=a2) || 
		     (b2>=a1*FACTOR   && b2*FACTOR <=a2) || 
		     (b2>=a2*FACTOR   && b1*FACTOR <=a1) 
		     ) ){
		connected=false;
		break;
	      }
	    }
	    if (connected) {
	      pListLabel[j]=currentLabel;
	      succ=true;
	    }
	  }
	} 
      } 
    } while (succ);   
    done=true;
    for (int i=0;done && i<pListSize;i++) if (pListLabel[i]==0) done=false;
    if (done) break;
  }
  for (int iLabel=1;iLabel<=currentLabel;iLabel++) // set empty
    for (int ip=0;ip<nParameter;ip++)
      pListHull[ip][iLabel].set(1.0,-1.0);
  for (int iLabel=1;iLabel<=currentLabel;iLabel++) {
    for (int j=0;j<pListSize;j++) {
      if (pListLabel[j]==iLabel){
	for (int ip=0;ip<nParameter;ip++) {
	  pListHull[ip][iLabel]=pListHull[ip][iLabel] | pList[ip][j];
	}
      }
    }
  }
  numberOfLabels=currentLabel;
}










// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// Estimate mean and variance from histogram
// Note that empty intervals are already considered in the called
// functions inf and sup in the class interval
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
void meanandvar(int jj){
  int BINS=300;
  if (model==2) BINS=500;
  for (int ipara=0;ipara<nParameter;ipara++){
    double binlength=(psearch[ipara][1]-psearch[ipara][0])/BINS;
    double ctot=0;
    double lnsum=0;
    int occur[groups];
    for (int igro=0;igro<groups;igro++)occur[igro]=0;
    for (int i=0;i<BINS;i++){
      double temval_lo=psearch[ipara][0]+i*binlength;
      double temval_mi=psearch[ipara][0]+(i+0.5)*binlength;
      double temval_hi=psearch[ipara][0]+(i+1)*binlength;
      for (int igro=0;igro<groups;igro++){
	if (inf(pexp[jj][igro][ipara])<temval_hi && sup(pexp[jj][igro][ipara])>=temval_lo) {
	  occur[igro]++;
	}
      }
    }
    for (int i=0;i<BINS;i++){
      double temval_lo=psearch[ipara][0]+i*binlength;
      double temval_mi=psearch[ipara][0]+(i+0.5)*binlength;
      double temval_hi=psearch[ipara][0]+(i+1)*binlength;
      double c=0.0;
      for (int igro=0;igro<groups;igro++){
	if (inf(pexp[jj][igro][ipara])<temval_hi && sup(pexp[jj][igro][ipara])>=temval_lo) {
	  double impact=1.0/occur[igro]; 
	  c+=impact;
	  ctot+=impact;
	  lnsum+=impact*log(temval_mi);
	}
      }
    }
    double mu=lnsum/ctot;
    double sum=0;
    for (int i=0;i<BINS;i++){
      double temval_lo=psearch[ipara][0]+i*binlength;
      double temval_mi=psearch[ipara][0]+(i+0.5)*binlength;
      double temval_hi=psearch[ipara][0]+(i+1)*binlength;
      for (int igro=0;igro<groups;igro++){
	if (inf(pexp[jj][igro][ipara])<temval_hi && sup(pexp[jj][igro][ipara])>=temval_lo) {
	  double impact=1.0/occur[igro];
	  sum+=impact*pow((log(temval_mi)-mu),2);
	}
      }
    }
    double sigma2=sum/(ctot-1);
    pest[ipara]=exp(mu);
    pestsigma2[ipara]=sigma2;
  }
}















// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// uniformly random numbers
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
double uniformRnd(double min, double max) {
  return ((double) rand()/RAND_MAX)*(max - min) + min;
}




